static int path_tainted = -1;

static char **origenviron;
#ifdef _WIN32
#define GET_ENVIRON(e) (e = rb_w32_get_environ())
#define FREE_ENVIRON(e) rb_w32_free_environ(e)
static char **my_environ;
#undef environ
#define environ my_environ
#elif defined(__APPLE__)
#undef environ
#define environ (*_NSGetEnviron())
#define GET_ENVIRON(e) (e)
#define FREE_ENVIRON(e)
#else
extern char **environ;
#define GET_ENVIRON(e) (e)
#define FREE_ENVIRON(e)
#endif
#ifdef ENV_IGNORECASE
#define ENVMATCH(s1, s2) (STRCASECMP(s1, s2) == 0)
#define ENVNMATCH(s1, s2, n) (STRNCASECMP(s1, s2, n) == 0)
#else
#define ENVMATCH(n1, n2) (strcmp(n1, n2) == 0)
#define ENVNMATCH(s1, s2, n) (memcmp(s1, s2, n) == 0)
#endif

static mrb_value
env_str_new(mrb_state *mrb, const char *ptr, long len)
{
    mrb_value str = rb_locale_str_new(mrb, ptr, len);

    //rb_obj_freeze(str);
    return str;
}

static mrb_value
env_str_new2(mrb_state *mrb, const char *ptr)
{
    if (!ptr) return mrb_nil_value();
    return env_str_new(mrb, ptr, strlen(ptr));
}

static mrb_value
env_delete(mrb_state *mrb, mrb_value obj, mrb_value name)
{
    char *nam, *val;

    rb_secure(4);
    //SafeStringValue(name);
    mrb_string_value(mrb, &name);
    nam = RSTRING_PTR(name);
    if (memchr(nam, '\0', RSTRING_LEN(name))) {
        mrb_raise(mrb, mrb->eArgumentError_class, "bad environment variable name");
    }
    val = getenv(nam);
    if (val) {
        mrb_value value = env_str_new2(mrb, val);

        ruby_setenv(mrb, nam, 0);
        if (ENVMATCH(nam, PATH_ENV)) {
            path_tainted = 0;
        }
        return value;
    }
    return mrb_nil_value();
}
#if 0
static mrb_value
env_delete_m(mrb_state *mrb, mrb_value obj, mrb_value name)
{
    mrb_value val;

    val = env_delete(mrb, obj, name);
    if (mrb_nil_p(val) && rb_block_given_p()) rb_yield(name);
    return val;
}
#endif

static int env_path_tainted(mrb_state *mrb, const char *);

static mrb_value
rb_f_getenv(mrb_state *mrb, mrb_value obj, mrb_value name)
{
    char *nam, *env;

    //rb_secure(4);
    //SafeStringValue(name);
    mrb_string_value(mrb, &name);
    nam = RSTRING_PTR(name);
    if (memchr(nam, '\0', RSTRING_LEN(name))) {
        mrb_raise(mrb, mrb->eArgumentError_class, "bad environment variable name");
    }
    env = getenv(nam);
    if (env) {
        if (ENVMATCH(nam, PATH_ENV) && !env_path_tainted(mrb, env)) {
            mrb_value str = rb_filesystem_str_new_cstr(mrb, env);

            //rb_obj_freeze(str);
            return str;
        }
        return env_str_new2(mrb, env);
    }
    return mrb_nil_value();
}

#if 0
static mrb_value
env_fetch(mrb_state *mrb, int argc, mrb_value *argv)
{
    mrb_value key, if_none;
    long block_given;
    char *nam, *env;

    rb_secure(4);
    rb_scan_args(argc, argv, "11", &key, &if_none);
    block_given = rb_block_given_p();
    if (block_given && argc == 2) {
        rb_warn("block supersedes default value argument");
    }
    //SafeStringValue(key);
    mrb_string_value(mrb, &key);
    nam = RSTRING_PTR(key);
    if (memchr(nam, '\0', RSTRING_LEN(key))) {
        mrb_raise(mrb, mrb->eArgumentError_class, "bad environment variable name");
    }
    env = getenv(nam);
    if (!env) {
        if (block_given) return rb_yield(key);
        if (argc == 1) {
            mrb_raise(mrb, mrb->eKeyError_class, "key not found");
        }
        return if_none;
    }
    if (ENVMATCH(nam, PATH_ENV) && !env_path_tainted(env))
        return rb_filesystem_str_new_cstr(env);
    return env_str_new2(mrb, env);
}
#endif

static void
path_tainted_p(mrb_state *mrb, const char *path)
{
    path_tainted = rb_path_check(mrb, path)?0:1;
}

static int
env_path_tainted(mrb_state *mrb, const char *path)
{
    if (path_tainted < 0) {
        path_tainted_p(mrb, path);
    }
    return path_tainted;
}

int
rb_env_path_tainted(mrb_state *mrb)
{
    if (path_tainted < 0) {
        path_tainted_p(mrb, getenv(PATH_ENV));
    }
    return path_tainted;
}

#if defined(_WIN32) || (defined(HAVE_SETENV) && defined(HAVE_UNSETENV))
#elif defined __sun__
static int
in_origenv(const char *str)
{
    char **env;
    for (env = origenviron; *env; ++env) {
        if (*env == str) return 1;
    }
    return 0;
}
#else
static int
envix(const char *nam)
{
    register int i, len = strlen(nam);
    char **env;

    env = GET_ENVIRON(environ);
    for (i = 0; env[i]; i++) {
        if (ENVNMATCH(env[i],nam,len) && env[i][len] == '=')
            break;                        /* memcmp must come first to avoid */
    }                                        /* potential SEGV's */
    FREE_ENVIRON(environ);
    return i;
}
#endif

#if defined(_WIN32)
#elif defined __sun__
#else  /* WIN32 */
char *
ruby_strdup(const char *str)
{
    char *tmp;
    size_t len = strlen(str) + 1;

    tmp = malloc(len);
    memcpy(tmp, str, len);

    return tmp;
}
#endif

void
ruby_setenv(mrb_state *mrb, const char *name, const char *value)
{
#if defined(_WIN32)
    int len;
    char *buf;
    int failed = 0;
    if (strchr(name, '=')) {
        errno = EINVAL;
        rb_sys_fail("ruby_setenv");
    }
    if (value) {
        len = strlen(name) + 1 + strlen(value) + 1;
        buf = ALLOCA_N(char, len);
        snprintf(buf, len, "%s=%s", name, value);
        failed = putenv(buf);

        /* putenv() doesn't handle empty value */
        if (!*value)
            failed = !SetEnvironmentVariable(name,value);
    }
    else {
        len = strlen(name) + 1 + 1;
        buf = ALLOCA_N(char, len);
        snprintf(buf, len, "%s=", name);
        putenv(buf);
        failed = !SetEnvironmentVariable(name, 0);
    }
    if (failed) {
        rb_warn("failed to set environment variable. Ruby 1.9.3 will raise SystemCallError in this case.");
    }
#elif defined(HAVE_SETENV) && defined(HAVE_UNSETENV)
#undef setenv
#undef unsetenv
    if (value) {
        if (setenv(name, value, 1))
            rb_sys_fail("setenv");
    } else {
#ifdef VOID_UNSETENV
        unsetenv(name);
#else
        if (unsetenv(name))
            rb_sys_fail("unsetenv");
#endif
    }
#elif defined __sun__
    size_t len;
    char **env_ptr, *str;
    if (strchr(name, '=')) {
        errno = EINVAL;
        rb_sys_fail("ruby_setenv");
    }
    len = strlen(name);
    for (env_ptr = GET_ENVIRON(environ); (str = *env_ptr) != 0; ++env_ptr) {
        if (!strncmp(str, name, len) && str[len] == '=') {
            if (!in_origenv(str)) free(str);
            while ((env_ptr[0] = env_ptr[1]) != 0) env_ptr++;
            break;
        }
    }
    if (value) {
        str = malloc(len += strlen(value) + 2);
        snprintf(str, len, "%s=%s", name, value);
        if (putenv(str))
            rb_sys_fail("putenv");
    }
#else  /* WIN32 */
    size_t len;
    int i;
    if (strchr(name, '=')) {
        errno = EINVAL;
        rb_sys_fail("ruby_setenv");
    }
    i=envix(name);                        /* where does it go? */

    if (environ == origenviron) {        /* need we copy environment? */
        int j;
        int max;
        char **tmpenv;

        for (max = i; environ[max]; max++) ;
        tmpenv = malloc(sizeof(char*)*(max+2));//ALLOC_N(char*, max+2);
        for (j=0; j<max; j++)                /* copy environment */
            tmpenv[j] = ruby_strdup(environ[j]);
        tmpenv[max] = 0;
        environ = tmpenv;                /* tell exec where it is now */
    }
    if (environ[i]) {
        char **envp = origenviron;
        while (*envp && *envp != environ[i]) envp++;
        if (!*envp)
            xfree(environ[i]);
        if (!value) {
            while (environ[i]) {
                environ[i] = environ[i+1];
                i++;
            }
            return;
        }
    }
    else {                        /* does not exist yet */
        if (!value) return;
        mrb_realloc(mrb, environ, sizeof(char*)*(i+2));        /* just expand it a bit */
        environ[i+1] = 0;        /* make sure it's null terminated */
    }
    len = strlen(name) + strlen(value) + 2;
    environ[i] = malloc(len);//ALLOC_N(char, len);
    snprintf(environ[i],len,"%s=%s",name,value); /* all that work just for this */
#endif /* WIN32 */
}

void
ruby_unsetenv(mrb_state *mrb, const char *name)
{
    ruby_setenv(mrb, name, 0);
}

#if 0
static mrb_value
env_aset(mrb_state *mrb, mrb_value obj, mrb_value nm, mrb_value val)
{
    char *name, *value;

    if (rb_safe_level() >= 4) {
        mrb_raise(mrb, mrb->eSecurityError_class, "can't change environment variable");
    }

    if (mrb_nil_p(val)) {
        env_delete(mrb, obj, nm);
        return mrb_nil_value();
    }
    StringValue(nm);
    StringValue(val);
    name = RSTRING_PTR(nm);
    value = RSTRING_PTR(val);
    if (memchr(name, '\0', RSTRING_LEN(nm)))
        mrb_raise(mrb, mrb->eArgumentError_class, "bad environment variable name");
    if (memchr(value, '\0', RSTRING_LEN(val)))
        mrb_raise(mrb, mrb->eArgumentError_class, "bad environment variable value");

    ruby_setenv(name, value);
    if (ENVMATCH(name, PATH_ENV)) {
        if (OBJ_TAINTED(val)) {
            /* already tainted, no check */
            path_tainted = 1;
            return val;
        }
        else {
            path_tainted_p(mrb, value);
        }
    }
    return val;
}
#endif

static mrb_value
env_keys(mrb_state *mrb)
{
    char **env;
    mrb_value ary;

    rb_secure(4);
    ary = mrb_ary_new(mrb);
    env = GET_ENVIRON(environ);
    while (*env) {
        char *s = strchr(*env, '=');
        if (s) {
            mrb_ary_push(mrb, ary, env_str_new(mrb, *env, s-*env));
        }
        env++;
    }
    FREE_ENVIRON(environ);
    return ary;
}

#if 0
static mrb_value
env_each_key(mrb_value ehash)
{
    mrb_value keys;
    long i;

    //RETURN_ENUMERATOR(ehash, 0, 0);
    keys = env_keys(mrb);        /* rb_secure(4); */
    for (i=0; i<RARRAY_LEN(keys); i++) {
        rb_yield(RARRAY_PTR(keys)[i]);
    }
    return ehash;
}
#endif

#if 0
static mrb_value
env_values(mrb_state *mrb)
{
    mrb_value ary;
    char **env;

    rb_secure(4);
    ary = mrb_ary_new(mrb);
    env = GET_ENVIRON(environ);
    while (*env) {
        char *s = strchr(*env, '=');
        if (s) {
            mrb_ary_push(mrb, ary, env_str_new2(mrb, s+1));
        }
        env++;
    }
    FREE_ENVIRON(environ);
    return ary;
}

static mrb_value
env_each_value(mrb_value ehash)
{
    mrb_value values;
    long i;

    //RETURN_ENUMERATOR(ehash, 0, 0);
    values = env_values();        /* rb_secure(4); */
    for (i=0; i<RARRAY_LEN(values); i++) {
        rb_yield(RARRAY_PTR(values)[i]);
    }
    return ehash;
}

static mrb_value
env_each_pair(mrb_state *mrb, mrb_value ehash)
{
    char **env;
    mrb_value ary;
    long i;

    //RETURN_ENUMERATOR(ehash, 0, 0);

    rb_secure(4);
    ary = mrb_ary_new(mrb);
    env = GET_ENVIRON(environ);
    while (*env) {
        char *s = strchr(*env, '=');
        if (s) {
            mrb_ary_push(mrb, ary, env_str_new(mrb, *env, s-*env));
            mrb_ary_push(mrb, ary, env_str_new2(mrb, s+1));
        }
        env++;
    }
    FREE_ENVIRON(environ);

    for (i=0; i<RARRAY_LEN(ary); i+=2) {
        rb_yield(rb_assoc_new(mrb, RARRAY_PTR(ary)[i], RARRAY_PTR(ary)[i+1]));
    }
    return ehash;
}

static mrb_value
env_reject_bang(mrb_state *mrb, mrb_value ehash)
{
    volatile mrb_value keys;
    long i;
    int del = 0;

    //RETURN_ENUMERATOR(ehash, 0, 0);
    keys = env_keys(mrb);        /* rb_secure(4); */
    for (i=0; i<RARRAY_LEN(keys); i++) {
        mrb_value val = rb_f_getenv(mrb, mrb_nil_value(), RARRAY_PTR(keys)[i]);
        if (!mrb_nil_p(val)) {
            if (RTEST(rb_yield_values(2, RARRAY_PTR(keys)[i], val))) {
                //FL_UNSET(RARRAY_PTR(keys)[i], FL_TAINT);
                env_delete(mrb, mrb_nil_value(), RARRAY_PTR(keys)[i]);
                del++;
            }
        }
    }
    if (del == 0) return mrb_nil_value();
    return envtbl;
}

static mrb_value
env_delete_if(mrb_value ehash)
{
    //RETURN_ENUMERATOR(ehash, 0, 0);
    env_reject_bang(ehash);
    return envtbl;
}

static mrb_value
env_values_at(int argc, mrb_value *argv)
{
    mrb_value result;
    long i;

    rb_secure(4);
    result = mrb_ary_new(mrb);
    for (i=0; i<argc; i++) {
        mrb_ary_push(mrb, result, rb_f_getenv(mrb, mrb_nil_value(), argv[i]));
    }
    return result;
}

static mrb_value
env_select(mrb_state *mrb, mrb_value ehash)
{
    mrb_value result;
    char **env;

    //RETURN_ENUMERATOR(ehash, 0, 0);
    rb_secure(4);
    result = rb_hash_new(mrb);
    env = GET_ENVIRON(environ);
    while (*env) {
        char *s = strchr(*env, '=');
        if (s) {
            mrb_value k = env_str_new(mrb, *env, s-*env);
            mrb_value v = env_str_new2(mrb, s+1);
            if (RTEST(rb_yield_values(2, k, v))) {
                mrb_hash_set/*rb_hash_aset*/(mrb, result, k, v);
            }
        }
        env++;
    }
    FREE_ENVIRON(environ);

    return result;
}

static mrb_value
env_select_bang(mrb_state *mrb, mrb_value ehash)
{
    volatile mrb_value keys;
    long i;
    int del = 0;

    //RETURN_ENUMERATOR(ehash, 0, 0);
    keys = env_keys(mrb);        /* rb_secure(4); */
    for (i=0; i<RARRAY_LEN(keys); i++) {
        mrb_value val = rb_f_getenv(mrb, mrb_nil_value(), RARRAY_PTR(keys)[i]);
        if (!mrb_nil_p(val)) {
            if (!RTEST(rb_yield_values(2, RARRAY_PTR(keys)[i], val))) {
                //FL_UNSET(RARRAY_PTR(keys)[i], FL_TAINT);
                env_delete(mrb, mrb_nil_value(), RARRAY_PTR(keys)[i]);
                del++;
            }
        }
    }
    if (del == 0) return mrb_nil_value();
    return envtbl;
}

static mrb_value
env_keep_if(mrb_value ehash)
{
    //RETURN_ENUMERATOR(ehash, 0, 0);
    env_select_bang(ehash);
    return envtbl;
}
#endif

mrb_value
rb_env_clear(mrb_state *mrb)
{
    volatile mrb_value keys;
    long i;

    keys = env_keys(mrb);        /* rb_secure(4); */
    for (i=0; i<RARRAY_LEN(keys); i++) {
        mrb_value val = rb_f_getenv(mrb, mrb_nil_value(), RARRAY_PTR(keys)[i]);
        if (!mrb_nil_p(val)) {
            env_delete(mrb, mrb_nil_value(), RARRAY_PTR(keys)[i]);
        }
    }
    return envtbl;
}

#if 0
static mrb_value
env_to_s(void)
{
    return rb_usascii_str_new2("ENV");
}

static mrb_value
env_inspect(void)
{
    char **env;
    mrb_value str, i;

    rb_secure(4);
    str = rb_str_buf_new2(mrb, "{");
    env = GET_ENVIRON(environ);
    while (*env) {
        char *s = strchr(*env, '=');

        if (env != environ) {
           mrb_str_buf_cat(mrb, str, ", ", strlen(", "));// rb_str_buf_cat2(str, ", ");
        }
        if (s) {
            mrb_str_buf_cat(mrb, str, "\"", strlen("\""));//rb_str_buf_cat2(str, "\"");
            mrb_str_buf_cat(mrb, str, *env, s-*env);
            mrb_str_buf_cat(mrb, str, "\"=>", strlen("\"=>"));//rb_str_buf_cat2(str, "\"=>");
            i = mrb_inspect(mrb, mrb_str_new2(mrb, s+1));
            rb_str_buf_append(mrb, str, i);
        }
        env++;
    }
    FREE_ENVIRON(environ);
    mrb_str_buf_cat(mrb, str, "}", strlen("}")); //rb_str_buf_cat2(str, "}");
    OBJ_TAINT(str);

    return str;
}

static mrb_value
env_to_a(mrb_state *mrb)
{
    char **env;
    mrb_value ary;

    rb_secure(4);
    ary = mrb_ary_new(mrb);
    env = GET_ENVIRON(environ);
    while (*env) {
        char *s = strchr(*env, '=');
        if (s) {
            mrb_ary_push(mrb, ary, rb_assoc_new(mrb, env_str_new(mrb, *env, s-*env),
                                          env_str_new2(mrb, s+1)));
        }
        env++;
    }
    FREE_ENVIRON(environ);
    return ary;
}

static mrb_value
env_none(void)
{
    return mrb_nil_value();
}

static mrb_value
env_size(void)
{
    int i;
    char **env;

    rb_secure(4);
    env = GET_ENVIRON(environ);
    for(i=0; env[i]; i++)
        ;
    FREE_ENVIRON(environ);
    return INT2FIX(i);
}

static mrb_value
env_empty_p(void)
{
    char **env;

    rb_secure(4);
    env = GET_ENVIRON(environ);
    if (env[0] == 0) {
        FREE_ENVIRON(environ);
        return mrb_true_value();
    }
    FREE_ENVIRON(environ);
    return mrb_false_value();
}

static mrb_value
env_has_key(mrb_value env, mrb_value key)
{
    char *s;

    rb_secure(4);
    s = StringValuePtr(key);
    if (memchr(s, '\0', RSTRING_LEN(key)))
        mrb_raise(mrb, mrb->eArgumentError_class, "bad environment variable name");
    if (getenv(s)) return mrb_true_value();
    return mrb_false_value();
}

static mrb_value
env_assoc(mrb_value env, mrb_value key)
{
    char *s, *e;

    rb_secure(4);
    s = StringValuePtr(key);
    if (memchr(s, '\0', RSTRING_LEN(key)))
        mrb_raise(mrb, mrb->eArgumentError_class, "bad environment variable name");
    e = getenv(s);
    if (e) return rb_assoc_new(mrb, key, rb_tainted_str_new2(e));
    return mrb_nil_value();
}

static mrb_value
env_has_value(mrb_value dmy, mrb_value obj)
{
    char **env;

    rb_secure(4);
    obj = rb_check_string_type(obj);
    if (mrb_nil_p(obj)) return mrb_nil_value();
    env = GET_ENVIRON(environ);
    while (*env) {
        char *s = strchr(*env, '=');
        if (s++) {
            long len = strlen(s);
            if (RSTRING_LEN(obj) == len && strncmp(s, RSTRING_PTR(obj), len) == 0) {
                FREE_ENVIRON(environ);
                return mrb_true_value();
            }
        }
        env++;
    }
    FREE_ENVIRON(environ);
    return mrb_false_value();
}

static mrb_value
env_rassoc(mrb_value dmy, mrb_value obj)
{
    char **env;

    rb_secure(4);
    obj = rb_check_string_type(obj);
    if (mrb_nil_p(obj)) return mrb_nil_value();
    env = GET_ENVIRON(environ);
    while (*env) {
        char *s = strchr(*env, '=');
        if (s++) {
            long len = strlen(s);
            if (RSTRING_LEN(obj) == len && strncmp(s, RSTRING_PTR(obj), len) == 0) {
                mrb_value result = rb_assoc_new(mrb, rb_tainted_str_new(*env, s-*env-1), obj);
                FREE_ENVIRON(environ);
                return result;
            }
        }
        env++;
    }
    FREE_ENVIRON(environ);
    return mrb_nil_value();
}

static mrb_value
env_key(mrb_value dmy, mrb_value value)
{
    char **env;
    mrb_value str;

    rb_secure(4);
    StringValue(value);
    env = GET_ENVIRON(environ);
    while (*env) {
        char *s = strchr(*env, '=');
        if (s++) {
            long len = strlen(s);
            if (RSTRING_LEN(value) == len && strncmp(s, RSTRING_PTR(value), len) == 0) {
                str = env_str_new(mrb, *env, s-*env-1);
                FREE_ENVIRON(environ);
                return str;
            }
        }
        env++;
    }
    FREE_ENVIRON(environ);
    return mrb_nil_value();
}

static mrb_value
env_index(mrb_value dmy, mrb_value value)
{
    rb_warn("ENV.index is deprecated; use ENV.key");
    return env_key(dmy, value);
}

static mrb_value
env_to_hash(mrb_state *mrb)
{
    char **env;
    mrb_value hash;

    rb_secure(4);
    hash = rb_hash_new(mrb);
    env = GET_ENVIRON(environ);
    while (*env) {
        char *s = strchr(*env, '=');
        if (s) {
            mrb_hash_set/*rb_hash_aset*/(mrb, hash, env_str_new(mrb, *env, s-*env),
                               env_str_new2(mrb, s+1));
        }
        env++;
    }
    FREE_ENVIRON(environ);
    return hash;
}

static mrb_value
env_reject(void)
{
    return rb_hash_delete_if(env_to_hash());
}

static mrb_value
env_shift(mrb_state *mrb)
{
    char **env;

    rb_secure(4);
    env = GET_ENVIRON(environ);
    if (*env) {
        char *s = strchr(*env, '=');
        if (s) {
            mrb_value key = env_str_new(mrb, *env, s-*env);
            mrb_value val = env_str_new2(mrb, getenv(RSTRING_PTR(key)));
            env_delete(mrb, mrb_nil_value(), key);
            return rb_assoc_new(mrb, key, val);
        }
    }
    FREE_ENVIRON(environ);
    return mrb_nil_value();
}

static mrb_value
env_invert(mrb_state *mrb)
{
    return rb_hash_invert(mrb, env_to_hash());
}

static int
env_replace_i(mrb_value key, mrb_value val, mrb_value keys)
{
    if (key != Qundef) {
        env_aset(mrb_nil_value(), key, val);
        if (rb_ary_includes(keys, key)) {
            rb_ary_delete(keys, key);
        }
    }
    return ST_CONTINUE;
}

static mrb_value
env_replace(mrb_state *mrb, mrb_value env, mrb_value hash)
{
    volatile mrb_value keys;
    long i;

    keys = env_keys(mrb);        /* rb_secure(4); */
    if (env == hash) return env;
    hash = to_hash(hash);
    rb_hash_foreach(hash, env_replace_i, &keys);

    for (i=0; i<RARRAY_LEN(keys); i++) {
        env_delete(mrb, env, RARRAY_PTR(keys)[i]);
    }
    return env;
}

static int
env_update_i(mrb_value key, mrb_value val)
{
    if (key != Qundef) {
        if (rb_block_given_p()) {
            val = rb_yield_values(3, key, rb_f_getenv(mrb, mrb_nil_value(), key), val);
        }
        env_aset(mrb_nil_value(), key, val);
    }
    return ST_CONTINUE;
}

static mrb_value
env_update(mrb_value env, mrb_value hash)
{
    rb_secure(4);
    if (env == hash) return env;
    hash = to_hash(hash);
    rb_hash_foreach(hash, env_update_i, 0);
    return env;
}
#endif

void
mrb_init_env(mrb_state *mrb)
{
  origenviron = environ;
  envtbl = rb_obj_alloc(rb_cObject);
  rb_extend_object(envtbl, rb_mEnumerable);

  rb_define_singleton_method(envtbl,"[]", rb_f_getenv, 1);
  rb_define_singleton_method(envtbl,"fetch", env_fetch, -1);
  rb_define_singleton_method(envtbl,"[]=", env_aset, 2);
  rb_define_singleton_method(envtbl,"store", env_aset, 2);
  rb_define_singleton_method(envtbl,"each", env_each_pair, 0);
  rb_define_singleton_method(envtbl,"each_pair", env_each_pair, 0);
  rb_define_singleton_method(envtbl,"each_key", env_each_key, 0);
  rb_define_singleton_method(envtbl,"each_value", env_each_value, 0);
  rb_define_singleton_method(envtbl,"delete", env_delete_m, 1);
  rb_define_singleton_method(envtbl,"delete_if", env_delete_if, 0);
  rb_define_singleton_method(envtbl,"keep_if", env_keep_if, 0);
  rb_define_singleton_method(envtbl,"clear", rb_env_clear, 0);
  rb_define_singleton_method(envtbl,"reject", env_reject, 0);
  rb_define_singleton_method(envtbl,"reject!", env_reject_bang, 0);
  rb_define_singleton_method(envtbl,"select", env_select, 0);
  rb_define_singleton_method(envtbl,"select!", env_select_bang, 0);
  rb_define_singleton_method(envtbl,"shift", env_shift, 0);
  rb_define_singleton_method(envtbl,"invert", env_invert, 0);
  rb_define_singleton_method(envtbl,"replace", env_replace, 1);
  rb_define_singleton_method(envtbl,"update", env_update, 1);
  rb_define_singleton_method(envtbl,"inspect", env_inspect, 0);
  rb_define_singleton_method(envtbl,"rehash", env_none, 0);
  rb_define_singleton_method(envtbl,"to_a", env_to_a, 0);
  rb_define_singleton_method(envtbl,"to_s", env_to_s, 0);
  rb_define_singleton_method(envtbl,"key", env_key, 1);
  rb_define_singleton_method(envtbl,"index", env_index, 1);
  rb_define_singleton_method(envtbl,"size", env_size, 0);
  rb_define_singleton_method(envtbl,"length", env_size, 0);
  rb_define_singleton_method(envtbl,"empty?", env_empty_p, 0);
  rb_define_singleton_method(envtbl,"keys", env_keys, 0);
  rb_define_singleton_method(envtbl,"values", env_values, 0);
  rb_define_singleton_method(envtbl,"values_at", env_values_at, -1);
  rb_define_singleton_method(envtbl,"include?", env_has_key, 1);
  rb_define_singleton_method(envtbl,"member?", env_has_key, 1);
  rb_define_singleton_method(envtbl,"has_key?", env_has_key, 1);
  rb_define_singleton_method(envtbl,"has_value?", env_has_value, 1);
  rb_define_singleton_method(envtbl,"key?", env_has_key, 1);
  rb_define_singleton_method(envtbl,"value?", env_has_value, 1);
  rb_define_singleton_method(envtbl,"to_hash", env_to_hash, 0);
  rb_define_singleton_method(envtbl,"assoc", env_assoc, 1);
  rb_define_singleton_method(envtbl,"rassoc", env_rassoc, 1);
  
  rb_define_global_const("ENV", envtbl);
}
