#include "mruby.h"

#ifdef INCLUDE_REGEXP
#include "class.h"
#include "range.h"
#include "variable.h"
#include "error.h"
#include "numeric.h"
#include "mio.h"
#include "mstring.h"
#include "array.h"
#include "encoding.h"
#include "hash.h"
#include <stdarg.h>

//#ifdef INCLUDE_REGEXP
#include "st.h"
//#endif
#include <string.h>
#include <stdio.h> /* stdin */
#define HAVE_STRING_H

#include <signal.h>      /* #define NSIG 23 */
//#include "mthread.h"
#include "eval_intern.h"

//#ifdef INCLUDE_REGEXP
  #define rb_usascii_str_new2 rb_usascii_str_new_cstr
//#endif

typedef struct rb_io_enc_t convconfig_t;
#define GetWriteIO(mrb, io) rb_io_get_write_io(mrb, io)

#if defined(RUBY_TEST_CRLF_ENVIRONMENT) || defined(_WIN32)
/* Windows */
# define NEED_NEWLINE_DECORATOR_ON_READ(fptr) (!(fptr->mode & FMODE_BINMODE))
# define NEED_NEWLINE_DECORATOR_ON_WRITE(fptr) (!(fptr->mode & FMODE_BINMODE))
# define TEXTMODE_NEWLINE_DECORATOR_ON_WRITE ECONV_CRLF_NEWLINE_DECORATOR
#else
/* Unix */
# define NEED_NEWLINE_DECORATOR_ON_READ(fptr) (fptr->mode & FMODE_TEXTMODE)
# define NEED_NEWLINE_DECORATOR_ON_WRITE(fptr) 0
#endif

#define NEED_READCONV(fptr) (fptr->encs.enc2 != NULL || NEED_NEWLINE_DECORATOR_ON_READ(fptr))
#define NEED_WRITECONV(mrb, fptr) ((fptr->encs.enc != NULL && fptr->encs.enc != rb_ascii8bit_encoding(mrb)) || NEED_NEWLINE_DECORATOR_ON_WRITE(fptr) || (fptr->encs.ecflags & (ECONV_DECORATOR_MASK|ECONV_STATEFUL_DECORATOR_MASK)))

#define READ_CHECK(mrb, fptr) do {\
    if (!READ_DATA_PENDING(fptr)) {\
      WAIT_FD_IN_WIN32(fptr);\
      rb_io_check_closed(mrb, fptr);\
     }\
} while(0)

#define MakeOpenFile(mrb, obj, fp) do {\
    if (RFILE(obj)->fptr) {\
      mrb_io_close_m(mrb, obj);\
      rb_io_fptr_finalize(mrb, RFILE(obj)->fptr);\
      RFILE(obj)->fptr = 0;\
    }\
    fp = 0;\
    fp = RFILE(obj)->fptr = mrb_malloc(mrb, sizeof(struct rb_io_t));\
    fp->fd = -1;\
    fp->stdio_file = NULL;\
    fp->mode = 0;\
    fp->pid = 0;\
    fp->lineno = 0;\
    fp->pathv = mrb_nil_value();\
    fp->finalize = 0;\
    fp->wbuf = NULL;\
    fp->wbuf_off = 0;\
    fp->wbuf_len = 0;\
    fp->wbuf_capa = 0;\
    fp->rbuf = NULL;\
    fp->rbuf_off = 0;\
    fp->rbuf_len = 0;\
    fp->rbuf_capa = 0;\
    fp->readconv = NULL;\
    fp->cbuf = NULL;\
    fp->cbuf_off = 0;\
    fp->cbuf_len = 0;\
    fp->cbuf_capa = 0;\
    fp->writeconv = NULL;\
    fp->writeconv_asciicompat = mrb_nil_value();\
    fp->writeconv_pre_ecflags = 0;\
    fp->writeconv_pre_ecopts = mrb_nil_value();\
    fp->writeconv_initialized = 0;\
    fp->tied_io_for_writing = mrb_fixnum_value(0);\
    fp->encs.enc = NULL;\
    fp->encs.enc2 = NULL;\
    fp->encs.ecflags = 0;\
    fp->encs.ecopts = mrb_nil_value();\
    fp->write_lock = mrb_fixnum_value(0);\
} while (0)

mrb_value rb_stdout;
mrb_value rb_stdin;
mrb_value rb_stderr;

static int gets_lineno;
//static mrb_value lineno = mrb_fixnum_value(0);
static mrb_value lineno;

struct RClass* rb_cProcessStatus;

#if !defined NOFILE
# define NOFILE 64
#endif

#if !HAVE_OFF_T && !defined(off_t)
# define off_t  long
#endif

#define FMODE_PREP (1<<16)
#define IS_PREP_STDIO(f) ((f)->mode & FMODE_PREP)
#define PREP_STDIO_NAME(f) (RSTRING_PTR((f)->pathv))

static int max_file_descriptor = NOFILE;
#define UPDATE_MAXFD(fd) \
    do { \
        if (max_file_descriptor < (fd)) max_file_descriptor = (fd); \
    } while (0)

#define MEMMOVE(p1,p2,type,n) memmove((p1), (p2), sizeof(type)*(n))
static void io_setstrbuf(mrb_state *mrb, mrb_value *str,long len);

struct sysopen_struct {
    mrb_value fname;
    int oflags;
    mode_t perm;
};

#define rb_sys_fail_path(path) rb_sys_fail(mrb_nil_p(path) ? 0 : RSTRING_PTR(path))

#define io_seek(mrb, fptr, ofs, whence) (errno = 0, lseek(flush_before_seek(mrb, fptr)->fd, ofs, whence))
#define io_tell(mrb, fptr) lseek(flush_before_seek(mrb, fptr)->fd, 0, SEEK_CUR)
void rb_io_check_closed(mrb_state *mrb, rb_io_t *fptr);

#define READ_DATA_PENDING(fptr) ((fptr)->rbuf_len)
#define READ_DATA_PENDING_COUNT(fptr) ((fptr)->rbuf_len)
#define READ_DATA_PENDING_PTR(fptr) ((fptr)->rbuf+(fptr)->rbuf_off)
#define READ_DATA_BUFFERED(fptr) READ_DATA_PENDING(fptr)

#define READ_CHAR_PENDING(fptr) ((fptr)->cbuf_len)
#define READ_CHAR_PENDING_COUNT(fptr) ((fptr)->cbuf_len)
#define READ_CHAR_PENDING_PTR(fptr) ((fptr)->cbuf+(fptr)->cbuf_off)

static mrb_value argf;
struct argf {
    mrb_value filename, current_file;
    int last_lineno;        /* $. */
    int lineno;
    int init_p, next_p;
    mrb_value argv;
    char *inplace;
    int binmode;
    struct rb_io_enc_t encs;
};
#define argf_of(obj) (*(struct argf *)DATA_PTR(obj))
#define ARGF argf_of(argf)

struct io_internal_struct {
    int fd;
    void *buf;
    size_t capa;
};

#define RUBY_UBF_IO ((rb_unblock_function_t *)-1)

#ifndef PIPE_BUF
# ifdef _POSIX_PIPE_BUF
#  define PIPE_BUF _POSIX_PIPE_BUF
# else
#  define PIPE_BUF 512 /* is this ok? */
# endif
#endif

#define IO_RBUF_CAPA_MIN  8192
#define IO_WBUF_CAPA_MIN  8192
#define IO_CBUF_CAPA_MIN  (128*1024)
#define IO_RBUF_CAPA_FOR(fptr) (NEED_READCONV(fptr) ? IO_CBUF_CAPA_MIN : IO_RBUF_CAPA_MIN)
#define MORE_CHAR_SUSPENDED mrb_true_value()
#define MORE_CHAR_FINISHED mrb_nil_value()

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

/*
  Solaris sys/select.h switches select to select_large_fdset to support larger
  file descriptors if FD_SETSIZE is larger than 1024 on 32bit environment.
  But Ruby doesn't change FD_SETSIZE because fd_set is allocated dynamically.
  So following definition is required to use select_large_fdset.
*/
#ifdef HAVE_SELECT_LARGE_FDSET
#define select(n, r, w, e, t) select_large_fdset(n, r, w, e, t)
#endif

#ifndef FALSE
#define FALSE   0
#endif

#ifndef TRUE
#define TRUE    1
#endif

/* ------------------------------------------------------------- */
void rb_io_check_char_readable(mrb_state *mrb, rb_io_t *fptr);

static rb_encoding* io_read_encoding(mrb_state *mrb, rb_io_t *fptr);
#ifdef INCLUDE_REGEXP
static mrb_value sym_extenc, sym_intenc, sym_encoding;
static mrb_value sym_textmode, sym_binmode;
static mrb_value sym_mode, sym_perm;
static void parse_mode_enc(mrb_state *mrb, const char *estr, rb_encoding **enc_p, rb_encoding **enc2_p, int *fmode_p);
static mrb_value sym_autoclose;
static mrb_value pop_last_hash(mrb_state *mrb, int *argc_p, mrb_value *argv);
static void rb_io_extract_modeenc(mrb_state *mrb, mrb_value *vmode_p, mrb_value *vperm_p, mrb_value opthash,
                                                  int *oflags_p, int *fmode_p, convconfig_t *convconfig_p);
static void validate_enc_binmode(mrb_state *mrb, int fmode, rb_encoding *enc, rb_encoding *enc2);
static void rb_io_ext_int_to_encs(mrb_state *mrb, rb_encoding *ext, rb_encoding *intern, rb_encoding **enc, rb_encoding **enc2);
#endif
static mrb_value fill_cbuf(mrb_state *mrb, rb_io_t *fptr, int ec_flags);
static int io_fflush(mrb_state *mrb, rb_io_t *fptr);
static void io_unread(mrb_state *mrb, rb_io_t *fptr);
static void clear_codeconv(rb_io_t *fptr);
static mrb_value io_puts_ary(mrb_state *mrb, mrb_value ary, mrb_value out, int recur);
static void io_ungetbyte(mrb_value str, rb_io_t *fptr);

//void native_cond_signal(pthread_cond_t *cond);
//void native_mutex_destroy(pthread_mutex_t *lock);
/* ------------------------------------------------------------- */


#if 0
static void
argf_mark(void *ptr)
{
  struct argf *p = ptr;
  rb_gc_mark(p->filename);
  rb_gc_mark(p->current_file);
  rb_gc_mark(p->argv);
  rb_gc_mark(p->encs.ecopts);
}

static void
argf_free(void *ptr)
{
  struct argf *p = ptr;
  xfree(p->inplace);
  xfree(p);
}
#endif

static inline void
argf_init(struct argf *p, mrb_value v)
{
  p->filename = mrb_nil_value();
  p->current_file = mrb_nil_value();
  p->lineno = 0;
  p->argv = v;
}

static mrb_value
argf_alloc(mrb_state *mrb, mrb_value klass)
{
  struct argf *p;
  mrb_value argf = Data_Make_Struct(mrb, klass, struct argf, argf_mark, argf_free, p);

  argf_init(p, mrb_nil_value());
  return argf;
}

#undef rb_argv

/* :nodoc: */
#if 0
static mrb_value
argf_initialize(mrb_value argf, mrb_value argv)
{
  memset(&ARGF, 0, sizeof(ARGF));
  argf_init(&ARGF, argv);

  return argf;
}
#endif

#ifdef INCLUDE_REGEXP
static int
io_encname_bom_p(const char *name, long len)
{
  static const char bom_prefix[] = "bom|utf-";
  enum {bom_prefix_len = (int)sizeof(bom_prefix) - 1};
  if (!len) {
    const char *p = strchr(name, ':');
    len = p ? (long)(p - name) : (long)strlen(name);
  }
  return len > bom_prefix_len && STRNCASECMP(name, bom_prefix, bom_prefix_len) == 0;
}

static void
io_encoding_set(mrb_state *mrb, rb_io_t *fptr, mrb_value v1, mrb_value v2, mrb_value opt)
{
  rb_encoding *enc, *enc2;
  int ecflags;
  mrb_value ecopts, tmp;

  if (!mrb_nil_p(v2)) {
    enc2 = rb_to_encoding(mrb, v1);
    tmp = mrb_check_string_type(mrb, v2);
    if (!mrb_nil_p(tmp)) {
        if (RSTRING_LEN(tmp) == 1 && RSTRING_PTR(tmp)[0] == '-') {
          /* Special case - "-" => no transcoding */
          enc = enc2;
          enc2 = NULL;
        }
        else
          enc = rb_to_encoding(mrb, v2);
        if (enc == enc2) {
          /* Special case - "-" => no transcoding */
          enc2 = NULL;
        }
    }
    else
      enc = rb_to_encoding(mrb, v2);
      ecflags = rb_econv_prepare_opts(mrb, opt, &ecopts);
  }
  else {
    if (mrb_nil_p(v1)) {
      /* Set to default encodings */
      rb_io_ext_int_to_encs(mrb, NULL, NULL, &enc, &enc2);
      ecflags = 0;
      ecopts = mrb_nil_value();
    }
    else {
      tmp = mrb_check_string_type(mrb, v1);
      if (!mrb_nil_p(tmp) && rb_enc_asciicompat(mrb, rb_enc_get(mrb, tmp))) {
        parse_mode_enc(mrb, RSTRING_PTR(tmp), &enc, &enc2, NULL);
        ecflags = rb_econv_prepare_opts(mrb, opt, &ecopts);
      }
      else {
        rb_io_ext_int_to_encs(mrb, rb_to_encoding(mrb, v1), NULL, &enc, &enc2);
        ecflags = 0;
        ecopts = mrb_nil_value();
      }
    }
  }
  validate_enc_binmode(mrb, fptr->mode, enc, enc2);
  fptr->encs.enc = enc;
  fptr->encs.enc2 = enc2;
  fptr->encs.ecflags = ecflags;
  fptr->encs.ecopts = ecopts;
  clear_codeconv(fptr);

}

static int
io_strip_bom(mrb_state *mrb, mrb_value io)
{
  int b1, b2, b3, b4;
  switch (b1 = mrb_fixnum_number(rb_io_getbyte(mrb, io))) {
    case 0xEF:
    b2 = mrb_fixnum_number(rb_io_getbyte(mrb, io));
    if (b2 == 0xBB) {
        b3 = mrb_fixnum_number(rb_io_getbyte(mrb, io));
        if (b3 == 0xBF) {
          return rb_utf8_encindex();
        }
        rb_io_ungetbyte(mrb, io, mrb_fixnum_value(b3));
    }
    rb_io_ungetbyte(mrb, io, mrb_fixnum_value(b2));
    break;

    case 0xFE:
    b2 = mrb_fixnum_number(rb_io_getbyte(mrb, io));
    if (b2 == 0xFF) {
        return rb_enc_find_index(mrb, "UTF-16BE");
    }
    rb_io_ungetbyte(mrb, io, mrb_fixnum_value(b2));
    break;

    case 0xFF:
    b2 = mrb_fixnum_number(rb_io_getbyte(mrb, io));
    if (b2 == 0xFE) {
        b3 = mrb_fixnum_number(rb_io_getbyte(mrb, io));
        if (b3 == 0) {
          b4 = mrb_fixnum_number(rb_io_getbyte(mrb, io));
          if (b4 == 0) {
              return rb_enc_find_index(mrb, "UTF-32LE");
          }
          rb_io_ungetbyte(mrb, io, mrb_fixnum_value(b4));
        }
        else {
          rb_io_ungetbyte(mrb, io, mrb_fixnum_value(b3));
          return rb_enc_find_index(mrb, "UTF-16LE");
        }
        rb_io_ungetbyte(mrb, io, mrb_fixnum_value(b3));
    }
    rb_io_ungetbyte(mrb, io, mrb_fixnum_value(b2));
    break;

    case 0:
    b2 = mrb_fixnum_number(rb_io_getbyte(mrb, io));
    if (b2 == 0) {
        b3 = mrb_fixnum_number(rb_io_getbyte(mrb, io));
        if (b3 == 0xFE) {
          b4 = mrb_fixnum_number(rb_io_getbyte(mrb, io));
          if (b4 == 0xFF) {
              return rb_enc_find_index(mrb, "UTF-32BE");
          }
          rb_io_ungetbyte(mrb, io, mrb_fixnum_value(b4));
        }
        rb_io_ungetbyte(mrb, io, mrb_fixnum_value(b3));
    }
    rb_io_ungetbyte(mrb, io, mrb_fixnum_value(b2));
    break;
  }
  rb_io_ungetbyte(mrb, io, mrb_fixnum_value(b1));
  return 0;
}


/*
 *  call-seq:
 *     io.internal_encoding   -> encoding
 *
 *  Returns the Encoding of the internal string if conversion is
 *  specified.  Otherwise returns nil.
 */

static mrb_value
rb_io_internal_encoding(mrb_state *mrb, mrb_value io)
{
  rb_io_t *fptr;

  GetOpenFile(mrb, io, fptr);
  if (!fptr->encs.enc2) return mrb_nil_value();
  return rb_enc_from_encoding(mrb, io_read_encoding(mrb, fptr));
}
#endif

static int
internal_read_func(void *ptr)
{
  struct io_internal_struct *iis = (struct io_internal_struct*)ptr;
  return read(iis->fd, iis->buf, iis->capa);
}

static ssize_t
rb_read_internal(mrb_state *mrb, int fd, void *buf, size_t count)
{
  struct io_internal_struct iis;
  iis.fd = fd;
  iis.buf = buf;
  iis.capa = count;
#if 0 /* THREAD */
  return (ssize_t)rb_thread_blocking_region(mrb, internal_read_func, &iis, RUBY_UBF_IO, 0);
#else /* NO THREAD */
  return (ssize_t)internal_read_func(&iis);
#endif
}

mrb_value
rb_str_locktmp(mrb_value str)
{
  /*if (FL_TEST(str, STR_TMPLOCK)) {
    rb_raise(rb_eRuntimeError, "temporal locking already locked string");
  }
  FL_SET(str, STR_TMPLOCK);
  */
  return str;
}

mrb_value
rb_str_unlocktmp(mrb_value str)
{
  /*if (!FL_TEST(str, STR_TMPLOCK)) {
    rb_raise(rb_eRuntimeError, "temporal unlocking already unlocked string");
  }
  FL_UNSET(str, STR_TMPLOCK);
  */
  return str;
}

static int
sysopen_func(void *ptr)
{
  const struct sysopen_struct *data = ptr;
  const char *fname = RSTRING_PTR(data->fname);
  return open(fname, data->oflags, data->perm);
}

static inline int
rb_sysopen_internal(mrb_state *mrb, struct sysopen_struct *data)
{
#if 0 /* kusuda */
  return (int)rb_thread_blocking_region(mrb, sysopen_func, data, RUBY_UBF_IO, 0);
#else
  return sysopen_func(data);
#endif
}

static int
internal_write_func(void *ptr)
{
  struct io_internal_struct *iis = (struct io_internal_struct*)ptr;
  return write(iis->fd, iis->buf, iis->capa);
}

void
rb_io_check_writable(mrb_state *mrb, rb_io_t *fptr)
{
  rb_io_check_closed(mrb, fptr);
  if (!(fptr->mode & FMODE_WRITABLE)) {
    mrb_raise(mrb, mrb->eIOError_class, "not opened for writing");
  }
  if (fptr->rbuf_len) {
      io_unread(mrb, fptr);
  }
}

static ssize_t
rb_write_internal(mrb_state *mrb, int fd, void *buf, size_t count)
{
  struct io_internal_struct iis;
  iis.fd = fd;
  iis.buf = buf;
  iis.capa = count;
#if 0 /* THREAD */
  return (ssize_t)rb_thread_blocking_region(mrb, internal_write_func, &iis, RUBY_UBF_IO, 0);
#else
  return (ssize_t)internal_write_func(&iis);
#endif
}

static long
remain_size(mrb_state *mrb, rb_io_t *fptr)
{
  struct stat st;
  off_t siz = READ_DATA_PENDING_COUNT(fptr);
  off_t pos;

  if (fstat(fptr->fd, &st) == 0  && S_ISREG(st.st_mode)
#if defined(__BEOS__) || defined(__HAIKU__)
      && (st.st_dev > 3)
#endif
      )
  {
      if (io_fflush(mrb, fptr) < 0)
          rb_sys_fail(0);
    pos = lseek(fptr->fd, 0, SEEK_CUR);
    if (st.st_size >= pos && pos >= 0) {
        siz += st.st_size - pos;
        if (siz > LONG_MAX) {
          mrb_raise(mrb, mrb->eIOError_class, "file too big for single read");
        }
    }
  }
  else {
    siz += BUFSIZ;
  }
  return (long)siz;
}

static mrb_value
io_enc_str(mrb_state *mrb, mrb_value str, rb_io_t *fptr)
{
  //OBJ_TAINT(str);
  rb_enc_associate(mrb, str, io_read_encoding(mrb, fptr));
  return str;
}

#ifdef INCLUDE_REGEXP
static void
validate_enc_binmode(mrb_state *mrb, int fmode, rb_encoding *enc, rb_encoding *enc2)
{
  if ((fmode & FMODE_READABLE) &&
      !enc2 &&
      !(fmode & FMODE_BINMODE) &&
      !rb_enc_asciicompat(mrb, enc ? enc : rb_default_external_encoding(mrb)))
    mrb_raise(mrb, mrb->eArgumentError_class, "ASCII incompatible encoding needs binmode");
}
#endif

static int
io_fillbuf(mrb_state *mrb, rb_io_t *fptr)
{
    ssize_t r;

    if (fptr->rbuf == NULL) {
      fptr->rbuf_off = 0;
      fptr->rbuf_len = 0;
      fptr->rbuf_capa = IO_RBUF_CAPA_FOR(fptr);
      fptr->rbuf = mrb_malloc(mrb, fptr->rbuf_capa);
#ifdef _WIN32
      fptr->rbuf_capa--;
#endif
    }
    if (fptr->rbuf_len == 0) {
      //retry:
      {
          r = rb_read_internal(mrb, fptr->fd, fptr->rbuf, fptr->rbuf_capa);
      }
        if (r < 0) {
            //if (rb_io_wait_readable(mrb, fptr->fd))
            //    goto retry;
            rb_sys_fail_path(fptr->pathv);
        }
        fptr->rbuf_off = 0;
        fptr->rbuf_len = (int)r; /* r should be <= rbuf_capa */
        if (r == 0)
            return -1; /* EOF */
    }
    return 0;
}

static int
wsplit_p(rb_io_t *fptr)
{
#if defined(HAVE_FCNTL) && defined(F_GETFL) && defined(O_NONBLOCK)
    int r;
#endif

    if (!(fptr->mode & FMODE_WSPLIT_INITIALIZED)) {
        struct stat buf;
        if (fstat(fptr->fd, &buf) == 0 &&
            !S_ISREG(buf.st_mode)
#if defined(HAVE_FCNTL) && defined(F_GETFL) && defined(O_NONBLOCK)
            && (r = fcntl(fptr->fd, F_GETFL)) != -1 &&
            !(r & O_NONBLOCK)
#endif
            ) {
            fptr->mode |= FMODE_WSPLIT;
        }
        fptr->mode |= FMODE_WSPLIT_INITIALIZED;
    }
    return fptr->mode & FMODE_WSPLIT;
}

static long
io_writable_length(rb_io_t *fptr, long l)
{
    if (PIPE_BUF < l &&
        //!rb_thread_alone() &&
        wsplit_p(fptr)) {
        l = PIPE_BUF;
    }
    return l;
}

static int
io_flush_buffer_sync(void *arg)
{
    rb_io_t *fptr = arg;
    long l = io_writable_length(fptr, fptr->wbuf_len);
    ssize_t r = write(fptr->fd, fptr->wbuf+fptr->wbuf_off, (size_t)l);

    if (fptr->wbuf_len <= r) {
      fptr->wbuf_off = 0;
      fptr->wbuf_len = 0;
      return (0);
    }
    if (0 <= r) {
      fptr->wbuf_off += (int)r;
      fptr->wbuf_len -= (int)r;
      errno = EAGAIN;
    }
    return (-1);
}

static int
io_flush_buffer_async(mrb_state *mrb, void */*rb_io_t **/arg)
{
#if 0 /* THREAD */
    return rb_thread_blocking_region(mrb, io_flush_buffer_sync, (void *)arg, RUBY_UBF_IO, 0);
#else
    return io_flush_buffer_sync(arg);
#endif
}

static void
make_readconv(mrb_state *mrb, rb_io_t *fptr, int size)
{
    if (!fptr->readconv) {
        int ecflags;
        mrb_value ecopts;
        const char *sname, *dname;
        ecflags = fptr->encs.ecflags;
        ecopts = fptr->encs.ecopts;
        if (NEED_NEWLINE_DECORATOR_ON_READ(fptr))
            ecflags |= ECONV_UNIVERSAL_NEWLINE_DECORATOR;
        if (fptr->encs.enc2) {
            sname = rb_enc_name(fptr->encs.enc2);
            dname = rb_enc_name(fptr->encs.enc);
        }
        else {
            sname = dname = "";
        }
        fptr->readconv = rb_econv_open_opts(mrb, sname, dname, ecflags, ecopts);
        if (!fptr->readconv)
            rb_exc_raise(mrb, rb_econv_open_exc(mrb, sname, dname, ecflags));
        fptr->cbuf_off = 0;
        fptr->cbuf_len = 0;
      if (size < IO_CBUF_CAPA_MIN) size = IO_CBUF_CAPA_MIN;
        fptr->cbuf_capa = size;
        fptr->cbuf = mrb_malloc(mrb, fptr->cbuf_capa);
    }
}

static void
make_writeconv(mrb_state *mrb, rb_io_t *fptr)
{
    if (!fptr->writeconv_initialized) {
        const char *senc, *denc;
        rb_encoding *enc;
        int ecflags;
        mrb_value ecopts;

        fptr->writeconv_initialized = 1;

        ecflags = fptr->encs.ecflags;
        ecopts = fptr->encs.ecopts;
#ifdef TEXTMODE_NEWLINE_DECORATOR_ON_WRITE
        if (NEED_NEWLINE_DECORATOR_ON_WRITE(fptr))
            ecflags |= TEXTMODE_NEWLINE_DECORATOR_ON_WRITE;
#endif

        if (!fptr->encs.enc || (fptr->encs.enc == rb_ascii8bit_encoding(mrb) && !fptr->encs.enc2)) {
            /* no encoding conversion */
            fptr->writeconv_pre_ecflags = 0;
            fptr->writeconv_pre_ecopts = mrb_nil_value();
            fptr->writeconv = rb_econv_open_opts(mrb, "", "", ecflags, ecopts);
            if (!fptr->writeconv)
                rb_exc_raise(mrb, rb_econv_open_exc(mrb, "", "", ecflags));
            fptr->writeconv_asciicompat = mrb_nil_value();
        }
        else {
            enc = fptr->encs.enc2 ? fptr->encs.enc2 : fptr->encs.enc;
            senc = rb_econv_asciicompat_encoding(rb_enc_name(enc));
            if (!senc && !(fptr->encs.ecflags & ECONV_STATEFUL_DECORATOR_MASK)) {
                /* single conversion */
                fptr->writeconv_pre_ecflags = ecflags;
                fptr->writeconv_pre_ecopts = ecopts;
                fptr->writeconv = NULL;
                fptr->writeconv_asciicompat = mrb_nil_value();
            }
            else {
                /* double conversion */
                fptr->writeconv_pre_ecflags = ecflags & ~ECONV_STATEFUL_DECORATOR_MASK;
                fptr->writeconv_pre_ecopts = ecopts;
                if (senc) {
                    denc = rb_enc_name(enc);
                    fptr->writeconv_asciicompat = mrb_str_new2(mrb, senc);
                }
                else {
                    senc = denc = "";
                    fptr->writeconv_asciicompat = mrb_str_new2(mrb, rb_enc_name(enc));
                }
                ecflags = fptr->encs.ecflags & (ECONV_ERROR_HANDLER_MASK|ECONV_STATEFUL_DECORATOR_MASK);
                ecopts = fptr->encs.ecopts;
                fptr->writeconv = rb_econv_open_opts(mrb, senc, denc, ecflags, ecopts);
                if (!fptr->writeconv)
                    rb_exc_raise(mrb, rb_econv_open_exc(mrb, senc, denc, ecflags));
            }
        }
    }
}

#ifdef INCLUDE_REGEXP
/*
 * Convert external/internal encodings to enc/enc2
 * NULL => use default encoding
 * mrb_nil_value() => no encoding specified (internal only)
 */
static void
rb_io_ext_int_to_encs(mrb_state *mrb, rb_encoding *ext, rb_encoding *intern, rb_encoding **enc, rb_encoding **enc2)
{
    int default_ext = 0;

    if (ext == NULL) {
      ext = rb_default_external_encoding(mrb);
      default_ext = 1;
    }
    if (intern == NULL && ext != rb_ascii8bit_encoding(mrb))
      /* If external is ASCII-8BIT, no default transcoding */
      intern = rb_default_internal_encoding(mrb);
    if (intern == NULL || intern == NULL || intern == ext) {
      /* No internal encoding => use external + no transcoding */
      *enc = (default_ext && intern != ext) ? NULL : ext;
      *enc2 = NULL;
    }
    else {
      *enc = intern;
      *enc2 = ext;
    }
}

static int
rb_io_fmode_oflags(int fmode)
{
    int oflags = 0;

    switch (fmode & FMODE_READWRITE) {
      case FMODE_READABLE:
        oflags |= O_RDONLY;
        break;
      case FMODE_WRITABLE:
        oflags |= O_WRONLY;
        break;
      case FMODE_READWRITE:
        oflags |= O_RDWR;
        break;
    }

    if (fmode & FMODE_APPEND) {
        oflags |= O_APPEND;
    }
    if (fmode & FMODE_TRUNC) {
        oflags |= O_TRUNC;
    }
    if (fmode & FMODE_CREATE) {
        oflags |= O_CREAT;
    }
#ifdef O_BINARY
    if (fmode & FMODE_BINMODE) {
        oflags |= O_BINARY;
    }
#endif

    return oflags;
}

int
rb_io_extract_encoding_option(mrb_state *mrb, mrb_value opt, rb_encoding **enc_p, rb_encoding **enc2_p, int *fmode_p)
{
    mrb_value encoding=mrb_nil_value(), extenc, intenc, tmp, Qundef;
    int extracted = 0;
    rb_encoding *extencoding = NULL;
    rb_encoding *intencoding = NULL;
    extenc.tt = 0; /* Qundef */
    intenc.tt = 0; /* Qundef */
    Qundef.tt = 0; /* Qundef */

    if (!mrb_nil_p(opt)) {
      mrb_value v;
      v = mrb_hash_getWithDef(mrb, opt, sym_encoding, mrb_nil_value());
      if (!mrb_obj_equal(mrb, v, mrb_nil_value())) encoding = v;
      v = mrb_hash_getWithDef(mrb, opt, sym_extenc, Qundef);
      if (!mrb_obj_equal(mrb, v, mrb_nil_value())) extenc = v;
      v = mrb_hash_getWithDef(mrb, opt, sym_intenc, Qundef);
      if (v.tt != 0) intenc = v;
    }
    if ((extenc.tt != 0 || intenc.tt != 0) && !mrb_nil_p(encoding)) {
      //if (!mrb_nil_p(ruby_verbose)) {
      //    int idx = rb_to_encoding_index(encoding);
      //    rb_warn("Ignoring encoding parameter '%s': %s_encoding is used",
      //          idx < 0 ? StringValueCStr(encoding) : rb_enc_name(rb_enc_from_index(mrb, idx)),
      //          extenc == Qundef ? "internal" : "external");
      //}
      encoding = mrb_nil_value();
    }
    if (extenc.tt != 0 && !mrb_nil_p(extenc)) {
      extencoding = rb_to_encoding(mrb, extenc);
    }
    if (intenc.tt != 0) {
      if (mrb_nil_p(intenc)) {
          /* internal_encoding: nil => no transcoding */
          intencoding = (rb_encoding *)NULL;
      }
      else if (!mrb_nil_p(tmp = mrb_check_string_type(mrb, intenc))) {
          //char *p = StringValueCStr(tmp);
          char *p = rb_string_value_cstr(mrb, &tmp);
          if (*p == '-' && *(p+1) == '\0') {
            /* Special case - "-" => no transcoding */
            intencoding = (rb_encoding *)NULL;
          }
          else {
            intencoding = rb_to_encoding(mrb, intenc);
          }
      }
      else {
          intencoding = rb_to_encoding(mrb, intenc);
      }
      if (extencoding == intencoding) {
          intencoding = (rb_encoding *)NULL;
      }
    }
    if (!mrb_nil_p(encoding)) {
      extracted = 1;
      if (!mrb_nil_p(tmp = mrb_check_string_type(mrb, encoding))) {
          //parse_mode_enc(mrb, StringValueCStr(tmp), enc_p, enc2_p, fmode_p);
          parse_mode_enc(mrb, rb_string_value_cstr(mrb, &tmp), enc_p, enc2_p, fmode_p);
      }
      else {
          rb_io_ext_int_to_encs(mrb, rb_to_encoding(mrb, encoding), NULL, enc_p, enc2_p);
      }
    }
    else if (extenc.tt != 0 || intenc.tt != 0) {
        extracted = 1;
      rb_io_ext_int_to_encs(mrb, extencoding, intencoding, enc_p, enc2_p);
    }
    return extracted;
}
#endif //#ifdef INCLUDE_REGEXP

void
rb_io_check_byte_readable(mrb_state *mrb, rb_io_t *fptr)
{
    rb_io_check_char_readable(mrb, fptr);
    if (READ_CHAR_PENDING(fptr)) {
      mrb_raise(mrb, mrb->eIOError_class, "byte oriented read for character buffered IO");
    }
}

void
rb_io_check_readable(mrb_state *mrb, rb_io_t *fptr)
{
    rb_io_check_byte_readable(mrb, fptr);
}

#ifdef INCLUDE_REGEXP
static void
parse_mode_enc(mrb_state *mrb, const char *estr, rb_encoding **enc_p, rb_encoding **enc2_p, int *fmode_p)
{
    const char *p;
    char encname[ENCODING_MAXNAMELEN+1];
    int idx, idx2;
    rb_encoding *ext_enc, *int_enc;

    /* parse estr as "enc" or "enc2:enc" or "enc:-" */

    p = strrchr(estr, ':');
    if (p) {
      long len = (p++) - estr;
      if (len == 0 || len > ENCODING_MAXNAMELEN)
          idx = -1;
      else {
          if (io_encname_bom_p(estr, len)) {
            if (fmode_p) *fmode_p |= FMODE_SETENC_BY_BOM;
            estr += 4;
            len -= 4;
          }
          memcpy(encname, estr, len);
          encname[len] = '\0';
          estr = encname;
          idx = rb_enc_find_index(mrb, encname);
      }
    }
    else {
      long len = strlen(estr);
      if (io_encname_bom_p(estr, len)) {
          if (fmode_p) *fmode_p |= FMODE_SETENC_BY_BOM;
          estr += 4;
          len -= 4;
          memcpy(encname, estr, len);
          encname[len] = '\0';
          estr = encname;
      }
      idx = rb_enc_find_index(mrb, estr);
    }

    if (idx >= 0)
      ext_enc = rb_enc_from_index(mrb, idx);
    else {
      if (idx != -2)
          rb_warn("Unsupported encoding %s ignored", estr);
      ext_enc = NULL;
    }

    int_enc = NULL;
    if (p) {
      if (*p == '-' && *(p+1) == '\0') {
          /* Special case - "-" => no transcoding */
          int_enc = (rb_encoding *)NULL;
      }
      else {
          idx2 = rb_enc_find_index(mrb, p);
          if (idx2 < 0)
            rb_warn("Unsupported encoding %s ignored", p);
          else if (idx2 == idx) {
            rb_warn("Ignoring internal encoding %s: it is identical to external encoding %s", p, estr);
            int_enc = (rb_encoding *)NULL;
          }
          else
            int_enc = rb_enc_from_index(mrb, idx2);
      }
    }

    rb_io_ext_int_to_encs(mrb, ext_enc, int_enc, enc_p, enc2_p);
}
#endif

/* reading functions */
static long
read_buffered_data(char *ptr, long len, rb_io_t *fptr)
{
    int n;

    n = READ_DATA_PENDING_COUNT(fptr);
    if (n <= 0) return 0;
    if (n > len) n = (int)len;
    MEMMOVE(ptr, fptr->rbuf+fptr->rbuf_off, char, n);
    fptr->rbuf_off += n;
    fptr->rbuf_len -= n;
    return n;
}

static long
io_fread(mrb_state *mrb, mrb_value str, long offset, rb_io_t *fptr)
{
    long len = RSTRING_LEN(str) - offset;
    long n = len;
    long c;

    rb_str_locktmp(str);
    if (READ_DATA_PENDING(fptr) == 0) {
      while (n > 0) {
          //again:
          c = rb_read_internal(mrb, fptr->fd, RSTRING_PTR(str)+offset, n);
          if (c == 0) break;
          if (c < 0) {
            //if (rb_io_wait_readable(mrb, fptr->fd))
            //  goto again;
            rb_sys_fail_path(fptr->pathv);
          }
          offset += c;
          if ((n -= c) <= 0) break;
          //rb_thread_wait_fd(mrb, fptr->fd);
      }
      rb_str_unlocktmp(str);
      return len - n;
    }

    while (n > 0) {
      c = read_buffered_data(RSTRING_PTR(str)+offset, n, fptr);
      if (c > 0) {
          offset += c;
          if ((n -= c) <= 0) break;
      }
      //rb_thread_wait_fd(mrb, fptr->fd);
      rb_io_check_closed(mrb, fptr);
      if (io_fillbuf(mrb, fptr) < 0) {
          break;
      }
    }
    rb_str_unlocktmp(str);
    return len - n;
}

/* writing functions */
struct binwrite_arg {
    rb_io_t *fptr;
    mrb_value str;
    long offset;
    long length;
};

#if 0
static int
io_binwrite_string(mrb_state *mrb, void *arg)
{
    struct binwrite_arg *p = (struct binwrite_arg *)arg;
    long l = io_writable_length(p->fptr, p->length);
    return rb_write_internal(mrb, p->fptr->fd, RSTRING_PTR(p->str)+p->offset, l);
}
#endif

static void
io_set_encoding_by_bom(mrb_state *mrb, mrb_value io)
{
    int idx = io_strip_bom(mrb, io);

    if (idx) {
      rb_io_t *fptr;
      GetOpenFile(mrb, io, fptr);
      io_encoding_set(mrb, fptr, rb_enc_from_encoding(mrb, rb_enc_from_index(mrb, idx)),
            rb_io_internal_encoding(mrb, io), mrb_nil_value());
    }
}

static mrb_value
io_shift_cbuf(mrb_state *mrb, rb_io_t *fptr, int len, mrb_value *strp)
{
    mrb_value str = mrb_nil_value();
    if (strp) {
      str = *strp;
      if (mrb_nil_p(str)) {
          *strp = str = mrb_str_new(mrb, fptr->cbuf+fptr->cbuf_off, len);
      }
      else {
        rb_str_cat(mrb, str, fptr->cbuf+fptr->cbuf_off, len);
      }
      //OBJ_TAINT(str);
      rb_enc_associate(mrb, str, fptr->encs.enc);
    }
    fptr->cbuf_off += len;
    fptr->cbuf_len -= len;
    /* xxx: set coderange */
    if (fptr->cbuf_len == 0)
        fptr->cbuf_off = 0;
    else if (fptr->cbuf_capa/2 < fptr->cbuf_off) {
        memmove(fptr->cbuf, fptr->cbuf+fptr->cbuf_off, fptr->cbuf_len);
        fptr->cbuf_off = 0;
    }
    return str;
}

#ifdef INCLUDE_REGEXP
static void
extract_binmode(mrb_state *mrb, mrb_value opthash, int *fmode)
{
    if (!mrb_nil_p(opthash)) {
      mrb_value v;
      v = mrb_hash_get(mrb, opthash, sym_textmode);
      if (!mrb_nil_p(v) && RTEST(v))
        *fmode |= FMODE_TEXTMODE;
      v = mrb_hash_get(mrb, opthash, sym_binmode);
      if (!mrb_nil_p(v) && RTEST(v))
        *fmode |= FMODE_BINMODE;

      if ((*fmode & FMODE_BINMODE) && (*fmode & FMODE_TEXTMODE))
        mrb_raise(mrb, mrb->eArgumentError_class, "both textmode and binmode specified");
    }
}
#endif

static void
io_check_tty(rb_io_t *fptr)
{
    if (isatty(fptr->fd))
        fptr->mode |= FMODE_TTY|FMODE_DUPLEX;
}

static void
clear_readconv(rb_io_t *fptr)
{
    if (fptr->readconv) {
        rb_econv_close(fptr->readconv);
        fptr->readconv = NULL;
    }
    if (fptr->cbuf) {
        free(fptr->cbuf);
        fptr->cbuf = NULL;
    }
}

static void
clear_writeconv(rb_io_t *fptr)
{
    if (fptr->writeconv) {
        rb_econv_close(fptr->writeconv);
        fptr->writeconv = NULL;
    }
    fptr->writeconv_initialized = 0;
}

static void
clear_codeconv(rb_io_t *fptr)
{
    clear_readconv(fptr);
    clear_writeconv(fptr);
}
static void
rscheck(mrb_state *mrb, const char *rsptr, long rslen, mrb_value rs)
{
    if (!rs.tt) return;
    if (RSTRING_PTR(rs) != rsptr && RSTRING_LEN(rs) != rslen)
      mrb_raise(mrb, mrb->eRuntimeError_class, "rs modified");
}

static mrb_value
more_char(mrb_state *mrb, rb_io_t *fptr)
{
    mrb_value v;
    v = fill_cbuf(mrb, fptr, ECONV_AFTER_OUTPUT);
    //if (v != MORE_CHAR_SUSPENDED && v != MORE_CHAR_FINISHED)
    if (!mrb_obj_equal(mrb, v, MORE_CHAR_SUSPENDED) && !mrb_obj_equal(mrb, v, MORE_CHAR_FINISHED))
        rb_exc_raise(mrb, v);
    return v;
}

static int
appendline(mrb_state *mrb, rb_io_t *fptr, int delim, mrb_value *strp, long *lp)
{
    mrb_value str = *strp;
    long limit = *lp;

    if (NEED_READCONV(fptr)) {
        make_readconv(mrb, fptr, 0);
        do {
            const char *p, *e;
            int searchlen;
            if (fptr->cbuf_len) {
                p = fptr->cbuf+fptr->cbuf_off;
                searchlen = fptr->cbuf_len;
                if (0 < limit && limit < searchlen)
                    searchlen = (int)limit;
                e = memchr(p, delim, searchlen);
                if (e) {
                int len = (int)(e-p+1);
                    if (mrb_nil_p(str))
                        *strp = str = mrb_str_new(mrb, p, len);
                    else
                        mrb_str_buf_cat(mrb, str, p, len);
                    fptr->cbuf_off += len;
                    fptr->cbuf_len -= len;
                    limit -= len;
                    *lp = limit;
                    return delim;
                }

                if (mrb_nil_p(str))
                    *strp = str = mrb_str_new(mrb, p, searchlen);
                else
                    mrb_str_buf_cat(mrb, str, p, searchlen);
                fptr->cbuf_off += searchlen;
                fptr->cbuf_len -= searchlen;
                limit -= searchlen;

                if (limit == 0) {
                    *lp = limit;
                    return (unsigned char)RSTRING_PTR(str)[RSTRING_LEN(str)-1];
                }
            }
        } while (!mrb_obj_equal(mrb, more_char(mrb, fptr), MORE_CHAR_FINISHED));
        clear_readconv(fptr);
        *lp = limit;
        return EOF;
    }

    do {
      long pending = READ_DATA_PENDING_COUNT(fptr);
      if (pending > 0) {
          const char *p = READ_DATA_PENDING_PTR(fptr);
          const char *e;
          long last;

          if (limit > 0 && pending > limit) pending = limit;
          e = memchr(p, delim, pending);
          if (e) pending = e - p + 1;
          if (!mrb_nil_p(str)) {
            last = RSTRING_LEN(str);
            mrb_str_resize(mrb, str, last + pending);
          }
          else {
                last = 0;
            *strp = str = mrb_str_buf_new(mrb, pending);
            rb_str_set_len(str, pending);
          }
          read_buffered_data(RSTRING_PTR(str) + last, pending, fptr); /* must not fail */
          limit -= pending;
          *lp = limit;
          if (e) return delim;
          if (limit == 0)
            return (unsigned char)RSTRING_PTR(str)[RSTRING_LEN(str)-1];
      }
      READ_CHECK(mrb, fptr);
    } while (io_fillbuf(mrb, fptr) >= 0);
    *lp = limit;
    return EOF;
}

static inline int
swallow(mrb_state *mrb, rb_io_t *fptr, int term)
{
    if (NEED_READCONV(fptr)) {
      rb_encoding *enc = io_read_encoding(mrb, fptr);
      int needconv = rb_enc_mbminlen(enc) != 1;
      make_readconv(mrb, fptr, 0);
      do {
          size_t cnt;
          while ((cnt = READ_CHAR_PENDING_COUNT(fptr)) > 0) {
            const char *p = READ_CHAR_PENDING_PTR(fptr);
            int i;
            if (!needconv) {
                if (*p != term) return TRUE;
                i = (int)cnt;
                while (--i && *++p == term);
            }
            else {
                const char *e = p + cnt;
                if (rb_enc_ascget(mrb, p, e, &i, enc) != term) return TRUE;
                while ((p += i) < e && rb_enc_ascget(mrb, p, e, &i, enc) == term);
                i = (int)(e - p);
            }
            io_shift_cbuf(mrb, fptr, (int)cnt - i, NULL);
          }
      } while (!mrb_obj_equal(mrb, more_char(mrb, fptr), MORE_CHAR_FINISHED));
      return FALSE;
    }

    do {
      size_t cnt;
      while ((cnt = READ_DATA_PENDING_COUNT(fptr)) > 0) {
          char buf[1024];
          const char *p = READ_DATA_PENDING_PTR(fptr);
          int i;
          if (cnt > sizeof buf) cnt = sizeof buf;
          if (*p != term) return TRUE;
          i = (int)cnt;
          while (--i && *++p == term);
          if (!read_buffered_data(buf, cnt - i, fptr)) /* must not fail */
            rb_sys_fail_path(fptr->pathv);
      }
      READ_CHECK(mrb, fptr);
    } while (io_fillbuf(mrb, fptr) == 0);
    return FALSE;
}

static mrb_value
rb_io_getline_fast(mrb_state *mrb, rb_io_t *fptr, rb_encoding *enc, mrb_value io)
{
    mrb_value str = mrb_nil_value();
    int len = 0;
    long pos = 0;
    int cr = 0;

    for (;;) {
      int pending = READ_DATA_PENDING_COUNT(fptr);

      if (pending > 0) {
          const char *p = READ_DATA_PENDING_PTR(fptr);
          const char *e;

          e = memchr(p, '\n', pending);
          if (e) {
                pending = (int)(e - p + 1);
          }
          if (mrb_nil_p(str)) {
            str = mrb_str_new(mrb, p, pending);
            fptr->rbuf_off += pending;
            fptr->rbuf_len -= pending;
          }
          else {
            mrb_str_resize(mrb, str, len + pending);
            read_buffered_data(RSTRING_PTR(str)+len, pending, fptr);
          }
          len += pending;
          if (cr != ENC_CODERANGE_BROKEN)
            pos += rb_str_coderange_scan_restartable(RSTRING_PTR(str) + pos, RSTRING_PTR(str) + len, enc, &cr);
          if (e) break;
      }
      READ_CHECK(mrb, fptr);
      if (io_fillbuf(mrb, fptr) < 0) {
          if (mrb_nil_p(str)) return mrb_nil_value();
          break;
      }
    }

    str = io_enc_str(mrb, str, fptr);
    ENC_CODERANGE_SET(str, cr);
    fptr->lineno++;
    if (mrb_obj_equal(mrb, io, ARGF.current_file)) {
      ARGF.lineno++;
      ARGF.last_lineno = ARGF.lineno;
    }
    else {
      ARGF.last_lineno = fptr->lineno;
    }

    return str;
}

static void
io_unread(mrb_state *mrb, rb_io_t *fptr)
{
  off_t r;
  rb_io_check_closed(mrb, fptr);
  if (fptr->rbuf_len == 0 || fptr->mode & FMODE_DUPLEX)
    return;
  /* xxx: target position may be negative if buffer is filled by ungetc */
  errno = 0;
  r = lseek(fptr->fd, -fptr->rbuf_len, SEEK_CUR);
  if (r < 0 && errno) {
    if (errno == ESPIPE)
      fptr->mode |= FMODE_DUPLEX;
    return;
  }
  fptr->rbuf_off = 0;
  fptr->rbuf_len = 0;
  return;
}

static inline int
io_flush_buffer(mrb_state *mrb, rb_io_t *fptr)
{
#if 0 /* THREAD */
  if (mrb_fixnum_number(fptr->write_lock) != 0) {
    return rb_mutex_synchronize(mrb, fptr->write_lock, io_flush_buffer_async, (void *)fptr);
  }
  else {
    return io_flush_buffer_async(mrb, fptr);
  }
#else
    return io_flush_buffer_async(mrb, fptr);
#endif
}

static int
io_fflush(mrb_state *mrb, rb_io_t *fptr)
{
  rb_io_check_closed(mrb, fptr);
  if (fptr->wbuf_len == 0)
    return 0;
  //if (!rb_thread_fd_writable(mrb, fptr->fd)) {
  //  rb_io_check_closed(mrb, fptr);
  //}
  while (fptr->wbuf_len > 0 && io_flush_buffer(mrb, fptr) != 0) {
    if (1/*!rb_io_wait_writable(mrb, fptr->fd)*/)
      return -1;
    rb_io_check_closed(mrb, fptr);
  }
  return 0;
}

static rb_io_t *
flush_before_seek(mrb_state *mrb, rb_io_t *fptr)
{
  if (io_fflush(mrb, fptr) < 0)
    rb_sys_fail(0);
  io_unread(mrb, fptr);
  errno = 0;
  return fptr;
}

void
mrb_io_check_initialized(mrb_state *mrb, rb_io_t *fptr)
{
  if (!fptr) {
    mrb_raise(mrb, mrb->eIOError_class, "uninitialized stream");
  }
}

mrb_value
rb_io_get_write_io(mrb_state *mrb, mrb_value io)
{
  mrb_value write_io;

  mrb_io_check_initialized(mrb, RFILE(io)->fptr);
  write_io = RFILE(io)->fptr->tied_io_for_writing;
  if (!mrb_obj_equal(mrb, write_io, mrb_fixnum_value(0))) {
    return write_io;
  }
  return io;
}

#ifdef INCLUDE_REGEXP
static int
rb_sysopen(mrb_state *mrb, mrb_value fname, int oflags, mode_t perm)
{
  int fd;
  struct sysopen_struct data;

#ifdef O_BINARY
  oflags |= O_BINARY;
#endif
  data.fname = rb_str_encode_ospath(mrb, fname);
  data.oflags = oflags;
  data.perm = perm;

  fd = rb_sysopen_internal(mrb, &data);
  if (fd < 0) {
    if (errno == EMFILE || errno == ENFILE) {
        //rb_gc();
        fd = rb_sysopen_internal(mrb, &data);
    }
    if (fd < 0) {
        rb_sys_fail(RSTRING_PTR(fname));
    }
  }
  UPDATE_MAXFD(fd);
  return fd;
}

static mrb_value
rb_file_open_generic(mrb_state *mrb, mrb_value io, mrb_value filename, int oflags, int fmode, convconfig_t *convconfig, mode_t perm)
{
  rb_io_t *fptr;
  convconfig_t cc;
  if (!convconfig) {
    /* Set to default encodings */
    rb_io_ext_int_to_encs(mrb, NULL, NULL, &cc.enc, &cc.enc2);
      cc.ecflags = 0;
      cc.ecopts = mrb_nil_value();
      convconfig = &cc;
  }
  validate_enc_binmode(mrb, fmode, convconfig->enc, convconfig->enc2);

  MakeOpenFile(mrb, io, fptr);
  fptr->mode = fmode;
  fptr->encs = *convconfig;
  fptr->pathv = rb_str_new_frozen(mrb, filename);
  fptr->fd = rb_sysopen(mrb, fptr->pathv, oflags, perm);
  io_check_tty(fptr);
  if (fmode & FMODE_SETENC_BY_BOM) io_set_encoding_by_bom(mrb, io);

  return io;
}

static void
rb_scan_open_args(mrb_state *mrb, int argc, mrb_value *argv,
        mrb_value *fname_p, int *oflags_p, int *fmode_p,
        convconfig_t *convconfig_p, mode_t *perm_p)
{
    mrb_value opt=mrb_nil_value(), fname, vmode, vperm;
    int oflags, fmode;
    mode_t perm;

    fname = mrb_nil_value();
    vmode = mrb_nil_value();
    vperm = mrb_nil_value();
    opt = pop_last_hash(mrb, &argc, argv);
    //rb_scan_args(argc, argv, "12", &fname, &vmode, &vperm);
    mrb_get_args(mrb, "ooo", &fname, &vmode, &vperm);
    FilePathValue(mrb, fname);

    rb_io_extract_modeenc(mrb, &vmode, &vperm, opt, &oflags, &fmode, convconfig_p);

    perm = mrb_nil_p(vperm) ? 0666 :  mrb_fixnum_number(vperm);

    *fname_p = fname;
    *oflags_p = oflags;
    *fmode_p = fmode;
    *perm_p = perm;
}
#endif

#ifndef INCLUDE_REGEXP
static mrb_value
rb_file_open_internal(mrb_state *mrb, mrb_value io, const char *fname, const char *mode)
{
    rb_io_t *fptr;

    MakeOpenFile(mrb, io, fptr);

    fptr->mode = rb_io_mode_flags(mrb, mode);
    fptr->path = ruby_strdup(fname);
    fptr->f = rb_fopen(fptr->path, rb_io_flags_mode(fptr->mode));

    return io;
}
#endif

mrb_value
rb_open_file(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
#ifdef INCLUDE_REGEXP
  mrb_value fname;
  int oflags, fmode;
  convconfig_t convconfig;
  mode_t perm;

  rb_scan_open_args(mrb, argc, argv, &fname, &oflags, &fmode, &convconfig, &perm);
  rb_file_open_generic(mrb, io, fname, oflags, fmode, &convconfig, perm);

  return io;
#else
  mrb_value fname, vmode, perm;
  const char *path, *mode;
  int flags;
  unsigned int fmode;
  //rb_scan_args(argc, argv, "12", &fname, &vmode, &perm);
  mrb_get_args(mrb, "ooo", &fname, &vmode, &perm);
  //SafeStringValue(fname);
  mrb_string_value(mrb, &fname);

  //path = StringValueCStr(fname);
  path = rb_string_value_cstr(mrb, &fname);
  if (FIXNUM_P(vmode) || !mrb_nil_p(perm)) {
    if (FIXNUM_P(vmode)) {
        flags = mrb_fixnum_number(vmode);
    }
    else {
        //SafeStringValue(vmode);
        mrb_string_value(mrb, &vmode);
        flags = rb_io_mode_modenum(mrb, RSTRING(vmode)->ptr);
    }
    fmode = mrb_nil_p(perm) ? 0666 :  mrb_fixnum_number(perm);

    rb_file_sysopen_internal(mrb, io, path, flags, fmode);
  }
  else {
    //mode = mrb_nil_p(vmode) ? "r" : StringValueCStr(vmode);
    mode = mrb_nil_p(vmode) ? "r" : rb_string_value_cstr(mrb, &vmode);
    rb_file_open_internal(mrb, io, path, mode);
  }
  return io;
#endif //#ifdef INCLUDE_REGEXP
}

#ifdef INCLUDE_REGEXP
static void
rb_io_extract_modeenc(mrb_state *mrb, mrb_value *vmode_p, mrb_value *vperm_p, mrb_value opthash,
        int *oflags_p, int *fmode_p, convconfig_t *convconfig_p)
{
  mrb_value vmode;
  int oflags, fmode;
  rb_encoding *enc, *enc2;
  int ecflags;
  mrb_value ecopts;
  int has_enc = 0, has_vmode = 0;
  mrb_value intmode;

  vmode = *vmode_p;

  /* Set to defaults */
  rb_io_ext_int_to_encs(mrb, NULL, NULL, &enc, &enc2);

  if (mrb_nil_p(vmode)) {
      fmode = FMODE_READABLE;
      oflags = O_RDONLY;
  }
  else if (!mrb_nil_p(intmode = rb_check_to_integer(mrb, vmode, "to_int"))) {
      vmode = intmode;
      oflags = mrb_fixnum_number(intmode);
      fmode = rb_io_oflags_fmode(oflags);
  }
  else {
    const char *p;

    vmode_handle:
    //SafeStringValue(vmode);
    mrb_string_value(mrb, &vmode);
    //p = StringValueCStr(vmode);
    p = rb_string_value_cstr(mrb, &vmode);
    fmode = rb_io_modestr_fmode(mrb, p);
    oflags = rb_io_fmode_oflags(fmode);
    p = strchr(p, ':');
    if (p) {
        has_enc = 1;
        parse_mode_enc(mrb, p+1, &enc, &enc2, &fmode);
    }
    else {
      rb_encoding *e;

      e = (fmode & FMODE_BINMODE) ? rb_ascii8bit_encoding(mrb) : NULL;
      rb_io_ext_int_to_encs(mrb, e, NULL, &enc, &enc2);
    }
  }

  if (mrb_nil_p(opthash)) {
    ecflags = 0;
    ecopts = mrb_nil_value();
  }
  else {
    mrb_value v;
    extract_binmode(mrb, opthash, &fmode);
#ifdef O_BINARY
    if (fmode & FMODE_BINMODE)
      oflags |= O_BINARY;
#endif
    if (!has_vmode) {
      v = mrb_hash_get(mrb, opthash, sym_mode);
      if (!mrb_nil_p(v)) {
        if (!mrb_nil_p(vmode)) {
            mrb_raise(mrb, mrb->eArgumentError_class, "mode specified twice");
        }
        has_vmode = 1;
        vmode = v;
        goto vmode_handle;
      }
    }
    v = mrb_hash_get(mrb, opthash, sym_perm);
    if (!mrb_nil_p(v)) {
      if (vperm_p) {
        if (!mrb_nil_p(*vperm_p)) {
            mrb_raise(mrb, mrb->eArgumentError_class, "perm specified twice");
        }
        *vperm_p = v;
      }
      else {
        /* perm no use, just ignore */
      }
    }
    ecflags = rb_econv_prepare_opts(mrb, opthash, &ecopts);

    if (rb_io_extract_encoding_option(mrb, opthash, &enc, &enc2, &fmode)) {
        if (has_enc) {
            mrb_raise(mrb, mrb->eArgumentError_class, "encoding specified twice");
        }
    }
  }

  validate_enc_binmode(mrb, fmode, enc, enc2);

  *vmode_p = vmode;

  *oflags_p = oflags;
  *fmode_p = fmode;
  convconfig_p->enc = enc;
  convconfig_p->enc2 = enc2;
  convconfig_p->ecflags = ecflags;
  convconfig_p->ecopts = ecopts;
}
#endif //#ifdef INCLUDE_REGEXP

#ifdef INCLUDE_REGEXP
/* function:
   1) Convert last_argument to hash_value
   2) decrement argc
*/
static mrb_value
pop_last_hash(mrb_state *mrb, int *argc_p, mrb_value *argv)
{
    mrb_value last, tmp;
    if (*argc_p == 0)
        return mrb_nil_value();
    last = argv[*argc_p-1];
    if (mrb_nil_p(last)) return mrb_nil_value();
    /* last argument -> hash_value */
    tmp = mrb_check_convert_type(mrb, last, MRB_TT_HASH, "Hash", "to_hash");
    if (mrb_nil_p(tmp))
        return mrb_nil_value();
    (*argc_p)--;
    return tmp;
}
#endif //#ifdef INCLUDE_REGEXP

void
rb_io_check_closed(mrb_state *mrb, rb_io_t *fptr)
{
  mrb_io_check_initialized(mrb, fptr);
  if (fptr->fd < 0) {
    mrb_raise(mrb, mrb->eIOError_class, "closed stream");
  }
}

/*
 *  call-seq:
 *     IO.new(fd [, mode] [, opt])   -> io
 *
 *  Returns a new <code>IO</code> object (a stream) for the given
 *  <code>IO</code> object or integer file descriptor and mode
 *  string. See also <code>IO.sysopen</code> and
 *  <code>IO.for_fd</code>.
 *
 *  === Parameters
 *  fd:: numeric file descriptor
 *  mode:: file mode. a string or an integer
 *  opt:: hash for specifying mode by name.
 *
 *  ==== Mode
 *  When <code>mode</code> is an integer it must be combination of
 *  the modes defined in <code>File::Constants</code>.
 *
 *  When <code>mode</code> is a string it must be in one of the
 *  following forms:
 *  - "fmode",
 *  - "fmode:extern",
 *  - "fmode:extern:intern".
 *  <code>extern</code> is the external encoding name for the IO.
 *  <code>intern</code> is the internal encoding.
 *  <code>fmode</code> must be combination of the directives. See
 *  the description of class +IO+ for a description of the directives.
 *
 *  When the mode of original IO is read only, the mode cannot be changed to
 *  be writable.  Similarly, the mode cannot be changed from write only to
 *  readable.
 *  If such a wrong change is directed, timing where the error actually occurs
 *  is different according to the platform.
 *
 *  ==== Options
 *  <code>opt</code> can have the following keys
 *  :mode ::
 *    same as <code>mode</code> parameter
 *  :external_encoding ::
 *    external encoding for the IO. "-" is a
 *    synonym for the default external encoding.
 *  :internal_encoding ::
 *    internal encoding for the IO.
 *    "-" is a synonym for the default internal encoding.
 *    If the value is nil no conversion occurs.
 *  :encoding ::
 *    specifies external and internal encodings as "extern:intern".
 *  :textmode ::
 *    If the value is truth value, same as "t" in argument <code>mode</code>.
 *  :binmode ::
 *    If the value is truth value, same as "b" in argument <code>mode</code>.
 *  :autoclose ::
 *    If the value is +false+, the _fd_ will be kept open after this
 *    +IO+ instance gets finalized.
 *
 *  Also <code>opt</code> can have same keys in <code>String#encode</code> for
 *  controlling conversion between the external encoding and the internal encoding.
 *
 *  === Example1
 *
 *     fd = IO.sysopen("/dev/tty", "w")
 *     a = IO.new(fd,"w")
 *     $stderr.puts "Hello"
 *     a.puts "World"
 *
 *  <em>produces:</em>
 *
 *     Hello
 *     World
 *
 *  === Example2
 *
 *     require 'fcntl'
 *
 *     fd = STDERR.fcntl(Fcntl::F_DUPFD)
 *     io = IO.new(fd, mode: 'w:UTF-16LE', cr_newline: true)
 *     io.puts "Hello, World!"
 *
 *     fd = STDERR.fcntl(Fcntl::F_DUPFD)
 *     io = IO.new(fd, mode: 'w', cr_newline: true, external_encoding: Encoding::UTF_16LE)
 *     io.puts "Hello, World!"
 *
 *  both of above print "Hello, World!" in UTF-16LE to standard error output with
 *  converting EOL generated by <code>puts</code> to CR.
 */

mrb_value
rb_io_initialize(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
#ifdef INCLUDE_REGEXP
  mrb_value fnum, vmode;
  rb_io_t *fp;
  int fd, fmode, oflags = O_RDONLY;
  convconfig_t convconfig;
  mrb_value opt;
//#if defined(HAVE_FCNTL) && defined(F_GETFL)
//    int ofmode;
//#else
  struct stat st;
//#endif

    //rb_secure(4);

  opt = pop_last_hash(mrb, &argc, argv);
  //rb_scan_args(argc, argv, "11", &fnum, &vmode);
  //?mrb_get_args(mrb, "oo", &fnum, &vmode);
  fnum = argv[0];
  vmode = argv[1];
  rb_io_extract_modeenc(mrb, &vmode, 0, opt, &oflags, &fmode, &convconfig);

  fd = mrb_fixnum_number(fnum);
//#if defined(HAVE_FCNTL) && defined(F_GETFL)
//    oflags = fcntl(fd, F_GETFL);
//    if (oflags == -1) rb_sys_fail(0);
//#else
  if (fstat(fd, &st) == -1) rb_sys_fail(0);
//#endif
  UPDATE_MAXFD(fd);
//#if defined(HAVE_FCNTL) && defined(F_GETFL)
//    ofmode = rb_io_oflags_fmode(oflags);
//    if (mrb_nil_p(vmode)) {
//    fmode = ofmode;
//    }
//  else if ((~ofmode & fmode) & FMODE_READWRITE) {
//    mrb_value error = mrb_fixnum_value(EINVAL);
//    rb_exc_raise(mrb, rb_class_new_instance(1, &error, rb_eSystemCallError));
//  }
//#endif
  if (!mrb_nil_p(opt) && mrb_obj_equal(mrb, mrb_hash_get(mrb, opt, sym_autoclose), mrb_false_value())) {
    fmode |= FMODE_PREP;
  }
  MakeOpenFile(mrb, io, fp);
  fp->fd = fd;
  fp->mode = fmode;
  fp->encs = convconfig;
  clear_codeconv(fp);
  io_check_tty(fp);
  if (fileno(stdin) == fd)
    fp->stdio_file = stdin;
  else if (fileno(stdout) == fd)
    fp->stdio_file = stdout;
  else if (fileno(stderr) == fd)
    fp->stdio_file = stderr;

  if (fmode & FMODE_SETENC_BY_BOM) io_set_encoding_by_bom(mrb, io);
  return io;
#else
  mrb_value fnum, mode;
  rb_io_t *fp;
  int fd, flags;

  //rb_secure(4);
  //rb_scan_args(argc, argv, "11", &fnum, &mode);
  fnum = argv[0];
  vmode = argv[1];
  fd = mrb_fixnum_number(fnum);
  if (argc == 2) {
    if (FIXNUM_P(mode)) {
      flags = mrb_fixnum_number(mode);
    }
    else {
      //SafeStringValue(mode);
      mrb_string_value(mrb, &mode);
      //flags = rb_io_mode_modenum(StringValueCStr(mode));
      flags = rb_io_mode_modenum(mrb, rb_string_value_cstr(mrb, &mode));
    }
  }
  else {
#if defined(HAVE_FCNTL) && defined(F_GETFL)
    flags = fcntl(fd, F_GETFL);
    if (flags == -1) rb_sys_fail(0);
#else
    flags = O_RDONLY;
#endif
  }
  MakeOpenFile(mrb, io, fp);
  fp->mode = rb_io_modenum_flags(flags);
  fp->f = rb_fdopen(fd, rb_io_modenum_mode(mrb, flags));

  return io;
#endif
}

mrb_value
mrb_io_initialize(mrb_state *mrb, mrb_value io)
{
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);
  return rb_io_initialize(mrb, argc, argv, io);
}

static long
io_binwrite(mrb_state *mrb, mrb_value str, rb_io_t *fptr, int nosync)
{
  long len, n, r, offset = 0;

  len = RSTRING_LEN(str);
  if ((n = len) <= 0) return n;
  if (fptr->wbuf == NULL && !(!nosync && (fptr->mode & FMODE_SYNC))) {
    fptr->wbuf_off = 0;
    fptr->wbuf_len = 0;
    fptr->wbuf_capa = IO_WBUF_CAPA_MIN;
    //fptr->wbuf = ALLOC_N(char, fptr->wbuf_capa);
    fptr->wbuf = malloc(fptr->wbuf_capa);
    //fptr->write_lock = rb_mutex_new(mrb);
  }
  if ((!nosync && (fptr->mode & (FMODE_SYNC|FMODE_TTY))) ||
      (fptr->wbuf && fptr->wbuf_capa <= fptr->wbuf_len + len)) {
    //struct binwrite_arg arg;

    /* xxx: use writev to avoid double write if available */
    if (fptr->wbuf_len && fptr->wbuf_len+len <= fptr->wbuf_capa) {
      if (fptr->wbuf_capa < fptr->wbuf_off+fptr->wbuf_len+len) {
          MEMMOVE(fptr->wbuf, fptr->wbuf+fptr->wbuf_off, char, fptr->wbuf_len);
          fptr->wbuf_off = 0;
      }
      MEMMOVE(fptr->wbuf+fptr->wbuf_off+fptr->wbuf_len, RSTRING_PTR(str)+offset, char, len);
      fptr->wbuf_len += (int)len;
      n = 0;
    }
    if (io_fflush(mrb, fptr) < 0)
        return -1L;
    if (n == 0)
        return len;
    /* avoid context switch between "a" and "\n" in STDERR.puts "a".
       [ruby-dev:25080] */
    //if (fptr->stdio_file != stderr && !rb_thread_fd_writable(mrb, fptr->fd)) {
    //  rb_io_check_closed(mrb, fptr);
    //}
    //arg.fptr = fptr;
    //arg.str = str;
    //retry:
    //arg.offset = offset;
    //arg.length = n;
    //if (mrb_fixnum_number(fptr->write_lock)) {
    //  r = rb_mutex_synchronize(mrb, fptr->write_lock, io_binwrite_string, (void *)&arg);
    //}
    //else {
      long l = io_writable_length(fptr, n);
      r = rb_write_internal(mrb, fptr->fd, RSTRING_PTR(str)+offset, l);
    //}
    /* xxx: other threads may modify given string. */
    if (r == n) return len;
    if (0 <= r) {
      offset += r;
      n -= r;
      errno = EAGAIN;
    }
    //if (rb_io_wait_writable(mrb, fptr->fd)) {
    //  rb_io_check_closed(mrb, fptr);
    //  if (offset < RSTRING_LEN(str))
    //    goto retry;
    //}
    return -1L;
  }

  if (fptr->wbuf_off) {
    if (fptr->wbuf_len)
      MEMMOVE(fptr->wbuf, fptr->wbuf+fptr->wbuf_off, char, fptr->wbuf_len);
    fptr->wbuf_off = 0;
  }
  MEMMOVE(fptr->wbuf+fptr->wbuf_off+fptr->wbuf_len, RSTRING_PTR(str)+offset, char, len);
  fptr->wbuf_len += (int)len;
  return len;
}

static mrb_value
do_writeconv(mrb_state *mrb, mrb_value str, rb_io_t *fptr)
{
  if (NEED_WRITECONV(mrb, fptr)) {
    mrb_value common_encoding = mrb_nil_value();

    make_writeconv(mrb, fptr);

    if (fptr->writeconv) {
        if (!mrb_nil_p(fptr->writeconv_asciicompat))
            common_encoding = fptr->writeconv_asciicompat;
        else if (!rb_enc_asciicompat(mrb, rb_enc_get(mrb, str))) {
            mrb_raise(mrb, mrb->eArgumentError_class, "ASCII incompatible string written for text mode IO without encoding conversion: %s",
                     rb_enc_name(rb_enc_get(mrb, str)));
        }
    }
    else {
        if (fptr->encs.enc2)
            common_encoding = rb_enc_from_encoding(mrb, fptr->encs.enc2);
        else if (fptr->encs.enc != rb_ascii8bit_encoding(mrb))
            common_encoding = rb_enc_from_encoding(mrb, fptr->encs.enc);
    }

    if (!mrb_nil_p(common_encoding)) {
        str = rb_str_encode(mrb, str, common_encoding,
            fptr->writeconv_pre_ecflags, fptr->writeconv_pre_ecopts);
    }

    if (fptr->writeconv) {
        str = rb_econv_str_convert(mrb, fptr->writeconv, str, ECONV_PARTIAL_INPUT);
    }
  }
  return str;
}

static long
io_fwrite(mrb_state *mrb, mrb_value str, rb_io_t *fptr, int nosync)
{
  str = do_writeconv(mrb, str, fptr);
  return io_binwrite(mrb, str, fptr, nosync);
}

static mrb_value
rb_io_check_io(mrb_state *mrb, mrb_value io)
{
    return mrb_check_convert_type(mrb, io, MRB_TT_FILE, "IO", "to_io");
}

mrb_value
io_write(mrb_state *mrb, mrb_value io, mrb_value str, int nosync)
{
  rb_io_t *fptr;
  long n;
  mrb_value tmp;

  //rb_secure(4);
  io = GetWriteIO(mrb, io);
  if (mrb_nil_p(str)) { /* kusuda */
    return mrb_nil_value();
  }
  str = mrb_obj_as_string(mrb, str);
  tmp = rb_io_check_io(mrb, io);
  if (mrb_nil_p(tmp)) {
    /* port is not IO, call write method for it. */
    return mrb_funcall(mrb, io, "write", 1, str);
  }
  io = tmp;
  if (RSTRING_LEN(str) == 0) return mrb_fixnum_value(0);

  GetOpenFile(mrb, io, fptr);
  rb_io_check_writable(mrb, fptr);

  n = io_fwrite(mrb, str, fptr, nosync);
  if (n == -1L) rb_sys_fail_path(fptr->pathv);

  return mrb_fixnum_value(n);
}

mrb_value
rb_io_write(mrb_state *mrb, mrb_value io, mrb_value str)
{
    return mrb_funcall(mrb, io, "write", 1, str);
}

static mrb_value
fill_cbuf(mrb_state *mrb, rb_io_t *fptr, int ec_flags)
{
  const unsigned char *ss, *sp, *se;
  unsigned char *ds, *dp, *de;
  rb_econv_result_t res;
  int putbackable;
  int cbuf_len0;
  mrb_value exc;

  ec_flags |= ECONV_PARTIAL_INPUT;

  if (fptr->cbuf_len == fptr->cbuf_capa)
      return MORE_CHAR_SUSPENDED; /* cbuf full */
  if (fptr->cbuf_len == 0)
      fptr->cbuf_off = 0;
  else if (fptr->cbuf_off + fptr->cbuf_len == fptr->cbuf_capa) {
      memmove(fptr->cbuf, fptr->cbuf+fptr->cbuf_off, fptr->cbuf_len);
      fptr->cbuf_off = 0;
  }

  cbuf_len0 = fptr->cbuf_len;

  while (1) {
      ss = sp = (const unsigned char *)fptr->rbuf + fptr->rbuf_off;
      se = sp + fptr->rbuf_len;
      ds = dp = (unsigned char *)fptr->cbuf + fptr->cbuf_off + fptr->cbuf_len;
      de = (unsigned char *)fptr->cbuf + fptr->cbuf_capa;
      res = rb_econv_convert(mrb, fptr->readconv, &sp, se, &dp, de, ec_flags);
      fptr->rbuf_off += (int)(sp - ss);
      fptr->rbuf_len -= (int)(sp - ss);
      fptr->cbuf_len += (int)(dp - ds);

      putbackable = rb_econv_putbackable(fptr->readconv);
      if (putbackable) {
          rb_econv_putback(fptr->readconv, (unsigned char *)fptr->rbuf + fptr->rbuf_off - putbackable, putbackable);
          fptr->rbuf_off -= putbackable;
          fptr->rbuf_len += putbackable;
      }

      exc = rb_econv_make_exception(mrb, fptr->readconv);
      if (!mrb_nil_p(exc))
          return exc;

      if (cbuf_len0 != fptr->cbuf_len)
          return MORE_CHAR_SUSPENDED;

      if (res == econv_finished) {
          return MORE_CHAR_FINISHED;
      }

      if (res == econv_source_buffer_empty) {
          if (fptr->rbuf_len == 0) {
      READ_CHECK(mrb, fptr);
              if (io_fillbuf(mrb, fptr) == -1) {
                  ds = dp = (unsigned char *)fptr->cbuf + fptr->cbuf_off + fptr->cbuf_len;
                  de = (unsigned char *)fptr->cbuf + fptr->cbuf_capa;
                  res = rb_econv_convert(mrb, fptr->readconv, NULL, NULL, &dp, de, 0);
                  fptr->cbuf_len += (int)(dp - ds);
                  rb_econv_check_error(mrb, fptr->readconv);
              }
          }
      }
  } /* end of while */
}

static mrb_value
read_all(mrb_state *mrb, rb_io_t *fptr, long siz, mrb_value str)
{
    long bytes;
    long n;
    long pos;
    rb_encoding *enc;
    int cr;

    if (NEED_READCONV(fptr)) {
      io_setstrbuf(mrb, &str, 0);
        make_readconv(mrb, fptr, 0);
        while (1) {
            mrb_value v;
            if (fptr->cbuf_len) {
                io_shift_cbuf(mrb, fptr, fptr->cbuf_len, &str);
            }
            v = fill_cbuf(mrb, fptr, 0);
            if (!mrb_obj_equal(mrb, v, MORE_CHAR_SUSPENDED) && !mrb_obj_equal(mrb, v, MORE_CHAR_FINISHED)) {
                if (fptr->cbuf_len) {
                    io_shift_cbuf(mrb, fptr, fptr->cbuf_len, &str);
                }
                rb_exc_raise(mrb, v);
            }
            if (mrb_obj_equal(mrb, v, MORE_CHAR_FINISHED)) {
                clear_readconv(fptr);
                return io_enc_str(mrb, str, fptr);
            }
        }
    }

    bytes = 0;
    pos = 0;

    enc = io_read_encoding(mrb, fptr);
    cr = 0;

    if (siz == 0) siz = BUFSIZ;
    io_setstrbuf(mrb, &str, siz);
    for (;;) {
      READ_CHECK(mrb, fptr);
      n = io_fread(mrb, str, bytes, fptr);
      if (n == 0 && bytes == 0) {
          break;
      }
      bytes += n;
      if (cr != ENC_CODERANGE_BROKEN)
          pos += rb_str_coderange_scan_restartable(RSTRING_PTR(str) + pos, RSTRING_PTR(str) + bytes, enc, &cr);
      if (bytes < siz) break;
      siz += BUFSIZ;
      mrb_str_resize(mrb, str, siz);
    }
    if (bytes != siz) mrb_str_resize(mrb, str, bytes);
    str = io_enc_str(mrb, str, fptr);
    ENC_CODERANGE_SET(str, cr);
    return str;
}

static void
prepare_getline_args(mrb_state *mrb, int argc, mrb_value *argv, mrb_value *rsp, long *limit, mrb_value io)
{
    mrb_value rs = rb_rs, lim = mrb_nil_value();
    rb_io_t *fptr;

    if (argc == 1) {
      mrb_value tmp = mrb_nil_value();

      if (mrb_type(argv[0]) == MRB_TT_FIXNUM) {
        lim = argv[0];
      }
      else {
        if (mrb_nil_p(argv[0]) || !mrb_nil_p(tmp = mrb_check_string_type(mrb, argv[0]))) {
          rs = tmp;
        }
        else {
          lim = argv[0];
        }
      }
    }
    else if (2 <= argc) {
      //rb_scan_args(argc, argv, "2", &rs, &lim);
      mrb_get_args(mrb, "oo", &rs, &lim);
      if (!mrb_nil_p(rs))
        //StringValue(rs);
        mrb_string_value(mrb, &rs);
    }
    if (!mrb_nil_p(rs)) {
      rb_encoding *enc_rs, *enc_io;

      GetOpenFile(mrb, io, fptr);
      enc_rs = rb_enc_get(mrb, rs);
      enc_io = io_read_encoding(mrb, fptr);
      if (enc_io != enc_rs &&
          (rb_enc_str_coderange(mrb, rs) != ENC_CODERANGE_7BIT ||
           (RSTRING_LEN(rs) > 0 && !rb_enc_asciicompat(mrb, enc_io)))) {
	rs = mrb_enc_str_new(mrb, 0, 0, enc_io);
	rb_str_buf_cat_ascii(mrb, rs, "\n");
      }
    }
    *rsp = rs;
    *limit = mrb_nil_p(lim) ? -1L : mrb_fixnum_number(lim);
}

void
rb_io_check_char_readable(mrb_state *mrb, rb_io_t *fptr)
{
  rb_io_check_closed(mrb, fptr);
  if (!(fptr->mode & FMODE_READABLE)) {
    mrb_raise(mrb, mrb->eIOError_class, "not opened for reading");
  }
  if (fptr->wbuf_len) {
    if (io_fflush(mrb, fptr) < 0)
        rb_sys_fail(0);
  }
  if (mrb_fixnum_number(fptr->tied_io_for_writing) != 0) {
    rb_io_t *wfptr;
    GetOpenFile(mrb, fptr->tied_io_for_writing, wfptr);
    if (io_fflush(mrb, wfptr) < 0)
      rb_sys_fail(0);
  }
}

static mrb_value
rb_io_getline_1(mrb_state *mrb, mrb_value rs, long limit, mrb_value io)
{
  mrb_value str = mrb_nil_value();
  rb_io_t *fptr;
  int nolimit = 0;
  rb_encoding *enc;

  GetOpenFile(mrb, io, fptr);
  rb_io_check_char_readable(mrb, fptr);
  if (mrb_nil_p(rs) && limit < 0) {
    str = read_all(mrb, fptr, 0, mrb_nil_value());
    if (RSTRING_LEN(str) == 0) return mrb_nil_value();
  }
  else if (limit == 0) {
    return mrb_enc_str_new(mrb, 0, 0, io_read_encoding(mrb, fptr));
  }
  else if (limit < 0 && !NEED_READCONV(fptr) &&
             rb_enc_asciicompat(mrb, enc = io_read_encoding(mrb, fptr))) {
    return rb_io_getline_fast(mrb, fptr, enc, io);
  }
  else {
    int c, newline = -1;
    const char *rsptr = 0;
    long rslen = 0;
    int rspara = 0;
    int extra_limit = 16;

    enc = io_read_encoding(mrb, fptr);

    if (!mrb_nil_p(rs)) {
      rslen = RSTRING_LEN(rs);
      if (rslen == 0) {
        rsptr = "\n\n";
        rslen = 2;
        rspara = 1;
        swallow(mrb, fptr, '\n');
        rs = mrb_str_new_cstr(mrb, "");
        if (!rb_enc_asciicompat(mrb, enc)) {
            rs = rb_usascii_str_new(mrb, rsptr, rslen);
            rs = rb_str_encode(mrb, rs, rb_enc_from_encoding(mrb, enc), 0, mrb_nil_value());
            //OBJ_FREEZE(rs);
            rsptr = RSTRING_PTR(rs);
            rslen = RSTRING_LEN(rs);
        }
      }
      else {
        rsptr = RSTRING_PTR(rs);
      }
      newline = (unsigned char)rsptr[rslen - 1];
    }

    /* MS - Optimisation */
    while ((c = appendline(mrb, fptr, newline, &str, &limit)) != EOF) {
      const char *s, *p, *pp, *e;

      if (c == newline) {
        if (RSTRING_LEN(str) < rslen) continue;
        s = RSTRING_PTR(str);
                e = s + RSTRING_LEN(str);
        p = e - rslen;
        pp = rb_enc_left_char_head(s, p, e, enc);
        if (pp != p) continue;
        if (!rspara) rscheck(mrb, rsptr, rslen, rs);
        if (memcmp(p, rsptr, rslen) == 0) break;
      }
      if (limit == 0) {
        s = RSTRING_PTR(str);
        p = s + RSTRING_LEN(str);
        pp = rb_enc_left_char_head(s, p-1, p, enc);
        if (extra_limit &&
            MBCLEN_NEEDMORE_P(rb_enc_precise_mbclen(pp, p, enc))) {
            /* relax the limit while incomplete character.
             * extra_limit limits the relax length */
            limit = 1;
            extra_limit--;
        }
        else {
            nolimit = 1;
            break;
        }
      }
    } /* end of whils */

    if (rspara) {
      if (c != EOF) {
        swallow(mrb, fptr, '\n');
      }
    }
    if (!mrb_nil_p(str))
      str = io_enc_str(mrb, str, fptr);
  } /* end of else */

  if (!mrb_nil_p(str)) {
    if (!nolimit) {
      fptr->lineno++;
      if (mrb_obj_equal(mrb, io, ARGF.current_file)) {
        ARGF.lineno++;
        ARGF.last_lineno = ARGF.lineno;
      }
      else {
        ARGF.last_lineno = fptr->lineno;
      }
    }
  }

  return str;
}

static mrb_value
rb_io_getline(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
    mrb_value rs;
    long limit;

    prepare_getline_args(mrb, argc, argv, &rs, &limit, io);
    return rb_io_getline_1(mrb, rs, limit, io);
}


static mrb_value
io_read(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
  rb_io_t *fptr;
  long n, len;
  mrb_value length, str;

  //rb_scan_args(argc, argv, "02", &length, &str);
  length = mrb_nil_value();
  str    = mrb_nil_value();
  //if (mrb_nil_p(length)) {
  if (argc == 0) {
    GetOpenFile(mrb, io, fptr);
    rb_io_check_char_readable(mrb, fptr);
    return read_all(mrb, fptr, remain_size(mrb, fptr), str);
  }

  length = argv[0];
  if (argc > 1) {
    str = argv[1];
  }
  len = mrb_fixnum_number(length);
  if (len < 0) {
    mrb_raise(mrb, mrb->eArgumentError_class, "negative length %ld given", len);
  }

  io_setstrbuf(mrb, &str, len);

  GetOpenFile(mrb, io, fptr);
  rb_io_check_byte_readable(mrb, fptr);
  if (len == 0) return str;

  READ_CHECK(mrb, fptr);
  n = io_fread(mrb, str, 0, fptr);
  if (n == 0) {
    if (fptr->fd < 0) return mrb_nil_value();
      mrb_str_resize(mrb, str, 0);
      return mrb_nil_value();
  }
  mrb_str_resize(mrb, str, n);
  //OBJ_TAINT(str);

  return str;
}

static void
io_setstrbuf(mrb_state *mrb, mrb_value *str,long len)
{
#ifdef _WIN32
    if (mrb_nil_p(*str)) {
      *str = mrb_str_new(mrb, 0, len+1);
      rb_str_set_len(*str,len);
    }
    else {
      //StringValue(*str);
      mrb_string_value(mrb, str);
      rb_str_modify(*str);
      mrb_str_resize(mrb, *str, len+1);
      rb_str_set_len(*str, len);
    }
#else
    if (mrb_nil_p(*str)) {
      *str = mrb_str_new(mrb, 0, len);
    }
    else {
      //StringValue(*str);
      mrb_string_value(mrb, str);
      rb_str_modify(*str);
      mrb_str_resize(mrb, *str, len);
    }
#endif
}

static int
str_end_with_asciichar(mrb_state *mrb, mrb_value str, int c)
{
  long len = RSTRING_LEN(str);
  const char *ptr = RSTRING_PTR(str);
  rb_encoding *enc = rb_enc_from_index(mrb, ENCODING_GET(mrb, str));
  int n;

  if (len == 0) return 0;
  if ((n = rb_enc_mbminlen(enc)) == 1) {
    return ptr[len - 1] == c;
  }
  return rb_enc_ascget(mrb, ptr + ((len - 1) / n) * n, ptr + len, &n, enc) == c;
}

mrb_value
io_puts(mrb_state *mrb, int argc, mrb_value *argv, mrb_value out)
{
  int i;
  mrb_value line;

  /* if no argument given, print newline. */
  if (argc == 0) {
    rb_io_write(mrb, out, rb_default_rs);
    return mrb_nil_value();
  }
  for (i=0; i<argc; i++) {
    if (mrb_type(argv[i]) == MRB_TT_STRING) {
        line = argv[i];
        goto string;
    }
    line = mrb_check_array_type(mrb, argv[i]);
    if (!mrb_nil_p(line)) {
        rb_exec_recursive(mrb, io_puts_ary, line, &out);
        continue;
    }
    line = mrb_obj_as_string(mrb, argv[i]);
    string:
    rb_io_write(mrb, out, line);
    if (RSTRING_LEN(line) == 0 ||
	!str_end_with_asciichar(mrb, line, '\n')) {
        rb_io_write(mrb, out, rb_default_rs);
    }
  }
  if (mrb_type(out) == MRB_TT_FILE) {
    mrb_io_flush(mrb, out);
  }

  return mrb_nil_value();
}

static mrb_value
io_puts_ary(mrb_state *mrb, mrb_value ary, mrb_value out, int recur)
{
  mrb_value tmp;
  long i;

  if (recur) {
    tmp = mrb_str_new2(mrb, "[...]");
    io_puts(mrb, 1, &tmp, out);
    return mrb_nil_value();
  }
  for (i=0; i<RARRAY_LEN(ary); i++) {
    tmp = RARRAY_PTR(ary)[i];
    io_puts(mrb, 1, &tmp, out);
  }
  return mrb_nil_value();
}

#ifdef INCLUDE_REGEXP
static mrb_value
io_getc(mrb_state *mrb, rb_io_t *fptr, rb_encoding *enc)
{
  int r, n, cr = 0;
  mrb_value str;

  if (NEED_READCONV(fptr)) {
    mrb_value str = mrb_nil_value();
    rb_encoding *read_enc = io_read_encoding(mrb, fptr);

    make_readconv(mrb, fptr, 0);

    while (1) {
      if (fptr->cbuf_len) {
          r = rb_enc_precise_mbclen(fptr->cbuf+fptr->cbuf_off,
                fptr->cbuf+fptr->cbuf_off+fptr->cbuf_len,
                read_enc);
              if (!MBCLEN_NEEDMORE_P(r))
                  break;
              if (fptr->cbuf_len == fptr->cbuf_capa) {
                  mrb_raise(mrb, mrb->eIOError_class, "too long character");
              }
      }

      if (mrb_obj_equal(mrb, more_char(mrb, fptr), MORE_CHAR_FINISHED)) {
        if (fptr->cbuf_len == 0) {
              clear_readconv(fptr);
              return mrb_nil_value();
        }
              /* return an unit of an incomplete character just before EOF */
        str = mrb_enc_str_new(mrb, fptr->cbuf+fptr->cbuf_off, 1, read_enc);
        fptr->cbuf_off += 1;
        fptr->cbuf_len -= 1;
        if (fptr->cbuf_len == 0) clear_readconv(fptr);
        ENC_CODERANGE_SET(str, ENC_CODERANGE_BROKEN);
        return str;
      }
    }
    if (MBCLEN_INVALID_P(r)) {
      r = rb_enc_mbclen(fptr->cbuf+fptr->cbuf_off,
                        fptr->cbuf+fptr->cbuf_off+fptr->cbuf_len,
                        read_enc);
      io_shift_cbuf(mrb, fptr, r, &str);
      cr = ENC_CODERANGE_BROKEN;
    }
    else {
      io_shift_cbuf(mrb, fptr, MBCLEN_CHARFOUND_LEN(r), &str);
      cr = ENC_CODERANGE_VALID;
    }
    str = io_enc_str(mrb, str, fptr);
    ENC_CODERANGE_SET(str, cr);
    return str;
  }

  if (io_fillbuf(mrb, fptr) < 0) {
    return mrb_nil_value();
  }
  if (rb_enc_asciicompat(mrb, enc) && ISASCII(fptr->rbuf[fptr->rbuf_off])) {
    str = mrb_str_new(mrb, fptr->rbuf+fptr->rbuf_off, 1);
    fptr->rbuf_off += 1;
    fptr->rbuf_len -= 1;
    cr = ENC_CODERANGE_7BIT;
  }
  else {
    r = rb_enc_precise_mbclen(fptr->rbuf+fptr->rbuf_off, fptr->rbuf+fptr->rbuf_off+fptr->rbuf_len, enc);
    if (MBCLEN_CHARFOUND_P(r) &&
        (n = MBCLEN_CHARFOUND_LEN(r)) <= fptr->rbuf_len) {
        str = mrb_str_new(mrb, fptr->rbuf+fptr->rbuf_off, n);
        fptr->rbuf_off += n;
        fptr->rbuf_len -= n;
        cr = ENC_CODERANGE_VALID;
    }
    else if (MBCLEN_NEEDMORE_P(r)) {
      str = mrb_str_new(mrb, fptr->rbuf+fptr->rbuf_off, fptr->rbuf_len);
      fptr->rbuf_len = 0;
      getc_needmore:
      if (io_fillbuf(mrb, fptr) != -1) {
        rb_str_cat(mrb, str, fptr->rbuf+fptr->rbuf_off, 1);
        fptr->rbuf_off++;
        fptr->rbuf_len--;
        r = rb_enc_precise_mbclen(RSTRING_PTR(str), RSTRING_PTR(str)+RSTRING_LEN(str), enc);
        if (MBCLEN_NEEDMORE_P(r)) {
            goto getc_needmore;
        }
        else if (MBCLEN_CHARFOUND_P(r)) {
            cr = ENC_CODERANGE_VALID;
        }
      }
    }
    else {
      str = mrb_str_new(mrb, fptr->rbuf+fptr->rbuf_off, 1);
      fptr->rbuf_off++;
      fptr->rbuf_len--;
    }
  }
  if (!cr) cr = ENC_CODERANGE_BROKEN;
  str = io_enc_str(mrb, str, fptr);
  ENC_CODERANGE_SET(str, cr);
  return str;
}
#endif

int
rb_ensure(int (*b_proc)(ANYARGS), void * data1, mrb_value (*e_proc)(ANYARGS), mrb_value data2)
{
#if 0 /* THREAD */
  int state;
  int result = (int)NULL;

  PUSH_TAG();
  if ((state = EXEC_TAG()) == 0) {
    result = (*b_proc) (data1);
  }
  POP_TAG();
  /* TODO: fix me */
  /* retval = prot_tag ? prot_tag->retval : mrb_nil_value(); */     /* save retval */
  (*e_proc) (data2);
  if (state)
    JUMP_TAG(state);
  return result;
#else
  int result;

  result = (*b_proc) (data1);
  (*e_proc) (data2);
  return result;
#endif
}

static mrb_value
io_call_close(mrb_state *mrb, mrb_value io)
{
  return mrb_funcall(mrb, io, "close", 0, 0);
}

static mrb_value
io_close(mrb_state *mrb, mrb_value io)
{
  return rb_rescue(mrb, io_call_close, &io, 0, (mrb_value*)0);
}

static int
ruby_dup(int orig)
{
  int fd;

  fd = dup(orig);
  if (fd < 0) {
    if (errno == EMFILE || errno == ENFILE || errno == ENOMEM) {
        //rb_gc();
        fd = dup(orig);
    }
    if (fd < 0) {
        rb_sys_fail(0);
    }
  }
  UPDATE_MAXFD(fd);
  return fd;
}

#if defined(__CYGWIN__) || !defined(HAVE_FORK)
static struct pipe_list {
    rb_io_t *fptr;
    struct pipe_list *next;
} *pipe_list;

static void
pipe_add_fptr(mrb_state *mrb, rb_io_t *fptr)
{
    struct pipe_list *list;

    //list = ALLOC(struct pipe_list);
    list = mrb_malloc(mrb, sizeof(struct pipe_list));
    list->fptr = fptr;
    list->next = pipe_list;
    pipe_list = list;
}

static void
pipe_del_fptr(rb_io_t *fptr)
{
    struct pipe_list *list = pipe_list;
    struct pipe_list *tmp;

    if (list->fptr == fptr) {
      pipe_list = list->next;
      free(list);
      return;
    }

    while (list->next) {
      if (list->next->fptr == fptr) {
          tmp = list->next;
          list->next = list->next->next;
          free(tmp);
          return;
      }
      list = list->next;
    }
}

static mrb_value
finish_writeconv(mrb_state *mrb, rb_io_t *fptr, int noalloc)
{
    unsigned char *ds, *dp, *de;
    rb_econv_result_t res;

    if (!fptr->wbuf) {
        unsigned char buf[1024];
        long r;

        res = econv_destination_buffer_full;
        while (res == econv_destination_buffer_full) {
            ds = dp = buf;
            de = buf + sizeof(buf);
            res = rb_econv_convert(mrb, fptr->writeconv, NULL, NULL, &dp, de, 0);
            while (dp-ds) {
              //retry:
                r = rb_write_internal(mrb, fptr->fd, ds, dp-ds);
                if (r == dp-ds)
                    break;
                if (0 <= r) {
                    ds += r;
                }
                //if (rb_io_wait_writable(mrb, fptr->fd)) {
                //    if (fptr->fd < 0)
                //        return noalloc ? mrb_true_value() : rb_exc_new3(mrb, mrb->eIOError_class, mrb_str_new_cstr(mrb, "closed stream"));
                //    goto retry;
                //}
                return noalloc ? mrb_true_value() : mrb_fixnum_value(errno);
            }
            if (res == econv_invalid_byte_sequence ||
                res == econv_incomplete_input ||
                res == econv_undefined_conversion) {
                return noalloc ? mrb_true_value() : rb_econv_make_exception(mrb, fptr->writeconv);
            }
        }

        return mrb_nil_value();
    }

    res = econv_destination_buffer_full;
    while (res == econv_destination_buffer_full) {
        if (fptr->wbuf_len == fptr->wbuf_capa) {
            if (io_fflush(mrb, fptr) < 0)
                return noalloc ? mrb_true_value() : mrb_fixnum_value(errno);
        }

        ds = dp = (unsigned char *)fptr->wbuf + fptr->wbuf_off + fptr->wbuf_len;
        de = (unsigned char *)fptr->wbuf + fptr->wbuf_capa;
        res = rb_econv_convert(mrb, fptr->writeconv, NULL, NULL, &dp, de, 0);
        fptr->wbuf_len += (int)(dp - ds);
        if (res == econv_invalid_byte_sequence ||
            res == econv_incomplete_input ||
            res == econv_undefined_conversion) {
            return noalloc ? mrb_true_value() : rb_econv_make_exception(mrb, fptr->writeconv);
        }
    }
    return mrb_nil_value();
}

struct finish_writeconv_arg {
    rb_io_t *fptr;
    int noalloc;
};

#if 0
static int
finish_writeconv_sync(mrb_state *mrb, void *arg)
{
    struct finish_writeconv_arg *p = (struct finish_writeconv_arg *)arg;
    return mrb_fixnum_number(finish_writeconv(mrb, p->fptr, p->noalloc));
}
#endif

static void
fptr_finalize(mrb_state *mrb, rb_io_t *fptr, int noraise)
{
  mrb_value err = mrb_nil_value();
  if (fptr->writeconv) {
    //if (fptr->write_lock.tt && !noraise) {
    //  struct finish_writeconv_arg arg;
    //  arg.fptr = fptr;
    //  arg.noalloc = noraise;
    //  err = mrb_fixnum_value(rb_mutex_synchronize(mrb, fptr->write_lock, finish_writeconv_sync, (void *)&arg));
    //}
    //else {
      err = finish_writeconv(mrb, fptr, noraise);
    //}
  }
  if (fptr->wbuf_len) {
    if (noraise) {
      if ((int)io_flush_buffer_sync(fptr) < 0 && mrb_nil_p(err))
        err = mrb_true_value();
    }
    else {
      if (io_fflush(mrb, fptr) < 0 && mrb_nil_p(err))
        err = mrb_fixnum_value(errno);
    }
  }
  if (IS_PREP_STDIO(fptr) || fptr->fd <= 2) {
    goto skip_fd_close;
  }
  if (fptr->stdio_file) {
    /* fptr->stdio_file is deallocated anyway
     * even if fclose failed.  */
    if (fclose(fptr->stdio_file) < 0 && mrb_nil_p(err))
      err = noraise ? mrb_true_value() : mrb_fixnum_value(errno);
  }
  else if (0 <= fptr->fd) {
    /* fptr->fd may be closed even if close fails.
     * POSIX doesn't specify it.
     * We assumes it is closed.  */
    if (close(fptr->fd) < 0 && mrb_nil_p(err))
      err = noraise ? mrb_true_value() : mrb_fixnum_value(errno);
  }
skip_fd_close:
  fptr->fd = -1;
  fptr->stdio_file = 0;
  fptr->mode &= ~(FMODE_READABLE|FMODE_WRITABLE);

  if (!mrb_nil_p(err) && !noraise) {
    switch(mrb_type(err)) {
      case MRB_TT_FIXNUM:
        errno = mrb_fixnum_number(err);
        rb_sys_fail_path(fptr->pathv);

      default:
        rb_exc_raise(mrb, err);
    }
  }
}

static void
rb_io_fptr_cleanup(mrb_state *mrb, rb_io_t *fptr, int noraise)
{
  if (fptr->finalize) {
    (*fptr->finalize)(fptr, noraise);
  }
  else {
    fptr_finalize(mrb, fptr, noraise);
  }
}

int
rb_io_fptr_finalize(mrb_state *mrb, rb_io_t *fptr)
{
  if (!fptr) return 0;
  fptr->pathv = mrb_nil_value();
  if (0 <= fptr->fd)
    rb_io_fptr_cleanup(mrb, fptr, TRUE);
  fptr->write_lock = mrb_fixnum_value(0);
  if (fptr->rbuf) {
    free(fptr->rbuf);
    fptr->rbuf = 0;
  }
  if (fptr->wbuf) {
    free(fptr->wbuf);
    fptr->wbuf = 0;
  }
  clear_codeconv(fptr);
  free(fptr);
  return 1;
}

#if 0
static void
pipe_atexit(void)
{
    struct pipe_list *list = pipe_list;
    struct pipe_list *tmp;

    while (list) {
      tmp = list->next;
      rb_io_fptr_finalize(mrb, list->fptr);
      list = tmp;
    }
}
#endif

static void
pipe_finalize(mrb_state *mrb, rb_io_t *fptr, int noraise)
{
#if !defined(HAVE_FORK) && !defined(_WIN32)
  int status = 0;
  if (fptr->stdio_file) {
    status = pclose(fptr->stdio_file);
  }
  fptr->fd = -1;
  fptr->stdio_file = 0;
  //rb_last_status_set(mrb, status, fptr->pid);
#else
  fptr_finalize(mrb, fptr, noraise);
#endif
  pipe_del_fptr(fptr);
}
#endif

static rb_encoding*
io_read_encoding(mrb_state *mrb, rb_io_t *fptr)
{
  if (fptr->encs.enc) {
    return fptr->encs.enc;
  }
  return rb_default_external_encoding(mrb);
}

#ifdef INCLUDE_REGEXP
static rb_encoding*
io_input_encoding(mrb_state *mrb, rb_io_t *fptr)
{
  if (fptr->encs.enc2) {
    return fptr->encs.enc2;
  }
  return io_read_encoding(mrb, fptr);
}
#endif

mrb_value
rb_io_binmode(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;

    GetOpenFile(mrb, io, fptr);
    if (fptr->readconv)
        rb_econv_binmode(fptr->readconv);
    if (fptr->writeconv)
        rb_econv_binmode(fptr->writeconv);
    fptr->mode |= FMODE_BINMODE;
    fptr->mode &= ~FMODE_TEXTMODE;
    fptr->writeconv_pre_ecflags &= ~(ECONV_UNIVERSAL_NEWLINE_DECORATOR|ECONV_CRLF_NEWLINE_DECORATOR|ECONV_CR_NEWLINE_DECORATOR);
    return io;
}

mrb_value
rb_io_get_io(mrb_state *mrb, mrb_value io)
{
  return mrb_convert_type(mrb, io, MRB_TT_FILE, "IO", "to_io");
}

/*
 *  call-seq:
 *     ios.getbyte   -> fixnum or nil
 *
 *  Gets the next 8-bit byte (0..255) from <em>ios</em>. Returns
 *  <code>nil</code> if called at end of file.
 *
 *     f = File.new("testfile")
 *     f.getbyte   #=> 84
 *     f.getbyte   #=> 104
 */

mrb_value
rb_io_getbyte(mrb_state *mrb, mrb_value io)
{
  rb_io_t *fptr;
  int c;

  GetOpenFile(mrb, io, fptr);
  rb_io_check_byte_readable(mrb, fptr);
  READ_CHECK(mrb, fptr);
  if (fptr->fd == 0 && (fptr->mode & FMODE_TTY) && mrb_type(rb_stdout) == MRB_TT_FILE) {
    rb_io_t *ofp;
    GetOpenFile(mrb, rb_stdout, ofp);
    if (ofp->mode & FMODE_TTY) {
        mrb_io_flush(mrb, rb_stdout);
    }
  }
  if (io_fillbuf(mrb, fptr) < 0) {
    return mrb_nil_value();
  }
  fptr->rbuf_off++;
  fptr->rbuf_len--;
  c = (unsigned char)fptr->rbuf[fptr->rbuf_off-1];
  return mrb_fixnum_value(c & 0xff);
}

int
rb_io_modestr_fmode(mrb_state *mrb, const char *modestr)
{
    int fmode = 0;
    const char *m = modestr, *p = NULL;

    switch (*m++) {
      case 'r':
      fmode |= FMODE_READABLE;
      break;
      case 'w':
      fmode |= FMODE_WRITABLE | FMODE_TRUNC | FMODE_CREATE;
      break;
      case 'a':
      fmode |= FMODE_WRITABLE | FMODE_APPEND | FMODE_CREATE;
      break;
      default:
      error:
      mrb_raise(mrb, mrb->eArgumentError_class, "invalid access mode %s", modestr);
    }

    while (*m) {
        switch (*m++) {
        case 'b':
            fmode |= FMODE_BINMODE;
            break;
        case 't':
            fmode |= FMODE_TEXTMODE;
            break;
        case '+':
            fmode |= FMODE_READWRITE;
            break;
        default:
            goto error;
        case ':':
          p = m;
            goto finished;
        }
    }

  finished:
    if ((fmode & FMODE_BINMODE) && (fmode & FMODE_TEXTMODE))
        goto error;
    if (p && io_encname_bom_p(p, 0))
      fmode |= FMODE_SETENC_BY_BOM;

    return fmode;
}

int
rb_io_oflags_fmode(int oflags)
{
    int fmode = 0;

    switch (oflags & (O_RDONLY|O_WRONLY|O_RDWR)) {
      case O_RDONLY:
      fmode = FMODE_READABLE;
      break;
      case O_WRONLY:
      fmode = FMODE_WRITABLE;
      break;
      case O_RDWR:
      fmode = FMODE_READWRITE;
      break;
    }

    if (oflags & O_APPEND) {
      fmode |= FMODE_APPEND;
    }
    if (oflags & O_TRUNC) {
      fmode |= FMODE_TRUNC;
    }
    if (oflags & O_CREAT) {
      fmode |= FMODE_CREATE;
    }
#ifdef O_BINARY
    if (oflags & O_BINARY) {
      fmode |= FMODE_BINMODE;
    }
#endif

    return fmode;
}

/*
 *  call-seq:
 *     ios.ungetbyte(string)   -> nil
 *     ios.ungetbyte(integer)   -> nil
 *
 *  Pushes back bytes (passed as a parameter) onto <em>ios</em>,
 *  such that a subsequent buffered read will return it. Only one byte
 *  may be pushed back before a subsequent read operation (that is,
 *  you will be able to read only the last of several bytes that have been pushed
 *  back). Has no effect with unbuffered reads (such as <code>IO#sysread</code>).
 *
 *     f = File.new("testfile")   #=> #<File:testfile>
 *     b = f.getbyte              #=> 0x38
 *     f.ungetbyte(b)             #=> nil
 *     f.getbyte                  #=> 0x38
 */

mrb_value
rb_io_ungetbyte(mrb_state *mrb, mrb_value io, mrb_value b)
{
    rb_io_t *fptr;

    GetOpenFile(mrb, io, fptr);
    rb_io_check_byte_readable(mrb, fptr);
    if (mrb_nil_p(b)) return mrb_nil_value();
    if (FIXNUM_P(b)) {
      char cc = mrb_fixnum_number(b);
      b = mrb_str_new(mrb, &cc, 1);
    }
    else {
      //SafeStringValue(b);
      mrb_string_value(mrb, &b);
    }
    io_ungetbyte(b, fptr);
    return mrb_nil_value();
}

void
rb_eof_error(mrb_state *mrb)
{
    mrb_raise(mrb, mrb->eEOFError_class, "end of file reached");
}

static void
io_ungetbyte(mrb_value str, rb_io_t *fptr)
{
  long len = RSTRING_LEN(str);

  if (fptr->rbuf == NULL) {
    const int min_capa = IO_RBUF_CAPA_FOR(fptr);
    fptr->rbuf_off = 0;
    fptr->rbuf_len = 0;
#if SIZEOF_LONG > SIZEOF_INT
    if (len > INT_MAX)
      rb_raise(rb_eIOError, "ungetbyte failed");
#endif
    if (len > min_capa)
      fptr->rbuf_capa = (int)len;
    else
      fptr->rbuf_capa = min_capa;
    //fptr->rbuf = ALLOC_N(char, fptr->rbuf_capa);
    fptr->rbuf = malloc(fptr->rbuf_capa);
  }
  if (fptr->rbuf_capa < len + fptr->rbuf_len) {
    rb_raise(rb_eIOError, "ungetbyte failed");
  }
  if (fptr->rbuf_off < len) {
    MEMMOVE(fptr->rbuf+fptr->rbuf_capa-fptr->rbuf_len,
            fptr->rbuf+fptr->rbuf_off,
            char, fptr->rbuf_len);
    fptr->rbuf_off = fptr->rbuf_capa-fptr->rbuf_len;
  }
  fptr->rbuf_off-=(int)len;
  fptr->rbuf_len+=(int)len;
  MEMMOVE(fptr->rbuf+fptr->rbuf_off, RSTRING_PTR(str), char, len);
}

static mrb_value
io_alloc(mrb_state *mrb, struct RClass *klass)
{
  //NEWOBJ(io, struct RFile);
  //OBJSETUP(io, klass, T_FILE);
  struct RFile* io;
  io = mrb_obj_alloc(mrb, MRB_TT_FILE, klass);

  io->fptr = 0;

  return mrb_obj_value(io);
}

static void
must_respond_to(mrb_state *mrb, mrb_sym mid, mrb_value val, mrb_sym id)
{
  if (!mrb_respond_to(mrb, val, mid)) {
    mrb_raise(mrb, mrb->eTypeError_class, "%s must have %s method, %s given",
        mrb_sym2name(mrb, id), mrb_sym2name(mrb, mid),//rb_id2name(id), rb_id2name(mid),
        mrb_obj_classname(mrb, val));
    }
}

static void
stdout_setter(mrb_state *mrb, mrb_value val, mrb_sym id, mrb_value *variable)
{
  must_respond_to(mrb, mrb_intern(mrb, "write"), val, id);
  *variable = val;
}

static mrb_value
prep_io(mrb_state *mrb, int fd, int fmode, struct RClass* klass, const char *path)
{
    rb_io_t *fp;
    mrb_value io = io_alloc(mrb, klass);

    MakeOpenFile(mrb, io, fp);
    fp->fd = fd;
#ifdef __CYGWIN__
    if (!isatty(fd)) {
        fmode |= FMODE_BINMODE;
        setmode(fd, O_BINARY);
    }
#endif
    fp->mode = fmode;
    io_check_tty(fp);
    if (path) fp->pathv = /*rb_obj_freeze*/(mrb_str_new_cstr(mrb, path));

    return io;
}

static mrb_value
prep_stdio(mrb_state *mrb, FILE *f, int fmode, struct RClass* klass, const char *path)
{
    rb_io_t *fptr;
    mrb_value io = prep_io(mrb, fileno(f), fmode|FMODE_PREP, klass, path);

    GetOpenFile(mrb, io, fptr);
    fptr->stdio_file = f;

    return io;
}

mrb_value
rb_io_close(mrb_state *mrb, mrb_value io)
{
  rb_io_t *fptr;
  int fd;
  mrb_value write_io;
  rb_io_t *write_fptr;

  write_io = GetWriteIO(mrb, io);
  if (!mrb_obj_equal(mrb, io, write_io)) {
      write_fptr = RFILE(write_io)->fptr;
      if (write_fptr && (0 <= write_fptr->fd)) {
          rb_io_fptr_cleanup(mrb, write_fptr, TRUE);
      }
  }

  fptr = RFILE(io)->fptr;
  if (!fptr) return mrb_nil_value();
  if (fptr->fd < 0) return mrb_nil_value();

  fd = fptr->fd;
  rb_io_fptr_cleanup(mrb, fptr, FALSE);
  //rb_thread_fd_close(fd);

  if (fptr->pid) {
    rb_syswait(fptr->pid);
    fptr->pid = 0;
  }

  return mrb_nil_value();
}

/* 15.2.20.4.1  */
/*
 *  Document-method: IO::open
 *
 *  call-seq:
 *     IO.open(fd, mode_string="r" [, opt] )               -> io
 *     IO.open(fd, mode_string="r" [, opt] ) {|io| block } -> obj
 *
 *  With no associated block, <code>open</code> is a synonym for
 *  <code>IO.new</code>. If the optional code block is given, it will
 *  be passed <i>io</i> as an argument, and the IO object will
 *  automatically be closed when the block terminates. In this instance,
 *  <code>IO.open</code> returns the value of the block.
 *
 */

static mrb_value
mrb_io_s_open(mrb_state *mrb, /*int argc, mrb_value *argv, */mrb_value klass)
{
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);
#if 0 /* original */
  mrb_value io = rb_class_new_instance(mrb, argc, argv, mrb_class_get(mrb, klass));
#else
  //SCLASS mrb_value io = mrb_obj_value(mrb_obj_alloc(mrb, mrb_class_get(mrb, klass)->tt, mrb_class_get(mrb, klass)));
  mrb_value io = mrb_obj_value(mrb_obj_alloc(mrb, MRB_TT_FILE, mrb_class_ptr(klass)));
  rb_obj_call_init(mrb, io, argc, argv);
#endif
  if (rb_block_given_p()) {
    return mrb_fixnum_value(rb_ensure(int_rb_yield, &io, io_close, io));
  }

  return io;
}

/* 15.2.20.5.1  */
/*
 *  call-seq:
 *     ios.close   -> nil
 *
 *  Closes <em>ios</em> and flushes any pending writes to the operating
 *  system. The stream is unavailable for any further data operations;
 *  an <code>IOError</code> is raised if such an attempt is made. I/O
 *  streams are automatically closed when they are claimed by the
 *  garbage collector.
 *
 *  If <em>ios</em> is opened by <code>IO.popen</code>,
 *  <code>close</code> sets <code>$?</code>.
 */

mrb_value
mrb_io_close_m(mrb_state *mrb, mrb_value io)
{
  //if (rb_safe_level() >= 4 && !OBJ_UNTRUSTED(io)) {
  //  mrb_raise(mrb, mrb->eSecurityError, "Insecure: can't close");
  //}
  rb_io_check_closed(mrb, RFILE(io)->fptr);
  rb_io_close(mrb, io);
  return mrb_nil_value();
}
/* 15.2.20.5.2  */
/*
 *  call-seq:
 *     ios.closed?    -> true or false
 *
 *  Returns <code>true</code> if <em>ios</em> is completely closed (for
 *  duplex streams, both reader and writer), <code>false</code>
 *  otherwise.
 *
 *     f = File.new("testfile")
 *     f.close         #=> nil
 *     f.closed?       #=> true
 *     f = IO.popen("/bin/sh","r+")
 *     f.close_write   #=> nil
 *     f.closed?       #=> false
 *     f.close_read    #=> nil
 *     f.closed?       #=> true
 */


static mrb_value
mrb_io_closed(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;
    mrb_value write_io;
    rb_io_t *write_fptr;

    write_io = GetWriteIO(mrb, io);
    if (!mrb_obj_equal(mrb, io, write_io)) {
        write_fptr = RFILE(write_io)->fptr;
        if (write_fptr && 0 <= write_fptr->fd) {
            return mrb_false_value();
        }
    }

    fptr = RFILE(io)->fptr;
    mrb_io_check_initialized(mrb, fptr);
    return 0 <= fptr->fd ? mrb_false_value() : mrb_true_value();
}

/* 15.2.20.5.3  */
/* 15.2.20.5.5  */
/*
 *  call-seq:
 *     ios.each(sep=$/) {|line| block }         -> ios
 *     ios.each(limit) {|line| block }          -> ios
 *     ios.each(sep,limit) {|line| block }      -> ios
 *     ios.each(...)                            -> an_enumerator
 *
 *     ios.each_line(sep=$/) {|line| block }    -> ios
 *     ios.each_line(limit) {|line| block }     -> ios
 *     ios.each_line(sep,limit) {|line| block } -> ios
 *     ios.each_line(...)                       -> an_enumerator
 *
 *     ios.lines(sep=$/) {|line| block }        -> ios
 *     ios.lines(limit) {|line| block }         -> ios
 *     ios.lines(sep,limit) {|line| block }     -> ios
 *     ios.lines(...)                           -> an_enumerator
 *
 *  Executes the block for every line in <em>ios</em>, where lines are
 *  separated by <i>sep</i>. <em>ios</em> must be opened for
 *  reading or an <code>IOError</code> will be raised.
 *
 *  If no block is given, an enumerator is returned instead.
 *
 *     f = File.new("testfile")
 *     f.each {|line| puts "#{f.lineno}: #{line}" }
 *
 *  <em>produces:</em>
 *
 *     1: This is line one
 *     2: This is line two
 *     3: This is line three
 *     4: And so on...
 */

static mrb_value
mrb_io_each_line(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value io)
{
  mrb_value str, rs;
  long limit;
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);

  //RETURN_ENUMERATOR(io, argc, argv);
  prepare_getline_args(mrb, argc, argv, &rs, &limit, io);
  while (!mrb_nil_p(str = rb_io_getline_1(mrb, rs, limit, io))) {
    rb_yield(str);
  }
  return io;
}

/* 15.2.20.5.4  */
/*
 *  call-seq:
 *     ios.bytes {|byte| block }      -> ios
 *     ios.bytes                      -> an_enumerator
 *
 *     ios.each_byte {|byte| block }  -> ios
 *     ios.each_byte                  -> an_enumerator
 *
 *  Calls the given block once for each byte (0..255) in <em>ios</em>,
 *  passing the byte as an argument. The stream must be opened for
 *  reading or an <code>IOError</code> will be raised.
 *
 *  If no block is given, an enumerator is returned instead.
 *
 *     f = File.new("testfile")
 *     checksum = 0
 *     f.each_byte {|x| checksum ^= x }   #=> #<File:testfile>
 *     checksum                           #=> 12
 */

static mrb_value
mrb_io_each_byte(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;
    char *p, *e;

    //RETURN_ENUMERATOR(io, 0, 0);
    GetOpenFile(mrb, io, fptr);

    for (;;) {
      p = fptr->rbuf+fptr->rbuf_off;
      e = p + fptr->rbuf_len;
      while (p < e) {
          fptr->rbuf_off++;
          fptr->rbuf_len--;
          rb_yield(mrb_fixnum_value(*p & 0xff));
          p++;
          errno = 0;
      }
      rb_io_check_byte_readable(mrb, fptr);
      READ_CHECK(mrb, fptr);
      if (io_fillbuf(mrb, fptr) < 0) {
          break;
      }
    }
    return io;
}

/* 15.2.20.5.6  */
/*
 *  call-seq:
 *     ios.eof     -> true or false
 *     ios.eof?    -> true or false
 *
 *  Returns true if <em>ios</em> is at end of file that means
 *  there are no more data to read.
 *  The stream must be opened for reading or an <code>IOError</code> will be
 *  raised.
 *
 *     f = File.new("testfile")
 *     dummy = f.readlines
 *     f.eof   #=> true
 *
 *  If <em>ios</em> is a stream such as pipe or socket, <code>IO#eof?</code>
 *  blocks until the other end sends some data or closes it.
 *
 *     r, w = IO.pipe
 *     Thread.new { sleep 1; w.close }
 *     r.eof?  #=> true after 1 second blocking
 *
 *     r, w = IO.pipe
 *     Thread.new { sleep 1; w.puts "a" }
 *     r.eof?  #=> false after 1 second blocking
 *
 *     r, w = IO.pipe
 *     r.eof?  # blocks forever
 *
 *  Note that <code>IO#eof?</code> reads data to the input byte buffer.
 *  So <code>IO#sysread</code> may not behave as you intend with
 *  <code>IO#eof?</code>, unless you call <code>IO#rewind</code>
 *  first (which is not available for some streams).
 */

mrb_value

mrb_io_eof(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;

    GetOpenFile(mrb, io, fptr);
    rb_io_check_char_readable(mrb, fptr);

    if (READ_CHAR_PENDING(fptr)) return mrb_false_value();
    if (READ_DATA_PENDING(fptr)) return mrb_false_value();
    READ_CHECK(mrb, fptr);
    if (io_fillbuf(mrb, fptr) < 0) {
      return mrb_true_value();
    }
    return mrb_false_value();
}

/* 15.2.20.5.7  */
/*
 *  call-seq:
 *     ios.flush    -> ios
 *
 *  Flushes any buffered data within <em>ios</em> to the underlying
 *  operating system (note that this is Ruby internal buffering only;
 *  the OS may buffer the data as well).
 *
 *     $stdout.print "no newline"
 *     $stdout.flush
 *
 *  <em>produces:</em>
 *
 *     no newline
 */

mrb_value
mrb_io_flush(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;

    if (mrb_type(io) != MRB_TT_FILE) {
        return mrb_funcall(mrb, io, "flush", 0);
    }

    io = GetWriteIO(mrb, io);
    GetOpenFile(mrb, io, fptr);

    if (fptr->mode & FMODE_WRITABLE) {
        if (io_fflush(mrb, fptr) < 0)
            rb_sys_fail(0);
#ifdef _WIN32
      fsync(fptr->fd);
#endif
    }
    if (fptr->mode & FMODE_READABLE) {
        io_unread(mrb, fptr);
    }

    return io;
}

/* 15.2.20.5.8  */
/*
 *  call-seq:
 *     ios.getc   -> string or nil
 *
 *  Reads a one-character string from <em>ios</em>. Returns
 *  <code>nil</code> if called at end of file.
 *
 *     f = File.new("testfile")
 *     f.getc   #=> "h"
 *     f.getc   #=> "e"
 */

static mrb_value
mrb_io_getc(mrb_state *mrb, mrb_value io)
{
  rb_io_t *fptr;
  rb_encoding *enc;

  GetOpenFile(mrb, io, fptr);
  rb_io_check_char_readable(mrb, fptr);

  enc = io_input_encoding(mrb, fptr);
  READ_CHECK(mrb, fptr);
  return io_getc(mrb, fptr, enc);
}

/* 15.2.20.5.9  */
/*
 *  call-seq:
 *     ios.gets(sep=$/)     -> string or nil
 *     ios.gets(limit)      -> string or nil
 *     ios.gets(sep, limit) -> string or nil
 *
 *  Reads the next ``line'' from the I/O stream; lines are separated by
 *  <i>sep</i>. A separator of <code>nil</code> reads the entire
 *  contents, and a zero-length separator reads the input a paragraph at
 *  a time (two successive newlines in the input separate paragraphs).
 *  The stream must be opened for reading or an <code>IOError</code>
 *  will be raised. The line read in will be returned and also assigned
 *  to <code>$_</code>. Returns <code>nil</code> if called at end of
 *  file.  If the first argument is an integer, or optional second
 *  argument is given, the returning string would not be longer than the
 *  given value in bytes.
 *
 *     File.new("testfile").gets   #=> "This is line one\n"
 *     $_                          #=> "This is line one\n"
 */

static mrb_value
mrb_io_gets_m(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value io)
{
  mrb_value str;
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);

  str = rb_io_getline(mrb, argc, argv, io);
  rb_lastline_set(str);

  return str;
}

/* 15.2.20.5.10 */
/* :nodoc: */
static mrb_value
mrb_io_init_copy(mrb_state *mrb, mrb_value dest)
{
  rb_io_t *fptr, *orig;
  int fd;
  mrb_value write_io;
  off_t pos;
  mrb_value io;
  mrb_value *pwrite_io;

  mrb_get_args(mrb, "o", &io);

  io = rb_io_get_io(mrb, io);
  if (mrb_obj_equal(mrb, dest, io)) return dest;
  GetOpenFile(mrb, io, orig);
  MakeOpenFile(mrb, dest, fptr);

  mrb_io_flush(mrb, io);

  /* copy rb_io_t structure */
  fptr->mode = orig->mode & ~FMODE_PREP;
  fptr->encs = orig->encs;
  fptr->pid = orig->pid;
  fptr->lineno = orig->lineno;
  if (!mrb_nil_p(orig->pathv)) fptr->pathv = orig->pathv;
  fptr->finalize = orig->finalize;
#if defined (__CYGWIN__) || !defined(HAVE_FORK)
  if ((void*)fptr->finalize == (void*)pipe_finalize)
    pipe_add_fptr(mrb, fptr);
#endif

  fd = ruby_dup(orig->fd);
  fptr->fd = fd;
  pos = io_tell(mrb, orig);
  if (0 <= pos)
    io_seek(mrb, fptr, pos, SEEK_SET);
  if (fptr->mode & FMODE_BINMODE) {
    rb_io_binmode(mrb, dest);
  }

  write_io = GetWriteIO(mrb, io);
  if (!mrb_obj_equal(mrb, io, write_io)) {
    pwrite_io = mrb_obj_alloc(mrb, mrb_type(write_io), mrb_obj_class(mrb, write_io));
    fptr->tied_io_for_writing = *pwrite_io;
    mrb_iv_set(mrb, dest, mrb_intern(mrb, "@tied_io_for_writing"),*pwrite_io);
  }

  return dest;
}

/* 15.2.20.5.11 */
/*
 *  call-seq:
 *     ios.print()             -> nil
 *     ios.print(obj, ...)     -> nil
 *
 *  Writes the given object(s) to <em>ios</em>. The stream must be
 *  opened for writing. If the output field separator (<code>$,</code>)
 *  is not <code>nil</code>, it will be inserted between each object.
 *  If the output record separator (<code>$\\</code>)
 *  is not <code>nil</code>, it will be appended to the output. If no
 *  arguments are given, prints <code>$_</code>. Objects that aren't
 *  strings will be converted by calling their <code>to_s</code> method.
 *  With no argument, prints the contents of the variable <code>$_</code>.
 *  Returns <code>nil</code>.
 *
 *     $stdout.print("This is ", 100, " percent.\n")
 *
 *  <em>produces:</em>
 *
 *     This is 100 percent.
 */

mrb_value
mrb_io_print(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value out)
{
  int i;
  mrb_value line;
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);

  /* if no argument given, print `$_' */
  if (argc == 0) {
    argc = 1;
    line = rb_lastline_get(mrb);
    argv[0] = line;
  }
  for (i=0; i<argc; i++) {
    if (mrb_type(out) == MRB_TT_FILE /*&&
      rb_method_basic_definition_p(CLASS_OF(rb_stdout), id_write)*/) {
      io_write(mrb, out, argv[i], 1);
    }
    else {
      rb_io_write(mrb, out, argv[i]);
    }
  }
  if (mrb_type(out) == MRB_TT_FILE) {
    mrb_io_flush(mrb, out);
  }

  return mrb_nil_value();
}

/* 15.2.20.5.12 */
/*
 *  call-seq:
 *     ios.putc(obj)    -> obj
 *
 *  If <i>obj</i> is <code>Numeric</code>, write the character whose code is
 *  the least-significant byte of <i>obj</i>, otherwise write the first byte
 *  of the string representation of <i>obj</i> to <em>ios</em>. Note: This
 *  method is not safe for use with multi-byte characters as it will truncate
 *  them.
 *
 *     $stdout.putc "A"
 *     $stdout.putc 65
 *
 *  <em>produces:</em>
 *
 *     AA
 */

static mrb_value
mrb_io_putc(mrb_state *mrb, mrb_value io)
{
  char c;
  mrb_value ch;

  mrb_get_args(mrb, "o", &ch);
  c = NUM2CHR(ch);

  rb_io_write(mrb, io, mrb_str_new(mrb, &c, 1));
  return ch;
}

/* 15.2.20.5.13 */
/*
 *  call-seq:
 *     ios.puts(obj, ...)    -> nil
 *
 *  Writes the given objects to <em>ios</em> as with
 *  <code>IO#print</code>. Writes a record separator (typically a
 *  newline) after any that do not already end with a newline sequence.
 *  If called with an array argument, writes each element on a new line.
 *  If called without arguments, outputs a single record separator.
 *
 *     $stdout.puts("this", "is", "a", "test")
 *
 *  <em>produces:</em>
 *
 *     this
 *     is
 *     a
 *     test
 */

mrb_value
mrb_io_puts(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value out)
{
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);

  return io_puts(mrb, argc, argv, out);
}

/* 15.2.20.5.14 */
/*
 *  call-seq:
 *     ios.read([length [, buffer]])    -> string, buffer, or nil
 *
 *  Reads <i>length</i> bytes from the I/O stream.
 *
 *  <i>length</i> must be a non-negative integer or <code>nil</code>.
 *
 *  If <i>length</i> is a positive integer,
 *  it try to read <i>length</i> bytes without any conversion (binary mode).
 *  It returns <code>nil</code> or a string whose length is 1 to <i>length</i> bytes.
 *  <code>nil</code> means it met EOF at beginning.
 *  The 1 to <i>length</i>-1 bytes string means it met EOF after reading the result.
 *  The <i>length</i> bytes string means it doesn't meet EOF.
 *  The resulted string is always ASCII-8BIT encoding.
 *
 *  If <i>length</i> is omitted or is <code>nil</code>,
 *  it reads until EOF and the encoding conversion is applied.
 *  It returns a string even if EOF is met at beginning.
 *
 *  If <i>length</i> is zero, it returns <code>""</code>.
 *
 *  If the optional <i>buffer</i> argument is present, it must reference
 *  a String, which will receive the data.
 *
 *  At end of file, it returns <code>nil</code> or <code>""</code>
 *  depend on <i>length</i>.
 *  <code><i>ios</i>.read()</code> and
 *  <code><i>ios</i>.read(nil)</code> returns <code>""</code>.
 *  <code><i>ios</i>.read(<i>positive-integer</i>)</code> returns <code>nil</code>.
 *
 *     f = File.new("testfile")
 *     f.read(16)   #=> "This is line one"
 *
 *     # reads whole file
 *     open("file") {|f|
 *       data = f.read # This returns a string even if the file is empty.
 *       ...
 *     }
 *
 *     # iterate over fixed length records.
 *     open("fixed-record-file") {|f|
 *       while record = f.read(256)
 *         ...
 *       end
 *     }
 *
 *     # iterate over variable length records.
 *     # record is prefixed by 32-bit length.
 *     open("variable-record-file") {|f|
 *       while len = f.read(4)
 *         len = len.unpack("N")[0] # 32-bit length
 *         record = f.read(len) # This returns a string even if len is 0.
 *       end
 *     }
 *
 *  Note that this method behaves like fread() function in C.
 *  If you need the behavior like read(2) system call,
 *  consider readpartial, read_nonblock and sysread.
 */

static mrb_value
mrb_io_read(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value io)
{
  mrb_value argv[16];
  int argc;

  //rb_scan_args(argc, argv, "02", &length, &str);
  mrb_get_args(mrb, "*", &argv, &argc);
  return io_read(mrb, argc, argv, io);

}

/* 15.2.20.5.15 */
/*
 *  call-seq:
 *     ios.readchar   -> string
 *
 *  Reads a one-character string from <em>ios</em>. Raises an
 *  <code>EOFError</code> on end of file.
 *
 *     f = File.new("testfile")
 *     f.readchar   #=> "h"
 *     f.readchar   #=> "e"
 */

static mrb_value
mrb_io_readchar(mrb_state *mrb, mrb_value io)
{
  mrb_value c = mrb_io_getc(mrb, io);

  if (mrb_nil_p(c)) {
    rb_eof_error(mrb);
  }
  return c;
}

/* 15.2.20.5.16 */
/*
 *  call-seq:
 *     ios.readline(sep=$/)     -> string
 *     ios.readline(limit)      -> string
 *     ios.readline(sep, limit) -> string
 *
 *  Reads a line as with <code>IO#gets</code>, but raises an
 *  <code>EOFError</code> on end of file.
 */

static mrb_value
mrb_io_readline(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value io)
{
  mrb_value line;

  line = mrb_io_gets_m(mrb, io);
  if (mrb_nil_p(line)) {
    rb_eof_error(mrb);
  }
  return line;
}

/* 15.2.20.5.17 */
/*
 *  call-seq:
 *     ios.readlines(sep=$/)     -> array
 *     ios.readlines(limit)      -> array
 *     ios.readlines(sep, limit) -> array
 *
 *  Reads all of the lines in <em>ios</em>, and returns them in
 *  <i>anArray</i>. Lines are separated by the optional <i>sep</i>. If
 *  <i>sep</i> is <code>nil</code>, the rest of the stream is returned
 *  as a single record.  If the first argument is an integer, or
 *  optional second argument is given, the returning string would not be
 *  longer than the given value in bytes. The stream must be opened for
 *  reading or an <code>IOError</code> will be raised.
 *
 *     f = File.new("testfile")
 *     f.readlines[0]   #=> "This is line one\n"
 */

static mrb_value
mrb_io_readlines(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value io)
{
  mrb_value line, ary, rs;
  long limit;
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);

  prepare_getline_args(mrb, argc, argv, &rs, &limit, io);
  ary = mrb_ary_new(mrb);
  while (!mrb_nil_p(line = rb_io_getline_1(mrb, rs, limit, io))) {
    mrb_ary_push(mrb, ary, line);
  }
  return ary;
}

/* 15.2.20.5.18 */
/*
 *  call-seq:
 *     ios.sync    -> true or false
 *
 *  Returns the current ``sync mode'' of <em>ios</em>. When sync mode is
 *  true, all output is immediately flushed to the underlying operating
 *  system and is not buffered by Ruby internally. See also
 *  <code>IO#fsync</code>.
 *
 *     f = File.new("testfile")
 *     f.sync   #=> false
 */

static mrb_value
mrb_io_sync(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;

    io = GetWriteIO(mrb, io);
    GetOpenFile(mrb, io, fptr);
    return (fptr->mode & FMODE_SYNC) ? mrb_true_value() : mrb_false_value();
}

/* 15.2.20.5.19 */
/*
 *  call-seq:
 *     ios.sync = boolean   -> boolean
 *
 *  Sets the ``sync mode'' to <code>true</code> or <code>false</code>.
 *  When sync mode is true, all output is immediately flushed to the
 *  underlying operating system and is not buffered internally. Returns
 *  the new state. See also <code>IO#fsync</code>.
 *
 *     f = File.new("testfile")
 *     f.sync = true
 *
 *  <em>(produces no output)</em>
 */

static mrb_value
mrb_io_set_sync(mrb_state *mrb, mrb_value io)
{
  rb_io_t *fptr;

  mrb_value sync;

  mrb_get_args(mrb, "o", &sync);
  io = GetWriteIO(mrb, io);
  GetOpenFile(mrb, io, fptr);
  if (mrb_test(sync)) {
    fptr->mode |= FMODE_SYNC;
  }
  else {
    fptr->mode &= ~FMODE_SYNC;
  }
  return sync;
}

/* 15.2.20.5.20 */
/*
 *  call-seq:
 *     ios.write(string)    -> integer
 *
 *  Writes the given string to <em>ios</em>. The stream must be opened
 *  for writing. If the argument is not a string, it will be converted
 *  to a string using <code>to_s</code>. Returns the number of bytes
 *  written.
 *
 *     count = $stdout.write( "This is a test\n" )
 *     puts "That was #{count} bytes of data"
 *
 *  <em>produces:</em>
 *
 *     This is a test
 *     That was 15 bytes of data
 */

static mrb_value
mrb_io_write_m(mrb_state *mrb, mrb_value io)
{
  mrb_value str;

  mrb_get_args(mrb, "o", &str);
  return io_write(mrb, io, str, 0);
}

/*
 *  call-seq:
 *     ios.to_io  ->  ios
 *
 *  Returns <em>ios</em>.
 */

static mrb_value
mrb_io_to_io(mrb_state *mrb, mrb_value io)
{
    return io;
}

static void
lineno_setter(mrb_value val, mrb_sym id, mrb_value *var)
{
    gets_lineno = mrb_fixnum_number(val);
    *var = mrb_fixnum_value(gets_lineno);
}

/* 15.2.20.5.23 (x)*/
/*
 * call-seq:
 *   ios.inspect   -> string
 *
 * Return a string describing this IO object.
 */

static mrb_value
mrb_io_inspect(mrb_state *mrb, mrb_value obj)
{
    rb_io_t *fptr;
    const char *cname;
    char fd_desc[4+sizeof(int)*3];
    const char *path;
    const char *st = "";

    fptr = RFILE(obj/*rb_io_taint_check(obj)*/)->fptr;
    if (!fptr) return mrb_any_to_s(mrb, obj);
    cname = mrb_obj_classname(mrb, obj);
    if (mrb_nil_p(fptr->pathv)) {
        if (fptr->fd < 0) {
            path = "";
            st = "(closed)";
        }
        else {
            sprintf(fd_desc, "fd %d", fptr->fd);
            path = fd_desc;
        }
    }
    else {
        path = RSTRING_PTR(fptr->pathv);
        if (fptr->fd < 0) {
            st = " (closed)";
        }
    }
    return rb_sprintf(mrb, "#<%s:%s%s>", cname, path, st);
}

void
mrb_init_io(mrb_state *mrb)
{
  struct RClass *c;
  c = mrb->io_class = mrb_define_class(mrb, "IO", mrb->object_class);

  mrb_define_class_method(mrb, c, "open",            mrb_io_s_open,     ARGS_ANY());    /* 15.2.20.4.1  */
  mrb_include_module(mrb, c, mrb->enumerable_module);
  mrb_define_method(mrb, c,       "close",           mrb_io_close_m,    ARGS_NONE());   /* 15.2.20.5.1  */
  mrb_define_method(mrb, c,       "closed?",         mrb_io_closed,     ARGS_NONE());   /* 15.2.20.5.2  */
  mrb_define_method(mrb, c,       "each",            mrb_io_each_line,  ARGS_ANY());    /* 15.2.20.5.3  */
  mrb_define_method(mrb, c,       "each_byte",       mrb_io_each_byte,  ARGS_NONE());   /* 15.2.20.5.4  */
  mrb_define_method(mrb, c,       "each_line",       mrb_io_each_line,  ARGS_ANY());    /* 15.2.20.5.5  */
  mrb_define_method(mrb, c,       "eof?",            mrb_io_eof,        ARGS_NONE());   /* 15.2.20.5.6  */
  mrb_define_method(mrb, c,       "flush",           mrb_io_flush,      ARGS_NONE());   /* 15.2.20.5.7  */
  mrb_define_method(mrb, c,       "getc",            mrb_io_getc,       ARGS_NONE());   /* 15.2.20.5.8  */
  mrb_define_method(mrb, c,       "gets",            mrb_io_gets_m,     ARGS_ANY());    /* 15.2.20.5.9  */
  mrb_define_method(mrb, c,       "initialize_copy", mrb_io_init_copy,  ARGS_REQ(1));   /* 15.2.20.5.10 */
  mrb_define_method(mrb, c,       "print",           mrb_io_print,      ARGS_ANY());    /* 15.2.20.5.11 */
  mrb_define_method(mrb, c,       "putc",            mrb_io_putc,       ARGS_REQ(1));   /* 15.2.20.5.12 */
  mrb_define_method(mrb, c,       "puts",            mrb_io_puts,       ARGS_ANY());    /* 15.2.20.5.13 */
  mrb_define_method(mrb, c,       "read",            mrb_io_read,       ARGS_ANY());    /* 15.2.20.5.14 */
  mrb_define_method(mrb, c,       "readchar",        mrb_io_readchar,   ARGS_NONE());   /* 15.2.20.5.15 */
  mrb_define_method(mrb, c,       "readline",        mrb_io_readline,   ARGS_ANY());    /* 15.2.20.5.16 */
  mrb_define_method(mrb, c,       "readlines",       mrb_io_readlines,  ARGS_ANY());    /* 15.2.20.5.17 */
  mrb_define_method(mrb, c,       "sync",            mrb_io_sync,       ARGS_NONE());   /* 15.2.20.5.18 */
  mrb_define_method(mrb, c,       "sync=",           mrb_io_set_sync,   ARGS_REQ(1));   /* 15.2.20.5.19 */
  mrb_define_method(mrb, c,       "write",           mrb_io_write_m,    ARGS_REQ(1));   /* 15.2.20.5.20 */

  mrb_define_method(mrb, c,       "initialize",      mrb_io_initialize, ARGS_ANY());    /* 15.2.20.5.21 (x)*/
  mrb_define_method(mrb, c,       "to_io",           mrb_io_to_io,      ARGS_NONE());   /* 15.2.20.5.22 (x)*/
  mrb_define_method(mrb, c,       "inspect",         mrb_io_inspect,    ARGS_NONE());   /* 15.2.20.5.23 (x)*/

  mrb_define_variable(mrb, "$stdin", &rb_stdin);
  rb_stdin  = prep_stdio(mrb, stdin, FMODE_READABLE, c, "<STDIN>");
  rb_define_hooked_variable(mrb, "$stdout", &rb_stdout, 0, stdout_setter);
  rb_stdout = prep_stdio(mrb, stdout, FMODE_WRITABLE, c, "<STDOUT>");
  rb_define_hooked_variable(mrb, "$stderr", &rb_stderr, 0, stdout_setter);
  rb_stderr = prep_stdio(mrb, stderr, FMODE_WRITABLE|FMODE_SYNC, c, "<STDERR>");
  rb_define_hooked_variable(mrb, "$>",  &rb_stdout,    0, stdout_setter);

  rb_define_hooked_variable(mrb, "$/",  &rb_rs,        0, rb_str_setter);
  rb_define_hooked_variable(mrb, "$-0", &rb_rs,        0, rb_str_setter);

  rb_define_hooked_variable(mrb, "$.",  &lineno,       0, lineno_setter);
  rb_define_virtual_variable(mrb, "$_", rb_lastline_get, rb_lastline_set);

  argf = argf_alloc(mrb, mrb_nil_value());

  lineno.tt = MRB_TT_FIXNUM;
  lineno.value.i = 0;

}

/* for IO */

#if defined(NFDBITS) && defined(HAVE_RB_FD_INIT)

/*
 * several Unix platforms support file descriptors bigger than FD_SETSIZE
 * in select(2) system call.
 *
 * - Linux 2.2.12 (?)
 * - NetBSD 1.2 (src/sys/kern/sys_generic.c:1.25)
 *   select(2) documents how to allocate fd_set dynamically.
 *   http://netbsd.gw.com/cgi-bin/man-cgi?select++NetBSD-4.0
 * - FreeBSD 2.2 (src/sys/kern/sys_generic.c:1.19)
 * - OpenBSD 2.0 (src/sys/kern/sys_generic.c:1.4)
 *   select(2) documents how to allocate fd_set dynamically.
 *   http://www.openbsd.org/cgi-bin/man.cgi?query=select&manpath=OpenBSD+4.4
 * - HP-UX documents how to allocate fd_set dynamically.
 *   http://docs.hp.com/en/B2355-60105/select.2.html
 * - Solaris 8 has select_large_fdset
 *
 * When fd_set is not big enough to hold big file descriptors,
 * it should be allocated dynamically.
 * Note that this assumes fd_set is structured as bitmap.
 *
 * rb_fd_init allocates the memory.
 * rb_fd_term free the memory.
 * rb_fd_set may re-allocates bitmap.
 *
 * So rb_fd_set doesn't reject file descriptors bigger than FD_SETSIZE.
 */

void
rb_fd_zero(rb_fdset_t *fds)
{
  if (fds->fdset) {
      memset(fds->fdset, 0, sizeof(fd_mask)*howmany(fds->maxfd, NFDBITS));
      FD_ZERO(fds->fdset);
  }
}

void
rb_fd_init(volatile rb_fdset_t *fds)
{
    fds->maxfd = 0;
    fds->fdset = malloc(sizeof(fd_set));
    FD_ZERO(fds->fdset);
}

void
rb_fd_term(rb_fdset_t *fds)
{
    if (fds->fdset) xfree(fds->fdset);
    fds->maxfd = 0;
    fds->fdset = 0;
}

static void
rb_fd_resize(int n, rb_fdset_t *fds)
{
    size_t m = howmany(n + 1, NFDBITS) * sizeof(fd_mask);
    size_t o = howmany(fds->maxfd, NFDBITS) * sizeof(fd_mask);

    if (m < sizeof(fd_set)) m = sizeof(fd_set);
    if (o < sizeof(fd_set)) o = sizeof(fd_set);

    if (m > o) {
    fds->fdset = xrealloc(fds->fdset, m);
    memset((char *)fds->fdset + o, 0, m - o);
    }
    if (n >= fds->maxfd) fds->maxfd = n + 1;
}

void
rb_fd_set(int n, rb_fdset_t *fds)
{
    rb_fd_resize(n, fds);
    FD_SET(n, fds->fdset);
}

void
rb_fd_clr(int n, rb_fdset_t *fds)
{
    if (n >= fds->maxfd) return;
    FD_CLR(n, fds->fdset);
}

int
rb_fd_isset(int n, const rb_fdset_t *fds)
{
    if (n >= fds->maxfd) return 0;
    return FD_ISSET(n, (const rb_fdset_t *)fds->fdset) != 0; /* "!= 0" avoids FreeBSD PR 91421 */
}

void
rb_fd_copy(rb_fdset_t *dst, const fd_set *src, int max)
{
    size_t size = howmany(max, NFDBITS) * sizeof(fd_mask);

    if (size < sizeof(fd_set)) size = sizeof(fd_set);
    dst->maxfd = max;
    dst->fdset = xrealloc(dst->fdset, size);
    memcpy(dst->fdset, src, size);
}
#endif

#endif //#ifdef INCLUDE_REGEXP
