/**********************************************************************

  rubyio.h -

  $Author: nobu $
  created at: Fri Nov 12 16:47:09 JST 1993

  Copyright (C) 1993-2007 Yukihiro Matsumoto

**********************************************************************/

#ifndef RUBY_IO_H
#define RUBY_IO_H 1

#if defined(__cplusplus)
extern "C" {
#if 0
} /* satisfy cc-mode */
#endif
#endif
/* ------------------------------------------------------------------------------------------- */

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#else
#include <io.h>
#endif
#include <errno.h>

#include "encoding.h"

#if defined(HAVE_STDIO_EXT_H)
#include <stdio_ext.h>
#endif

#define HAVE_FCNTL_H  /* kusuda */

#if defined(HAVE_FCNTL_H)
#include <fcntl.h>
#elif defined(HAVE_SYS_FCNTL_H)
#include <sys/fcntl.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>

//#include "st.h" /* define ANYARGS */

extern mrb_value rb_stdin;
extern mrb_value rb_stdout;
extern mrb_value rb_stderr;

typedef unsigned long rb_pid_t;

#ifdef INCLUDE_REGEXP
typedef struct rb_io_t {
    int fd;                  /* file descriptor */
    FILE *stdio_file;        /* stdio ptr for read/write if available */
    int mode;                /* mode flags: FMODE_XXXs */
    rb_pid_t pid;            /* child's pid (for pipes) */
    int lineno;              /* number of lines read */
    mrb_value pathv;         /* pathname for file */
    void (*finalize)(struct rb_io_t*,int); /* finalize proc */

    char *wbuf;                 /* wbuf_off + wbuf_len <= wbuf_capa */
    int wbuf_off;
    int wbuf_len;
    int wbuf_capa;

    char *rbuf;                 /* rbuf_off + rbuf_len <= rbuf_capa */
    int rbuf_off;
    int rbuf_len;
    int rbuf_capa;

    mrb_value tied_io_for_writing;

    /*
     * enc  enc2 read action                      write action
     * NULL NULL force_encoding(default_external) write the byte sequence of str
     * e1   NULL force_encoding(e1)               convert str.encoding to e1
     * e1   e2   convert from e2 to e1            convert str.encoding to e2
     */
    struct rb_io_enc_t {
        rb_encoding *enc;
        rb_encoding *enc2;
        int ecflags;
        mrb_value ecopts;
    } encs;

    rb_econv_t *readconv;
    char *cbuf;                /* cbuf_off + cbuf_len <= cbuf_capa */
    int cbuf_off;
    int cbuf_len;
    int cbuf_capa;

    rb_econv_t *writeconv;
    mrb_value writeconv_asciicompat;
    int writeconv_pre_ecflags;
    mrb_value writeconv_pre_ecopts;
    int writeconv_initialized;

    mrb_value write_lock;
} rb_io_t;
#else
typedef struct rb_io_t {
    FILE *f;               /* stdio ptr for read/write */
    FILE *f2;              /* additional ptr for rw pipes */
    int mode;              /* mode flags */
    int pid;               /* child's pid (for pipes) */
    int lineno;            /* number of lines read */
    char *path;            /* pathname for file */
    void (*finalize) (struct rb_io_t*,int); /* finalize proc */
} rb_io_t;
#endif

struct RFile {
    struct RBasic basic;
    struct rb_io_t *fptr;
};
#define RFILE(obj)   ((struct RFile*)((obj).value.p))

#define HAVE_RB_IO_T 1

#define FMODE_READABLE              0x00000001
#define FMODE_WRITABLE              0x00000002
#define FMODE_READWRITE             (FMODE_READABLE|FMODE_WRITABLE)
#define FMODE_BINMODE               0x00000004
#define FMODE_SYNC                  0x00000008
#define FMODE_TTY                   0x00000010
#define FMODE_DUPLEX                0x00000020
#define FMODE_APPEND                0x00000040
#define FMODE_CREATE                0x00000080
/* #define FMODE_NOREVLOOKUP        0x00000100 */
#define FMODE_WSPLIT                0x00000200
#define FMODE_WSPLIT_INITIALIZED    0x00000400
#define FMODE_TRUNC                 0x00000800
#define FMODE_TEXTMODE              0x00001000
/* #define FMODE_PREP               0x00010000 */
#define FMODE_SETENC_BY_BOM         0x00100000

#define GetOpenFile(mrb, obj, fp) rb_io_check_closed(mrb, (fp) = RFILE(obj)->fptr)

#ifndef NORETURN
# define NORETURN(x) x
#endif
#ifndef DEPRECATED
# define DEPRECATED(x) x
#endif

FILE *rb_io_stdio_file(rb_io_t *fptr);

FILE *rb_fdopen(int, const char*);
int rb_io_modestr_fmode(mrb_state *mrb, const char *modestr);
int rb_io_modestr_oflags(const char *modestr);
int rb_io_oflags_fmode(int oflags);
void rb_io_check_writable(mrb_state *mrb, rb_io_t*);
void rb_io_check_readable(mrb_state *mrb, rb_io_t*);
int rb_io_fptr_finalize(mrb_state *mrb, rb_io_t*);
void rb_io_synchronized(rb_io_t*);
void mrb_io_check_initialized(mrb_state *mrb, rb_io_t*);
void rb_io_check_closed(mrb_state *mrb, rb_io_t*);
int rb_io_wait_readable(mrb_state *mrb, int);
int rb_io_wait_writable(mrb_state *mrb, int);
void rb_io_set_nonblock(rb_io_t *fptr);
mrb_value mrb_io_close_m(mrb_state *mrb, mrb_value io);
mrb_value mrb_io_print(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value out);
mrb_value rb_io_write(mrb_state *mrb, mrb_value io, mrb_value str);
mrb_value io_write(mrb_state *mrb, mrb_value io, mrb_value str, int nosync);

/* compatibility for ruby 1.8 and older */
#define rb_io_mode_flags(mrb, modestr) rb_io_modestr_fmode(mrb, modestr)
#define rb_io_modenum_flags(oflags) rb_io_oflags_fmode(oflags)

mrb_value rb_io_taint_check(mrb_value);
NORETURN(void rb_eof_error(mrb_state *mrb));

void rb_io_read_check(rb_io_t*);
int rb_io_read_pending(rb_io_t*);
DEPRECATED(void rb_read_check(mrb_state *mrb, FILE*));

mrb_value rb_io_initialize(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io);
mrb_value rb_open_file(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io);
mrb_value rb_str_encode_ospath(mrb_state *mrb, mrb_value path);

#if __GNUC__ >= 3
#define UNINITIALIZED_VAR(x) x = x
#else
#define UNINITIALIZED_VAR(x) x
#endif

mrb_value rb_get_path(mrb_state *mrb, mrb_value);
#if 0
#define FilePathValue(mrb, v) (RB_GC_GUARD(v) = rb_get_path(mrb, v))
#else
#define FilePathValue(mrb, v) ((v) = rb_get_path(mrb, v))
#endif
int rb_path_check(mrb_state *mrb, const char *path);

void ruby_xfree(void *x);

/* for IO */
#undef FD_ZERO
#undef FD_SET
#undef FD_CLR
#undef FD_ISSET

#define FD_ZERO(f)    rb_fd_zero(f)
#define FD_SET(i, f)    rb_fd_set(i, f)
#define FD_CLR(i, f)    rb_fd_clr(i, f)
#define FD_ISSET(i, f)    rb_fd_isset(i, f)

#undef _WIN32
#if defined(_WIN32)
#define WAIT_FD_IN_WIN32(fptr) \
    (rb_w32_has_cancel_io() ? 0 : rb_thread_wait_fd(mrb, (fptr)->fd))
#else
#define WAIT_FD_IN_WIN32(fptr)
#endif

int rb_ensure(int (*b_proc)(ANYARGS), void * data1, mrb_value (*e_proc)(ANYARGS), mrb_value data2);
mrb_value rb_rescue(mrb_state *mrb, mrb_value (* b_proc)(ANYARGS), mrb_value *data1,
                          mrb_value (* r_proc)(ANYARGS), mrb_value *data2);

void rb_syswait(rb_pid_t pid);

/* ------------------------------------------------------------------------------------------- */
#if defined(__cplusplus)
#if 0
{ /* satisfy cc-mode */
#endif
}  /* extern "C" { */
#endif

#endif /* RUBY_IO_H */
