#include "../../mruby.h"
#include "string.h"
#include "rtos.h"

// Task

mrb_value
tsk_create(mrb_state *mrb, mrb_value self)
{
  void *entry;
  mrb_int priority;
  mrb_int stksiz;
  mrb_int act;
  mrb_value vid;
  mrb_value vargs[10];
  int iargs[10];
  mrb_int tid;
  int i;

  mrb_get_args(mrb, "iiiiooooooooooo", &entry, &priority, &stksiz, &act, &vid,
    &vargs[0], &vargs[1], &vargs[2], &vargs[3], &vargs[4], &vargs[5], &vargs[6], &vargs[7], &vargs[8], &vargs[9]);

  for(i=0; i<10; i++) {
    iargs[i] = mrb_bil_p(vargs[i]) ? 0 : vargs[i].value.i;
  }

  if(mrb_nil_p(vid)) {
    tid = rtos_tsk_create_auto_id(entry, priority, stksiz, act, iargs);
  }
  else {
    tid = vid.value.i;
    tid = rtos_tsk_create_spec_id(tid, entry, priority, stksiz, act, iargs);
  }

  return mrb_fixnum_value(tid);
}

mrb_value
tsk_sleep(mrb_state *mrb, mrb_value self)
{
  mrb_int tmout;

  mrb_get_args(mrb, "i", &tmout);

  return mrb_fixnum_value(rtos_tsk_sleep(tmout));
}

mrb_value
tsk_lock(mrb_state *mrb, mrb_value self)
{
  return mrb_fixnum_value(rtos_tsk_lock());
}

mrb_value
tsk_unlock(mrb_state *mrb, mrb_value self)
{
  return mrb_fixnum_value(rtos_tsk_unlock());
}

mrb_value
tsk_activate(mrb_state *mrb, mrb_value self)
{
  mrb_int tid;

  mrb_get_args(mrb, "i", &tid);

  return mrb_fixnum_value(rtos_tsk_activate(tid));
}

mrb_value
tsk_suspend(mrb_state *mrb, mrb_value self)
{
  mrb_int tid;

  mrb_get_args(mrb, "i", &tid);

  return mrb_fixnum_value(rtos_tsk_suspend(tid));
}

mrb_value
tsk_resume(mrb_state *mrb, mrb_value self)
{
  mrb_int tid;

  mrb_get_args(mrb, "i", &tid);

  return mrb_fixnum_value(rtos_tsk_resume(tid));
}

mrb_value
tsk_restart(mrb_state *mrb, mrb_value self)
{
  return mrb_fixnum_value(rtos_tsk_restart());
}

mrb_value
tsk_delete(mrb_state *mrb, mrb_value self)
{
  mrb_int tid;

  mrb_get_args(mrb, "i", &tid);

  return mrb_fixnum_value(rtos_tsk_delete(tid));
}




// EventFlag

mrb_value
evt_create(mrb_state *mrb, mrb_value self)
{
  mrb_value vid;
  mrb_int inival;
  mrb_int manrst;
  mrb_int eid;

  mrb_get_args(mrb, "oii", &vid, &inival, &manrst);

  if(mrb_nil_p(vid)) {
    eid = rtos_evt_create_auto_id(inival, manrst);
  }
  else {
    eid = vid.value.i;
    eid = rtos_evt_create_spec_id(eid, inival, manrst);
  }

  return mrb_fixnum_value(eid);
}

mrb_value
evt_wait(mrb_state *mrb, mrb_value self)
{
  mrb_int eid;
  mrb_int tmout;
  mrb_int waiflg;
  mrb_int waityp;

  mrb_get_args(mrb, "iiii", &eid, &tmout, &waiflg, &waityp);

  return mrb_fixnum_value(rtos_evt_wait(eid, tmout, waiflg, waityp));
}

mrb_value
evt_set(mrb_state *mrb, mrb_value self)
{
  mrb_int eid;
  mrb_int flg;

  mrb_get_args(mrb, "ii", &eid, &flg);

  return mrb_fixnum_value(rtos_evt_set(eid, flg));
}

mrb_value
evt_reset(mrb_state *mrb, mrb_value self)
{
  mrb_int eid;
  mrb_int flg;

  mrb_get_args(mrb, "ii", &eid, &flg);

  return mrb_fixnum_value(rtos_evt_reset(eid, flg));
}



// Mutex

mrb_value
mut_create(mrb_state *mrb, mrb_value self)
{
  mrb_value vid;
  mrb_int iniown;
  mrb_int mid;

  mrb_get_args(mrb, "oi", &vid, &iniown);

  if(mrb_nil_p(vid)) {
    mid = rtos_mut_create_auto_id(iniown);
  }
  else {
    mid = vid.value.i;
    mid = rtos_mut_create_spec_id(mid, iniown);
  }

  return mrb_fixnum_value(mid);
}

mrb_value
mut_wait(mrb_state *mrb, mrb_value self)
{
  mrb_int mid;
  mrb_int tmout;

  mrb_get_args(mrb, "ii", &mid, &tmout);

  return mrb_fixnum_value(rtos_mut_wait(mid, tmout));
}

mrb_value
mut_release(mrb_state *mrb, mrb_value self)
{
  mrb_int mid;

  mrb_get_args(mrb, "i", &mid);

  return mrb_fixnum_value(rtos_mut_release(mid));
}



// Semaphore

mrb_value
sem_create(mrb_state *mrb, mrb_value self)
{
  mrb_value vid;
  mrb_int maxcnt;
  mrb_int inicnt;
  mrb_int sid;

  mrb_get_args(mrb, "oii", &vid, &maxcnt, &inicnt);

  if(mrb_nil_p(vid)) {
    sid = rtos_sem_create_auto_id(maxcnt, inicnt);
  }
  else {
    sid = vid.value.i;
    sid = rtos_sem_create_spec_id(sid, maxcnt, inicnt);
  }

  return mrb_fixnum_value(sid);
}

mrb_value
sem_wait(mrb_state *mrb, mrb_value self)
{
  mrb_int sid;
  mrb_int tmout;
  mrb_int cnt;

  mrb_get_args(mrb, "iii", &sid, &tmout, &cnt);

  return mrb_fixnum_value(rtos_sem_wait(sid, tmout, cnt));
}

mrb_value
sem_release(mrb_state *mrb, mrb_value self)
{
  mrb_int sid;
  mrb_int cnt;

  mrb_get_args(mrb, "ii", &sid, &cnt);

  return mrb_fixnum_value(rtos_sem_release(sid, cnt));
}



// MailBox

mrb_value
mbx_create(mrb_state *mrb, mrb_value self)
{
  mrb_value vid;
  mrb_int maxmsg;
  mrb_int maxlen;
  mrb_int prique;
  mrb_int mid;

  mrb_get_args(mrb, "oiii", &vid, &maxmsg, &maxlen, &prique);

  if(mrb_nil_p(vid)) {
    mid = rtos_mbx_create_auto_id(maxmsg, maxlen, prique);
  }
  else {
    mid = vid.value.i;
    mid = rtos_mbx_create_spec_id(mid, maxmsg, maxlen, prique);
  }

  return mrb_fixnum_value(mid);
}

mrb_value
mbx_wait(mrb_state *mrb, mrb_value self)
{
  mrb_int mid;
  mrb_int tmout;
  mrb_int maxlen;
  char *msg;
  mrb_int rc;
  mrb_value rv;

  mrb_get_args(mrb, "iii", &mid, &tmout, &maxlen);

  msg = (char*)mrb_malloc(mrb, maxlen + 1);
  if(msg == 0) {
    return mrb_fixnum_value(-1);	// TODO: Define return value
  }
  memset(msg, 0, (size_t)(maxlen + 1));

  rc = rtos_mbx_wait(mid, tmout, msg, maxlen);
  if(rc >= 0) {
    rv = mrb_str_new_cstr(mrb, msg);
  }
  else {
    rv = mrb_fixnum_value(rc);
  }
  mrb_free(mrb, msg);

  return rv;
}

mrb_value
mbx_send(mrb_state *mrb, mrb_value self)
{
  mrb_int mid;
  mrb_value msg;

  mrb_get_args(mrb, "io", &mid, &msg);

  return mrb_fixnum_value(rtos_mbx_send(mid, (char*)msg.value.p));
}



// WatchdogTimer

mrb_value
wdt_create(mrb_state *mrb, mrb_value self)
{
  mrb_value vid;
  mrb_int wid;

  mrb_get_args(mrb, "o", &vid);

  if(mrb_nil_p(vid)) {
    wid = rtos_wdt_create_auto_id();
  }
  else {
    wid = vid.value.i;
    wid = rtos_wdt_create_spec_id(wid);
  }

  return mrb_fixnum_value(wid);
}

mrb_value
wdt_start(mrb_state *mrb, mrb_value self)
{
  mrb_int wid;
  mrb_int delay;
  void *tmofnc;
  mrb_int arg;

  mrb_get_args(mrb, "iiii", &wid, &delay, &tmofnc, &arg);

  return mrb_fixnum_value(rtos_wdt_start(wid, delay, tmofnc, arg));
}

mrb_value
wdt_cancel(mrb_state *mrb, mrb_value self)
{
  mrb_int wid;

  mrb_get_args(mrb, "i", &wid);

  return mrb_fixnum_value(rtos_wdt_cancel(wid));
}



// Others

mrb_value
tick_get(mrb_state *mrb, mrb_value self)
{
  return mrb_fixnum_value(rtos_tick_get());
}




void
mrb_init_rtos(mrb_state *mrb)
{
  struct RClass *rtos;

  rtos = mrb_define_module(mrb, "RTOS");
  // Task
  mrb_define_method(mrb, rtos, "tsk_create", tsk_create, ARGS_REQ(15));
  mrb_define_method(mrb, rtos, "tsk_sleep", tsk_sleep, ARGS_REQ(1));
  mrb_define_method(mrb, rtos, "tsk_lock", tsk_lock, ARGS_NONE());
  mrb_define_method(mrb, rtos, "tsk_unlock", tsk_unlock, ARGS_NONE());
  mrb_define_method(mrb, rtos, "tsk_activate", tsk_activate, ARGS_REQ(1));
  mrb_define_method(mrb, rtos, "tsk_suspend", tsk_suspend, ARGS_REQ(1));
  mrb_define_method(mrb, rtos, "tsk_resume", tsk_resume, ARGS_REQ(1));
  mrb_define_method(mrb, rtos, "tsk_restart", tsk_restart, ARGS_NONE());
  mrb_define_method(mrb, rtos, "tsk_delete", tsk_delete, ARGS_REQ(1));
  // EventFlag
  mrb_define_method(mrb, rtos, "evt_create", evt_create, ARGS_REQ(3));
  mrb_define_method(mrb, rtos, "evt_wait", evt_wait, ARGS_REQ(4));
  mrb_define_method(mrb, rtos, "evt_set", evt_set, ARGS_REQ(2));
  mrb_define_method(mrb, rtos, "evt_reset", evt_reset, ARGS_REQ(2));
  // Mutex
  mrb_define_method(mrb, rtos, "mut_create", mut_create, ARGS_REQ(2));
  mrb_define_method(mrb, rtos, "mut_wait", mut_wait, ARGS_REQ(2));
  mrb_define_method(mrb, rtos, "mut_release", mut_release, ARGS_REQ(1));
  // Semaphore
  mrb_define_method(mrb, rtos, "sem_create", sem_create, ARGS_REQ(3));
  mrb_define_method(mrb, rtos, "sem_wait", mut_wait, ARGS_REQ(3));
  mrb_define_method(mrb, rtos, "sem_release", mut_release, ARGS_REQ(2));
  // MailBox
  mrb_define_method(mrb, rtos, "mbx_create", mbx_create, ARGS_REQ(4));
  mrb_define_method(mrb, rtos, "mbx_wait", mbx_wait, ARGS_REQ(2));
  mrb_define_method(mrb, rtos, "mbx_send", mbx_send, ARGS_REQ(2));
  // WatchdogTimer
  mrb_define_method(mrb, rtos, "wdt_create", wdt_create, ARGS_REQ(1));
  mrb_define_method(mrb, rtos, "wdt_start", wdt_start, ARGS_REQ(4));
  mrb_define_method(mrb, rtos, "wdt_cancel", wdt_cancel, ARGS_REQ(1));
  // Others
  mrb_define_method(mrb, rtos, "tick_get", tick_get, ARGS_NONE());

  mrb_include_module(mrb, mrb->object_class, rtos);
}
