/************************************************

  init.c -

  created at: Thu Mar 31 12:21:29 JST 1994

  Copyright (C) 1993-2007 Yukihiro Matsumoto

************************************************/

#include "rubysocket.h"
#include "stdio.h" /* BUFSIZ */

VALUE
rsock_init_sock(mrb_state *mrb, VALUE sock, int fd)
{
  mrb_iv_set(mrb, sock, mrb_intern(mrb, "socket_fd"), mrb_fixnum_value(fd));
  return sock;
}

VALUE
rsock_send_blocking(void *data)
{
    struct rsock_send_arg *arg = data;
    VALUE mesg = arg->mesg;

    return mrb_fixnum_value(send(arg->fd, RSTRING_PTR(mesg), RSTRING_LEN(mesg),
                                    arg->flags));
}

struct recvfrom_arg {
    int fd, flags;
    VALUE str;
    socklen_t alen;
    struct sockaddr_storage buf;
};

static int
recvfrom_blocking(void *data)
{
    struct recvfrom_arg *arg = data;
    return recvfrom(arg->fd, RSTRING_PTR(arg->str), RSTRING_LEN(arg->str),
                           arg->flags, (struct sockaddr*)&arg->buf, &arg->alen);
}

VALUE
rsock_s_recvfrom(mrb_state *mrb, mrb_value self, enum sock_recv_type from)
{
  mrb_value fd;
  VALUE str;
  struct recvfrom_arg arg;
  VALUE len, flg;
  long buflen;
  long slen;

  mrb_get_args(mrb, "oo", &len, &flg);

  if (mrb_nil_p(flg))
    arg.flags = 0;
  else
    arg.flags = mrb_fixnum(flg);
  buflen = mrb_fixnum(len);

  fd = mrb_iv_get(mrb, self, mrb_intern(mrb, "socket_fd"));
  arg.fd = mrb_fixnum(fd);
  arg.alen = (socklen_t)sizeof(arg.buf);
  arg.str = str = mrb_tainted_str_new(mrb, 0, buflen);

  slen = (long)recvfrom_blocking(&arg);
  if (slen < RSTRING_LEN(str)) {
    mrb_str_set_len(str, slen);
  }

  return str;
}

VALUE
rsock_s_read(mrb_state *mrb, mrb_value self)
{
  mrb_value fd;
  struct recvfrom_arg arg;
  VALUE len, str;
  long buflen;
  long slen;

  mrb_get_args(mrb, "oo", &len, &str);

  arg.flags = 0;

  buflen = FIXNUM_P(len) ? mrb_fixnum(len) : BUFSIZ;
  if (buflen < 0) {
    mrb_raise(mrb, mrb->eArgumentError_class, "negative length given");
  }

  str = mrb_nil_p(str) ? mrb_tainted_str_new(mrb, 0, buflen) : mrb_str_new(mrb, RSTRING_PTR(str), buflen);

  fd = mrb_iv_get(mrb, self, mrb_intern(mrb, "socket_fd"));
  arg.fd = mrb_fixnum(fd);
  arg.alen = (socklen_t)sizeof(arg.buf);
  arg.str = str;

  slen = recvfrom_blocking(&arg);

  if (slen == 0) {
    mrb_str_resize(mrb, str, slen);
    return mrb_nil_value();
  } else if (slen < RSTRING_LEN(str)) {
    mrb_str_resize(mrb, str, slen);
  }

  return str;
}

int
rsock_socket(int domain, int type, int proto)
{
  int fd;

  fd = socket(domain, type, proto);
  if (fd < 0) {
    if (errno == EMFILE || errno == ENFILE) {
      /* TODO force garbage collect */
//    mrb_gc();
//    fd = socket(domain, type, proto);
    }
  }
  return fd;
}

struct accept_arg {
    int fd;
    struct sockaddr *sockaddr;
    socklen_t *len;
};

static int
accept_blocking(void *data)
{
    struct accept_arg *arg = data;
    return accept(arg->fd, arg->sockaddr, arg->len);
}

VALUE
rsock_s_accept(mrb_state *mrb, VALUE klass, int fd, struct sockaddr *sockaddr, socklen_t *len)
{
  int fd2;
  struct accept_arg arg;
  mrb_value c;

  arg.fd = fd;
  arg.sockaddr = sockaddr;
  arg.len = len;

  fd2 = accept_blocking(&arg);
  if (fd2 < 0) {
    if (errno == EMFILE || errno == ENFILE) {
      /* TODO force garbage collect */
//    mrb_gc();
//    fd2 = accept_blocking(&arg);
    }

    if (fd2 < 0)
      mrb_sys_fail("accept");
  }

  if (!mrb_object(klass))
    return mrb_fixnum_value(fd2);

  c = mrb_const_get(mrb, mrb_obj_value(mrb->object_class), mrb_intern(mrb, "TCPSocket"));
  return rsock_init_sock(mrb, mrb_instance_new(mrb, c), fd2);
}

/*
 * SocketError is the error class for socket.
 */
void
rsock_init_socket_init(mrb_state *mrb)
{
  rsock_init_tcpsocket(mrb);
  rsock_init_tcpserver(mrb);
}
