/************************************************

  ipsocket.c -

  created at: Thu Mar 31 12:21:29 JST 1994

  Copyright (C) 1993-2007 Yukihiro Matsumoto

************************************************/

#include "rubysocket.h"

struct inetsock_arg
{
  VALUE sock;
  struct {
    VALUE host, serv;
    struct addrinfo *res;
  } remote, local;
  int type;
  int fd;
};

static VALUE
inetsock_cleanup(mrb_state *mrb, struct inetsock_arg *arg)
{
  if (arg->remote.res) {
    freeaddrinfo(arg->remote.res);
    arg->remote.res = 0;
  }
  if (arg->local.res) {
    freeaddrinfo(arg->local.res);
    arg->local.res = 0;
  }
  if (arg->fd >= 0) {
  #ifdef __NIOS2__
    close_s(arg->fd);
  #else
    close(arg->fd);
  #endif
  }

  return mrb_nil_value();
}

static VALUE
init_inetsock_internal(mrb_state *mrb, struct inetsock_arg *arg)
{
  int type = arg->type;
  struct addrinfo *res;
  int fd, status = 0;
  const char *syscall = 0;

  // return dummy addrinfo
  arg->remote.res = rsock_addrinfo(mrb, arg->remote.host, arg->remote.serv, SOCK_STREAM,
                                  (type == INET_SERVER) ? AI_PASSIVE : 0);

  arg->fd = fd = -1;
  for (res = arg->remote.res; res; res = res->ai_next) {
#if !defined(INET6) && defined(AF_INET6)
    if (res->ai_family == AF_INET6)
        continue;
#endif
    status = rsock_socket(res->ai_family,res->ai_socktype,res->ai_protocol);
    // syscall = "socket(2)";
    fd = status;
    if (fd < 0) {
        continue;
    }
    arg->fd = fd;
    if (type == INET_SERVER) {
#if !defined(_WIN32) && !defined(__CYGWIN__)
      status = 1;
      setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
                 (char*)&status, (socklen_t)sizeof(status));
#endif
      status = bind(fd, res->ai_addr, res->ai_addrlen);
      syscall = "bind(2)";
    } else {
      mrb_raise(mrb, mrb->eStandardError_class, "not support client\n");
      return mrb_nil_value();
    }

    if (status < 0) {
      close(fd);
      arg->fd = fd = -1;
      continue;
    } else
      break;
  }

  arg->fd = -1;

  if (type == INET_SERVER) {
    status = listen(fd, 5);
    if (status < 0) {
      close(fd);
      syscall = "listen(2)";
    }
  }

  if (status < 0)
    mrb_sys_fail(syscall);

  /* create new instance */
  return rsock_init_sock(mrb, mrb_instance_new(mrb, arg->sock), fd);
}

VALUE
rsock_init_inetsock(mrb_state *mrb, VALUE sock, VALUE remote_host, VALUE remote_serv,
	            VALUE local_host, VALUE local_serv, int type)
{
  struct inetsock_arg arg;
  mrb_value o;

  arg.sock = sock;
  arg.remote.host = remote_host;
  arg.remote.serv = remote_serv;
  arg.remote.res = 0;
  arg.local.host = local_host;
  arg.local.serv = local_serv;
  arg.local.res = 0;
  arg.type = type;
  arg.fd = -1;

  o = init_inetsock_internal(mrb, &arg);
  inetsock_cleanup(mrb, &arg);

  return o;
}
