/************************************************

  ainfo.c -

  created at: Thu Mar 31 12:21:29 JST 1994

  Copyright (C) 1993-2007 Yukihiro Matsumoto

************************************************/

#include "rubysocket.h"

#include <string.h> /* for memset */
#include <stdio.h> /* for sscanf */

static int
inet_pton(const char *hostname, void *pton)
{
  struct in_addr in;
  int d1, d2, d3, d4;
  char ch;

  if (sscanf(hostname, "%d.%d.%d.%d%c", &d1, &d2, &d3, &d4, &ch) == 4 &&
    0 <= d1 && d1 <= 255 && 0 <= d2 && d2 <= 255 &&
    0 <= d3 && d3 <= 255 && 0 <= d4 && d4 <= 255) {
    in.s_addr = htonl(
      ((long) d1 << 24) | ((long) d2 << 16) |
      ((long) d3 << 8) | ((long) d4 << 0));
  }
  else {
    return 0;
  }

  memcpy(pton, &in, sizeof(in));
  return 1;
}

struct addrinfo*
rsock_addrinfo(mrb_state *mrb, VALUE host, VALUE port, int socktype, int flags)
{
  struct addrinfo *res = NULL;
  struct sockaddr_in *ai_addr = NULL;
  in_addr_t s_addr;

  if (!FIXNUM_P(port))
    return 0;

  if (mrb_nil_p(host))
    s_addr = INADDR_ANY;
  else if (!inet_pton(RSTRING_PTR(host), &s_addr))
    return 0;

  if ((res = mrb_malloc(mrb, sizeof(*res))) == 0)
    return 0;

  if ((ai_addr = mrb_malloc(mrb, sizeof(*ai_addr))) == 0)
    return 0;

  memset(res, 0, sizeof(*res));
  memset(ai_addr, 0, sizeof(*ai_addr));

  res->ai_flags = flags | AI_NUMERICHOST | AI_NUMERICSERV;
  res->ai_family = PF_INET;
  res->ai_socktype = socktype;
  res->ai_protocol = IPPROTO_TCP;
  res->ai_addrlen = sizeof(*ai_addr);
  res->ai_canonname = NULL;
  res->ai_addr = (struct sockaddr *)ai_addr;
  res->ai_next = NULL;

  ai_addr->sin_family = AF_INET;
  ai_addr->sin_addr.s_addr = s_addr;
  ai_addr->sin_port = htons(mrb_fixnum(port));

  return res;

  /* TODO rsock_getaddrinfo() */
//  struct addrinfo hints;
//
//  memset(&hints, 0, sizeof(hints));
//  hints.ai_family = AF_UNSPEC;
//  hints.ai_socktype = socktype;
//  hints.ai_flags = flags;
//
//  return rsock_getaddrinfo(mrb, host, port, &hints, 1);
}

