/************************************************

  tcpserver.c -

  created at: Thu Mar 31 12:21:29 JST 1994

  Copyright (C) 1993-2007 Yukihiro Matsumoto

************************************************/

#include "rubysocket.h"

/*
 * call-seq:
 *   TCPServer.new([hostname,] port)                    => tcpserver
 *
 * Creates a new server socket bound to _port_.
 *
 * If _hostname_ is given, the socket is bound to it.
 *
 *   serv = TCPServer.new("127.0.0.1", 28561)
 *   s = serv.accept
 *   s.puts Time.now
 *   s.close
 */
static VALUE
tcp_svr_init(mrb_state *mrb, VALUE self)
{
  VALUE host, serv;

  mrb_get_args(mrb, "oo", &host, &serv);
  return rsock_init_inetsock(mrb, self, host, serv, mrb_nil_value(), mrb_nil_value(), INET_SERVER);
}

/*
 * call-seq:
 *   tcpserver.accept => tcpsocket
 *
 *   TCPServer.open("127.0.0.1", 14641) {|serv|
 *     s = serv.accept
 *     s.puts Time.now
 *     s.close
 *   }
 *
 */
static VALUE
tcp_accept(mrb_state *mrb, VALUE self)
{
  mrb_value fd;
  struct sockaddr_storage from;
  socklen_t fromlen;

  fd = mrb_iv_get(mrb, self, mrb_intern(mrb, "socket_fd"));
  fromlen = sizeof(from);
  return rsock_s_accept(mrb, self, mrb_fixnum(fd),
                        (struct sockaddr*)&from, &fromlen);
}

/*
 * Document-class: ::TCPServer < TCPSocket
 *
 * TCPServer represents a TCP/IP server socket.
 */
void
rsock_init_tcpserver(mrb_state *mrb)
{
  struct RClass *cTCPSocket, *cTCPServer;

  cTCPSocket = mrb_class_ptr(mrb_const_get(mrb, mrb_obj_value(mrb->object_class), mrb_intern(mrb, "TCPSocket")));
  cTCPServer = mrb_define_class(mrb, "TCPServer", cTCPSocket);

  mrb_define_class_method(mrb, cTCPServer, "new",       tcp_svr_init,           ARGS_REQ(1) | ARGS_OPT(1));
  mrb_define_class_method(mrb, cTCPServer, "open",      tcp_svr_init,           ARGS_REQ(1) | ARGS_OPT(1));

  mrb_define_method(mrb, cTCPServer, "accept",          tcp_accept,             ARGS_NONE());
//  mrb_define_method(mrb, cTCPServer, "initialize",      tcp_svr_init,           ARGS_REQ(1) | ARGS_OPT(1));
  mrb_define_method(mrb, cTCPServer, "listen",          rsock_sock_listen,      ARGS_REQ(1)); /* in socket.c */
}
