/************************************************

  tcpsocket.c -

  created at: Thu Mar 31 12:21:29 JST 1994

  Copyright (C) 1993-2007 Yukihiro Matsumoto

************************************************/

#include "rubysocket.h"

/*
 * call-seq:
 *    TCPSocket.new(remote_host, remote_port, local_host=nil, local_port=nil)
 *
 * Opens a TCP connection to +remote_host+ on +remote_port+.  If +local_host+
 * and +local_port+ are specified, then those parameters are used on the local
 * end to establish the connection.
 */
static VALUE
tcp_init(mrb_state *mrb, mrb_value sock)
{
  VALUE remote_host, remote_serv;
  VALUE local_host, local_serv;

  mrb_get_args(mrb, "oooo", &remote_host, &remote_serv,
                             &local_host, &local_serv);

  return rsock_init_inetsock(mrb, sock, remote_host, remote_serv,
                             local_host, local_serv, INET_CLIENT);
}

/*
 * Document-class: ::TCPSocket < IPSocket
 *
 * TCPSocket represents a TCP/IP client socket.
 */
void
rsock_init_tcpsocket(mrb_state *mrb)
{
  struct RClass *cBasicSocket, *cTCPSocket;

  cBasicSocket = mrb_class_ptr(mrb_const_get(mrb, mrb_obj_value(mrb->object_class), mrb_intern(mrb, "BasicSocket")));
  cTCPSocket = mrb_define_class(mrb, "TCPSocket", cBasicSocket);

  mrb_define_class_method(mrb, cTCPSocket, "open", tcp_init, ARGS_REQ(2) | ARGS_OPT(2));
  mrb_define_class_method(mrb, cTCPSocket, "new",  tcp_init, ARGS_REQ(2) | ARGS_OPT(2));

//  mrb_define_method(mrb, cTCPSocket, "initialize", tcp_init, ARGS_REQ(2) | ARGS_OPT(2));
}
