#ifndef MTHREAD_H
#define MTHREAD_H

#define HAVE_PTHREAD_H 1
//#define _UWIN
#undef _WIN32
#define linux

#if   defined(_WIN32)
#include "thread_win32.h"
#elif defined(HAVE_PTHREAD_H)
#include "thread_pthread.h"
#else
#error "unsupported thread type"
#endif

#include "method.h"
#include "mdata.h"

#define HAVE_SYS_TIME_H  /* kusuda */
#if defined(HAVE_SYS_TIME_H)
#include <sys/time.h>
#elif !defined(_WIN32)
#define time_t long
struct timeval {
    time_t tv_sec;	/* seconds */
    long tv_usec;	/* microseconds */
};
#endif

#define	NBBY	8		/* number of bits in a byte */
typedef	long	fd_mask;
#define	NFDBITS	(sizeof (fd_mask) * NBBY)	/* bits per mask */
#define HAVE_RB_FD_INIT
#include <sys/types.h>

#if defined(NFDBITS) && defined(HAVE_RB_FD_INIT)
/* Add --> kusuda */
#ifndef	FD_SETSIZE
#define	FD_SETSIZE	64
#endif

#ifndef	howmany
#define	howmany(x,y)	(((x)+((y)-1))/(y))
#endif

/* We use a macro for fd_set so that including Sockets.h afterwards
   can work.  */
typedef	struct _types_fd_set {
	fd_mask	fds_bits[howmany(FD_SETSIZE, NFDBITS)];
} _types_fd_set;
#define fd_set _types_fd_set
/* Add --< kusuda */
typedef struct struct_rb_fdset_t{
    int maxfd;
    struct struct_rb_fdset_t *fdset;
} rb_fdset_t;

int select(int n, fd_set *r, fd_set *w, fd_set *e, struct timeval *timeout);

void rb_fd_init(volatile rb_fdset_t *);
void rb_fd_term(rb_fdset_t *);
void rb_fd_zero(rb_fdset_t *);
void rb_fd_set(int, rb_fdset_t *);
void rb_fd_clr(int, rb_fdset_t *);
int rb_fd_isset(int, const rb_fdset_t *);
void rb_fd_copy(rb_fdset_t *, const fd_set *, int);
int rb_fd_select(int, rb_fdset_t *, rb_fdset_t *, rb_fdset_t *, struct timeval *);

#define rb_fd_ptr(f)	((f)->fdset)
#define rb_fd_max(f)	((f)->maxfd)

#elif defined(_WIN32)

typedef struct {
    int capa;
    fd_set *fdset;
} rb_fdset_t;

void rb_fd_init(volatile rb_fdset_t *);
void rb_fd_term(rb_fdset_t *);
#define rb_fd_zero(f)		((f)->fdset->fd_count = 0)
void rb_fd_set(int, rb_fdset_t *);
#define rb_fd_clr(n, f)		rb_w32_fdclr(n, (f)->fdset)
#define rb_fd_isset(n, f)	rb_w32_fdisset(n, (f)->fdset)
#define rb_fd_select(n, rfds, wfds, efds, timeout)	rb_w32_select(n, (rfds) ? ((rb_fdset_t*)rfds)->fdset : NULL, (wfds) ? ((rb_fdset_t*)wfds)->fdset : NULL, (efds) ? ((rb_fdset_t*)efds)->fdset: NULL, timeout)
#define rb_fd_resize(n, f)	(void)(f)

#define rb_fd_ptr(f)	((f)->fdset)
#define rb_fd_max(f)	((f)->fdset->fd_count)

#else

#define fd_set _types_fd_set

typedef fd_set rb_fdset_t;
#define rb_fd_zero(f)	FD_ZERO(f)
#define rb_fd_set(n, f)	FD_SET(n, f)
#define rb_fd_clr(n, f)	FD_CLR(n, f)
#define rb_fd_isset(n, f) FD_ISSET(n, f)
#define rb_fd_copy(d, s, n) (*(d) = *(s))
#define rb_fd_resize(n, f)	(void)(f)
#define rb_fd_ptr(f)	(f)
#define rb_fd_init(f)	FD_ZERO(f)
#define rb_fd_term(f)	(void)(f)
#define rb_fd_max(f)	FD_SETSIZE
#define rb_fd_select(n, rfds, wfds, efds, timeout)	select(n, rfds, wfds, efds, timeout)

#endif

#define _SIGMAX (64)
#ifndef NSIG
# define NSIG (_SIGMAX + 1)      /* For QNX */
#endif


#define RUBY_NSIG NSIG

enum ruby_special_exceptions {
    ruby_error_reenter,
    ruby_error_nomemory,
    ruby_error_sysstack,
    ruby_special_error_count
};

typedef struct rb_vm_struct {
    struct st_table *living_threads;
#if 0
    rb_thread_lock_t global_vm_lock;
#else
    mrb_value self;

    rb_thread_lock_t global_vm_lock;

    struct rb_thread_struct *main_thread;
    struct rb_thread_struct *running_thread;

    mrb_value thgroup_default;

    int running;
    int thread_abort_on_exception;
    unsigned long trace_flag;
    volatile int sleeper;

    /* object management */
    mrb_value mark_object_ary;



    mrb_value special_exceptions[ruby_special_error_count];

    /* load */
    mrb_value top_self;
    mrb_value load_path;
    mrb_value loaded_features;
    struct st_table *loading_table;

    /* signal */
    struct {
	mrb_value cmd;
	int safe;
    } trap_list[RUBY_NSIG];

    /* hook */
    rb_event_hook_t *event_hooks;

    int src_encoding_index;

    mrb_value verbose, debug, progname;
    mrb_value coverages;

    struct unlinked_method_entry_list_entry *unlinked_method_entry_list;

  #if defined(ENABLE_VM_OBJSPACE) && ENABLE_VM_OBJSPACE
    struct rb_objspace *objspace;
  #endif
#endif
} rb_vm_t;

struct rb_iseq_struct {
    /***************/
    /* static data */
    /***************/

    mrb_value type;          /* instruction sequence type */
    mrb_value name;	         /* String: iseq name */
    mrb_value filename;      /* file information where this sequence from */
    mrb_value filepath;      /* real file path or nil */
    mrb_value *iseq;         /* iseq (insn number and openrads) */
    mrb_value *iseq_encoded; /* encoded iseq */
    unsigned long iseq_size;
    mrb_value mark_ary;	/* Array: includes operands which should be GC marked */
    mrb_value coverage;     /* coverage array */
    unsigned short line_no;

    /* insn info, must be freed */
    struct iseq_insn_info_entry *insn_info_table;
    size_t insn_info_size;

    mrb_sym *local_table;		/* must free */
    int local_table_size;

    /* method, class frame: sizeof(vars) + 1, block frame: sizeof(vars) */
    int local_size;

    struct iseq_inline_cache_entry *ic_entries;
    int ic_size;

    /**
     * argument information
     *
     *  def m(a1, a2, ..., aM,                    # mandatory
     *        b1=(...), b2=(...), ..., bN=(...),  # optinal
     *        *c,                                 # rest
     *        d1, d2, ..., dO,                    # post
     *        &e)                                 # block
     * =>
     *
     *  argc           = M
     *  arg_rest       = M+N+1 // or -1 if no rest arg
     *  arg_opts       = N
     *  arg_opts_tbl   = [ (N entries) ]
     *  arg_post_len   = O // 0 if no post arguments
     *  arg_post_start = M+N+2
     *  arg_block      = M+N + 1 + O + 1 // -1 if no block arg
     *  arg_simple     = 0 if not simple arguments.
     *                 = 1 if no opt, rest, post, block.
     *                 = 2 if ambiguos block parameter ({|a|}).
     *  arg_size       = argument size.
     */

    int argc;
    int arg_simple;
    int arg_rest;
    int arg_block;
    int arg_opts;
    int arg_post_len;
    int arg_post_start;
    int arg_size;
    mrb_value *arg_opt_table;

    size_t stack_max; /* for stack overflow check */

    /* catch table */
    struct iseq_catch_table_entry *catch_table;
    int catch_table_size;

    /* for child iseq */
    struct rb_iseq_struct *parent_iseq;
    struct rb_iseq_struct *local_iseq;

    /****************/
    /* dynamic data */
    /****************/

    mrb_value self;
    mrb_value orig;			/* non-NULL if its data have origin */

    /* block inlining */
    /*
     * NODE *node;
     * void *special_block_builder;
     * void *cached_special_block_builder;
     * mrb_value cached_special_block;
     */

    /* klass/module nest information stack (cref) */
    NODE *cref_stack;
    mrb_value klass;

    /* misc */
    mrb_sym defined_method_id;	/* for define_method */

    /* used at compile time */
    struct iseq_compile_data *compile_data;
};

typedef struct {
    mrb_value *pc;			/* cfp[0] */
    mrb_value *sp;			/* cfp[1] */
    mrb_value *bp;			/* cfp[2] */
    rb_iseq_t *iseq;		/* cfp[3] */
    mrb_value flag;			/* cfp[4] */
    mrb_value self;			/* cfp[5] / block[0] */
    mrb_value *lfp;			/* cfp[6] / block[1] */
    mrb_value *dfp;			/* cfp[7] / block[2] */
    rb_iseq_t *block_iseq;	/* cfp[8] / block[3] */
    mrb_value proc;			/* cfp[9] / block[4] */
    const rb_method_entry_t *me;/* cfp[10] */
} rb_control_frame_t;

enum rb_thread_status {
    THREAD_TO_KILL,
    THREAD_RUNNABLE,
    THREAD_STOPPED,
    THREAD_STOPPED_FOREVER,
    THREAD_KILLED
};

typedef void rb_unblock_function_t(void *);
struct rb_unblock_callback {
    rb_unblock_function_t *func;
    void *arg;
};

typedef int rb_blocking_function_t(void *);
void rb_thread_check_ints(mrb_state *mrb);
int rb_thread_interrupted(mrb_value thval);
int rb_thread_blocking_region(mrb_state *mrb,
    rb_blocking_function_t *func, void *data1,
    rb_unblock_function_t *ubf, void *data2);

#define RUBY_UBF_IO ((rb_unblock_function_t *)-1)
#define RUBY_UBF_PROCESS ((rb_unblock_function_t *)-1)
mrb_value rb_mutex_new(mrb_state *mrb);
mrb_value rb_mutex_locked_p(mrb_state *mrb, mrb_value mutex);
mrb_value rb_mutex_try_lock(mrb_value mutex);
mrb_value rb_mutex_lock(mrb_state *mrb, mrb_value mutex);
mrb_value rb_mutex_unlock(mrb_state *mrb, mrb_value mutex);
mrb_value rb_mutex_sleep(mrb_state *mrb, mrb_value self, mrb_value timeout);
int rb_mutex_synchronize(mrb_state *mrb, mrb_value mutex, int (*func)(mrb_state *mrb, void *arg), void *arg);
mrb_value rb_barrier_new(mrb_state *mrb);
mrb_value rb_barrier_wait(mrb_state *mrb, mrb_value self);
mrb_value rb_barrier_release(mrb_state *mrb, mrb_value self);
mrb_value rb_barrier_destroy(mrb_state *mrb, mrb_value self);
void rb_thread_start_timer_thread(mrb_state *mrb);
void rb_thread_stop_timer_thread(void);
void rb_thread_reset_timer_thread(void);

void rb_thread_wait_fd(mrb_state *mrb, int);
int rb_thread_fd_writable(mrb_state *mrb, int);
void rb_thread_fd_close(int);
int rb_thread_alone(void);
void rb_thread_polling(mrb_state *mrb);
void rb_thread_sleep(mrb_state *mrb, int);
void rb_thread_sleep_forever(mrb_state *mrb);
mrb_value rb_thread_stop(mrb_state *mrb, mrb_value);
mrb_value rb_thread_wakeup(mrb_state *mrb, mrb_value);
mrb_value rb_thread_run(mrb_state *mrb, mrb_value);
mrb_value rb_thread_kill(mrb_state *mrb, mrb_value);


#define rb_jmpbuf_t void*   /* kusuda */
typedef struct rb_block_struct {
    mrb_value self;         /* share with method frame if it's only block */
    mrb_value *lfp;         /* share with method frame if it's only block */
    mrb_value *dfp;         /* share with method frame if it's only block */
    rb_iseq_t *iseq;
    mrb_value proc;
} rb_block_t;
typedef struct rb_thread_struct
{
    mrb_sym self;
    rb_vm_t *vm;

    /* execution information */
    mrb_value *stack;		/* must free, must mark */
    unsigned long stack_size;
    rb_control_frame_t *cfp;
    int safe_level;
    int raised_flag;
    mrb_value last_status; /* $? */

    /* passing state */
    int state;

    /* for rb_iterate */
    const rb_block_t *passed_block;

    /* for bmethod */
    //const rb_method_entry_t *passed_me;

    /* for load(true) */
    mrb_value top_self;
    int top_wrapper;

    /* eval env */
    //rb_block_t *base_block;

    mrb_value *local_lfp;
    mrb_value local_svar;

    /* thread control */
    rb_thread_id_t thread_id;
    enum rb_thread_status status;
    int priority;
    int slice;

    native_thread_data_t native_thread_data;
    void *blocking_region_buffer;

    mrb_value thgroup;
    mrb_value value;

    mrb_value errinfo;
    mrb_value thrown_errinfo;
    int exec_signal;

    int interrupt_flag;
    rb_thread_lock_t interrupt_lock;
    struct rb_unblock_callback unblock;
    mrb_value locking_mutex;
    struct rb_mutex_struct *keeping_mutexes;
    int transition_for_lock;

    struct rb_vm_tag *tag;
    struct rb_vm_protect_tag *protect_tag;

    int parse_in_eval;
    int mild_compile_error;

    /* storage */
    void *local_storage;
  #if USE_VALUE_CACHE
    mrb_value value_cache[RUBY_VM_VALUE_CACHE_SIZE + 1];
    mrb_value *value_cache_ptr;
  #endif

    struct rb_thread_struct *join_list_next;
    struct rb_thread_struct *join_list_head;

    mrb_value first_proc;
    mrb_value first_args;
    mrb_value (*first_func)(ANYARGS);

    /* for GC */
    mrb_value *machine_stack_start;
    mrb_value *machine_stack_end;
    size_t machine_stack_maxsize;
  #ifdef __ia64
    mrb_value *machine_register_stack_start;
    mrb_value *machine_register_stack_end;
    size_t machine_register_stack_maxsize;
  #endif
    jmp_buf machine_regs;
    int mark_stack_len;

    /* statistics data for profiler */
    mrb_value stat_insn_usage;

    /* tracer */
    rb_event_hook_t *event_hooks;
    rb_event_flag_t event_flags;
    int tracing;

    /* fiber */
    mrb_value fiber;
    mrb_value root_fiber;
    rb_jmpbuf_t root_jmpbuf;

    /* misc */
    int method_missing_reason;
    int abort_on_exception;
  #ifdef USE_SIGALTSTACK
    void *altstack;
  #endif
} rb_thread_t;

#ifndef FUNC_FASTCALL
#define FUNC_FASTCALL(x) x
#endif

typedef rb_control_frame_t *
  (FUNC_FASTCALL(*rb_insn_func_t))(rb_thread_t *, rb_control_frame_t *);


//typedef unsigned long rb_thread_cond_t;       /* identify a condition variable */
//typedef struct rb_mutex_struct
//{
//    rb_thread_lock_t lock;
//    rb_thread_cond_t cond;
//    struct rb_thread_struct volatile *th;
//    volatile int cond_waiting, cond_notified;
//    struct rb_mutex_struct *next_mutex;
//} mutex_t;

void *rb_check_typeddata(mrb_state *mrb, mrb_value, const rb_data_type_t *);
#define TypedData_Get_Struct(mrb,obj,type,data_type,sval) do {\
    sval = (type*)rb_check_typeddata(mrb, obj, data_type); \
} while (0)

#define rb_setjmp(env) RUBY_SETJMP(env)
#define ruby_setjmp(env) RUBY_SETJMP(env)
#define ruby_longjmp(env,val) RUBY_LONGJMP(env,val)
//typedef RUBY_JMP_BUF rb_jmpbuf_t;
#define rb_jmp_buf rb_jmpbuf_t
struct rb_vm_tag {
    rb_jmpbuf_t buf;
    int tag;
    int retval;
    struct rb_vm_tag *prev;
};

//int RUBY_SETJMP(rb_jmpbuf_t env);
//void RUBY_LONGJMP(rb_jmpbuf_t env, int val);


//#if RUBY_VM_THREAD_MODEL == 2
extern rb_thread_t *ruby_current_thread;
extern rb_vm_t *ruby_current_vm;

//#define FIXNUM_MAX (LONG_MAX>>1)
#define POSFIXABLE(f) ((f) < FIXNUM_MAX+1)

#define GET_VM() ruby_current_vm
#define GET_THREAD() ruby_current_thread
#define rb_thread_set_current_raw(th) (void)(ruby_current_thread = (th))
#define rb_thread_set_current(th) do { \
    rb_thread_set_current_raw(th); \
    th->vm->running_thread = th; \
} while (0)

//#else
//#error "unsupported thread model"
//#endif

#define RUBY_VM_SET_INTERRUPT(th) ((th)->interrupt_flag |= 0x02)
#define RUBY_VM_SET_TIMER_INTERRUPT(th) ((th)->interrupt_flag |= 0x01)
#define RUBY_VM_SET_FINALIZER_INTERRUPT(th) ((th)->interrupt_flag |= 0x04)
#define RUBY_VM_INTERRUPTED(th) ((th)->interrupt_flag & 0x02)

#define TH_PUSH_TAG(th) do { \
  rb_thread_t * const _th = th; \
  struct rb_vm_tag _tag; \
  _tag.tag = 0; \
  _tag.prev = _th->tag; \
  _th->tag = &_tag;

#define TH_POP_TAG() \
  _th->tag = _tag.prev; \
} while (0)

#define TH_POP_TAG2() \
  _th->tag = _tag.prev

#define PUSH_TAG() TH_PUSH_TAG(GET_THREAD())
#define POP_TAG()      TH_POP_TAG()

#define TH_EXEC_TAG() ruby_setjmp(_th->tag->buf)

#define EXEC_TAG() \
  TH_EXEC_TAG()

#define TH_JUMP_TAG(th, st) do { \
  ruby_longjmp(th->tag->buf,(st)); \
} while (0)

#define JUMP_TAG(st) TH_JUMP_TAG(GET_THREAD(), st)

mrb_value convert_type(mrb_state *mrb, mrb_value val, const char *tname, const char *method, mrb_int raise);
//void native_cond_initialize(pthread_cond_t *cond);
//void native_cond_destroy(pthread_cond_t *cond);
//void native_mutex_initialize(pthread_mutex_t *lock);
//static void native_mutex_destroy(pthread_mutex_t *lock);
//void native_mutex_lock(pthread_mutex_t *lock);
//void native_mutex_unlock(pthread_mutex_t *lock);

int st_strcasecmp(const char *s1, const char *s2);
int st_strncasecmp(const char *s1, const char *s2, size_t n);
#define STRCASECMP(s1, s2) (st_strcasecmp(s1, s2))
#define STRNCASECMP(s1, s2, n) (st_strncasecmp(s1, s2, n))


void rb_enable_interrupt(void);
void rb_disable_interrupt(void);

#define EXEC_EVENT_HOOK(mrb, th, flag, self, id, klass) do { \
    rb_event_flag_t wait_event__ = th->event_flags; \
    if (UNLIKELY(wait_event__)) { \
	if (wait_event__ & (flag | RUBY_EVENT_VM)) { \
	    rb_threadptr_exec_event_hooks(mrb, th, flag, self, id, klass); \
	} \
    } \
} while (0)

/* likely */
#if __GNUC__ >= 3
#define LIKELY(x)   (__builtin_expect((x), 1))
#define UNLIKELY(x) (__builtin_expect((x), 0))
#else /* __GNUC__ >= 3 */
#define LIKELY(x)   (x)
#define UNLIKELY(x) (x)
#endif /* __GNUC__ >= 3 */

#if __GNUC__ >= 3
#define UNINITIALIZED_VAR(x) x = x
#else
#define UNINITIALIZED_VAR(x) x
#endif

enum ruby_value_type {
    RUBY_T_NONE   = 0x00,

    RUBY_T_OBJECT = 0x01,
    RUBY_T_CLASS  = 0x02,
    RUBY_T_MODULE = 0x03,
    RUBY_T_FLOAT  = 0x04,
    RUBY_T_STRING = 0x05,
    RUBY_T_REGEXP = 0x06,
    RUBY_T_ARRAY  = 0x07,
    RUBY_T_HASH   = 0x08,
    RUBY_T_STRUCT = 0x09,
    RUBY_T_BIGNUM = 0x0a,
    RUBY_T_FILE   = 0x0b,
    RUBY_T_DATA   = 0x0c,
    RUBY_T_MATCH  = 0x0d,
    RUBY_T_COMPLEX  = 0x0e,
    RUBY_T_RATIONAL = 0x0f,

    RUBY_T_NIL    = 0x11,
    RUBY_T_TRUE   = 0x12,
    RUBY_T_FALSE  = 0x13,
    RUBY_T_SYMBOL = 0x14,
    RUBY_T_FIXNUM = 0x15,

    RUBY_T_UNDEF  = 0x1b,
    RUBY_T_NODE   = 0x1c,
    RUBY_T_ICLASS = 0x1d,
    RUBY_T_ZOMBIE = 0x1e,

    RUBY_T_MASK   = 0x1f
};
#define T_NODE   RUBY_T_NODE
#define T_MASK   RUBY_T_MASK
#define BUILTIN_TYPE(x) (int)(((struct RBasic*)(x))->flags & T_MASK)
#define RUBY_VM_IFUNC_P(ptr)        (BUILTIN_TYPE(ptr) == T_NODE)
void rb_threadptr_execute_interrupts(mrb_state *mrb, rb_thread_t *th);
#define RUBY_VM_CHECK_INTS_TH(mrb, th) do { \
  if (UNLIKELY(th->interrupt_flag)) { \
    rb_threadptr_execute_interrupts(mrb, th); \
  } \
} while (0)
#define RUBY_VM_CHECK_INTS(mrb) \
  RUBY_VM_CHECK_INTS_TH(mrb, GET_THREAD())

void rb_thread_schedule(mrb_state * mrb);
int rb_thread_method_id_and_class(mrb_state *mrb,
                  rb_thread_t *th, mrb_sym *idp, mrb_value *klassp);
void rb_threadptr_exec_event_hooks(mrb_state *mrb, rb_thread_t *th, rb_event_flag_t flag, mrb_value self, mrb_sym id, mrb_value klass);

int rb_thread_select(mrb_state *mrb, int max, fd_set * read, fd_set * write, fd_set * except,
         struct timeval *timeout);
mrb_sym rb_frame_this_func(mrb_state *mrb);
int ruby_cleanup(mrb_state *mrb, volatile int ex);
int rb_obj_id(mrb_value obj);


typedef unsigned long rb_num_t;
typedef struct {
    rb_block_t block;

    mrb_value envval;		/* for GC mark */
    mrb_value blockprocval;
    int safe_level;
    int is_from_method;
    int is_lambda;
} rb_proc_t;

#if 1
#define GetCoreDataFromValue(obj, type, ptr) do { \
    ptr = (type*)DATA_PTR(obj); \
} while (0)
#else
#define GetCoreDataFromValue(obj, type, ptr) Data_Get_Struct(obj, type, ptr)
#endif

#define RUBY_VM_THREAD_STACK_SIZE (128 * 1024)
#define GetProcPtr(obj, ptr) \
  GetCoreDataFromValue(obj, rb_proc_t, ptr)
#define GetThreadPtr(obj, ptr) \
  GetCoreDataFromValue(obj, rb_thread_t, ptr)
mrb_value rb_thread_alloc(mrb_state *mrb, struct RClass* klass);

#if defined PRIdPTR && !defined PRI_VALUE_PREFIX
#define PRIdVALUE PRIdPTR
#define PRIiVALUE PRIiPTR
#define PRIoVALUE PRIoPTR
#define PRIuVALUE PRIuPTR
#define PRIxVALUE PRIxPTR
#define PRIXVALUE PRIXPTR
#else
#define PRIdVALUE PRI_VALUE_PREFIX"d"
#define PRIiVALUE PRI_VALUE_PREFIX"i"
#define PRIoVALUE PRI_VALUE_PREFIX"o"
#define PRIuVALUE PRI_VALUE_PREFIX"u"
#define PRIxVALUE PRI_VALUE_PREFIX"x"
#define PRIXVALUE PRI_VALUE_PREFIX"X"
#endif
#ifndef PRI_VALUE_PREFIX
# define PRI_VALUE_PREFIX ""
#endif

void rb_threadptr_signal_raise(mrb_state *mrb, rb_thread_t *th, int sig);
void rb_threadptr_signal_exit(mrb_state *mrb, rb_thread_t *th);
void rb_trap_exit(mrb_state *mrb);
mrb_value rb_vm_top_self(void);
mrb_value rb_exec_recursive_paired(mrb_state *mrb, mrb_value (*func) (mrb_state *, mrb_value, mrb_value, int),
                                  mrb_value obj, mrb_value paired_obj, void* arg);
mrb_value rb_exec_recursive_outer(mrb_state *mrb, mrb_value (*func) (mrb_state *, mrb_value, mrb_value, int),
                                  mrb_value obj, mrb_value arg);
mrb_sym rb_to_id(mrb_state *mrb, mrb_value name);
mrb_sym rb_frame_callee(mrb_state *mrb);
void ruby_sig_finalize(void);

#endif //MTHREAD_H
