#include "../../mruby.h"
#include "mecrobot.h"

#define GYRO_OFFSET  600 /* ジャイロセンサオフセット値(角速度0[deg/sec]時) */

static void
balance_run(int forward, int turn)
{
  signed char pwm_L, pwm_R; /* 左右モータPWM出力 */

  if ( abs(forward) > 99 || abs(turn) > 99)
    forward = turn = 0;

  /* 倒立振子制御(forward = 0, turn = 0で静止バランス) */
  balance_control(
          (float)forward,                              /* 前後進命令(+:前進, -:後進) */
          (float)turn,                                 /* 旋回命令(+:右旋回, -:左旋回) */
          (float)ecrobot_get_gyro_sensor(NXT_PORT_S4), /* ジャイロセンサ値 */
          (float)GYRO_OFFSET,                          /* ジャイロセンサオフセット値 */
          (float)nxt_motor_get_count(NXT_PORT_C),      /* 左モータ回転角度[deg] */
          (float)nxt_motor_get_count(NXT_PORT_B),      /* 右モータ回転角度[deg] */
          (float)ecrobot_get_battery_voltage(),        /* バッテリ電圧[mV] */
          &pwm_L,                                      /* 左モータPWM出力値 */
          &pwm_R);                                     /* 右モータPWM出力値 */

  if ( abs(pwm_L) > 99 && abs(pwm_R) > 99 ) {
      pwm_L = pwm_R = 0;
  }
  nxt_motor_set_speed(NXT_PORT_C, pwm_L, 1); /* 左モータPWM出力セット(-100～100) */
  nxt_motor_set_speed(NXT_PORT_B, pwm_R, 1); /* 右モータPWM出力セット(-100～100) */

  return;
}

static mrb_value
run(mrb_state *mrb, mrb_value self)
{
  mrb_int forward, turn;
#ifdef TIME_CNT
  static int msec = 0;
  static int sec = 0;
#endif

  mrb_get_args(mrb, "ii", &forward, &turn);

  balance_run(forward, turn);

  //  ecrobot_show_int(forward*1000 + turn);

  systick_wait_ms(4); /* 4msecウェイト */

#ifdef TIME_CNT
  msec += 4;
  if ( msec >= 1000 ) {
    sec++;
    msec = 0;
  }

  if ( (msec % 250) == 0 )
  ecrobot_show_int((sec*1000) + msec);
#endif

  return mrb_nil_value();
}


mrb_value
get_light_sensor(mrb_state *mrb, mrb_value self)
{
  mrb_int light_sensor;
  mrb_int port;

  mrb_get_args(mrb, "i", &port);

  light_sensor = ecrobot_get_light_sensor(NXT_PORT_S3);
  //  ecrobot_show_int(light_sensor);

  return mrb_fixnum_value(light_sensor);
}

void
mrb_init_ecrobot(mrb_state *mrb)
{
  struct RClass *ecrobot;

  ecrobot = mrb_define_class(mrb, "ECRobot", mrb->object_class);

  mrb_define_method(mrb, ecrobot, "run", run, ARGS_REQ(2));
  mrb_define_method(mrb, ecrobot, "light", get_light_sensor, ARGS_NONE());

}
