/**********************************************************************

  hash.c -

  $Author: yugui $
  created at: Mon Nov 22 18:51:18 JST 1993

  Copyright (C) 1993-2007 Yukihiro Matsumoto
  Copyright (C) 2000  Network Applied Communication Laboratory, Inc.
  Copyright (C) 2000  Information-technology Promotion Agency, Japan

**********************************************************************/

#include "mruby.h"
#include "mhash.h"
#include "khash.h"
#include "mclass.h"
#include "marray.h"
#include "error.h"
#include "mstring.h"
#include "mnumeric.h"
#include "mstruct.h"
#include "st.h"
#include <errno.h>
#include <string.h>

#ifdef __APPLE__
#include <crt_externs.h>
#endif

KHASH_MAP_INIT_INT(iv, mrb_value);
#define mrb_secure(x)

mrb_value mrb_exec_recursive_paired(mrb_state *mrb, mrb_value (*func) (mrb_state *, mrb_value, mrb_value, int),
                                  mrb_value obj, mrb_value paired_obj, void* arg);

#ifdef INCLUDE_REGEXP
  #define mrb_usascii_str_new2 mrb_usascii_str_new_cstr
#else
  #define mrb_usascii_str_new2        mrb_str_new_cstr
  #define mrb_filesystem_str_new_cstr mrb_str_new_cstr
  #define mrb_locale_str_new          mrb_tainted_str_new
  #define mrb_str_hash_cmp            mrb_str_cmp
#endif

#ifndef FALSE
#define FALSE   0
#endif

#ifndef TRUE
#define TRUE    1
#endif

#define PATH_ENV "PATH"

//static mrb_value mrb_hash_s_try_convert(mrb_value, mrb_value);

//#define HASH_DELETED  FL_USER1
//#define HASH_PROC_DEFAULT FL_USER2
static void mrb_hash_modify(mrb_state *mrb, mrb_value hash);

mrb_value
mrb_hash_new_capa(mrb_state *mrb, size_t capa)
{
  struct RHash *h;

  h = mrb_obj_alloc(mrb, MRB_TT_HASH, mrb->hash_class);
  h->buf = mrb_malloc(mrb, sizeof(mrb_value)*2*capa);
  h->capa = capa;
  h->size = 0;
  h->ifnone = mrb_nil_value(); /* kusuda */
#ifndef MRB_SLIM_RHASH //change by yamashita
  h->procdefault = mrb_nil_value(); /* kusuda */
#endif
  return mrb_obj_value(h);
}

mrb_value
mrb_hash_new(mrb_state *mrb, int capa)
{
  return mrb_hash_new_capa(mrb, 0);
}

mrb_value
mrb_hash_get(mrb_state *mrb, mrb_value hash, mrb_value key) /* mrb_hash_aref */ /* mrb_hash_lookup */
{
  struct RHash *h = mrb_hash_ptr(hash);
  int i;

  for (i=0; i<h->size; i++) {
    if (mrb_equal(mrb, h->buf[i].key, key)) {
      return h->buf[i].val;
    }
  }
  /* not found */
  if (mrb_nil_p(RHASH_PROCDEFAULT(hash))) {
    return RHASH_IFNONE(hash);
  }
  else {
    return mrb_funcall(mrb, RHASH_PROCDEFAULT(hash), "call", 2, hash, key);
  }
}

mrb_value
mrb_hash_getWithDef(mrb_state *mrb, mrb_value hash, mrb_value vkey, mrb_value def) /* mrb_hash_lookup2 */
{
  struct RHash *h = mrb_hash_ptr(hash);
  int i;

  //mrb_value vkey = mrb_fixnum_value(key);
  for (i=0; i<h->size; i++) {
    if (mrb_obj_equal(mrb, h->buf[i].key, vkey)) {
      return h->buf[i].val;
    }
  }
  /* not found */
  return def;
}

void
mrb_hash_set(mrb_state *mrb, mrb_value hash, mrb_value key, mrb_value val) /* mrb_hash_aset */
{
  struct RHash *h = mrb_hash_ptr(hash);
  int i;

  mrb_hash_modify(mrb, hash);
  for (i=0; i<h->size; i++) {
    if (mrb_equal(mrb, h->buf[i].key, key)) {
      h->buf[i].val = val;
      mrb_write_barrier(mrb, (struct RBasic*)h);
      return;
    }
  }
  /* expand */
  if (h->size == h->capa) {
    /* resize */
    h->buf = mrb_realloc(mrb, h->buf, sizeof(mrb_value)*2*(h->capa+1));
    h->capa++;
  }
  h->buf[h->size].key = key;
  h->buf[h->size].val = val;
  mrb_write_barrier(mrb, (struct RBasic*)h);
  h->size++; /* kusuda */
}

mrb_value
mrb_hash_freeze(mrb_value hash)
{
  //return mrb_obj_freeze(hash);
  return (hash);
}


static int
mrb_any_cmp(mrb_state *mrb, mrb_value a, mrb_value b)
{
    if (mrb_obj_equal(mrb, a, b)) return 0;
    if (FIXNUM_P(a) && FIXNUM_P(b)) {
        return (mrb_fixnum(a) != mrb_fixnum(b));
    }
    if (mrb_type(a) == MRB_TT_STRING && RBASIC(a)->c == mrb->string_class &&
        mrb_type(b) == MRB_TT_STRING && RBASIC(b)->c == mrb->string_class) {
        return mrb_str_hash_cmp(mrb, a, b);
    }
    if (a.tt == 0/*a == Qundef*/ || b.tt == 0/*b == Qundef*/) return -1;
    if (SYMBOL_P(a) && SYMBOL_P(b)) {
        return (!mrb_obj_equal(mrb, a, b));
    }

    return !mrb_eql(mrb, a, b);
}

mrb_value
mrb_hash(mrb_state *mrb, mrb_value obj)
{
  mrb_value hval = mrb_funcall(mrb, obj, "Hash", 0);
retry:
  switch (mrb_type(hval)) {
    case MRB_TT_FIXNUM:
      return hval;

    default:
      hval = mrb_to_int(mrb, hval);
      goto retry;
  }
}

st_index_t
mrb_hash_start(st_index_t h)
{
    return st_hash_start(h);
}

static st_index_t
mrb_any_hash(mrb_state *mrb, mrb_value a)
{
    mrb_value hval;
    st_index_t hnum;

    switch (mrb_type(a)) {
      case MRB_TT_FIXNUM:
      case MRB_TT_SYMBOL:
      //case MRB_TT_NIL:
      case MRB_TT_FALSE:
      case MRB_TT_TRUE:
        hnum = mrb_hash_end(mrb_hash_start((unsigned int)mrb_fixnum(a)));
        break;

      case MRB_TT_STRING:
        hnum = mrb_str_hash(mrb, a);
        break;

      default:
        hval = mrb_hash(mrb, a);
        hnum = mrb_fixnum(hval);
    }
    hnum <<= 1;
    return (st_index_t)RSHIFT(hnum, 1);
}

static const struct st_hash_type objhash = {
    mrb_any_cmp,
    mrb_any_hash,
};

static const struct st_hash_type identhash = {
    st_numcmp,
    st_numhash,
};

typedef int st_foreach_func(mrb_sym, void*, void *);

struct foreach_safe_arg {
    st_table *tbl;
    st_foreach_func *func;
    void *arg;
};

static int
foreach_safe_i(mrb_state *mrb, mrb_sym key, void* value, struct foreach_safe_arg *arg)
{
    int status;

    if (key == 0xffffffff/*key == Qundef*/) return ST_CONTINUE;
    status = (*arg->func)(key, value, arg->arg);
    if (status == ST_CONTINUE) {
        return ST_CHECK;
    }
    return status;
}

void
st_foreach_safe(mrb_state *mrb, void *table, int (*func)(ANYARGS), void* a)
{
    struct foreach_safe_arg arg;

    arg.tbl = table;
    arg.func = (st_foreach_func *)func;
    arg.arg = a;
    if (st_foreach(table, foreach_safe_i, (st_data_t)&arg)) {
        mrb_raise(mrb, mrb->eRuntimeError_class, "hash modified during iteration");
    }
}

typedef int mrb_foreach_func(mrb_value, mrb_value, void*);

struct hash_foreach_arg {
    mrb_value hash;
    mrb_foreach_func *func;
    void * arg;
};

static int
hash_foreach_iter(mrb_state *mrb, st_data_t key, st_data_t value, struct hash_foreach_arg *arg)
{
  int status;
  st_table *tbl;
  //mrb_value Qundef;
  //Qundef.tt = MRB_TT_FREE;

  tbl = (st_table *)RHASH(arg->hash)->iv;//ntbl;
  if (key == 0xffffffff/*Qundef*/) return ST_CONTINUE;
  status = (arg->func)(mrb_fixnum_value(key), mrb_fixnum_value(value), arg->arg);
  if ((st_table *)RHASH(arg->hash)->iv/*ntbl*/ != tbl) {
      mrb_raise(mrb, mrb->eRuntimeError_class, "rehash occurred during iteration");
  }
  switch (status) {
    case ST_DELETE:
      st_delete_safe(tbl, &key, 0, 0xffffffff/*Qundef*/);
      //FL_SET(arg->hash, HASH_DELETED);
    case ST_CONTINUE:
      break;
    case ST_STOP:
      return ST_STOP;
  }
  return ST_CHECK;
}

static mrb_value
hash_foreach_ensure(mrb_value hash)
{
  mrb_value Qundef;
  Qundef.tt = MRB_TT_FREE;

  RHASH(hash)->size--;//RHASH(hash)->iter_lev--;

  if (RHASH(hash)->size == 0) {//if (RHASH(hash)->iter_lev == 0) {
    //if (FL_TEST(hash, HASH_DELETED)) {
    //    st_cleanup_safe(RHASH(hash)->iv/*ntbl*/, Qundef);
    //    FL_UNSET(hash, HASH_DELETED);
    //}
  }
  return mrb_fixnum_value(0);
}

#if HASH_CLASS_METHOD == BASICHASH
int
bs_foreach(struct RHash* table, int (*func)(ANYARGS), void* arg)
{
//enum st_retval {ST_CONTINUE, ST_STOP, ST_DELETE, ST_CHECK};
    int/*enum st_retval*/ retval;
    st_index_t i;

    if (table->size) {
      for (i = 0; i < table->size; i++) {
        mrb_value key, val;
        key = table->buf[i].key;
        val = table->buf[i].val;
        retval = (*func)(key, val, arg);
        switch (retval) {
          case 0/*ST_CONTINUE*/:
            break;
          case 1/*ST_STOP*/:
            return 0;
        }
      }
      return 0;
    }
  return 0;
}
#endif

static int
hash_foreach_call(mrb_state *mrb, struct hash_foreach_arg *arg)
{
#if HASH_CLASS_METHOD == BASICHASH
  if (bs_foreach(RHASH(arg->hash), hash_foreach_iter, arg)) {
#else
  if (st_foreach((st_table *)RHASH(arg->hash)->iv/*ntbl*/, hash_foreach_iter, (st_data_t)arg)) {
#endif
    mrb_raise(mrb, mrb->eRuntimeError_class, "hash modified during iteration");
  }
  return 0/*mrb_nil_value()*/;
}

void
mrb_hash_foreach(mrb_value hash, int (*func)(ANYARGS), void *farg)
{
  struct hash_foreach_arg arg;

  if (!RHASH(hash)->iv/*ntbl*/)
    return;
  //RHASH(hash)->iter_lev++;
  arg.hash = hash;
  arg.func = (mrb_foreach_func *)func;
  arg.arg  = farg;
  hash_foreach_call((mrb_state*)0, &arg);
  hash_foreach_ensure(hash);
}

static mrb_value
hash_alloc(mrb_state *mrb, mrb_value self)
{
  mrb_value hash;

  mrb_get_args(mrb, "o", &hash);
  struct RHash *h = RHASH(hash);
  int capa=0;
  h->buf = mrb_malloc(mrb, sizeof(mrb_value)*2*capa);
  h->capa = capa;
  h->size = 0;
  h->ifnone = mrb_nil_value(); /* kusuda */

#ifndef MRB_SLIM_RHASH //added by yamashita
  h->procdefault = mrb_nil_value(); /* kusuda */
#endif
  return mrb_obj_value(h);
}

static mrb_value
hash_s_new(mrb_state *mrb, mrb_value klass)
{
  mrb_value *argv;
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);
  mrb_value hash = mrb_hash_new_capa(mrb, 0);
  mrb_obj_call_init(mrb, hash, argc, argv);
  return hash;
}

mrb_value
rb_hash_new(mrb_state *mrb, mrb_value hash)
{
    return hash_alloc(mrb, hash);
}

mrb_value
mrb_hash_dup(mrb_state *mrb, mrb_value hash)
{
  struct RHash* ret;

  //NEWOBJ(ret, struct RHash);
  //DUPSETUP(ret, hash);
  ret = mrb_obj_alloc(mrb, MRB_TT_HASH, mrb->hash_class);

  if (!RHASH_EMPTY_P(hash))
      ret->iv/*ntbl*/ = (struct kh_iv *)st_copy((st_table *)RHASH(hash)->iv/*ntbl*/);
  //if (FL_TEST(hash, HASH_PROC_DEFAULT)) {
  //    FL_SET(ret, HASH_PROC_DEFAULT);
  //}
  //RHASH_IFNONE(ret) = RHASH_IFNONE(hash);
  return mrb_obj_value(ret);
}

static void
mrb_hash_modify_check(mrb_state *mrb, mrb_value hash)
{
  //if (OBJ_FROZEN(hash)) mrb_error_frozen("hash");
  //if (!OBJ_UNTRUSTED(hash) && mrb_safe_level() >= 4)
  //  mrb_raise(mrb, mrb->eSecurityError_class, "Insecure: can't modify hash");
}

struct st_table *
mrb_hash_tbl(mrb_value hash)
{
    if (!RHASH(hash)->iv/*ntbl*/) {
        RHASH(hash)->iv/*ntbl*/ = (struct kh_iv *)st_init_table(&objhash);
    }
    return (st_table *)RHASH(hash)->iv/*ntbl*/;
}

static void
mrb_hash_modify(mrb_state *mrb, mrb_value hash)
{
  //mrb_hash_modify_check(mrb, hash);
  mrb_hash_tbl(hash);
}

/* 15.2.13.4.16 */
/*
 *  call-seq:
 *     Hash.new                          -> new_hash
 *     Hash.new(obj)                     -> new_hash
 *     Hash.new {|hash, key| block }     -> new_hash
 *
 *  Returns a new, empty hash. If this hash is subsequently accessed by
 *  a key that doesn't correspond to a hash entry, the value returned
 *  depends on the style of <code>new</code> used to create the hash. In
 *  the first form, the access returns <code>nil</code>. If
 *  <i>obj</i> is specified, this single object will be used for
 *  all <em>default values</em>. If a block is specified, it will be
 *  called with the hash object and the key, and should return the
 *  default value. It is the block's responsibility to store the value
 *  in the hash if required.
 *
 *     h = Hash.new("Go Fish")
 *     h["a"] = 100
 *     h["b"] = 200
 *     h["a"]           #=> 100
 *     h["c"]           #=> "Go Fish"
 *     # The following alters the single default object
 *     h["c"].upcase!   #=> "GO FISH"
 *     h["d"]           #=> "GO FISH"
 *     h.keys           #=> ["a", "b"]
 *
 *     # While this creates a new default object each time
 *     h = Hash.new { |hash, key| hash[key] = "Go Fish: #{key}" }
 *     h["c"]           #=> "Go Fish: c"
 *     h["c"].upcase!   #=> "GO FISH: C"
 *     h["d"]           #=> "Go Fish: d"
 *     h.keys           #=> ["c", "d"]
 *
 */

static mrb_value
mrb_hash_init_core(mrb_state *mrb, mrb_value hash)
{
  mrb_value block;
  mrb_value *argv;
  int argc;

  mrb_get_args(mrb, "o*", &block, &argv, &argc);

  mrb_hash_modify(mrb, hash);

  if (mrb_nil_p(block)) {
    if (argc > 0) {
      RHASH_IFNONE(hash) = argv[0];
    }
    else {
      RHASH_IFNONE(hash) = mrb_nil_value();
    }
    RHASH_PROCDEFAULT(hash) = mrb_nil_value();
  }
  else {
    if (argc > 0) {
      mrb_raise(mrb, mrb->eArgumentError_class, "wrong number of arguments");
    }
    RHASH_IFNONE(hash) = mrb_nil_value();
    RHASH_PROCDEFAULT(hash) = block;
  }

  return hash;
}

/*
 *  call-seq:
 *     Hash[ key, value, ... ]         -> new_hash
 *     Hash[ [ [key, value], ... ] ]   -> new_hash
 *     Hash[ object ]                  -> new_hash
 *
 *  Creates a new hash populated with the given objects. Equivalent to
 *  the literal <code>{ <i>key</i> => <i>value</i>, ... }</code>. In the first
 *  form, keys and values occur in pairs, so there must be an even number of arguments.
 *  The second and third form take a single argument which is either
 *  an array of key-value pairs or an object convertible to a hash.
 *
 *     Hash["a", 100, "b", 200]             #=> {"a"=>100, "b"=>200}
 *     Hash[ [ ["a", 100], ["b", 200] ] ]   #=> {"a"=>100, "b"=>200}
 *     Hash["a" => 100, "b" => 200]         #=> {"a"=>100, "b"=>200}
 */

static mrb_value
to_hash(mrb_state *mrb, mrb_value hash)
{
    return mrb_convert_type(mrb, hash, MRB_TT_HASH, "Hash", "to_hash");
}

/*
 *  call-seq:
 *     Hash.try_convert(obj) -> hash or nil
 *
 *  Try to convert <i>obj</i> into a hash, using to_hash method.
 *  Returns converted hash or nil if <i>obj</i> cannot be converted
 *  for any reason.
 *
 *     Hash.try_convert({1=>2})   # => {1=>2}
 *     Hash.try_convert("1=>2")   # => nil
 */

/* 15.2.13.4.2  */
/*
 *  call-seq:
 *     hsh[key]    ->  value
 *
 *  Element Reference---Retrieves the <i>value</i> object corresponding
 *  to the <i>key</i> object. If not found, returns the default value (see
 *  <code>Hash::new</code> for details).
 *
 *     h = { "a" => 100, "b" => 200 }
 *     h["a"]   #=> 100
 *     h["c"]   #=> nil
 *
 */
mrb_value
mrb_hash_aget(mrb_state *mrb, mrb_value self)
{
  mrb_value key;

  mrb_get_args(mrb, "o", &key);
  return mrb_hash_get(mrb, self, key);
}

mrb_value
mrb_hash_lookup2(mrb_value hash, mrb_value key, mrb_value def)
{
    mrb_value val;

    if (!RHASH(hash)->iv/*ntbl*/ || !st_lookup((st_table *)RHASH(hash)->iv/*ntbl*/, mrb_fixnum(key), (st_data_t *)&val)) {
        return def; /* without Hash#default */
    }
    return val;
}

mrb_value
mrb_hash_lookup(mrb_value hash, mrb_value key)
{
    return mrb_hash_lookup2(hash, key, mrb_nil_value());
}

/*
 *  call-seq:
 *     hsh.fetch(key [, default] )       -> obj
 *     hsh.fetch(key) {| key | block }   -> obj
 *
 *  Returns a value from the hash for the given key. If the key can't be
 *  found, there are several options: With no other arguments, it will
 *  raise an <code>KeyError</code> exception; if <i>default</i> is
 *  given, then that will be returned; if the optional code block is
 *  specified, then that will be run and its result returned.
 *
 *     h = { "a" => 100, "b" => 200 }
 *     h.fetch("a")                            #=> 100
 *     h.fetch("z", "go fish")                 #=> "go fish"
 *     h.fetch("z") { |el| "go fish, #{el}"}   #=> "go fish, z"
 *
 *  The following example shows that an exception is raised if the key
 *  is not found and a default value is not supplied.
 *
 *     h = { "a" => 100, "b" => 200 }
 *     h.fetch("z")
 *
 *  <em>produces:</em>
 *
 *     prog.rb:2:in `fetch': key not found (KeyError)
 *      from prog.rb:2
 *
 */

/* 15.2.13.4.5  */
/*
 *  call-seq:
 *     hsh.default(key=nil)   -> obj
 *
 *  Returns the default value, the value that would be returned by
 *  <i>hsh</i>[<i>key</i>] if <i>key</i> did not exist in <i>hsh</i>.
 *  See also <code>Hash::new</code> and <code>Hash#default=</code>.
 *
 *     h = Hash.new                            #=> {}
 *     h.default                               #=> nil
 *     h.default(2)                            #=> nil
 *
 *     h = Hash.new("cat")                     #=> {}
 *     h.default                               #=> "cat"
 *     h.default(2)                            #=> "cat"
 *
 *     h = Hash.new {|h,k| h[k] = k.to_i*10}   #=> {}
 *     h.default                               #=> nil
 *     h.default(2)                            #=> 20
 */

static mrb_value
mrb_hash_default(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value hash)
{
  mrb_value *argv;
  int argc;
  mrb_value key;//, ifnone;

  mrb_get_args(mrb, "*", &argv, &argc);
#ifndef MRB_SLIM_RHASH //added by yamashita
  if (!mrb_nil_p(RHASH_PROCDEFAULT(hash))) {
#else
    if (!mrb_nil_p(RHASH_IFNONE(hash))) {
#endif
    if (argc == 0) return mrb_nil_value();
    key = argv[0];
#ifndef MRB_SLIM_RHASH //added by yamashita
    return RHASH_PROCDEFAULT(hash);
#else
    return RHASH_IFNONE(hash);
#endif
  }
  else {
    return RHASH_IFNONE(hash);
  }
  //mrb_scan_args(argc, argv, "01", &key);
  //ifnone = RHASH_IFNONE(hash);
  //if (FL_TEST(hash, HASH_PROC_DEFAULT)) {
  //  if (argc == 0) return mrb_nil_value();
  //  return mrb_funcall(ifnone, id_yield, 2, hash, key);
  //}
  //return ifnone;
  //return hash;
}

/* 15.2.13.4.6  */
/*
 *  call-seq:
 *     hsh.default = obj     -> obj
 *
 *  Sets the default value, the value returned for a key that does not
 *  exist in the hash. It is not possible to set the default to a
 *  <code>Proc</code> that will be executed on each key lookup.
 *
 *     h = { "a" => 100, "b" => 200 }
 *     h.default = "Go fish"
 *     h["a"]     #=> 100
 *     h["z"]     #=> "Go fish"
 *     # This doesn't do what you might hope...
 *     h.default = proc do |hash, key|
 *       hash[key] = key + key
 *     end
 *     h[2]       #=> #<Proc:0x401b3948@-:6>
 *     h["cat"]   #=> #<Proc:0x401b3948@-:6>
 */

static mrb_value
mrb_hash_set_default(mrb_state *mrb, mrb_value hash)
{
  mrb_value ifnone;
  mrb_get_args(mrb, "o", &ifnone);

  mrb_hash_modify(mrb, hash);
  RHASH_IFNONE(hash) = ifnone;
  //FL_UNSET(hash, HASH_PROC_DEFAULT);

#ifndef MRB_SLIM_RHASH //added by yamashita
  RHASH_PROCDEFAULT(hash) = mrb_nil_value();
#endif
  return ifnone;
}

/* 15.2.13.4.7  */
/*
 *  call-seq:
 *     hsh.default_proc -> anObject
 *
 *  If <code>Hash::new</code> was invoked with a block, return that
 *  block, otherwise return <code>nil</code>.
 *
 *     h = Hash.new {|h,k| h[k] = k*k }   #=> {}
 *     p = h.default_proc                 #=> #<Proc:0x401b3d08@-:1>
 *     a = []                             #=> []
 *     p.call(a, 2)
 *     a                                  #=> [nil, nil, 4]
 */


static mrb_value
mrb_hash_default_proc(mrb_state *mrb, mrb_value hash)
{
    //if (FL_TEST(hash, HASH_PROC_DEFAULT)) {
    //    return RHASH_IFNONE(hash);
    //}
    //return mrb_nil_value();
#ifndef MRB_SLIM_RHASH
    return RHASH_PROCDEFAULT(hash);
#else
    return RHASH_IFNONE(hash);
#endif
}

mrb_value mrb_obj_is_proc(mrb_value proc);

/*
 *  call-seq:
 *     hsh.default_proc = proc_obj     -> proc_obj
 *
 *  Sets the default proc to be executed on each key lookup.
 *
 *     h.default_proc = proc do |hash, key|
 *       hash[key] = key + key
 *     end
 *     h[2]       #=> 4
 *     h["cat"]   #=> "catcat"
 */

mrb_value
mrb_hash_delete_key(mrb_state *mrb, mrb_value hash, mrb_value key)
{
  mrb_value Qundef;
  Qundef.tt = MRB_TT_FREE;

#if HASH_CLASS_METHOD == BASICHASH
  struct RHash *h = mrb_hash_ptr(hash);
  mrb_value delVal;
  int i;
  int len;
  if (h->size > 0) {
    for (i=0; i<h->size; i++) {
      if (mrb_equal(mrb, h->buf[i].key, key)) {
        delVal = h->buf[i].val;
        h->size--;
        len = sizeof(struct mrb_value) * 2*(h->size-i);
        memmove(&h->buf[i], &h->buf[i+1],
                      len);
        h->buf[h->size].key = mrb_nil_value();
        h->buf[h->size].val = mrb_nil_value();
        return delVal;;
      }
    }
  }
  return Qundef;
#else
  st_data_t ktmp = mrb_fixnum(key);
  st_data_t val;

  if (!RHASH(hash)->iv/*ntbl*/)
    return Qundef;
  if (RHASH(hash)->size > 0) {//if (RHASH(hash)->size/*iter_lev*/ > 0) {
    //if (st_delete_safe(RHASH(hash)->iv/*ntbl*/, &ktmp, &val, Qundef)) {
    if (st_delete_safe((st_table *)(RHASH(hash)->iv), &ktmp, &val, 0/*Qundef*/)) {
      //FL_SET(hash, HASH_DELETED);
      return mrb_fixnum_value(val);
    }
  }
  else if (st_delete((st_table *)RHASH(hash)->iv/*ntbl*/, &ktmp, &val))
    return mrb_fixnum_value(val);
  return Qundef;
#endif
}

/* 15.2.13.4.8  */
/*
 *  call-seq:
 *     hsh.delete(key)                   -> value
 *     hsh.delete(key) {| key | block }  -> value
 *
 *  Deletes and returns a key-value pair from <i>hsh</i> whose key is
 *  equal to <i>key</i>. If the key is not found, returns the
 *  <em>default value</em>. If the optional code block is given and the
 *  key is not found, pass in the key and return the result of
 *  <i>block</i>.
 *
 *     h = { "a" => 100, "b" => 200 }
 *     h.delete("a")                              #=> 100
 *     h.delete("z")                              #=> nil
 *     h.delete("z") { |el| "#{el} not found" }   #=> "z not found"
 *
 */
mrb_value
mrb_hash_delete(mrb_state *mrb, mrb_value self)
{
  mrb_value key, val;

  mrb_get_args(mrb, "o", &key);

  //mrb_hash_modify(mrb, self);
  val = mrb_hash_delete_key(mrb, self, key);
  if (val.tt != MRB_TT_FREE) return val;
  //if (mrb_block_given_p()) {
  //  return mrb_yield(key);
  //}
  return mrb_nil_value();
}
struct shift_var {
    mrb_value key;
    mrb_value val;
};


/* 15.2.13.4.24 */
/*
 *  call-seq:
 *     hsh.shift -> anArray or obj
 *
 *  Removes a key-value pair from <i>hsh</i> and returns it as the
 *  two-item array <code>[</code> <i>key, value</i> <code>]</code>, or
 *  the hash's default value if the hash is empty.
 *
 *     h = { 1 => "a", 2 => "b", 3 => "c" }
 *     h.shift   #=> [1, "a"]
 *     h         #=> {2=>"b", 3=>"c"}
 */

static mrb_value
mrb_hash_shift(mrb_state *mrb, mrb_value hash)
{
  mrb_hash_modify(mrb, hash);
#if HASH_CLASS_METHOD == BASICHASH
  struct RHash *h = mrb_hash_ptr(hash);
  mrb_value delKey,delVal;
  int len;
  if (h->size > 0) {
    delKey = h->buf[0].key;
    delVal = h->buf[0].val;
    len = sizeof(struct mrb_value) * 2*(h->size-1);
    memmove(&h->buf[0], &h->buf[1], len);
    h->size--;
    h->buf[h->size].key = mrb_nil_value();
    h->buf[h->size].val = mrb_nil_value();
    mrb_value result = mrb_ary_new_capa(mrb, 2);
    mrb_ary_push(mrb, result, delKey);
    mrb_ary_push(mrb, result, delVal);
    return result;
  }
  return RHASH_IFNONE(hash);
#else
  mrb_value Qundef;
  struct shift_var var;
  Qundef.tt = MRB_TT_FREE;

  var.key = Qundef;
  mrb_hash_foreach(hash, RHASH(hash)->size/*iter_lev*/ > 0 ? shift_i_safe : shift_i,
          &var);

  if (var.key.tt != 0 /*var.key!= Qundef*/) {
      if (RHASH(hash)->size/*iter_lev*/ > 0) {
          mrb_hash_delete_key(mrb, hash, var.key);
      }
      return mrb_assoc_new(mrb, var.key, var.val);
  }
  //else if (FL_TEST(hash, HASH_PROC_DEFAULT)) {
  //    return mrb_funcall(RHASH_IFNONE(hash), id_yield, 2, hash, mrb_nil_value());
  //}
  else {
      return hash;//RHASH_IFNONE(hash);
  }
#endif
}

static int
delete_if_i(mrb_state *mrb, mrb_value key, mrb_value value, void * hash)
{
  if (key.tt == 0/*key == Qundef*/) return ST_CONTINUE;
  if (RTEST(mrb_yield_values(2, key, value))) {
      mrb_hash_delete_key(mrb, *(mrb_value*)hash, key);
  }
  return ST_CONTINUE;
}

/*
 *  call-seq:
 *     hsh.delete_if {| key, value | block }  -> hsh
 *     hsh.delete_if                          -> an_enumerator
 *
 *  Deletes every key-value pair from <i>hsh</i> for which <i>block</i>
 *  evaluates to <code>true</code>.
 *
 *  If no block is given, an enumerator is returned instead.
 *
 *     h = { "a" => 100, "b" => 200, "c" => 300 }
 *     h.delete_if {|key, value| key >= "b" }   #=> {"a"=>100}
 *
 */

mrb_value
mrb_hash_delete_if(mrb_state *mrb, mrb_value hash)
{
  //RETURN_ENUMERATOR(hash, 0, 0);
  mrb_hash_modify(mrb, hash);
  mrb_hash_foreach(hash, delete_if_i, &hash);
  return hash;
}

/*
 *  call-seq:
 *     hsh.reject! {| key, value | block }  -> hsh or nil
 *     hsh.reject!                          -> an_enumerator
 *
 *  Equivalent to <code>Hash#delete_if</code>, but returns
 *  <code>nil</code> if no changes were made.
 */

mrb_value
mrb_hash_reject_bang(mrb_state *mrb, mrb_value hash)
{
    st_index_t n;

    //RETURN_ENUMERATOR(hash, 0, 0);
    mrb_hash_modify(mrb, hash);
    if (!RHASH(hash)->iv/*ntbl*/)
        return mrb_nil_value();
    n = (st_index_t)RHASH(hash)->iv/*ntbl*/->n_occupied;
    mrb_hash_foreach(hash, delete_if_i, &hash);
    if (n == (st_index_t)RHASH(hash)->iv/*ntbl*/->n_occupied) return mrb_nil_value();
    return hash;
}

/*
 *  call-seq:
 *     hsh.reject {| key, value | block }  -> a_hash
 *
 *  Same as <code>Hash#delete_if</code>, but works on (and returns) a
 *  copy of the <i>hsh</i>. Equivalent to
 *  <code><i>hsh</i>.dup.delete_if</code>.
 *
 */

/*
 * call-seq:
 *   hsh.values_at(key, ...)   -> array
 *
 * Return an array containing the values associated with the given keys.
 * Also see <code>Hash.select</code>.
 *
 *   h = { "cat" => "feline", "dog" => "canine", "cow" => "bovine" }
 *   h.values_at("cow", "cat")  #=> ["bovine", "feline"]
 */

mrb_value
mrb_hash_values_at(mrb_state *mrb, int argc, mrb_value *argv, mrb_value hash)
{
    mrb_value result = mrb_ary_new_capa(mrb, argc);//mrb_ary_new2(argc);
    long i;

    for (i=0; i<argc; i++) {
        mrb_ary_push(mrb, result, mrb_hash_get(mrb, hash, argv[i]));
    }
    return result;
}

static int
select_i(mrb_state *mrb, mrb_value key, mrb_value value, void * result)
{
    if (key.tt == 0/*key == Qundef*/) return ST_CONTINUE;
    if (RTEST(mrb_yield_values(2, key, value)))
        mrb_hash_set/*mrb_hash_aset*/(mrb, *(mrb_value*)result, key, value);
    return ST_CONTINUE;
}

/*
 *  call-seq:
 *     hsh.select {|key, value| block}   -> a_hash
 *     hsh.select                        -> an_enumerator
 *
 *  Returns a new hash consisting of entries for which the block returns true.
 *
 *  If no block is given, an enumerator is returned instead.
 *
 *     h = { "a" => 100, "b" => 200, "c" => 300 }
 *     h.select {|k,v| k > "a"}  #=> {"b" => 200, "c" => 300}
 *     h.select {|k,v| v < 200}  #=> {"a" => 100}
 */

mrb_value
mrb_hash_select(mrb_state *mrb, mrb_value hash)
{
    mrb_value result;

    //RETURN_ENUMERATOR(hash, 0, 0);
    result = rb_hash_new(mrb, hash);
    mrb_hash_foreach(hash, select_i, &result);
    return result;
}

static int
keep_if_i(mrb_value key, mrb_value value, void *hash)
{
    if (key.tt == 0/*key == Qundef*/) return ST_CONTINUE;
    if (!RTEST(mrb_yield_values(2, key, value))) {
        return ST_DELETE;
    }
    return ST_CONTINUE;
}

/*
 *  call-seq:
 *     hsh.select! {| key, value | block }  -> hsh or nil
 *     hsh.select!                          -> an_enumerator
 *
 *  Equivalent to <code>Hash#keep_if</code>, but returns
 *  <code>nil</code> if no changes were made.
 */

mrb_value
mrb_hash_select_bang(mrb_state *mrb, mrb_value hash)
{
    st_index_t n;

    //RETURN_ENUMERATOR(hash, 0, 0);
    mrb_hash_modify(mrb, hash);
    if (!RHASH(hash)->iv/*ntbl*/)
        return mrb_nil_value();
    n = RHASH(hash)->iv/*ntbl*/->n_occupied;
    mrb_hash_foreach(hash, keep_if_i, &hash);
    if (n == RHASH(hash)->iv/*ntbl*/->n_occupied) return mrb_nil_value();
    return hash;
}

/*
 *  call-seq:
 *     hsh.keep_if {| key, value | block }  -> hsh
 *     hsh.keep_if                          -> an_enumerator
 *
 *  Deletes every key-value pair from <i>hsh</i> for which <i>block</i>
 *  evaluates to false.
 *
 *  If no block is given, an enumerator is returned instead.
 *
 */

mrb_value
mrb_hash_keep_if(mrb_state *mrb, mrb_value hash)
{
    //RETURN_ENUMERATOR(hash, 0, 0);
    mrb_hash_modify(mrb, hash);
    mrb_hash_foreach(hash, keep_if_i, &hash);
    return hash;
}

/* 15.2.13.4.4  */
/*
 *  call-seq:
 *     hsh.clear -> hsh
 *
 *  Removes all key-value pairs from <i>hsh</i>.
 *
 *     h = { "a" => 100, "b" => 200 }   #=> {"a"=>100, "b"=>200}
 *     h.clear                          #=> {}
 *
 */

static mrb_value
mrb_hash_clear(mrb_state *mrb, mrb_value hash)
{
  mrb_hash_modify_check(mrb, hash);
#if HASH_CLASS_METHOD == BASICHASH
  struct RHash *h = mrb_hash_ptr(hash);
  int i;
  if (h->size > 0) {
    for (i=0; i<h->size; i++) {
      h->buf[i].key = mrb_nil_value();
      h->buf[i].val = mrb_nil_value();
    }
    h->size = 0;
  }
#else
  if (!RHASH(hash)->iv/*ntbl*/)
      return hash;
  if (RHASH(hash)->iv/*ntbl*/->n_occupied > 0) {
      if (RHASH(hash)->size/*iter_lev*/ > 0)
          mrb_hash_foreach(hash, clear_i, 0);
      else
          st_clear((st_table *)RHASH(hash)->iv/*ntbl*/);
  }
#endif
  return hash;
}

/* 15.2.13.4.3  */
/* 15.2.13.4.26 */
/*
 *  call-seq:
 *     hsh[key] = value        -> value
 *     hsh.store(key, value)   -> value
 *
 *  Element Assignment---Associates the value given by
 *  <i>value</i> with the key given by <i>key</i>.
 *  <i>key</i> should not have its value changed while it is in
 *  use as a key (a <code>String</code> passed as a key will be
 *  duplicated and frozen).
 *
 *     h = { "a" => 100, "b" => 200 }
 *     h["a"] = 9
 *     h["c"] = 4
 *     h   #=> {"a"=>9, "b"=>200, "c"=>4}
 *
 */
mrb_value
mrb_hash_aset(mrb_state *mrb, mrb_value self)
{
  mrb_value key, val;

  mrb_get_args(mrb, "oo", &key, &val);
  mrb_hash_set(mrb, self, key, val);
  return val;
}

/* 15.2.13.4.17 */
/* 15.2.13.4.23 */
/*
 *  call-seq:
 *     hsh.replace(other_hash) -> hsh
 *
 *  Replaces the contents of <i>hsh</i> with the contents of
 *  <i>other_hash</i>.
 *
 *     h = { "a" => 100, "b" => 200 }
 *     h.replace({ "c" => 300, "d" => 400 })   #=> {"c"=>300, "d"=>400}
 *
 */

static mrb_value
mrb_hash_replace(mrb_state *mrb, mrb_value hash)
{
  mrb_value hash2;
  mrb_get_args(mrb, "o", &hash2);

  mrb_hash_modify_check(mrb, hash);
  hash2 = to_hash(mrb, hash2);
  if (mrb_obj_equal(mrb, hash, hash2)) return hash;
  mrb_hash_clear(mrb, hash);
#if HASH_CLASS_METHOD == BASICHASH
  struct RHash *h2 = mrb_hash_ptr(hash2);
  int i;
  if (h2->size > 0) {
    for (i=0; i<h2->size; i++) {
      mrb_hash_set(mrb, hash, h2->buf[i].key, h2->buf[i].val);
    }
  }
#else
  if (RHASH(hash2)->iv/*ntbl*/) {
      mrb_hash_tbl(hash);
      //RHASH(hash)->iv/*ntbl*/->type = RHASH(hash2)->iv/*ntbl*/->type;
  }
  mrb_hash_foreach(hash2, replace_i, &hash);
  //RHASH_IFNONE(hash) = RHASH_IFNONE(hash2);
  //if (FL_TEST(hash2, HASH_PROC_DEFAULT)) {
  //    FL_SET(hash, HASH_PROC_DEFAULT);
  //}
  //else {
  //    FL_UNSET(hash, HASH_PROC_DEFAULT);
  //}
#endif
  return hash;
}

/* 15.2.13.4.20 */
/* 15.2.13.4.25 */
/*
 *  call-seq:
 *     hsh.length    ->  fixnum
 *     hsh.size      ->  fixnum
 *
 *  Returns the number of key-value pairs in the hash.
 *
 *     h = { "d" => 100, "a" => 200, "v" => 300, "e" => 400 }
 *     h.length        #=> 4
 *     h.delete("a")   #=> 200
 *     h.length        #=> 3
 */
static mrb_value
mrb_hash_size_m(mrb_state *mrb, mrb_value self)
{
  struct RHash *h = mrb_hash_ptr(self);

  return mrb_fixnum_value(h->size);
}

/* 15.2.13.4.12 */
/*
 *  call-seq:
 *     hsh.empty?    -> true or false
 *
 *  Returns <code>true</code> if <i>hsh</i> contains no key-value pairs.
 *
 *     {}.empty?   #=> true
 *
 */
static mrb_value
mrb_hash_empty_p(mrb_state *mrb, mrb_value self)
{
  struct RHash *h = mrb_hash_ptr(self);

  if (h->size == 0)
    return mrb_true_value();
  return mrb_false_value();
}

static int
each_value_i(mrb_value key, void *value)
{
    if (key.tt == 0/*key == Qundef*/) return ST_CONTINUE;
    mrb_yield(*(mrb_value *)value);
    return ST_CONTINUE;
}

/* 15.2.13.4.11 */
/*
 *  call-seq:
 *     hsh.each_value {| value | block } -> hsh
 *     hsh.each_value                    -> an_enumerator
 *
 *  Calls <i>block</i> once for each key in <i>hsh</i>, passing the
 *  value as a parameter.
 *
 *  If no block is given, an enumerator is returned instead.
 *
 *     h = { "a" => 100, "b" => 200 }
 *     h.each_value {|value| puts value }
 *
 *  <em>produces:</em>
 *
 *     100
 *     200
 */

static mrb_value
mrb_hash_each_value(mrb_state *mrb, mrb_value hash)
{
    //RETURN_ENUMERATOR(hash, 0, 0);
    mrb_hash_foreach(hash, each_value_i, 0);
    return hash;
}

static int
each_key_i(mrb_value key, void *value)
{
    if (key.tt == 0/*key == Qundef*/) return ST_CONTINUE;
    mrb_yield(key);
    return ST_CONTINUE;
}

/* 15.2.13.4.10 */
/*
 *  call-seq:
 *     hsh.each_key {| key | block } -> hsh
 *     hsh.each_key                  -> an_enumerator
 *
 *  Calls <i>block</i> once for each key in <i>hsh</i>, passing the key
 *  as a parameter.
 *
 *  If no block is given, an enumerator is returned instead.
 *
 *     h = { "a" => 100, "b" => 200 }
 *     h.each_key {|key| puts key }
 *
 *  <em>produces:</em>
 *
 *     a
 *     b
 */
static mrb_value
mrb_hash_each_key(mrb_state *mrb, mrb_value hash)
{
    //RETURN_ENUMERATOR(hash, 0, 0);
    mrb_hash_foreach(hash, each_key_i, 0);
    return hash;
}

static int
each_pair_i(mrb_state *mrb, mrb_value key, void *value)
{
    if (key.tt == 0/*key == Qundef*/) return ST_CONTINUE;
    mrb_yield(mrb_assoc_new(mrb, key, *(mrb_value *)value));
    return ST_CONTINUE;
}

/* 15.2.13.4.9  */
/*
 *  call-seq:
 *     hsh.each      {| key, value | block } -> hsh
 *     hsh.each_pair {| key, value | block } -> hsh
 *     hsh.each                              -> an_enumerator
 *     hsh.each_pair                         -> an_enumerator
 *
 *  Calls <i>block</i> once for each key in <i>hsh</i>, passing the key-value
 *  pair as parameters.
 *
 *  If no block is given, an enumerator is returned instead.
 *
 *     h = { "a" => 100, "b" => 200 }
 *     h.each {|key, value| puts "#{key} is #{value}" }
 *
 *  <em>produces:</em>
 *
 *     a is 100
 *     b is 200
 *
 */

static mrb_value
mrb_hash_each_pair(mrb_state *mrb, mrb_value hash)
{
    //RETURN_ENUMERATOR(hash, 0, 0);
    mrb_hash_foreach(hash, each_pair_i, 0);
    return hash;
}

static mrb_value
inspect_hash(mrb_state *mrb, mrb_value hash, mrb_value dummy, int recur)
{
  mrb_value str;

  if (recur) return mrb_usascii_str_new2(mrb, "{...}");
  str = mrb_str_new_cstr(mrb, "{");//mrb_str_buf_new2(mrb, "{");
#if HASH_CLASS_METHOD == BASICHASH
  struct RHash *h = mrb_hash_ptr(hash);
  mrb_value str2;
  int i;
  if (h->size > 0) {
    for (i=0; i<h->size; i++) {
      int ai = mrb_gc_arena_save(mrb);

      if (RSTRING_LEN(str) > 1) {
        mrb_str_cat2(mrb, str, ", ");
      }
      str2 = mrb_inspect(mrb, h->buf[i].key);
      //      mrb_str_buf_append(mrb, str, str2);//deleted by yamashita
      rb_str_buf_append(mrb, str, str2);// added by yamashita

      mrb_str_buf_cat(mrb, str, "=>", strlen("=>"));//mrb_str_buf_cat2(str, "=>");
      str2 = mrb_inspect(mrb, h->buf[i].val);
      //      mrb_str_buf_append(mrb, str, str2);//deleted by yamashita
      rb_str_buf_append(mrb, str, str2); // added by yamashita
      mrb_gc_arena_restore(mrb, ai);
    }
  }
#else
  mrb_hash_foreach(hash, inspect_i, &str);
#endif
  mrb_str_buf_cat(mrb, str, "}", strlen("}"));//mrb_str_buf_cat2(str, "}");
  //OBJ_INFECT(str, hash);

  return str;
}

/* 15.2.13.4.30 (x)*/
/*
 * call-seq:
 *   hsh.to_s     -> string
 *   hsh.inspect  -> string
 *
 * Return the contents of this hash as a string.
 *
 *     h = { "c" => 300, "a" => 100, "d" => 400, "c" => 300  }
 *     h.to_s   #=> "{\"c\"=>300, \"a\"=>100, \"d\"=>400}"
 */

static mrb_value
mrb_hash_inspect(mrb_state *mrb, mrb_value hash)
{
  if (RHASH_EMPTY_P(hash))
      return mrb_usascii_str_new2(mrb, "{}");
  return inspect_hash(mrb, hash, hash, 0);
}

/* 15.2.13.4.29 (x)*/
/*
 * call-seq:
 *    hsh.to_hash   => hsh
 *
 * Returns +self+.
 */

static mrb_value
mrb_hash_to_hash(mrb_state *mrb, mrb_value hash)
{
    return hash;
}

/* 15.2.13.4.19 */
/*
 *  call-seq:
 *     hsh.keys    -> array
 *
 *  Returns a new array populated with the keys from this hash. See also
 *  <code>Hash#values</code>.
 *
 *     h = { "a" => 100, "b" => 200, "c" => 300, "d" => 400 }
 *     h.keys   #=> ["a", "b", "c", "d"]
 *
 */

static mrb_value
mrb_hash_keys(mrb_state *mrb, mrb_value hash)
{
  mrb_value ary;

  ary = mrb_ary_new(mrb);
#if HASH_CLASS_METHOD == BASICHASH
  struct RHash *h = mrb_hash_ptr(hash);
  int i;
  if (h->size > 0) {
    for (i=0; i<h->size; i++) {
      mrb_ary_push(mrb, ary, h->buf[i].key);
    }
  }
#else
  mrb_hash_foreach(hash, keys_i, &ary);
#endif
  return ary;
}

/* 15.2.13.4.28 */
/*
 *  call-seq:
 *     hsh.values    -> array
 *
 *  Returns a new array populated with the values from <i>hsh</i>. See
 *  also <code>Hash#keys</code>.
 *
 *     h = { "a" => 100, "b" => 200, "c" => 300 }
 *     h.values   #=> [100, 200, 300]
 *
 */

static mrb_value
mrb_hash_values(mrb_state *mrb, mrb_value hash)
{
    mrb_value ary;

    ary = mrb_ary_new(mrb);
#if HASH_CLASS_METHOD == BASICHASH
  struct RHash *h = mrb_hash_ptr(hash);
  int i;
  if (h->size > 0) {
    for (i=0; i<h->size; i++) {
      mrb_ary_push(mrb, ary, h->buf[i].val);
    }
  }
#else
    mrb_hash_foreach(hash, values_i, &ary);
#endif

    return ary;
}

static mrb_value
mrb_hash_has_keyWithKey(mrb_state *mrb, mrb_value hash, mrb_value key)
{
#if HASH_CLASS_METHOD == BASICHASH
  struct RHash *h = mrb_hash_ptr(hash);
  int i;
  if (h->size > 0) {
    for (i=0; i<h->size; i++) {
      if (mrb_equal(mrb, h->buf[i].key, key)) {
        return mrb_true_value();
      }
    }
  }
  return mrb_false_value();
#else
  if (!RHASH(hash)->iv/*ntbl*/)
      return mrb_false_value();
  if (st_lookup((st_table *)RHASH(hash)->iv/*ntbl*/, mrb_fixnum(key), 0)) {
      return mrb_true_value();
  }
  return mrb_false_value();
#endif
}

/* 15.2.13.4.13 */
/* 15.2.13.4.15 */
/* 15.2.13.4.18 */
/* 15.2.13.4.21 */
/*
 *  call-seq:
 *     hsh.has_key?(key)    -> true or false
 *     hsh.include?(key)    -> true or false
 *     hsh.key?(key)        -> true or false
 *     hsh.member?(key)     -> true or false
 *
 *  Returns <code>true</code> if the given key is present in <i>hsh</i>.
 *
 *     h = { "a" => 100, "b" => 200 }
 *     h.has_key?("a")   #=> true
 *     h.has_key?("z")   #=> false
 *
 */

static mrb_value
mrb_hash_has_key(mrb_state *mrb, mrb_value hash)
{
  mrb_value key;

  mrb_get_args(mrb, "o", &key);
  return mrb_hash_has_keyWithKey(mrb, hash, key);
}

/* 15.2.13.4.14 */
/* 15.2.13.4.27 */
/*
 *  call-seq:
 *     hsh.has_value?(value)    -> true or false
 *     hsh.value?(value)        -> true or false
 *
 *  Returns <code>true</code> if the given value is present for some key
 *  in <i>hsh</i>.
 *
 *     h = { "a" => 100, "b" => 200 }
 *     h.has_value?(100)   #=> true
 *     h.has_value?(999)   #=> false
 */

static mrb_value
mrb_hash_has_value(mrb_state *mrb, mrb_value hash)
{
  mrb_value val;
  mrb_get_args(mrb, "o", &val);

#if HASH_CLASS_METHOD == BASICHASH
  struct RHash *h = mrb_hash_ptr(hash);
  int i;
  if (h->size > 0) {
    for (i=0; i<h->size; i++) {
      if (mrb_equal(mrb, h->buf[i].val, val)) {
        return mrb_true_value();
      }
    }
  }
  return mrb_false_value();
#else
  mrb_value data[2];
  data[0] = mrb_false_value();
  data[1] = val;
  mrb_hash_foreach(hash, rb_hash_search_value, &data);
  return data[0];
#endif
}

struct equal_data {
    mrb_value result;
    st_table *tbl;
    int eql;
};

static mrb_value
recursive_eql(mrb_state *mrb, mrb_value hash, mrb_value dt, int recur)
{
  mrb_value key;
  struct RHash *h1 = mrb_hash_ptr(hash);
  struct RHash *h2 = mrb_hash_ptr(dt);
  int i,j;
  int findF;

  for (i=0; i<h1->size; i++) {
    key = h1->buf[i].key;
    findF = 0;
    for (j=0; j<h2->size; j++) {
      if (mrb_equal(mrb, h2->buf[j].key, key)) {
        if (mrb_equal(mrb, h1->buf[i].val, h2->buf[j].val)) {
          findF = 1;
          break; /* next key */
        }
        else {
          return mrb_false_value();
        }
      }
    }
    if (findF == 0) {
      return mrb_false_value();
    }
  }
  return mrb_true_value();
}

static mrb_value
hash_equal(mrb_state *mrb, mrb_value hash1, mrb_value hash2, int eql)
{
    //struct equal_data data;

    if (mrb_obj_equal(mrb, hash1, hash2)) return mrb_true_value();
    if (mrb_type(hash2) != MRB_TT_HASH) {
        if (!mrb_respond_to(mrb, hash2, mrb_intern(mrb, "to_hash"))) {
            return mrb_false_value();
        }
        if (eql)
            return mrb_fixnum_value(mrb_eql(mrb, hash2, hash1));
        else
            return mrb_fixnum_value(mrb_equal(mrb, hash2, hash1));
    }
    if (RHASH_SIZE(hash1) != RHASH_SIZE(hash2))
        return mrb_false_value();
    if (!RHASH(hash1)->iv/*ntbl*/ || !RHASH(hash2)->iv/*ntbl*/)
        return mrb_true_value();
    //if (RHASH(hash1)->iv/*ntbl*/->type != RHASH(hash2)->iv/*ntbl*/->type)
    //    return mrb_false_value();
    //if (!(mrb_equal(mrb, RHASH_IFNONE(hash1), RHASH_IFNONE(hash2)) &&
    //      FL_TEST(hash1, HASH_PROC_DEFAULT) == FL_TEST(hash2, HASH_PROC_DEFAULT)))
    //    return mrb_false_value();

    //data.tbl = (st_table*)RHASH(hash2);
    //data.eql = eql;
    return mrb_exec_recursive_paired(mrb, recursive_eql, hash1, hash2, (void*)0);
}

/* 15.2.13.4.1  */
/*
 *  call-seq:
 *     hsh == other_hash    -> true or false
 *
 *  Equality---Two hashes are equal if they each contain the same number
 *  of keys and if each key-value pair is equal to (according to
 *  <code>Object#==</code>) the corresponding elements in the other
 *  hash.
 *
 *     h1 = { "a" => 1, "c" => 2 }
 *     h2 = { 7 => 35, "c" => 2, "a" => 1 }
 *     h3 = { "a" => 1, "c" => 2, 7 => 35 }
 *     h4 = { "a" => 1, "d" => 2, "f" => 35 }
 *     h1 == h2   #=> false
 *     h2 == h3   #=> true
 *     h3 == h4   #=> false
 *
 */

static mrb_value
mrb_hash_equal(mrb_state *mrb, mrb_value hash1)
{
  mrb_value hash2;
  mrb_get_args(mrb, "o", &hash2);
  return hash_equal(mrb, hash1, hash2, FALSE);
}

/* 15.2.13.4.32 (x)*/
/*
 *  call-seq:
 *     hash.eql?(other)  -> true or false
 *
 *  Returns <code>true</code> if <i>hash</i> and <i>other</i> are
 *  both hashes with the same content.
 */

static mrb_value
mrb_hash_eql(mrb_state *mrb, mrb_value hash1)
{
  mrb_value hash2;
  mrb_get_args(mrb, "o", &hash2);
  return hash_equal(mrb, hash1, hash2, TRUE);
}

/*
 *  call-seq:
 *     hsh.merge!(other_hash)                                 -> hsh
 *     hsh.update(other_hash)                                 -> hsh
 *     hsh.merge!(other_hash){|key, oldval, newval| block}    -> hsh
 *     hsh.update(other_hash){|key, oldval, newval| block}    -> hsh
 *
 *  Adds the contents of <i>other_hash</i> to <i>hsh</i>.  If no
 *  block is specified, entries with duplicate keys are overwritten
 *  with the values from <i>other_hash</i>, otherwise the value
 *  of each duplicate key is determined by calling the block with
 *  the key, its value in <i>hsh</i> and its value in <i>other_hash</i>.
 *
 *     h1 = { "a" => 100, "b" => 200 }
 *     h2 = { "b" => 254, "c" => 300 }
 *     h1.merge!(h2)   #=> {"a"=>100, "b"=>254, "c"=>300}
 *
 *     h1 = { "a" => 100, "b" => 200 }
 *     h2 = { "b" => 254, "c" => 300 }
 *     h1.merge!(h2) { |key, v1, v2| v1 }
 *                     #=> {"a"=>100, "b"=>200, "c"=>300}
 */

static mrb_value
mrb_hash_update(mrb_state *mrb, mrb_value hash1, mrb_value hash2)
{
  mrb_hash_modify(mrb, hash1);
  hash2 = to_hash(mrb, hash2);
#if HASH_CLASS_METHOD == BASICHASH
  struct RHash *h1 = mrb_hash_ptr(hash1);
  struct RHash *h2 = mrb_hash_ptr(hash2);
  mrb_value key1;
  mrb_value key2;
  int i,j;
  int FindF;
  if (h1->size > 0) {
    for (i=0; i<h1->size; i++) {
      key1 = h1->buf[i].key;
      /* same key -> replace val */
      if (h2->size > 0) {
        for (j=0; j<h2->size; j++) {
          if (mrb_equal(mrb, h2->buf[j].key, key1)) {
            h1->buf[i].val = h2->buf[j].val;
            mrb_field_write_barrier_value(mrb,
              (struct RBasic*)h1, h2->buf[j].val);
            break;
          }
        }
      }
    }
    /* unique key of hash2 -> add pair(key,val) */
    if (h2->size > 0) {
      for (j=0; j<h2->size; j++) {
        key2 = h2->buf[j].key;
        FindF = 0;
        for (i=0; i<h1->size; i++) {
          if (mrb_equal(mrb, h1->buf[j].key, key2)) {
            FindF = 1;
            break;
          }
        }
        if (FindF == 0) {
          mrb_hash_set(mrb, hash1, h2->buf[j].key, h2->buf[j].val);
        }
      }
    }
  }
  else {
    if (h2->size > 0) {
      for (j=0; j<h2->size; j++) {
        mrb_hash_set(mrb, hash1, h2->buf[j].key, h2->buf[j].val);
      }
    }
  }
  return hash1;
#else
  if (mrb_block_given_p()) {
    mrb_hash_foreach(hash2, mrb_hash_update_block_i, &hash1);
  }
  else {
    mrb_hash_foreach(hash2, mrb_hash_update_i, &hash1);
  }
  return hash1;
#endif
}

/* 15.2.13.4.22 */
/*
 *  call-seq:
 *     hsh.merge(other_hash)                              -> new_hash
 *     hsh.merge(other_hash){|key, oldval, newval| block} -> new_hash
 *
 *  Returns a new hash containing the contents of <i>other_hash</i> and
 *  the contents of <i>hsh</i>. If no block is specified, the value for
 *  entries with duplicate keys will be that of <i>other_hash</i>. Otherwise
 *  the value for each duplicate key is determined by calling the block
 *  with the key, its value in <i>hsh</i> and its value in <i>other_hash</i>.
 *
 *     h1 = { "a" => 100, "b" => 200 }
 *     h2 = { "b" => 254, "c" => 300 }
 *     h1.merge(h2)   #=> {"a"=>100, "b"=>254, "c"=>300}
 *     h1.merge(h2){|key, oldval, newval| newval - oldval}
 *                    #=> {"a"=>100, "b"=>54,  "c"=>300}
 *     h1             #=> {"a"=>100, "b"=>200}
 *
 */

static mrb_value
mrb_hash_merge(mrb_state *mrb, mrb_value hash1)
{
  mrb_value hash2;
  mrb_get_args(mrb, "o", &hash2);

  return mrb_hash_update(mrb, mrb_obj_dup(mrb, hash1), hash2);
}

static int
assoc_i(mrb_state *mrb, mrb_value key, mrb_value val, void *arg)
{
    mrb_value *args = arg;

    if (key.tt == 0/*key == Qundef*/) return ST_CONTINUE;
    if (/*RTEST*/(mrb_equal(mrb, args[0], key))) {
        args[1] = mrb_assoc_new(mrb, key, val);
        return ST_STOP;
    }
    return ST_CONTINUE;
}

/*
 *  call-seq:
 *     hash.assoc(obj)   ->  an_array  or  nil
 *
 *  Searches through the hash comparing _obj_ with the key using <code>==</code>.
 *  Returns the key-value pair (two elements array) or +nil+
 *  if no match is found.  See <code>Array#assoc</code>.
 *
 *     h = {"colors"  => ["red", "blue", "green"],
 *          "letters" => ["a", "b", "c" ]}
 *     h.assoc("letters")  #=> ["letters", ["a", "b", "c"]]
 *     h.assoc("foo")      #=> nil
 */

mrb_value
mrb_hash_assoc(mrb_value hash, mrb_value obj)
{
    mrb_value args[2];

    args[0] = obj;
    args[1] = mrb_nil_value();
    mrb_hash_foreach(hash, assoc_i, args);
    return args[1];
}

static int
rassoc_i(mrb_state *mrb, mrb_value key, mrb_value val, void * arg)
{
    mrb_value *args = arg;

    if (key.tt == 0/*key == Qundef*/) return ST_CONTINUE;
    if (/*RTEST*/(mrb_equal(mrb, args[0], val))) {
        args[1] = mrb_assoc_new(mrb, key, val);
        return ST_STOP;
    }
    return ST_CONTINUE;
}

/*
 *  call-seq:
 *     hash.rassoc(key) -> an_array or nil
 *
 *  Searches through the hash comparing _obj_ with the value using <code>==</code>.
 *  Returns the first key-value pair (two-element array) that matches. See
 *  also <code>Array#rassoc</code>.
 *
 *     a = {1=> "one", 2 => "two", 3 => "three", "ii" => "two"}
 *     a.rassoc("two")    #=> [2, "two"]
 *     a.rassoc("four")   #=> nil
 */

mrb_value
mrb_hash_rassoc(mrb_value hash, mrb_value obj)
{
    mrb_value args[2];

    args[0] = obj;
    args[1] = mrb_nil_value();
    mrb_hash_foreach(hash, rassoc_i, args);
    return args[1];
}

/*
 *  call-seq:
 *     hash.flatten -> an_array
 *     hash.flatten(level) -> an_array
 *
 *  Returns a new array that is a one-dimensional flattening of this
 *  hash. That is, for every key or value that is an array, extract
 *  its elements into the new array.  Unlike Array#flatten, this
 *  method does not flatten recursively by default.  The optional
 *  <i>level</i> argument determines the level of recursion to flatten.
 *
 *     a =  {1=> "one", 2 => [2,"two"], 3 => "three"}
 *     a.flatten    # => [1, "one", 2, [2, "two"], 3, "three"]
 *     a.flatten(2) # => [1, "one", 2, 2, "two", 3, "three"]
 */

/*
 *  A <code>Hash</code> is a collection of key-value pairs. It is
 *  similar to an <code>Array</code>, except that indexing is done via
 *  arbitrary keys of any object type, not an integer index. Hashes enumerate
 *  their values in the order that the corresponding keys were inserted.
 *
 *  Hashes have a <em>default value</em> that is returned when accessing
 *  keys that do not exist in the hash. By default, that value is
 *  <code>nil</code>.
 *
 */

void
mrb_init_hash(mrb_state *mrb)
{
  struct RClass *h;

  h = mrb->hash_class = mrb_define_class(mrb, "Hash", mrb->object_class);

  //mrb_define_class_method(mrb, h, "new",       hash_s_new,           ARGS_ANY());
  mrb_include_module(mrb, h, mrb_class_get(mrb, "Enumerable"));
  mrb_define_method(mrb, h, "==",              mrb_hash_equal,       ARGS_REQ(1)); /* 15.2.13.4.1  */
  mrb_define_method(mrb, h, "[]",              mrb_hash_aget,        ARGS_REQ(1)); /* 15.2.13.4.2  */
  mrb_define_method(mrb, h, "[]=",             mrb_hash_aset,        ARGS_REQ(2)); /* 15.2.13.4.3  */
  mrb_define_method(mrb, h, "clear",           mrb_hash_clear,       ARGS_NONE()); /* 15.2.13.4.4  */
  mrb_define_method(mrb, h, "default",         mrb_hash_default,     ARGS_ANY());  /* 15.2.13.4.5  */
  mrb_define_method(mrb, h, "default=",        mrb_hash_set_default, ARGS_REQ(1)); /* 15.2.13.4.6  */
  mrb_define_method(mrb, h, "default_proc",    mrb_hash_default_proc,ARGS_NONE()); /* 15.2.13.4.7  */
  mrb_define_method(mrb, h, "__delete",        mrb_hash_delete,      ARGS_REQ(1)); /* core of 15.2.13.4.8  */
  mrb_define_method(mrb, h, "each",            mrb_hash_each_pair,   ARGS_NONE()); /* 15.2.13.4.9  */
  mrb_define_method(mrb, h, "each_key",        mrb_hash_each_key,    ARGS_NONE()); /* 15.2.13.4.10 */
  mrb_define_method(mrb, h, "each_value",      mrb_hash_each_value,  ARGS_NONE()); /* 15.2.13.4.11 */
  mrb_define_method(mrb, h, "empty?",          mrb_hash_empty_p,     ARGS_NONE()); /* 15.2.13.4.12 */
  mrb_define_method(mrb, h, "has_key?",        mrb_hash_has_key,     ARGS_REQ(1)); /* 15.2.13.4.13 */
  mrb_define_method(mrb, h, "has_value?",      mrb_hash_has_value,   ARGS_REQ(1)); /* 15.2.13.4.14 */
  mrb_define_method(mrb, h, "include?",        mrb_hash_has_key,     ARGS_REQ(1)); /* 15.2.13.4.15 */
  mrb_define_method(mrb, h, "__init_core",     mrb_hash_init_core,   ARGS_ANY());  /* core of 15.2.13.4.16 */
  mrb_define_method(mrb, h, "initialize_copy", mrb_hash_replace,     ARGS_REQ(1)); /* 15.2.13.4.17 */
  mrb_define_method(mrb, h, "key?",            mrb_hash_has_key,     ARGS_REQ(1)); /* 15.2.13.4.18 */
  mrb_define_method(mrb, h, "keys",            mrb_hash_keys,        ARGS_NONE()); /* 15.2.13.4.19 */
  mrb_define_method(mrb, h, "length",          mrb_hash_size_m,      ARGS_NONE()); /* 15.2.13.4.20 */
  mrb_define_method(mrb, h, "member?",         mrb_hash_has_key,     ARGS_REQ(1)); /* 15.2.13.4.21 */
  mrb_define_method(mrb, h, "merge",           mrb_hash_merge,       ARGS_REQ(1)); /* 15.2.13.4.22 */
  mrb_define_method(mrb, h, "replace",         mrb_hash_replace,     ARGS_REQ(1)); /* 15.2.13.4.23 */
  mrb_define_method(mrb, h, "shift",           mrb_hash_shift,       ARGS_NONE()); /* 15.2.13.4.24 */
  mrb_define_method(mrb, h, "size",            mrb_hash_size_m,      ARGS_NONE()); /* 15.2.13.4.25 */
  mrb_define_method(mrb, h, "store",           mrb_hash_aset,        ARGS_REQ(2)); /* 15.2.13.4.26 */
  mrb_define_method(mrb, h, "value?",          mrb_hash_has_value,   ARGS_REQ(1)); /* 15.2.13.4.27 */
  mrb_define_method(mrb, h, "values",          mrb_hash_values,      ARGS_NONE()); /* 15.2.13.4.28 */

  mrb_define_method(mrb, h, "to_hash",         mrb_hash_to_hash,     ARGS_NONE()); /* 15.2.13.4.29 (x)*/
  mrb_define_method(mrb, h, "inspect",         mrb_hash_inspect,     ARGS_NONE()); /* 15.2.13.4.30 (x)*/
  mrb_define_alias(mrb,  h, "to_s",            "inspect");                         /* 15.2.13.4.31 (x)*/
  mrb_define_method(mrb, h, "eql?",            mrb_hash_eql,         ARGS_REQ(1)); /* 15.2.13.4.32 (x)*/
}
