/**********************************************************************

  mdata.h -


  Copyright (C) 2007 Yukihiro Matsumoto

**********************************************************************/

#ifndef RUBY_DATA_H
#define RUBY_DATA_H 1

#if defined(__cplusplus)
extern "C" {
#endif

struct RData {
    struct RBasic basic;
    void (*dmark)(void*);
    void (*dfree)(void*);
    void *data;
};

typedef struct mrb_data_type_struct {
    const char *wrap_struct_name;
    void (*dmark)(mrb_state *mrb, void*);
    void (*dfree)(void*);
    size_t (*dsize)(const void *);
    void *reserved[3]; /* For future extension.
                          This array *must* be filled with ZERO. */
    void *data;        /* This area can be used for any purpose
                          by a programmer who define the type. */
} mrb_data_type_t;

struct RTypedData {
    struct RBasic basic;
    const mrb_data_type_t *type;
    unsigned int/*VALUE*/ typed_flag; /* 1 or not */
    void *data;
};

void *mrb_check_typeddata(mrb_state *mrb, mrb_value, const mrb_data_type_t *);
#define TypedData_Get_Struct(mrb,obj,type,data_type,sval) do {\
    sval = (type*)mrb_check_typeddata(mrb, obj, data_type); \
} while (0)

#define RDATA(obj)         ((struct RData *)((obj).value.p))
#define DATA_PTR(dta)      (RDATA(dta)->data)

#define RTYPEDDATA(obj)    ((struct RTypedData*)((obj).value.p))
#define RTYPEDDATA_P(v)    (RTYPEDDATA(v)->typed_flag == 1)
#define RTYPEDDATA_TYPE(v) (RTYPEDDATA(v)->type)
#define RTYPEDDATA_DATA(v) (RTYPEDDATA(v)->data)

#if defined(__cplusplus)
}  /* extern "C" { */
#endif

#endif /* RUBY_DATA_H */

