#include "mruby.h"

/* helloworld.c for TOPPERS/ATK(OSEK) */
#include "kernel.h"
#include "kernel_id.h"
#include "ecrobot_interface.h"
#include "balancer.h" /* 倒立振子制御用ヘッダファイル */

void debug_disp(char *c , int flag)//flag =1 stop
{
  display_goto_xy(0, 6);
  display_string(c);
  display_string("               ");
  display_update();
  if (1 == flag){
    while(1){;}
  }
}

/* nxtOSEK hook to be invoked from an ISR in category 2 */
void user_1ms_isr_type2(void){ /* do nothing */ }

//*****************************************************************************
// 関数名 : ecrobot_device_initialize
// 引数 : なし
// 戻り値 : なし
// 概要 : ECROBOTデバイス初期化処理フック関数
//*****************************************************************************
void
ecrobot_device_initialize()
{
  ecrobot_set_light_sensor_active(NXT_PORT_S3); /* 光センサ赤色LEDをON */
  ecrobot_init_sonar_sensor(NXT_PORT_S2); /* 超音波センサ(I2C通信)を初期化 */
  nxt_motor_set_count(NXT_PORT_A, 0); /* 完全停止用モータエンコーダリセット */
}

//*****************************************************************************
// 関数名 : ecrobot_device_terminate
// 引数 : なし
// 戻り値 : なし
// 概要 : ECROBOTデバイス終了処理フック関数
//*****************************************************************************
void
ecrobot_device_terminate()
{
  ecrobot_set_light_sensor_inactive(NXT_PORT_S3); /* 光センサ赤色LEDをOFF */
  ecrobot_term_sonar_sensor(NXT_PORT_S2); /* 超音波センサ(I2C通信)を終了 */
}

#ifdef HEAP_PROFILE

int nxt_alloc_size = 0;
int max_alloc_size = 0;

static size_t
get_memsize(size_t size)
{
  /* This function depends on the platform */
  size_t memsize;

  if (size == 0)
    size++;

  if (size <= 4)
    size += 4;

  memsize = (((size+3) / 8) + 1) * 8 + 1;

  return (memsize);
}

void *
my_malloc(size_t size)
{
  nxt_alloc_size += get_memsize(size);

  if (max_alloc_size < nxt_alloc_size)
    max_alloc_size = nxt_alloc_size;
  return malloc(size);
}
void *
my_realloc(void *p, size_t size)
{
  if (p) {
    size_t free_size = *(size_t *)(p - sizeof(size_t));
    nxt_alloc_size -= get_memsize(free_size);
  }
  nxt_alloc_size += get_memsize(size);

  if (max_alloc_size < nxt_alloc_size)
    max_alloc_size = nxt_alloc_size;
  return realloc(p, size);
}

void *
my_calloc(size_t nmemb, size_t size)
{
  nxt_alloc_size += get_memsize(size);
  if (max_alloc_size < nxt_alloc_size)
    max_alloc_size = nxt_alloc_size;
  return calloc(nmemb, size);
}

void
my_free(void *p)
{
  size_t size;

  if (!p) return;

  size = *(size_t *)(p - sizeof(size_t));
  nxt_alloc_size -= size;

  if (max_alloc_size < nxt_alloc_size)
    max_alloc_size = nxt_alloc_size;
  return free(p);
}
#endif

void debug_touch(char *c)
{
#ifdef PC_SIMULATION
  #include <stdio.h>
#ifdef HEAP_PROFILE
  printf("%s, %d\n", c, nxt_alloc_size);
#endif
#else
  display_goto_xy(0, 6);
  display_string(c);
  display_string("               ");
  display_update();
#ifdef HEAP_PROFILE
  ecrobot_show_int(max_alloc_size);
#endif
  ecrobot_show_int(ecrobot_get_gyro_sensor(NXT_PORT_S4));

  while(1)
  {
    if (ecrobot_get_touch_sensor(NXT_PORT_S1) == 1)
      break; /* タッチセンサが押された */

    systick_wait_ms(10); /* 10msecウェイト */
  }
#endif
}

TASK(OSEK_Task_Background)
{
  mrb_state *mrb;

  balance_init(); // 倒立振子制御初期化 

  nxt_motor_set_count(NXT_PORT_C, 0); // 左モータエンコーダリセット
  nxt_motor_set_count(NXT_PORT_A, 0); // 右モータエンコーダリセット
  mrb = mrb_open();

  debug_touch("mrb_open end");

  extern void balancer(mrb_state *mrb);
  balancer(mrb);
  
  debug_touch("balancer end");

  while(1)
  {
    systick_wait_ms(500); // 500msec wait
  }

}
