/* Ritehash.h */

/*
  Sample program from 'khash.h'

  #include "khash.h"
  KHASH_MAP_INIT_INT(32, char)
  int main() {
  int ret, is_missing;
  khiter_t k;
  khash_t(32) *h = kh_init(32);
  k = kh_put(32, h, 5, &ret);
  if (!ret) kh_del(32, h, k);
  kh_value(h, k) = 10;
  k = kh_get(32, h, 10);
  is_missing = (k == kh_end(h));
  k = kh_get(32, h, 5);
  kh_del(32, h, k);
  for (k = kh_begin(h); k != kh_end(h); ++k)
  if (kh_exist(h, k)) kh_value(h, k) = 1;
  kh_destroy(32, h);
  return 0;
  }
*/


#ifndef __AC_KHASH_H
#define __AC_KHASH_H

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
// #include <stdio.h>

typedef uint32_t khint_t;
typedef khint_t khiter_t;

#define __ac_isempty(flag, i) ( (flag[i>>4] >> ((i&0x0f)*2) ) & 0x02)
#define __ac_isdel(flag, i) ( (flag[i>>4] >> ((i&0x0f)*2) ) & 0x01)
#define __ac_iseither(flag, i) ( (flag[i>>4] >> ((i&0x0f)*2) ) & 0x03)
#define __ac_set_isdel_false(flag, i) (flag[i>>4] &= ~(0x01 << ((i&0x0f)*2)))
#define __ac_set_isempty_false(flag, i) (flag[i>>4] &= ~(0x02 << ((i&0x0f)*2)))
#define __ac_set_isboth_false(flag, i) (flag[i>>4] &= ~( 0x03 << ((i&0x0f)*2)))
#define __ac_set_isdel_true(flag, i) (flag[i>>4] |= 0x01 << ((i&0x0f)*2))

#define KHASH_INIT(name, khkey_t, khval_t, kh_is_map, __hash_func, __hash_equal) \
  typedef struct kh_##name {                                            \
    khint_t n_buckets, size, n_occupied, upper_bound;                   \
    uint32_t *flags;                                                    \
    khkey_t *keys;                                                      \
    khval_t *vals;                                                      \
  } kh_##name##_t;                                                      \
  static inline void kh_clear_##name(kh_##name##_t *h)                  \
  {                                                                     \
    if (h && h->flags) {                                                \
      memset(h->flags, 0xaa, ((h->n_buckets>>4) + 1) * sizeof(uint32_t)); \
      h->size = h->n_occupied = 0;                                      \
    }                                                                   \
  }                                                                     \
  static inline kh_##name##_t *kh_init_##name() {                       \
    kh_##name##_t *h=(kh_##name##_t*)calloc(1, sizeof(kh_##name##_t));  \
    h->n_buckets = 32;                                                  \
    h->upper_bound = 24;                                                \
    h->flags = (uint32_t*)malloc(4 * sizeof(uint32_t));                 \
    h->keys = (khkey_t*)malloc(32 * sizeof(khkey_t));                   \
    h->vals = (khval_t*)malloc(32 * sizeof(khval_t));                   \
    kh_clear_##name(h);                                                 \
    return h;                                                           \
  }                                                                     \
  static inline void kh_destroy_##name(kh_##name##_t *h)                \
  {                                                                     \
    if (h) {                                                            \
      free(h->keys); free(h->flags);                                    \
      free(h->vals);                                                    \
      free(h);                                                          \
    }                                                                   \
  }                                                                     \
  static inline khint_t kh_get_##name(kh_##name##_t *h, khkey_t key)    \
  {                                                                     \
    if (h->n_buckets) {                                                 \
      khint_t inc, k, i, last;                                          \
      k = __hash_func(key); i = k & (h->n_buckets-1);                   \
      inc = 1 + k % (h->n_buckets - 1); last = i;                       \
      while (!__ac_isempty(h->flags, i) && (__ac_isdel(h->flags, i) || !__hash_equal(h->keys[i], key))) { \
        i += inc;                                                       \
        if ( i >= h->n_buckets) i -= h->n_buckets;                      \
        if (i == last) return h->n_buckets;                             \
      }                                                                 \
      return __ac_iseither(h->flags, i)? h->n_buckets : i;              \
    } else return 0;                                                    \
  }                                                                     \
  static inline void kh_resize_##name(kh_##name##_t *h, khint_t new_n_buckets) \
  {                                                                     \
    uint32_t *new_flags = 0;                                            \
    khint_t j = 1;                                                      \
    {                                                                   \
      khint_t new_size = 1;                                             \
      while( new_n_buckets > 0 ){                                       \
        new_n_buckets >>= 1;                                            \
        new_size <<= 1;                                                 \
      }                                                                 \
      new_n_buckets = new_size;                                         \
      if (h->size >= (khint_t)((new_n_buckets>>2)+(new_n_buckets>>1)+1)) j = 0; \
      else {                                                            \
        new_flags = (uint32_t*)malloc(((new_n_buckets>>4) + 1) * sizeof(uint32_t)); \
        memset(new_flags, 0xaa, ((new_n_buckets>>4) + 1) * sizeof(uint32_t)); \
        if (h->n_buckets < new_n_buckets) {                             \
          h->keys = (khkey_t*)realloc(h->keys, new_n_buckets * sizeof(khkey_t)); \
          if (kh_is_map)                                                \
            h->vals = (khval_t*)realloc(h->vals, new_n_buckets * sizeof(khval_t)); \
        }                                                               \
      }                                                                 \
    }                                                                   \
    if (j) {                                                            \
      for (j = 0; j != h->n_buckets; ++j) {                             \
        if (__ac_iseither(h->flags, j) == 0) {                          \
          khkey_t key = h->keys[j];                                     \
          khval_t val;                                                  \
          if (kh_is_map) val = h->vals[j];                              \
          __ac_set_isdel_true(h->flags, j);                             \
          while (1) {                                                   \
            khint_t inc, k, i;                                          \
            k = __hash_func(key);                                       \
            i = k & (new_n_buckets - 1);                                \
            inc = 1 + k % (new_n_buckets - 1);                          \
            while (!__ac_isempty(new_flags, i)) {                       \
              i += inc;                                                 \
              if ( i >= new_n_buckets) i -= new_n_buckets;              \
            }                                                           \
            __ac_set_isempty_false(new_flags, i);                       \
            if (i < h->n_buckets && __ac_iseither(h->flags, i) == 0) {  \
              { khkey_t tmp = h->keys[i]; h->keys[i] = key; key = tmp; } \
              if (kh_is_map) { khval_t tmp = h->vals[i]; h->vals[i] = val; val = tmp; } \
              __ac_set_isdel_true(h->flags, i);                         \
            } else {                                                    \
              h->keys[i] = key;                                         \
              if (kh_is_map) h->vals[i] = val;                          \
              break;                                                    \
            }                                                           \
          }                                                             \
        }                                                               \
      }                                                                 \
      if (h->n_buckets > new_n_buckets) {                               \
        h->keys = (khkey_t*)realloc(h->keys, new_n_buckets * sizeof(khkey_t)); \
        if (kh_is_map)                                                  \
          h->vals = (khval_t*)realloc(h->vals, new_n_buckets * sizeof(khval_t)); \
      }                                                                 \
      free(h->flags);                                                   \
      h->flags = new_flags;                                             \
      h->n_buckets = new_n_buckets;                                     \
      h->n_occupied = h->size;                                          \
      h->upper_bound = (khint_t)((h->n_buckets>>2)+(h->n_buckets>>1)+1); \
    }                                                                   \
  }                                                                     \
  static inline khint_t kh_put_##name(kh_##name##_t *h, khkey_t key, int *ret) \
  {                                                                     \
    khint_t x;                                                          \
    if (h->n_occupied >= h->upper_bound) {                              \
      if (h->n_buckets > (h->size<<1)) kh_resize_##name(h, h->n_buckets - 1); \
      else kh_resize_##name(h, h->n_buckets + 1);                       \
    }                                                                   \
    {                                                                   \
      khint_t inc, k, i, site, last;                                    \
      x = site = h->n_buckets; k = __hash_func(key);                    \
      i = k & (h->n_buckets-1);                                         \
      /*      printf("#P");  */                                         \
      if (__ac_isempty(h->flags, i)) x = i;                             \
      else {                                                            \
        inc = 1 + k % (h->n_buckets - 1); last = i;                     \
        while (!__ac_isempty(h->flags, i) && (__ac_isdel(h->flags, i) || !__hash_equal(h->keys[i], key))) { \
          if (__ac_isdel(h->flags, i)) site = i;                        \
          i += inc;                                                     \
          /*    printf("#C"); */                                        \
          if ( i >= h->n_buckets) i -= h->n_buckets;                    \
          if (i == last) { x = site; break; }                           \
        }                                                               \
        if (x == h->n_buckets) {                                        \
          if (__ac_isempty(h->flags, i) && site != h->n_buckets) x = site; \
          else x = i;                                                   \
        }                                                               \
      }                                                                 \
    }                                                                   \
    if (__ac_isempty(h->flags, x)) {                                    \
      h->keys[x] = key;                                                 \
      __ac_set_isboth_false(h->flags, x);                               \
      ++h->size; ++h->n_occupied;                                       \
      *ret = 1;                                                         \
    } else if (__ac_isdel(h->flags, x)) {                               \
      h->keys[x] = key;                                                 \
      __ac_set_isboth_false(h->flags, x);                               \
      ++h->size;                                                        \
      *ret = 2;                                                         \
    } else *ret = 0;                                                    \
    return x;                                                           \
  }                                                                     \
  static inline void kh_del_##name(kh_##name##_t *h, khint_t x)         \
  {                                                                     \
    if (x != h->n_buckets && !__ac_iseither(h->flags, x)) {             \
      __ac_set_isdel_true(h->flags, x);                                 \
      --h->size;                                                        \
    }                                                                   \
  }

/* --- BEGIN OF HASH FUNCTIONS --- */

/*
#define kh_int_hash_func(key) (uint32_t)( ((key)>>1 ^ (key) ^ (key)<<1 ^ (key)<<3) )
*/

static inline uint32_t kh_int_hash_func(uint32_t key)
{
  // khash.h 
  //return key;

  // sample
  return (key>>2) ^ (key) ^ (key<<3);

  // Barrel shifter Ver.1 Simple
  /*
  uint32_t k = (key>>16) ^ (key&0xffff);             
  uint32_t k2 = ((k&0x0003 << 2) |             
                 (k&0x000C >> 2) |             
                 (k&0x0070 << 3) |              
                 (k&0x0380 >> 3) |              
                 (k&0x1C00 << 3) |              
                 (k&0xE000 >> 3) );             
  return k2 ^ k2>>4;
  */

  // Barrel shifter Ver.2 HDL
  /*
  uint32_t k2 = (key>>16) ^ (key&0xffff);
  uint32_t k3 =
    ((k2&0x0008)<<12) | ((k2&0x8000)>>1) |
    ((k2&0x08000)<<) | ((k2&0x8000)>>1) |
  */
}


#define kh_int_hash_equal(a, b) (a == b)
#define kh_int64_hash_func(key) (uint32_t)(key ^ (key<<2) ^ (key>>2))
#define kh_int64_hash_equal(a, b) (a == b)
static inline khint_t __ac_X31_hash_string(const char *s)
{
    khint_t h = *s;
    if (h) for (++s ; *s; ++s) h = (h << 5) - h + *s;
    return h;
}
#define kh_str_hash_func(key) __ac_X31_hash_string(key)
#define kh_str_hash_equal(a, b) (strcmp(a, b) == 0)

/* --- END OF HASH FUNCTIONS --- */

/* Other necessary macros... */

#define khash_t(name) kh_##name##_t

#define kh_init(name) kh_init_##name()
#define kh_destroy(name, h) kh_destroy_##name(h)
#define kh_clear(name, h) kh_clear_##name(h)
#define kh_resize(name, h, s) kh_resize_##name(h, s)
#define kh_put(name, h, k, r) kh_put_##name(h, k, r)
#define kh_get(name, h, k) kh_get_##name(h, k)
#define kh_del(name, h, k) kh_del_##name(h, k)

#define kh_exist(h, x) (!__ac_iseither((h)->flags, (x)))
#define kh_key(h, x) ((h)->keys[x])
#define kh_val(h, x) ((h)->vals[x])
#define kh_value(h, x) ((h)->vals[x])
#define kh_begin(h) (khint_t)(0)
#define kh_end(h) ((h)->n_buckets)
#define kh_size(h) ((h)->size)
#define kh_n_buckets(h) ((h)->n_buckets)

/* More conenient interfaces */

#define KHASH_SET_INIT_INT(name)                                        \
  KHASH_INIT(name, uint32_t, char, 0, kh_int_hash_func, kh_int_hash_equal)

#define KHASH_MAP_INIT_INT(name, khval_t)                               \
  KHASH_INIT(name, uint32_t, khval_t, 1, kh_int_hash_func, kh_int_hash_equal)

#define KHASH_SET_INIT_INT64(name)                                      \
  KHASH_INIT(name, uint64_t, char, 0, kh_int64_hash_func, kh_int64_hash_equal)

#define KHASH_MAP_INIT_INT64(name, khval_t)                             \
  KHASH_INIT(name, uint64_t, khval_t, 1, kh_int64_hash_func, kh_int64_hash_equal)

typedef const char *kh_cstr_t;
#define KHASH_SET_INIT_STR(name)                                        \
  KHASH_INIT(name, kh_cstr_t, char, 0, kh_str_hash_func, kh_str_hash_equal)

#define KHASH_MAP_INIT_STR(name, khval_t)                               \
  KHASH_INIT(name, kh_cstr_t, khval_t, 1, kh_str_hash_func, kh_str_hash_equal)

#endif /* __AC_KHASH_H */
