/**********************************************************************

  io.c -

  $Author: shyouhei $
  $Date: 2010-06-08 18:02:21 +0900 (Tue, 08 Jun 2010) $
  created at: Fri Oct 15 18:08:59 JST 1993

  Copyright (C) 1993-2003 Yukihiro Matsumoto
  Copyright (C) 2000  Network Applied Communication Laboratory, Inc.
  Copyright (C) 2000  Information-technology Promotion Agency, Japan

**********************************************************************/

#include "mruby.h"

#ifndef INCLUDE_REGEXP

#if defined(__VMS)
#define _XOPEN_SOURCE
#define _POSIX_C_SOURCE 2
#endif

#include "mio.h"
/*----- rubyio.h -----*/
#define FMODE_WBUF     16
#define FMODE_RBUF     32
/*----- -------- -----*/

//#include "rubysig.h"
#if 0
  #ifdef _WIN32
typedef LONG rb_atomic_t;

# define ATOMIC_TEST(var) InterlockedExchange(&(var), 0)
# define ATOMIC_SET(var, val) InterlockedExchange(&(var), (val))
# define ATOMIC_INC(var) InterlockedIncrement(&(var))
# define ATOMIC_DEC(var) InterlockedDecrement(&(var))

/* Windows doesn't allow interrupt while system calls */
# define TRAP_BEG(mrb) do {\
    int saved_errno = 0;\
    rb_atomic_t trap_immediate = ATOMIC_SET(rb_trap_immediate, 1)
# define TRAP_END(mrb)\
    ATOMIC_SET(rb_trap_immediate, trap_immediate);\
    saved_errno = errno;\
    CHECK_INTS(mrb);\
    errno = saved_errno;\
} while (0)
# define RUBY_CRITICAL(statements) do {\
    rb_w32_enter_critical();\
    statements;\
    rb_w32_leave_critical();\
} while (0)
  #else
typedef int rb_atomic_t;

# define ATOMIC_TEST(var) ((var) ? ((var) = 0, 1) : 0)
# define ATOMIC_SET(var, val) ((var) = (val))
# define ATOMIC_INC(var) (++(var))
# define ATOMIC_DEC(var) (--(var))

# define TRAP_BEG(mrb) do {\
    int saved_errno = 0;\
    int trap_immediate = rb_trap_immediate;\
    rb_trap_immediate = 1
# define TRAP_END(mrb) rb_trap_immediate = trap_immediate;\
    saved_errno = errno;\
    CHECK_INTS(mrb);\
    errno = saved_errno;\
} while (0)

# define RUBY_CRITICAL(statements) do {\
    int trap_immediate = rb_trap_immediate;\
    rb_trap_immediate = 0;\
    statements;\
    rb_trap_immediate = trap_immediate;\
} while (0)
  #endif
extern rb_atomic_t rb_trap_immediate;
extern rb_atomic_t rb_trap_pending;

#else

# define ATOMIC_TEST(var)
# define ATOMIC_SET(var, val)
# define ATOMIC_INC(var)
# define ATOMIC_DEC(var)
# define TRAP_BEG(mrb)
# define TRAP_END(mrb)
# define RUBY_CRITICAL(statements)
#endif

#if defined(HAVE_SETITIMER) || defined(_THREAD_SAFE)
# define CHECK_INTS(mrb) do {\
    if (!(rb_prohibit_interrupt)) {\
	if (rb_thread_pending) rb_thread_schedule(mrb);\
    }\
} while (0)
#else
/* pseudo preemptive thread switching */
extern int rb_thread_tick;
#define THREAD_TICK 500
#define CHECK_INTS(mrb) do {\
    if (!(rb_prohibit_interrupt)) {\
	if (rb_thread_pending || rb_thread_tick-- <= 0) {\
	    rb_thread_tick = THREAD_TICK;\
	    rb_thread_schedule(mrb);\
	}\
    }\
} while (0)
#endif
extern int rb_prohibit_interrupt;
extern int rb_thread_pending;

//#include "env.h"
#include "re.h"
#include <ctype.h>
#include <errno.h>
#include "class.h"
#include "range.h"
#include "variable.h"
#include "error.h"
#include "numeric.h"
#include "mio.h"
#include "mstring.h"
#include "array.h"
#include "encoding.h"
#include "hash.h"
#include <stdarg.h>

#include <string.h>
#include <stdio.h> /* stdin */
//#define HAVE_STRING_H

//#include <signal.h>      /* #define NSIG 23 */
//#include "mthread.h"
//#include "eval_intern.h"

#if defined(MSDOS) || defined(__BOW__) || defined(__CYGWIN__) || defined(_WIN32) || defined(__human68k__) || defined(__EMX__) || defined(__BEOS__)
# define NO_SAFE_RENAME
#endif

#if defined(MSDOS) || defined(__CYGWIN__) || defined(_WIN32)
# define NO_LONG_FNAME
#endif

#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__) || defined(__DragonFly__) || defined(sun) || defined(_nec_ews)
# define USE_SETVBUF
#endif

#ifndef BSD_STDIO
# if defined(__MACH__) || defined(__DARWIN__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__) || defined(__bsdi__)
#   define BSD_STDIO 1
# else
#   define BSD_STDIO 0
# endif
#endif

#ifdef __QNXNTO__
#include "unix.h"
#endif

#include <sys/types.h>
#if defined(HAVE_SYS_IOCTL_H) && !defined(DJGPP) && !defined(_WIN32) && !defined(__human68k__)
#include <sys/ioctl.h>
#endif
#if defined(HAVE_FCNTL_H) || defined(_WIN32)
#include <fcntl.h>
#elif defined(HAVE_SYS_FCNTL_H)
#include <sys/fcntl.h>
#endif
#ifdef __CYGWIN__
#include <io.h>
#endif

#if !HAVE_OFF_T && !defined(off_t)
# define off_t  long
#endif
#if !HAVE_FSEEKO && !defined(fseeko)
# define fseeko  fseek
#endif
#if !HAVE_FTELLO && !defined(ftello)
# define ftello  ftell
#endif

#include <sys/stat.h>

/* EMX has sys/param.h, but.. */
#if defined(HAVE_SYS_PARAM_H) && !(defined(__EMX__) || defined(__HIUX_MPP__))
# include <sys/param.h>
#endif

#if !defined NOFILE
# define NOFILE 64
#endif

#ifdef HAVE_UNISTD_H
#ifdef HAVE_SYSCALL_H
#include <syscall.h>
#elif defined HAVE_SYS_SYSCALL_H
#include <sys/syscall.h>
#endif

#include <unistd.h>
#endif

extern void Init_File (void);

#ifdef __BEOS__
# ifndef NOFILE
#  define NOFILE (OPEN_MAX)
# endif
#include <net/socket.h>
#endif

//#include "util.h"

#ifndef O_ACCMODE
#define O_ACCMODE (O_RDONLY | O_WRONLY | O_RDWR)
#endif

#if SIZEOF_OFF_T > SIZEOF_LONG && !defined(HAVE_LONG_LONG)
# error off_t is bigger than long, but you have no long long...
#endif

#ifndef PIPE_BUF
# ifdef _POSIX_PIPE_BUF
#  define PIPE_BUF _POSIX_PIPE_BUF
# else
#  define PIPE_BUF 512 /* is this ok? */
# endif
#endif

#define GetReadFile(fptr) ((fptr)->f)
#define GetWriteFile(fptr) (((fptr)->f2) ? (fptr)->f2 : (fptr)->f)

#define preserving_errno(stmts) \
      do {int saved_errno = errno; stmts; errno = saved_errno;} while (0)

#ifndef FALSE
#define FALSE   0
#endif

#ifndef TRUE
#define TRUE    1
#endif

mrb_value rb_cIO;
mrb_value rb_eEOFError;
//mrb_value rb_eIOError;

mrb_value rb_stdin, rb_stdout, rb_stderr;
mrb_value rb_deferr;            /* rescue VIM plugin */
//static mrb_value orig_stdout, orig_stderr;
mrb_value orig_stdout;

//static mrb_value argf;

//static mrb_sym id_write, id_read, id_getc;

extern char *ruby_inplace_mode;

struct timeval rb_time_interval (mrb_value);

//static mrb_value filename, current_file;
static int gets_lineno;
//static int init_p = 0, next_p = 0;
static mrb_value lineno;// = mrb_fixnum_value(0);

#ifdef _STDIO_USES_IOSTREAM  /* GNU libc */
#  ifdef _IO_fpos_t
#    define READ_DATA_PENDING(fp) ((fp)->_IO_read_ptr != (fp)->_IO_read_end)
#    define READ_DATA_PENDING_COUNT(fp) ((fp)->_IO_read_end - (fp)->_IO_read_ptr)
#    define READ_DATA_PENDING_PTR(fp) ((fp)->_IO_read_ptr)
#  else
#    define READ_DATA_PENDING(fp) ((fp)->_gptr < (fp)->_egptr)
#    define READ_DATA_PENDING_COUNT(fp) ((fp)->_egptr - (fp)->_gptr)
#    define READ_DATA_PENDING_PTR(fp) ((fp)->_gptr)
#  endif
#elif defined(_LP64) && (defined(__sun__) || defined(__sun))
typedef struct _FILE64 {
  unsigned char        *_ptr;  /* next character from/to here in buffer */
  unsigned char        *_base; /* the buffer */
  unsigned char        *_end;  /* the end of the buffer */
  ssize_t      _cnt;   /* number of available characters in buffer */
  int          _file;  /* UNIX System file descriptor */
  unsigned int _flag;  /* the state of the stream */
  char         __fill[80];     /* filler to bring size to 128 bytes */
} FILE64;
#  define READ_DATA_PENDING(fp) (((FILE64*)(fp))->_cnt > 0)
#  define READ_DATA_PENDING_COUNT(fp) (((FILE64*)(fp))->_cnt)
#  define READ_DATA_PENDING_PTR(fp) ((char *)((FILE64*)(fp))->_ptr)
#elif defined(FILE_COUNT)
#  define READ_DATA_PENDING(fp) ((fp)->FILE_COUNT > 0)
#  define READ_DATA_PENDING_COUNT(fp) ((fp)->FILE_COUNT)
#elif defined(FILE_READEND)
#  define READ_DATA_PENDING(fp) ((fp)->FILE_READPTR < (fp)->FILE_READEND)
#  define READ_DATA_PENDING_COUNT(fp) ((fp)->FILE_READEND - (fp)->FILE_READPTR)
#elif defined(__BEOS__)
#  define READ_DATA_PENDING(fp) (fp->_state._eof == 0)
#elif defined(__VMS)
#  define READ_DATA_PENDING_COUNT(fp) ((unsigned int)(*(fp))->_cnt)
#  define READ_DATA_PENDING(fp)       (((unsigned int)(*(fp))->_cnt) > 0)
#  define READ_DATA_BUFFERED(fp) 0
#elif defined(__DragonFly__)
/* FILE is an incomplete struct type since DragonFly BSD 1.4.0 */
#  define READ_DATA_PENDING(fp) (((struct __FILE_public *)(fp))->_r > 0)
#  define READ_DATA_PENDING_COUNT(fp) (((struct __FILE_public *)(fp))->_r)
#else
/* requires systems own version of the ReadDataPending() */
extern int ReadDataPending();
#  define READ_DATA_PENDING(fp) (!feof(fp))
#  define READ_DATA_BUFFERED(fp) 0
#endif
#ifndef READ_DATA_BUFFERED
#  define READ_DATA_BUFFERED(fp) READ_DATA_PENDING(fp)
#endif

#ifndef READ_DATA_PENDING_PTR
# ifdef FILE_READPTR
#  define READ_DATA_PENDING_PTR(fp) ((char *)(fp)->FILE_READPTR)
# endif
#endif

#if defined __DJGPP__
# undef READ_DATA_PENDING_COUNT
# undef READ_DATA_PENDING_PTR
#endif

#define MakeOpenFile(mrb, obj, fp) do {\
  if (RFILE(obj)->fptr) {\
    rb_io_close(mrb, obj);\
    free(RFILE(obj)->fptr);\
    RFILE(obj)->fptr = 0;\
  }\
  fp = 0;\
  fp = RFILE(obj)->fptr = mrb_malloc(mrb, sizeof(struct rb_io_t));\
  fp->f = fp->f2 = NULL;\
  fp->mode = 0;\
  fp->pid = 0;\
  fp->lineno = 0;\
  fp->path = NULL;\
  fp->finalize = 0;\
} while (0)

/****************************************************************/
//#define F_SETFL         1
#define uid_t int
mrb_value io_puts(mrb_state *mrb, int argc, mrb_value *argv, mrb_value out);
mrb_value rb_protect_inspect(mrb_value (*func)(ANYARGS), mrb_value obj, mrb_value arg);
void rb_thread_schedule(mrb_state * mrb);
//void rb_trap_exec (void);

#define	NBBY	8		/* number of bits in a byte */
typedef	long	fd_mask;
//#define	NFDBITS	(sizeof (fd_mask) * NBBY)	/* bits per mask */
#define HAVE_RB_FD_INIT
#include <sys/types.h>

#if defined(NFDBITS) && defined(HAVE_RB_FD_INIT)
/* Add --> kusuda */
#ifndef	FD_SETSIZE
#define	FD_SETSIZE	64
#endif
/* We use a macro for fd_set so that including Sockets.h afterwards
   can work.  */
#ifndef	howmany
#define	howmany(x,y)	(((x)+((y)-1))/(y))
#endif
typedef	struct _types_fd_set {
	fd_mask	fds_bits[howmany(FD_SETSIZE, NFDBITS)];
} _types_fd_set;
#define fd_set _types_fd_set
/* Add --< kusuda */
typedef struct struct_rb_fdset_t{
    int maxfd;
    struct struct_rb_fdset_t *fdset;
} rb_fdset_t;

int select(int n, fd_set *r, fd_set *w, fd_set *e, struct timeval *timeout);

void rb_fd_init(rb_fdset_t *);
void rb_fd_term(rb_fdset_t *);
void rb_fd_zero(rb_fdset_t *);
void rb_fd_set(int, rb_fdset_t *);
void rb_fd_clr(int, rb_fdset_t *);
int rb_fd_isset(int, const rb_fdset_t *);
void rb_fd_copy(rb_fdset_t *, const fd_set *, int);
int rb_fd_select(int, rb_fdset_t *, rb_fdset_t *, rb_fdset_t *, struct timeval *);

#define rb_fd_ptr(f)	((f)->fdset)
#define rb_fd_max(f)	((f)->maxfd)

#elif defined(_WIN32)

typedef struct {
    int capa;
    fd_set *fdset;
} rb_fdset_t;

void rb_fd_init(volatile rb_fdset_t *);
void rb_fd_term(rb_fdset_t *);
#define rb_fd_zero(f)		((f)->fdset->fd_count = 0)
void rb_fd_set(int, rb_fdset_t *);
#define rb_fd_clr(n, f)		rb_w32_fdclr(n, (f)->fdset)
#define rb_fd_isset(n, f)	rb_w32_fdisset(n, (f)->fdset)
#define rb_fd_select(n, rfds, wfds, efds, timeout)	rb_w32_select(n, (rfds) ? ((rb_fdset_t*)rfds)->fdset : NULL, (wfds) ? ((rb_fdset_t*)wfds)->fdset : NULL, (efds) ? ((rb_fdset_t*)efds)->fdset: NULL, timeout)
#define rb_fd_resize(n, f)	(void)(f)

#define rb_fd_ptr(f)	((f)->fdset)
#define rb_fd_max(f)	((f)->fdset->fd_count)

#else

#define fd_set _types_fd_set

//typedef fd_set rb_fdset_t;
#define rb_fd_zero(f)	FD_ZERO(f)
#define rb_fd_set(n, f)	FD_SET(n, f)
#define rb_fd_clr(n, f)	FD_CLR(n, f)
#define rb_fd_isset(n, f) FD_ISSET(n, f)
#define rb_fd_copy(d, s, n) (*(d) = *(s))
#define rb_fd_resize(n, f)	(void)(f)
#define rb_fd_ptr(f)	(f)
#define rb_fd_init(f)	FD_ZERO(f)
#define rb_fd_term(f)	(void)(f)
#define rb_fd_max(f)	FD_SETSIZE
#define rb_fd_select(n, rfds, wfds, efds, timeout)	select(n, rfds, wfds, efds, timeout)

#endif
int rb_inspecting_p(mrb_value);
//#define mbclen(c)   (re_mbctab[(unsigned char)(c)]+1)
//int rb_thread_select(mrb_state *mrb, int max, fd_set * read, fd_set * write, fd_set * except,
//         struct timeval *timeout);

/****************************************************************/
#define READ_CHECK(mrb,fp) do {\
    if (!READ_DATA_PENDING(fp)) {\
        rb_io_check_closed(mrb, fptr);\
     }\
} while(0)

void
rb_eof_error(mrb_state *mrb)
{
    mrb_raise(mrb, mrb->eEOFError_class, "end of file reached");
}

mrb_value
rb_io_taint_check(mrb_value io)
{
    //if (!OBJ_TAINTED(io) && rb_safe_level() >= 4)
    //  mrb_raise(mrb, mrb->eSecurityError_class, "Insecure: operation on untainted IO");
    //rb_check_frozen(io);
    return io;
}

void
mrb_io_check_initialized(mrb_state *mrb, rb_io_t *fptr)
{
  if (!fptr) {
    mrb_raise(mrb, mrb->eIOError_class, "uninitialized stream");
  }
}

void
rb_io_check_closed(mrb_state *mrb, rb_io_t *fptr)
{
  mrb_io_check_initialized(mrb, fptr);
  if (!fptr->f && !fptr->f2) {
    mrb_raise(mrb, mrb->eIOError_class, "closed stream");
  }
}

static void io_fflush (mrb_state *mrb, FILE *, rb_io_t *);

static rb_io_t *
flush_before_seek(mrb_state *mrb, rb_io_t *fptr)
{
    if (fptr->mode & FMODE_WBUF) {
      io_fflush(mrb, GetWriteFile(fptr), fptr);
    }
    errno = 0;
    return fptr;
}

#define io_seek(fptr, ofs, whence) fseeko(flush_before_seek(mrb, fptr)->f, ofs, whence)
#define io_tell(fptr) ftello(flush_before_seek(mrb, fptr)->f)

#ifndef SEEK_CUR
# define SEEK_SET 0
# define SEEK_CUR 1
# define SEEK_END 2
#endif

#define FMODE_SYNCWRITE (FMODE_SYNC|FMODE_WRITABLE)

void
rb_io_check_readable(mrb_state *mrb, rb_io_t *fptr)
{
    rb_io_check_closed(mrb, fptr);
    if (!(fptr->mode & FMODE_READABLE)) {
      mrb_raise(mrb, mrb->eIOError_class, "not opened for reading");
    }
#ifdef NEED_IO_SEEK_BETWEEN_RW
    if (((fptr->mode & FMODE_WBUF) ||
       (fptr->mode & (FMODE_SYNCWRITE|FMODE_RBUF)) == FMODE_SYNCWRITE) &&
      !feof(fptr->f) &&
      !fptr->f2) {
      io_seek(fptr, 0, SEEK_CUR);
    }
#endif
    fptr->mode |= FMODE_RBUF;
}

void
rb_io_check_writable(mrb_state *mrb, rb_io_t *fptr)
{
    rb_io_check_closed(mrb, fptr);
    if (!(fptr->mode & FMODE_WRITABLE)) {
      mrb_raise(mrb, mrb->eIOError_class, "not opened for writing");
    }
    if ((fptr->mode & FMODE_RBUF) && !feof(fptr->f) && !fptr->f2) {
      io_seek(fptr, 0, SEEK_CUR);
    }
    if (!fptr->f2) {
      fptr->mode &= ~FMODE_RBUF;
    }
}

int
rb_read_pending(FILE *fp)
{
    return READ_DATA_PENDING(fp);
}

void
rb_read_check(mrb_state *mrb, FILE *fp)
{
    if (!READ_DATA_PENDING(fp)) {
      //rb_thread_wait_fd(mrb, fileno(fp));
    }
}

static int
ruby_dup(int orig)
{
    int fd;

    fd = dup(orig);
    if (fd < 0) {
      if (errno == EMFILE || errno == ENFILE || errno == ENOMEM) {
          //rb_gc();
          fd = dup(orig);
      }
      if (fd < 0) {
          rb_sys_fail(0);
      }
    }
    return fd;
}

static mrb_value
io_alloc(mrb_state *mrb, struct RClass *klass)
{
  //NEWOBJ(io, struct RFile);
  //OBJSETUP(io, klass, MRB_TT_FILE);
  struct RFile* io;
  io = mrb_obj_alloc(mrb, MRB_TT_FILE, klass);

  io->fptr = 0;

  return mrb_obj_value(io);
}

static void
io_fflush(mrb_state *mrb, FILE *f, rb_io_t *fptr)
{
    int n;

    //if (!rb_thread_fd_writable(mrb, fileno(f))) {
    //    rb_io_check_closed(mrb, fptr);
    //}
    for (;;) {
      TRAP_BEG(mrb);
      n = fflush(f);
      TRAP_END(mrb);
      if (n != EOF) break;
      //if (!rb_io_wait_writable(mrb, fileno(f)))
      rb_sys_fail(fptr->path);
    }
    fptr->mode &= ~FMODE_WBUF;
}

#ifndef S_ISREG
#   define S_ISREG(m) ((m & S_IFMT) == S_IFREG)
#endif

static int
wsplit_p(rb_io_t *fptr)
{
    FILE *f = GetWriteFile(fptr);
#if defined(HAVE_FCNTL) && defined(F_GETFL) && defined(O_NONBLOCK)
    int r;
#endif

    if (!(fptr->mode & FMODE_WSPLIT_INITIALIZED)) {
        struct stat buf;
        if (fstat(fileno(f), &buf) == 0 &&
            !S_ISREG(buf.st_mode)
#if defined(HAVE_FCNTL) && defined(F_GETFL) && defined(O_NONBLOCK)
            && (r = fcntl(fileno(f), F_GETFL)) != -1 &&
            !(r & O_NONBLOCK)
#endif
            ) {
            fptr->mode |= FMODE_WSPLIT;
        }
        fptr->mode |= FMODE_WSPLIT_INITIALIZED;
    }
    return fptr->mode & FMODE_WSPLIT;
}

/* writing functions */
static long
io_fwrite(mrb_state *mrb, mrb_value str, rb_io_t *fptr)
{
    long len, n, r, l, offset = 0;
    FILE *f = GetWriteFile(fptr);

    len = RSTRING(str)->len;
    if ((n = len) <= 0) return n;
    if (fptr->mode & FMODE_SYNC) {
      io_fflush(mrb, f, fptr);
      //if (!rb_thread_fd_writable(mrb, fileno(f))) {
      //    rb_io_check_closed(mrb, fptr);
      //}
      //retry:
        l = n;
        if (PIPE_BUF < l &&
            wsplit_p(fptr)) {
            l = PIPE_BUF;
        }
        TRAP_BEG(mrb);
      r = write(fileno(f), RSTRING(str)->buf+offset, l);
        TRAP_END(mrb);
#if BSD_STDIO
      preserving_errno(fseeko(f, lseek(fileno(f), (off_t)0, SEEK_CUR), SEEK_SET));
#endif
        if (r == n) return len;
        if (0 <= r) {
            offset += r;
            n -= r;
            errno = EAGAIN;
        }
        //if (rb_io_wait_writable(mrb, fileno(f))) {
        //    rb_io_check_closed(mrb, fptr);
        //  if (offset < RSTRING(str)->len)
        //    goto retry;
        //}
        return -1L;
    }
#if defined(__human68k__) || defined(__vms)
    do {
      if (fputc(RSTRING(str)->buf[offset++], f) == EOF) {
          if (ferror(f)) return -1L;
          break;
      }
    } while (--n > 0);
#else
    while (errno = 0, offset += (r = fwrite(RSTRING(str)->buf+offset, 1, n, f)), (n -= r) > 0) {
      if (ferror(f)
#if defined __BORLANDC__
          || errno
#endif
      ) {
#ifdef __hpux
          if (!errno) errno = EAGAIN;
#elif defined(_WIN32) && !defined(__BORLANDC__)
          /* workaround for MSVCRT's bug */
          if (!errno) {
            if (GetLastError() == ERROR_NO_DATA)
                errno = EPIPE;
            else
                errno = EBADF;
          }
#endif
          //if (rb_io_wait_writable(mrb, fileno(f))) {
          //  rb_io_check_closed(mrb, fptr);
          //  clearerr(f);
          //  if (offset < RSTRING(str)->len)
          //      continue;
          //}
          return -1L;
      }
    }
#endif
    return len - n;
}

long
rb_io_fwrite(mrb_state *mrb, const char *ptr, long len, FILE *f)
{
    rb_io_t of;

    of.f = f;
    of.f2 = NULL;
    of.mode = FMODE_WRITABLE;
    of.path = NULL;
    return io_fwrite(mrb, mrb_str_new(mrb, ptr, len), &of);
}

/*
 *  call-seq:
 *     ios.write(string)    => integer
 *
 *  Writes the given string to <em>ios</em>. The stream must be opened
 *  for writing. If the argument is not a string, it will be converted
 *  to a string using <code>to_s</code>. Returns the number of bytes
 *  written.
 *
 *     count = $stdout.write( "This is a test\n" )
 *     puts "That was #{count} bytes of data"
 *
 *  <em>produces:</em>
 *
 *     This is a test
 *     That was 15 bytes of data
 */

mrb_value
io_write(mrb_state *mrb, mrb_value io, mrb_value str, int nosync)
{
    rb_io_t *fptr;
    long n;

    //rb_secure(4);
    if (mrb_type(str) != MRB_TT_STRING)
      str = mrb_obj_as_string(mrb, str);

    if (mrb_type(io) != MRB_TT_FILE) {
      /* port is not IO, call write method for it. */
      //return rb_funcall(io, "write", 1, str);
      return mrb_funcall(mrb, io, "write", 1, str);
    }
    if (RSTRING(str)->len == 0) return mrb_fixnum_value(0);

    GetOpenFile(mrb, io, fptr);
    rb_io_check_writable(mrb, fptr);

    n = io_fwrite(mrb, str, fptr);
    if (n == -1L) rb_sys_fail(fptr->path);
    if (!(fptr->mode & FMODE_SYNC)) {
      fptr->mode |= FMODE_WBUF;
    }

    return mrb_fixnum_value(n);
}

mrb_value
rb_io_write(mrb_state *mrb, mrb_value io, mrb_value str)
{
    //return rb_funcall(io, "write", 1, str);
    return mrb_funcall(mrb, io, "write", 1, str);
}

/* 15.2.20.5.20 */
/*
 *  call-seq:
 *     ios.write(string)    -> integer
 *
 *  Writes the given string to <em>ios</em>. The stream must be opened
 *  for writing. If the argument is not a string, it will be converted
 *  to a string using <code>to_s</code>. Returns the number of bytes
 *  written.
 *
 *     count = $stdout.write( "This is a test\n" )
 *     puts "That was #{count} bytes of data"
 *
 *  <em>produces:</em>
 *
 *     This is a test
 *     That was 15 bytes of data
 */

static mrb_value
mrb_io_write_m(mrb_state *mrb, mrb_value io)
{
  mrb_value str;

  mrb_get_args(mrb, "o", &str);
  return io_write(mrb, io, str, 0);
}
/*
 *  call-seq:
 *     ios << obj     => ios
 *
 *  String Output---Writes <i>obj</i> to <em>ios</em>.
 *  <i>obj</i> will be converted to a string using
 *  <code>to_s</code>.
 *
 *     $stdout << "Hello " << "world!\n"
 *
 *  <em>produces:</em>
 *
 *     Hello world!
 */

mrb_value
rb_io_addstr(mrb_state *mrb, mrb_value io, mrb_value str)
{
    rb_io_write(mrb, io, str);
    return io;
}

/* 15.2.20.5.7  */
/*
 *  call-seq:
 *     ios.flush    => ios
 *
 *  Flushes any buffered data within <em>ios</em> to the underlying
 *  operating system (note that this is Ruby internal buffering only;
 *  the OS may buffer the data as well).
 *
 *     $stdout.print "no newline"
 *     $stdout.flush
 *
 *  <em>produces:</em>
 *
 *     no newline
 */

mrb_value
mrb_io_flush(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;
    FILE *f;

    GetOpenFile(mrb, io, fptr);
    rb_io_check_writable(mrb, fptr);
    f = GetWriteFile(fptr);

    io_fflush(mrb, f, fptr);
#ifdef _WIN32
    fsync(fileno(f));
#endif

    return io;
}

#if 0
/*
 *  call-seq:
 *     ios.pos     => integer
 *     ios.tell    => integer
 *
 *  Returns the current offset (in bytes) of <em>ios</em>.
 *
 *     f = File.new("testfile")
 *     f.pos    #=> 0
 *     f.gets   #=> "This is line one\n"
 *     f.pos    #=> 17
 */

static mrb_value
rb_io_tell(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;
    off_t pos;

    GetOpenFile(mrb, io, fptr);
    pos = io_tell(fptr);
    if (pos < 0 && errno) rb_sys_fail(fptr->path);
    return OFFT2NUM(pos);
}
#endif

#if 0
static mrb_value
rb_io_seek(mrb_state *mrb, mrb_value io, mrb_value offset, int whence)
{
    rb_io_t *fptr;
    off_t pos;

    pos = NUM2OFFT(offset);
    GetOpenFile(mrb, io, fptr);
    pos = io_seek(fptr, pos, whence);
    if (pos < 0 && errno) rb_sys_fail(fptr->path);
    clearerr(fptr->f);

    return mrb_fixnum_value(0);
}

/*
 *  call-seq:
 *     ios.seek(amount, whence=SEEK_SET) -> 0
 *
 *  Seeks to a given offset <i>anInteger</i> in the stream according to
 *  the value of <i>whence</i>:
 *
 *    IO::SEEK_CUR  | Seeks to _amount_ plus current position
 *    --------------+----------------------------------------------------
 *    IO::SEEK_END  | Seeks to _amount_ plus end of stream (you probably
 *                  | want a negative value for _amount_)
 *    --------------+----------------------------------------------------
 *    IO::SEEK_SET  | Seeks to the absolute location given by _amount_
 *
 *  Example:
 *
 *     f = File.new("testfile")
 *     f.seek(-13, IO::SEEK_END)   #=> 0
 *     f.readline                  #=> "And so on...\n"
 */

static mrb_value
rb_io_seek_m(int argc, mrb_value *argv, mrb_value io)
{
    mrb_value offset, ptrname;
    int whence = SEEK_SET;

    if (rb_scan_args(argc, argv, "11", &offset, &ptrname) == 2) {
      whence = mrb_fixnum_number(ptrname);
    }

    return rb_io_seek(io, offset, whence);
}
#endif

#if 0
/*
 *  call-seq:
 *     ios.pos = integer    => integer
 *
 *  Seeks to the given position (in bytes) in <em>ios</em>.
 *
 *     f = File.new("testfile")
 *     f.pos = 17
 *     f.gets   #=> "This is line two\n"
 */

static mrb_value
rb_io_set_pos(mrb_state *mrb, mrb_value io, mrb_value offset)
{
    rb_io_t *fptr;
    off_t pos;

    pos = NUM2OFFT(offset);
    GetOpenFile(mrb, io, fptr);
    pos = io_seek(fptr, pos, SEEK_SET);
    if (pos != 0) rb_sys_fail(fptr->path);
    clearerr(fptr->f);

    return OFFT2NUM(pos);
}

/*
 *  call-seq:
 *     ios.rewind    => 0
 *
 *  Positions <em>ios</em> to the beginning of input, resetting
 *  <code>lineno</code> to zero.
 *
 *     f = File.new("testfile")
 *     f.readline   #=> "This is line one\n"
 *     f.rewind     #=> 0
 *     f.lineno     #=> 0
 *     f.readline   #=> "This is line one\n"
 */

static mrb_value
rb_io_rewind(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;

    GetOpenFile(mrb, io, fptr);
    if (io_seek(fptr, 0L, 0) != 0) rb_sys_fail(fptr->path);
    clearerr(fptr->f);
    if (io == current_file) {
      gets_lineno -= fptr->lineno;
    }
    fptr->lineno = 0;

    return mrb_fixnum_value(0);
}
#endif

/* 15.2.20.5.6  */
/*
 *  call-seq:
 *     ios.eof     => true or false
 *     ios.eof?    => true or false
 *
 *  Returns true if <em>ios</em> is at end of file that means
 *  there are no more data to read.
 *  The stream must be opened for reading or an <code>IOError</code> will be
 *  raised.
 *
 *     f = File.new("testfile")
 *     dummy = f.readlines
 *     f.eof   #=> true
 *
 *  If <em>ios</em> is a stream such as pipe or socket, <code>IO#eof?</code>
 *  blocks until the other end sends some data or closes it.
 *
 *     r, w = IO.pipe
 *     Thread.new { sleep 1; w.close }
 *     r.eof?  #=> true after 1 second blocking
 *
 *     r, w = IO.pipe
 *     Thread.new { sleep 1; w.puts "a" }
 *     r.eof?  #=> false after 1 second blocking
 *
 *     r, w = IO.pipe
 *     r.eof?  # blocks forever
 *
 *  Note that <code>IO#eof?</code> reads data to a input buffer.
 *  So <code>IO#sysread</code> doesn't work with <code>IO#eof?</code>.
 */

mrb_value
mrb_io_eof(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;
    int ch;

    GetOpenFile(mrb, io, fptr);
    rb_io_check_readable(mrb, fptr);

    if (feof(fptr->f)) return mrb_true_value();
    if (READ_DATA_PENDING(fptr->f)) return mrb_false_value();
    READ_CHECK(mrb, fptr->f);
    clearerr(fptr->f);
    TRAP_BEG(mrb);
    ch = getc(fptr->f);
    TRAP_END(mrb);

    if (ch != EOF) {
      ungetc(ch, fptr->f);
      return mrb_false_value();
    }
    rb_io_check_closed(mrb, fptr);
    clearerr(fptr->f);
    return mrb_true_value();
}

/* 15.2.20.5.18 */
/*
 *  call-seq:
 *     ios.sync    => true or false
 *
 *  Returns the current ``sync mode'' of <em>ios</em>. When sync mode is
 *  true, all output is immediately flushed to the underlying operating
 *  system and is not buffered by Ruby internally. See also
 *  <code>IO#fsync</code>.
 *
 *     f = File.new("testfile")
 *     f.sync   #=> false
 */

static mrb_value
mrb_io_sync(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;

    GetOpenFile(mrb, io, fptr);
    return (fptr->mode & FMODE_SYNC) ? mrb_true_value() : mrb_false_value();
}

/* 15.2.20.5.19 */
/*
 *  call-seq:
 *     ios.sync = boolean   => boolean
 *
 *  Sets the ``sync mode'' to <code>true</code> or <code>false</code>.
 *  When sync mode is true, all output is immediately flushed to the
 *  underlying operating system and is not buffered internally. Returns
 *  the new state. See also <code>IO#fsync</code>.
 *
 *     f = File.new("testfile")
 *     f.sync = true
 *
 *  <em>(produces no output)</em>
 */

static mrb_value
mrb_io_set_sync(mrb_state *mrb, mrb_value io)
{
  rb_io_t *fptr;
  mrb_value mode;

  mrb_get_args(mrb, "o", &mode);

    GetOpenFile(mrb, io, fptr);
    if (mrb_test(mode)) {
      fptr->mode |= FMODE_SYNC;
    }
    else {
      fptr->mode &= ~FMODE_SYNC;
    }
    return mode;
}

#if 0
/*
 *  call-seq:
 *     ios.fsync   => 0 or nil
 *
 *  Immediately writes all buffered data in <em>ios</em> to disk.
 *  Returns <code>nil</code> if the underlying operating system does not
 *  support <em>fsync(2)</em>. Note that <code>fsync</code> differs from
 *  using <code>IO#sync=</code>. The latter ensures that data is flushed
 *  from Ruby's buffers, but doesn't not guarantee that the underlying
 *  operating system actually writes it to disk.
 */

static mrb_value
rb_io_fsync(mrb_value io)
{
#ifdef HAVE_FSYNC
    rb_io_t *fptr;
    FILE *f;

    GetOpenFile(mrb, io, fptr);
    f = GetWriteFile(fptr);

    io_fflush(mrb, f, fptr);
    if (fsync(fileno(f)) < 0)
      rb_sys_fail(fptr->path);
    return mrb_fixnum_value(0);
#else
    rb_notimplement();
    return mrb_nil_value();            /* not reached */
#endif
}

/*
 *  call-seq:
 *     ios.fileno    => fixnum
 *     ios.to_i      => fixnum
 *
 *  Returns an integer representing the numeric file descriptor for
 *  <em>ios</em>.
 *
 *     $stdin.fileno    #=> 0
 *     $stdout.fileno   #=> 1
 */

static mrb_value
rb_io_fileno(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;
    int fd;

    GetOpenFile(mrb, io, fptr);
    fd = fileno(fptr->f);
    return mrb_fixnum_value(fd);
}

/*
 *  call-seq:
 *     ios.pid    => fixnum
 *
 *  Returns the process mrb_sym of a child process associated with
 *  <em>ios</em>. This will be set by <code>IO::popen</code>.
 *
 *     pipe = IO.popen("-")
 *     if pipe
 *       $stderr.puts "In parent, child pid is #{pipe.pid}"
 *     else
 *       $stderr.puts "In child, pid is #{$$}"
 *     end
 *
 *  <em>produces:</em>
 *
 *     In child, pid is 26209
 *     In parent, child pid is 26209
 */

static mrb_value
rb_io_pid(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;

    GetOpenFile(mrb, io, fptr);
    if (!fptr->pid)
      return mrb_nil_value();
    return mrb_fixnum_value(fptr->pid);
}
#endif

/* 15.2.20.5.23 (x)*/
/*
 * call-seq:
 *   ios.inspect   => string
 *
 * Return a string describing this IO object.
 */

static mrb_value
mrb_io_inspect(mrb_state *mrb, mrb_value obj)
{
    rb_io_t *fptr;
    char *buf;
    const char *cname, *st = "";
    long len;

    fptr = RFILE(rb_io_taint_check(obj))->fptr;
    if (!fptr || !fptr->path) return mrb_any_to_s(mrb, obj);
    cname = mrb_obj_classname(mrb, obj);
    len = strlen(cname) + strlen(fptr->path) + 5;
    if (!(fptr->f || fptr->f2)) {
      st = " (closed)";
      len += 9;
    }
    //buf = ALLOCA_N(char, len);
    buf = mrb_malloc(mrb, len);
    snprintf(buf, len, "#<%s:%s%s>", cname, fptr->path, st);
    return mrb_str_new2(mrb, buf);
}

/* 15.2.20.5.22 (x)*/
/*
 *  call-seq:
 *     ios.to_io -> ios
 *
 *  Returns <em>ios</em>.
 */

static mrb_value
mrb_io_to_io(mrb_state *mrb, mrb_value io)
{
    return io;
}

/* reading functions */
static long
read_buffered_data(char *ptr, long len, FILE *f)
{
    long n;

#ifdef READ_DATA_PENDING_COUNT
    n = READ_DATA_PENDING_COUNT(f);
    if (n <= 0) return 0;
    if (n > len) n = len;
    return fread(ptr, 1, n, f);
#else
    int c;

    for (n = 0; n < len && READ_DATA_PENDING(f) && (c = getc(f)) != EOF; ++n) {
      *ptr++ = c;
    }
    return n;
#endif
}

static long
io_fread(mrb_state *mrb, char *ptr, long len, rb_io_t *fptr)
{
    long n = len;
    int c;
    int saved_errno;

    while (n > 0) {
        c = read_buffered_data(ptr, n, fptr->f);
        if (c < 0) goto eof;
        if (c > 0) {
            ptr += c;
            if ((n -= c) <= 0) break;
        }
        //rb_thread_wait_fd(mrb, fileno(fptr->f));
        rb_io_check_closed(mrb, fptr);
      clearerr(fptr->f);
      TRAP_BEG(mrb);
      c = getc(fptr->f);
      TRAP_END(mrb);
      if (c == EOF) {
        eof:
          if (ferror(fptr->f)) {
            switch (errno) {
              case EINTR:
#if defined(ERESTART)
              case ERESTART:
#endif
                clearerr(fptr->f);
                continue;
              case EAGAIN:
#if defined(EWOULDBLOCK) && EWOULDBLOCK != EAGAIN
              case EWOULDBLOCK:
#endif
                if (len > n) {
                  clearerr(fptr->f);
                }
                    saved_errno = errno;
                    rb_warning("nonblocking IO#read is obsolete; use IO#readpartial or IO#sysread");
                    errno = saved_errno;
            }
            if (len == n) return 0;
          }
          break;
      }
      *ptr++ = c;
      n--;
    }
    return len - n;
}

long
rb_io_fread(mrb_state *mrb, char *ptr, long len, FILE *f)
{
    rb_io_t of;

    of.f = f;
    of.f2 = NULL;
    return io_fread(mrb, ptr, len, &of);
}

#define SMALLBUF 100

static long
remain_size(mrb_state *mrb, rb_io_t *fptr)
{
    struct stat st;
    off_t siz = BUFSIZ;
    off_t pos;

    if (feof(fptr->f)) return 0;
    if (fstat(fileno(fptr->f), &st) == 0  && S_ISREG(st.st_mode)
#ifdef __BEOS__
      && (st.st_dev > 3)
#endif
      )
    {
      pos = io_tell(fptr);
      if (st.st_size >= pos && pos >= 0) {
          siz = st.st_size - pos + 1;
          if (siz > LONG_MAX) {
            mrb_raise(mrb, mrb->eIOError_class, "file too big for single read");
          }
      }
    }
    return (long)siz;
}

mrb_value
rb_str_locktmp(mrb_value str)
{
  /*if (FL_TEST(str, STR_TMPLOCK)) {
    rb_raise(rb_eRuntimeError, "temporal locking already locked string");
  }
  FL_SET(str, STR_TMPLOCK);
  */
  return str;
}

mrb_value
rb_str_unlocktmp(mrb_value str)
{
  /*if (!FL_TEST(str, STR_TMPLOCK)) {
    rb_raise(rb_eRuntimeError, "temporal unlocking already unlocked string");
  }
  FL_UNSET(str, STR_TMPLOCK);
  */
  return str;
}

static mrb_value
read_all(mrb_state *mrb, rb_io_t *fptr, long siz, mrb_value str)
{
    long bytes = 0;
    long n;

    if (siz == 0) siz = BUFSIZ;
    if (mrb_nil_p(str)) {
      str = mrb_str_new(mrb, 0, siz);
    }
    else {
      mrb_str_resize(mrb, str, siz);
    }
    for (;;) {
      rb_str_locktmp(str);
      READ_CHECK(mrb, fptr->f);
      n = io_fread(mrb, RSTRING(str)->buf+bytes, siz-bytes, fptr);
      rb_str_unlocktmp(str);
      if (n == 0 && bytes == 0) {
          if (!fptr->f) break;
          if (feof(fptr->f)) break;
          if (!ferror(fptr->f)) break;
          rb_sys_fail(fptr->path);
      }
      bytes += n;
      if (bytes < siz) break;
      siz += BUFSIZ;
      mrb_str_resize(mrb, str, siz);
    }
    if (bytes != siz) mrb_str_resize(mrb, str, bytes);
    //OBJ_TAINT(str);

    return str;
}

#if 0
void rb_io_set_nonblock(rb_io_t *fptr)
{
    int flags;
#ifdef F_GETFL
    flags = fcntl(fileno(fptr->f), F_GETFL);
    if (flags == -1) {
        rb_sys_fail(fptr->path);
    }
#else
    flags = 0;
#endif
    if ((flags & O_NONBLOCK) == 0) {
        flags |= O_NONBLOCK;
        if (fcntl(fileno(fptr->f), F_SETFL, flags) == -1) {
            rb_sys_fail(fptr->path);
        }
    }
    if (fptr->f2) {
#ifdef F_GETFL
        flags = fcntl(fileno(fptr->f2), F_GETFL);
        if (flags == -1) {
            rb_sys_fail(fptr->path);
        }
#else
        flags = 0;
#endif
        if ((flags & O_NONBLOCK) == 0) {
            flags |= O_NONBLOCK;
            if (fcntl(fileno(fptr->f2), F_SETFL, flags) == -1) {
                rb_sys_fail(fptr->path);
            }
        }
    }
}

static mrb_value
io_getpartial(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io, int nonblock)
{
    rb_io_t *fptr;
    mrb_value length, str;
    long n, len;

    rb_scan_args(argc, argv, "11", &length, &str);

    if ((len = mrb_fixnum_number(length)) < 0) {
        mrb_raise(mrb, mrb->eArgumentError_class, "negative length %ld given", len);
    }

    if (mrb_nil_p(str)) {
        str = mrb_str_new(mrb, 0, len);
    }
    else {
        //StringValue(str);
        mrb_string_value(mrb, &str);
        rb_str_modify(str);
        mrb_str_resize(mrb, str, len);
    }
    OBJ_TAINT(str);

    GetOpenFile(mrb, io, fptr);
    rb_io_check_readable(mrb, fptr);

    if (len == 0)
        return str;

    if (!nonblock) {
        READ_CHECK(mrb, fptr->f);
    }
    if (RSTRING(str)->len != len) {
      modified:
        mrb_raise(mrb, mrb->eRuntimeError_class, "buffer string modified");
    }
    n = read_buffered_data(RSTRING(str)->buf, len, fptr->f);
    if (n <= 0) {
again:
      if (RSTRING(str)->len != len) goto modified;
      if (nonblock) {
          rb_io_set_nonblock(fptr);
          n = read(fileno(fptr->f), RSTRING(str)->buf, len);
      }
      else {
          TRAP_BEG(mrb);
          n = read(fileno(fptr->f), RSTRING(str)->buf, len);
          TRAP_END(mrb);
      }
      if (n < 0) {
          //if (!nonblock && rb_io_wait_readable(mrb, fileno(fptr->f)))
          //    goto again;
          rb_sys_fail(fptr->path);
      }
      if (fptr->f) /* update pos in FILE structure [ruby-core:21561] */
          fflush(fptr->f);
    }
    mrb_str_resize(mrb, str, n);

    if (n == 0)
        return mrb_nil_value();
    else
        return str;
}
#endif

#if 0
/*
 *  call-seq:
 *     ios.readpartial(maxlen)              => string
 *     ios.readpartial(maxlen, outbuf)      => outbuf
 *
 *  Reads at most <i>maxlen</i> bytes from the I/O stream.
 *  It blocks only if <em>ios</em> has no data immediately available.
 *  It doesn't block if some data available.
 *  If the optional <i>outbuf</i> argument is present,
 *  it must reference a String, which will receive the data.
 *  It raises <code>EOFError</code> on end of file.
 *
 *  readpartial is designed for streams such as pipe, socket, tty, etc.
 *  It blocks only when no data immediately available.
 *  This means that it blocks only when following all conditions hold.
 *  * the buffer in the IO object is empty.
 *  * the content of the stream is empty.
 *  * the stream is not reached to EOF.
 *
 *  When readpartial blocks, it waits data or EOF on the stream.
 *  If some data is reached, readpartial returns with the data.
 *  If EOF is reached, readpartial raises EOFError.
 *
 *  When readpartial doesn't blocks, it returns or raises immediately.
 *  If the buffer is not empty, it returns the data in the buffer.
 *  Otherwise if the stream has some content,
 *  it returns the data in the stream.
 *  Otherwise if the stream is reached to EOF, it raises EOFError.
 *
 *     r, w = IO.pipe           #               buffer          pipe content
 *     w << "abc"               #               ""              "abc".
 *     r.readpartial(4096)      #=> "abc"       ""              ""
 *     r.readpartial(4096)      # blocks because buffer and pipe is empty.
 *
 *     r, w = IO.pipe           #               buffer          pipe content
 *     w << "abc"               #               ""              "abc"
 *     w.close                  #               ""              "abc" EOF
 *     r.readpartial(4096)      #=> "abc"       ""              EOF
 *     r.readpartial(4096)      # raises EOFError
 *
 *     r, w = IO.pipe           #               buffer          pipe content
 *     w << "abc\ndef\n"        #               ""              "abc\ndef\n"
 *     r.gets                   #=> "abc\n"     "def\n"         ""
 *     w << "ghi\n"             #               "def\n"         "ghi\n"
 *     r.readpartial(4096)      #=> "def\n"     ""              "ghi\n"
 *     r.readpartial(4096)      #=> "ghi\n"     ""              ""
 *
 *  Note that readpartial behaves similar to sysread.
 *  The differences are:
 *  * If the buffer is not empty, read from the buffer instead of "sysread for buffered IO (IOError)".
 *  * It doesn't cause Errno::EAGAIN and Errno::EINTR.  When readpartial meets EAGAIN and EINTR by read system call, readpartial retry the system call.
 *
 *  The later means that readpartial is nonblocking-flag insensitive.
 *  It blocks on the situation IO#sysread causes Errno::EAGAIN as if the fd is blocking mode.
 *
 */

static mrb_value
io_readpartial(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
    mrb_value ret;

    ret = io_getpartial(argc, argv, io, 0);
    if (mrb_nil_p(ret))
        rb_eof_error(mrb);
    else
        return ret;
}

/*
 *  call-seq:
 *     ios.read_nonblock(maxlen)              => string
 *     ios.read_nonblock(maxlen, outbuf)      => outbuf
 *
 *  Reads at most <i>maxlen</i> bytes from <em>ios</em> using
 *  read(2) system call after O_NONBLOCK is set for
 *  the underlying file descriptor.
 *
 *  If the optional <i>outbuf</i> argument is present,
 *  it must reference a String, which will receive the data.
 *
 *  read_nonblock just calls read(2).
 *  It causes all errors read(2) causes: EAGAIN, EINTR, etc.
 *  The caller should care such errors.
 *
 *  read_nonblock causes EOFError on EOF.
 *
 *  If the read buffer is not empty,
 *  read_nonblock reads from the buffer like readpartial.
 *  In this case, read(2) is not called.
 *
 */

static mrb_value
io_read_nonblock(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
    mrb_value ret;

    ret = io_getpartial(argc, argv, io, 1);
    if (mrb_nil_p(ret))
        rb_eof_error(mrb);
    else
        return ret;
}

/*
 *  call-seq:
 *     ios.write_nonblock(string)   => integer
 *
 *  Writes the given string to <em>ios</em> using
 *  write(2) system call after O_NONBLOCK is set for
 *  the underlying file descriptor.
 *
 *  write_nonblock just calls write(2).
 *  It causes all errors write(2) causes: EAGAIN, EINTR, etc.
 *  The result may also be smaller than string.length (partial write).
 *  The caller should care such errors and partial write.
 *
 */

static mrb_value
rb_io_write_nonblock(mrb_state *mrb, mrb_value io, mrb_value str)
{
    rb_io_t *fptr;
    FILE *f;
    long n;

    rb_secure(4);
    if (mrb_type(str) != T_STRING)
      str = mrb_obj_as_string(mrb, str);

    GetOpenFile(mrb, io, fptr);
    rb_io_check_writable(mrb, fptr);

    f = GetWriteFile(fptr);

    rb_io_set_nonblock(fptr);
    n = write(fileno(f), RSTRING(str)->buf, RSTRING(str)->len);

    if (n == -1) rb_sys_fail(fptr->path);

    return mrb_fixnum_value(n);
}
#endif

/*
 *  call-seq:
 *     ios.read([length [, buffer]])    => string, buffer, or nil
 *
 *  Reads at most <i>length</i> bytes from the I/O stream, or to the
 *  end of file if <i>length</i> is omitted or is <code>nil</code>.
 *  <i>length</i> must be a non-negative integer or nil.
 *  If the optional <i>buffer</i> argument is present, it must reference
 *  a String, which will receive the data.
 *
 *  At end of file, it returns <code>nil</code> or <code>""</code>
 *  depend on <i>length</i>.
 *  <code><i>ios</i>.read()</code> and
 *  <code><i>ios</i>.read(nil)</code> returns <code>""</code>.
 *  <code><i>ios</i>.read(<i>positive-integer</i>)</code> returns nil.
 *
 *     f = File.new("testfile")
 *     f.read(16)   #=> "This is line one"
 */

static mrb_value
io_read(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
    rb_io_t *fptr;
    long n, len;
    mrb_value length, str;

    //rb_scan_args(argc, argv, "02", &length, &str);
    length = argv[0];
    str    = argv[1];

    if (mrb_nil_p(length)) {
      //if (!mrb_nil_p(str)) StringValue(str);
      if (!mrb_nil_p(str)) mrb_string_value(mrb, &str);
      GetOpenFile(mrb, io, fptr);
      rb_io_check_readable(mrb, fptr);
      return read_all(mrb, fptr, remain_size(mrb, fptr), str);
    }
    len = mrb_fixnum_number(length);
    if (len < 0) {
      mrb_raise(mrb, mrb->eArgumentError_class, "negative length %ld given", len);
    }

    if (mrb_nil_p(str)) {
      str = rb_tainted_str_new(mrb, 0, len);
    }
    else {
      //StringValue(str);
      mrb_string_value(mrb, &str);
      rb_str_modify(str);
      mrb_str_resize(mrb, str,len);
    }

    GetOpenFile(mrb, io, fptr);
    rb_io_check_readable(mrb, fptr);
    if (feof(fptr->f)) return mrb_nil_value();
    if (len == 0) return str;

    rb_str_locktmp(str);
    READ_CHECK(mrb, fptr->f);
    if (RSTRING(str)->len != len) {
      mrb_raise(mrb, mrb->eRuntimeError_class, "buffer string modified");
    }
    n = io_fread(mrb, RSTRING(str)->buf, len, fptr);
    rb_str_unlocktmp(str);
    if (n == 0) {
      if (!fptr->f) return mrb_nil_value();
      if (feof(fptr->f)) {
          mrb_str_resize(mrb, str, 0);
          return mrb_nil_value();
      }
      if (len > 0) rb_sys_fail(fptr->path);
    }
    mrb_str_resize(mrb, str, n);
    RSTRING(str)->len = n;
    RSTRING(str)->buf[n] = '\0';
    //OBJ_TAINT(str);

    return str;
}

static int
appendlineOld(mrb_state *mrb, rb_io_t *fptr, int delim, mrb_value *strp)
{
    FILE *f = fptr->f;
    mrb_value str = *strp;
    int c = EOF;
#ifndef READ_DATA_PENDING_PTR
    char buf[8192];
    char *bp = buf, *bpe = buf + sizeof buf - 3;
    int update = FALSE;
#endif

    do {
#ifdef READ_DATA_PENDING_PTR
      long pending = READ_DATA_PENDING_COUNT(f);
      if (pending > 0) {
          const char *p = READ_DATA_PENDING_PTR(f);
          const char *e = memchr(p, delim, pending);
          long last = 0, len = (c != EOF);
          if (e) pending = e - p + 1;
          len += pending;
          if (!mrb_nil_p(str)) {
            last = RSTRING(str)->len;
            mrb_str_resize(mrb, str, last + len);
          }
          else {
            *strp = str = mrb_str_buf_new(mrb, len);
            RSTRING(str)->len = len;
            RSTRING(str)->buf[len] = '\0';
          }
          if (c != EOF) {
            RSTRING(str)->buf[last++] = c;
          }
          fread(RSTRING(str)->buf + last, 1, pending, f); /* must not fail */
          if (e) return delim;
      }
      else if (c != EOF) {
          if (!mrb_nil_p(str)) {
            char ch = c;
            mrb_str_buf_cat(mrb, str, &ch, 1);
          }
          else {
            *strp = str = mrb_str_buf_new(mrb, 1);
            RSTRING(str)->buf[RSTRING(str)->len++] = c;
          }
      }
      //rb_thread_wait_fd(mrb, fileno(f));
      rb_io_check_closed(mrb, fptr);
#else
      READ_CHECK(mrb, f);
#endif
      clearerr(f);
      TRAP_BEG(mrb);
      c = getc(f);
      TRAP_END(mrb);
      if (c == EOF) {
          if (ferror(f)) {
            clearerr(f);
            //if (!rb_io_wait_readable(mrb, fileno(f)))
            rb_sys_fail(fptr->path);
            continue;
          }
#ifdef READ_DATA_PENDING_PTR
          return c;
#endif
      }
#ifndef READ_DATA_PENDING_PTR
      if (c == EOF || (*bp++ = c) == delim || bp == bpe) {
          int cnt = bp - buf;

          if (cnt > 0) {
            if (!mrb_nil_p(str))
                rb_str_cat(mrb, str, buf, cnt);
            else
                *strp = str = mrb_str_new(mrb, buf, cnt);
          }
          if (c == EOF) {
            if (update) {
              if (!mrb_nil_p(str))
                return (int)RSTRING(str)->buf[RSTRING(str)->len-1];
            }
            return c;
          }
          bp = buf;
      }
      update = TRUE;
#endif
    } while (c != delim);

#ifdef READ_DATA_PENDING_PTR
    {
      char ch = c;
      if (!mrb_nil_p(str)) {
          rb_str_cat(mrb, str, &ch, 1);
      }
      else {
          *strp = str = mrb_str_new(mrb, &ch, 1);
      }
    }
#endif

    return c;
}

static int
appendline(mrb_state *mrb, rb_io_t *fptr, int delim, mrb_value *strp, long *lp)
{
    FILE *f = fptr->f;
    mrb_value str = *strp;
    int c = EOF;
#ifndef READ_DATA_PENDING_PTR
    long limit = *lp;
    char buf[8192];
    char *bp = buf, *bpe = buf + sizeof buf - 3;
    int update = FALSE;
#endif

    do {
#ifdef READ_DATA_PENDING_PTR
      long pending = READ_DATA_PENDING_COUNT(f);
      if (pending > 0) {
          const char *p = READ_DATA_PENDING_PTR(f);
          const char *e = memchr(p, delim, pending);
          long last = 0, len = (c != EOF);
          if (e) pending = e - p + 1;
          len += pending;
          if (!mrb_nil_p(str)) {
            last = RSTRING(str)->len;
            mrb_str_resize(mrb, str, last + len);
          }
          else {
            *strp = str = mrb_str_buf_new(mrb, len);
            RSTRING(str)->len = len;
            RSTRING(str)->buf[len] = '\0';
          }
          if (c != EOF) {
            RSTRING(str)->buf[last++] = c;
          }
          fread(RSTRING(str)->buf + last, 1, pending, f); /* must not fail */
          if (e) return delim;
      }
      else if (c != EOF) {
          if (!mrb_nil_p(str)) {
            char ch = c;
            mrb_str_buf_cat(mrb, str, &ch, 1);
          }
          else {
            *strp = str = mrb_str_buf_new(mrb, 1);
            RSTRING(str)->buf[RSTRING(str)->len++] = c;
          }
      }
      //rb_thread_wait_fd(mrb, fileno(f));
      rb_io_check_closed(mrb, fptr);
#else
      READ_CHECK(mrb, f);
#endif
      clearerr(f);
      TRAP_BEG(mrb);
      c = getc(f);
      TRAP_END(mrb);
      if (c == EOF) {
          if (ferror(f)) {
            clearerr(f);
            //if (!rb_io_wait_readable(mrb, fileno(f)))
            rb_sys_fail(fptr->path);
            continue;
          }
#ifdef READ_DATA_PENDING_PTR
          return c;
#endif
      }
#ifndef READ_DATA_PENDING_PTR
      if (c == EOF || (*bp++ = c) == delim || bp == bpe || (bp==(buf+limit))) {
          int cnt = bp - buf;

          if (cnt > 0) {
            if (!mrb_nil_p(str))
                rb_str_cat(mrb, str, buf, cnt);
            else
                *strp = str = mrb_str_new(mrb, buf, cnt);
          }
          if (c == EOF) {
            if (update) {
              if (!mrb_nil_p(str))
                return (int)RSTRING(str)->buf[RSTRING(str)->len-1];
            }
            return c;
          }
          else {
            *lp = limit - (bp - buf);
            return c;
          }
          bp = buf;
      }
      update = TRUE;
#endif
    } while (c != delim);

#ifdef READ_DATA_PENDING_PTR
    {
      char ch = c;
      if (!mrb_nil_p(str)) {
          rb_str_cat(mrb, str, &ch, 1);
      }
      else {
          *strp = str = mrb_str_new(mrb, &ch, 1);
      }
    }
#endif

    return c;}

static inline int
swallow(mrb_state *mrb, rb_io_t *fptr, int term)
{
    FILE *f = fptr->f;
    int c;

    do {
#ifdef READ_DATA_PENDING_PTR
      long cnt;
      while ((cnt = READ_DATA_PENDING(f)) > 0) {
          char buf[1024];
          const char *p = READ_DATA_PENDING_PTR(f);
          int i;
          if (cnt > sizeof buf) cnt = sizeof buf;
          if (*p != term) return TRUE;
          i = cnt;
          while (--i && *++p == term);
          if (!fread(buf, 1, cnt - i, f)) /* must not fail */
            rb_sys_fail(fptr->path);
      }
      //rb_thread_wait_fd(mrb, fileno(f));
      rb_io_check_closed(mrb, fptr);
#else
      READ_CHECK(mrb, f);
#endif
      clearerr(f);
      TRAP_BEG(mrb);
      c = getc(f);
      TRAP_END(mrb);
      if (c != term) {
          ungetc(c, f);
          return TRUE;
      }
    } while (c != EOF);
    return FALSE;
}

static mrb_value
rb_io_getline_fast(mrb_state *mrb, rb_io_t *fptr, unsigned char delim)
{
    mrb_value str = mrb_nil_value();
    int c;

    while ((c = appendlineOld(mrb, fptr, delim, &str)) != EOF && (c != delim));

    if (!mrb_nil_p(str)) {
      fptr->lineno++;
      lineno = mrb_fixnum_value(fptr->lineno);
      //OBJ_TAINT(str);
    }

    return str;
}

static int
rscheck(mrb_state *mrb, const char *rsptr, long rslen, mrb_value rs)
{
    if (RSTRING(rs)->buf != rsptr && RSTRING(rs)->len != rslen)
      mrb_raise(mrb, mrb->eRuntimeError_class, "rs modified");
    return 1;
}

static mrb_value
rb_io_getline_1(mrb_state *mrb, mrb_value rs, long limit, mrb_value io)
{
  mrb_value str = mrb_nil_value();
  rb_io_t *fptr;
  int nolimit = 0;
  long svlimit = limit;

  GetOpenFile(mrb, io, fptr);
  rb_io_check_readable(mrb, fptr);
  if (mrb_nil_p(rs) && limit < 0) {
    str = read_all(mrb, fptr, 0, mrb_nil_value());
    if (RSTRING_LEN(str) == 0) return mrb_nil_value();
  }
  else if (limit == 0) {
    return mrb_str_new(mrb, 0, 0);
  }
  else if (limit < 0) {
    return rb_io_getline_fast(mrb, fptr, '\n');
  }
  else {
    int c, newline = -1;
    const char *rsptr = 0;
    long rslen = 0;
    int rspara = 0;
    int extra_limit = 16;
    const char *s, *p, *e, *pp;

    if (!mrb_nil_p(rs)) {
      rslen = RSTRING_LEN(rs);
      if (rslen == 0) {
        rsptr = "\n\n";
        rslen = 2;
        rspara = 1;
        swallow(mrb, fptr, '\n');
        rs = mrb_str_new_cstr(mrb, "");
      }
      else {
        rsptr = RSTRING_PTR(rs);
      }
      newline = (unsigned char)rsptr[rslen - 1];
    }

    /* MS - Optimisation */
    while ((c = appendline(mrb, fptr, newline, &str, &limit)) != EOF) {

      if (c == newline) {
        if (RSTRING_LEN(str) < rslen) continue;
        s = RSTRING_PTR(str);
                e = s + RSTRING_LEN(str);
        p = e - rslen;
        //pp = rb_enc_left_char_head(s, p, e, enc);
        //if (pp != p) continue;
        if (!rspara) rscheck(mrb, rsptr, rslen, rs);
        if (memcmp(p, rsptr, rslen) == 0) break;
      }
      if (limit == 0) {
        limit = svlimit;
        s = RSTRING_PTR(str);
        p = s + RSTRING_LEN(str);
        pp = p-1;//rb_enc_left_char_head(s, p-1, p, enc);
        if (extra_limit && ((p - pp) < -1)
            /* MBCLEN_NEEDMORE_P(rb_enc_precise_mbclen(pp, p, enc))*/) {
            /* relax the limit while incomplete character.
             * extra_limit limits the relax length */
            limit = 1;
            extra_limit--;
        }
        else {
            nolimit = 1;
            break;
        }
      }
    } /* end of whils */

    if (rspara) {
      if (c != EOF) {
        swallow(mrb, fptr, '\n');
      }
    }
  } /* end of else */

  if (!mrb_nil_p(str)) {
    if (!nolimit) {
      fptr->lineno++;
#if 1
      lineno = mrb_fixnum_value(fptr->lineno);
#else
      if (mrb_obj_equal(mrb, io, ARGF.current_file)) {
        ARGF.lineno++;
        ARGF.last_lineno = ARGF.lineno;
      }
      else {
        ARGF.last_lineno = fptr->lineno;
      }
#endif
    }
  }

  return str;
}

static void
prepare_getline_args(mrb_state *mrb, int argc, mrb_value *argv, mrb_value *rsp, long *limit, mrb_value io);

static mrb_value
rb_io_getline(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
    mrb_value rs;
    long limit;

    prepare_getline_args(mrb, argc, argv, &rs, &limit, io);
    return rb_io_getline_1(mrb, rs, limit, io);
}

static mrb_value
rb_io_gets_m(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
  mrb_value str;

  str = rb_io_getline(mrb, argc, argv, io);
  rb_lastline_set(str);

  return str;
}

/* 15.2.20.5.9  */
/*
 *  call-seq:
 *     ios.gets(sep_string=$/)   => string or nil
 *
 *  Reads the next ``line'' from the I/O stream; lines are separated by
 *  <i>sep_string</i>. A separator of <code>nil</code> reads the entire
 *  contents, and a zero-length separator reads the input a paragraph at
 *  a time (two successive newlines in the input separate paragraphs).
 *  The stream must be opened for reading or an <code>IOError</code>
 *  will be raised. The line read in will be returned and also assigned
 *  to <code>$_</code>. Returns <code>nil</code> if called at end of
 *  file.
 *
 *     File.new("testfile").gets   #=> "This is line one\n"
 *     $_                          #=> "This is line one\n"
 */

static mrb_value
mrb_io_gets_m(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value io)
{
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);
  return rb_io_gets_m(mrb, argc, argv, io);
}

#if 0
/*
 *  call-seq:
 *     ios.lineno    => integer
 *
 *  Returns the current line number in <em>ios</em>. The stream must be
 *  opened for reading. <code>lineno</code> counts the number of times
 *  <code>gets</code> is called, rather than the number of newlines
 *  encountered. The two values will differ if <code>gets</code> is
 *  called with a separator other than newline. See also the
 *  <code>$.</code> variable.
 *
 *     f = File.new("testfile")
 *     f.lineno   #=> 0
 *     f.gets     #=> "This is line one\n"
 *     f.lineno   #=> 1
 *     f.gets     #=> "This is line two\n"
 *     f.lineno   #=> 2
 */

static mrb_value
rb_io_lineno(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;

    GetOpenFile(mrb, io, fptr);
    rb_io_check_readable(mrb, fptr);
    return mrb_fixnum_value(fptr->lineno);
}

/*
 *  call-seq:
 *     ios.lineno = integer    => integer
 *
 *  Manually sets the current line number to the given value.
 *  <code>$.</code> is updated only on the next read.
 *
 *     f = File.new("testfile")
 *     f.gets                     #=> "This is line one\n"
 *     $.                         #=> 1
 *     f.lineno = 1000
 *     f.lineno                   #=> 1000
 *     $. # lineno of last read   #=> 1
 *     f.gets                     #=> "This is line two\n"
 *     $. # lineno of last read   #=> 1001
 */

static mrb_value
rb_io_set_lineno(mrb_state *mrb, mrb_value io, mrb_value lineno)
{
    rb_io_t *fptr;

    GetOpenFile(mrb, io, fptr);
    rb_io_check_readable(mrb, fptr);
    fptr->lineno = mrb_fixnum_number(lineno);
    return lineno;
}
#endif

static void
lineno_setter(mrb_value val, mrb_sym id, mrb_value *var)
{
    gets_lineno = mrb_fixnum_number(val);
    *var = mrb_fixnum_value(gets_lineno);
}

#if 0
static mrb_value
argf_set_lineno(mrb_value argf, mrb_value val)
{
    gets_lineno = mrb_fixnum_number(val);
    lineno = mrb_fixnum_value(gets_lineno);
    return mrb_nil_value();
}

static mrb_value
argf_lineno()
{
    return lineno;
}
#endif

/* 15.2.20.5.14 */
/*
 *  call-seq:
 *     ios.read([length [, buffer]])    -> string, buffer, or nil
 *
 *  Reads <i>length</i> bytes from the I/O stream.
 *
 *  <i>length</i> must be a non-negative integer or <code>nil</code>.
 *
 *  If <i>length</i> is a positive integer,
 *  it try to read <i>length</i> bytes without any conversion (binary mode).
 *  It returns <code>nil</code> or a string whose length is 1 to <i>length</i> bytes.
 *  <code>nil</code> means it met EOF at beginning.
 *  The 1 to <i>length</i>-1 bytes string means it met EOF after reading the result.
 *  The <i>length</i> bytes string means it doesn't meet EOF.
 *  The resulted string is always ASCII-8BIT encoding.
 *
 *  If <i>length</i> is omitted or is <code>nil</code>,
 *  it reads until EOF and the encoding conversion is applied.
 *  It returns a string even if EOF is met at beginning.
 *
 *  If <i>length</i> is zero, it returns <code>""</code>.
 *
 *  If the optional <i>buffer</i> argument is present, it must reference
 *  a String, which will receive the data.
 *
 *  At end of file, it returns <code>nil</code> or <code>""</code>
 *  depend on <i>length</i>.
 *  <code><i>ios</i>.read()</code> and
 *  <code><i>ios</i>.read(nil)</code> returns <code>""</code>.
 *  <code><i>ios</i>.read(<i>positive-integer</i>)</code> returns <code>nil</code>.
 *
 *     f = File.new("testfile")
 *     f.read(16)   #=> "This is line one"
 *
 *     # reads whole file
 *     open("file") {|f|
 *       data = f.read # This returns a string even if the file is empty.
 *       ...
 *     }
 *
 *     # iterate over fixed length records.
 *     open("fixed-record-file") {|f|
 *       while record = f.read(256)
 *         ...
 *       end
 *     }
 *
 *     # iterate over variable length records.
 *     # record is prefixed by 32-bit length.
 *     open("variable-record-file") {|f|
 *       while len = f.read(4)
 *         len = len.unpack("N")[0] # 32-bit length
 *         record = f.read(len) # This returns a string even if len is 0.
 *       end
 *     }
 *
 *  Note that this method behaves like fread() function in C.
 *  If you need the behavior like read(2) system call,
 *  consider readpartial, read_nonblock and sysread.
 */

static mrb_value
mrb_io_read(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value io)
{
  mrb_value argv[16];
  int argc;

  //rb_scan_args(argc, argv, "02", &length, &str);
  mrb_get_args(mrb, "*", &argv, &argc);
  return io_read(mrb, argc, argv, io);

}

/* 15.2.20.5.16 */
/*
 *  call-seq:
 *     ios.readline(sep_string=$/)   => string
 *
 *  Reads a line as with <code>IO#gets</code>, but raises an
 *  <code>EOFError</code> on end of file.
 */

static mrb_value
mrb_io_readline(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value io)
{
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);
  mrb_value line = rb_io_gets_m(mrb, argc, argv, io);

  if (mrb_nil_p(line)) {
    rb_eof_error(mrb);
  }
  return line;
}

static void
prepare_getline_args(mrb_state *mrb, int argc, mrb_value *argv, mrb_value *rsp, long *limit, mrb_value io)
{
    mrb_value rs = rb_rs, lim = mrb_nil_value();

    if (argc == 1) {
      mrb_value tmp = mrb_nil_value();

      if (mrb_type(argv[0]) == MRB_TT_FIXNUM) {
        lim = argv[0];
      }
      else {
        if (mrb_nil_p(argv[0]) || !mrb_nil_p(tmp = mrb_check_string_type(mrb, argv[0]))) {
          rs = tmp;
        }
        else {
          lim = argv[0];
        }
      }
    }
    else if (2 <= argc) {
      //rb_scan_args(argc, argv, "2", &rs, &lim);
      mrb_get_args(mrb, "oo", &rs, &lim);
      if (!mrb_nil_p(rs))
        //StringValue(rs);
        mrb_string_value(mrb, &rs);
    }
    *rsp = rs;
    *limit = mrb_nil_p(lim) ? -1L : mrb_fixnum_number(lim);
}

/* 15.2.20.5.17 */
/*
 *  call-seq:
 *     ios.readlines(sep_string=$/)  =>   array
 *
 *  Reads all of the lines in <em>ios</em>, and returns them in
 *  <i>anArray</i>. Lines are separated by the optional
 *  <i>sep_string</i>. If <i>sep_string</i> is <code>nil</code>, the
 *  rest of the stream is returned as a single record.
 *  The stream must be opened for reading or an
 *  <code>IOError</code> will be raised.
 *
 *     f = File.new("testfile")
 *     f.readlines[0]   #=> "This is line one\n"
 */

static mrb_value
mrb_io_readlines(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value io)
{
  mrb_value line, ary, rs;
  long limit;
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);
  prepare_getline_args(mrb, argc, argv, &rs, &limit, io);
  ary = mrb_ary_new(mrb);
  while (!mrb_nil_p(line = rb_io_getline_1(mrb, rs, limit, io))) {
    mrb_ary_push(mrb, ary, line);
  }
  return ary;
}

/* 15.2.20.5.3  */
/* 15.2.20.5.5  */
/*
 *  call-seq:
 *     ios.each(sep_string=$/)      {|line| block }  => ios
 *     ios.each_line(sep_string=$/) {|line| block }  => ios
 *
 *  Executes the block for every line in <em>ios</em>, where lines are
 *  separated by <i>sep_string</i>. <em>ios</em> must be opened for
 *  reading or an <code>IOError</code> will be raised.
 *
 *     f = File.new("testfile")
 *     f.each {|line| puts "#{f.lineno}: #{line}" }
 *
 *  <em>produces:</em>
 *
 *     1: This is line one
 *     2: This is line two
 *     3: This is line three
 *     4: And so on...
 */

static mrb_value
mrb_io_each_line(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value io)
{
  mrb_value str;
  mrb_value rs;
  long limit;
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);

  //RETURN_ENUMERATOR(io, argc, argv);
  prepare_getline_args(mrb, argc, argv, &rs, &limit, io);
  while (!mrb_nil_p(str = rb_io_getline_1(mrb, rs, limit, io))) {
    rb_yield(str);
  }
  return io;
}

/* 15.2.20.5.4  */
/*
 *  call-seq:
 *     ios.each_byte {|byte| block }  => ios
 *
 *  Calls the given block once for each byte (0..255) in <em>ios</em>,
 *  passing the byte as an argument. The stream must be opened for
 *  reading or an <code>IOError</code> will be raised.
 *
 *     f = File.new("testfile")
 *     checksum = 0
 *     f.each_byte {|x| checksum ^= x }   #=> #<File:testfile>
 *     checksum                           #=> 12
 */

static mrb_value
mrb_io_each_byte(mrb_state *mrb, mrb_value io)
{
  rb_io_t *fptr;
  FILE *f;
  int c;

  //RETURN_ENUMERATOR(io, 0, 0);
  GetOpenFile(mrb, io, fptr);

  for (;;) {
    rb_io_check_readable(mrb, fptr);
    f = fptr->f;
    READ_CHECK(mrb, f);
    clearerr(f);
    TRAP_BEG(mrb);
    c = getc(f);
    TRAP_END(mrb);
    if (c == EOF) {
        if (ferror(f)) {
          clearerr(f);
          //if (!rb_io_wait_readable(mrb, fileno(f)))
          rb_sys_fail(fptr->path);
          continue;
        }
        break;
    }
    rb_yield(mrb_fixnum_value(c & 0xff));
  }
  if (ferror(f)) rb_sys_fail(fptr->path);
  return io;
}

mrb_value mrb_io_getc (mrb_state *, mrb_value);

#if 0
/*
 *  call-seq:
 *     ios.each_char {|c| block }  => ios
 *
 *  Calls the given block once for each character in <em>ios</em>,
 *  passing the character as an argument. The stream must be opened for
 *  reading or an <code>IOError</code> will be raised.  Multibyte
 *  characters are dealt with according to $KCODE.
 *
 *     f = File.new("testfile")
 *     f.each_char {|c| print c, ' ' }   #=> #<File:testfile>
 */

static mrb_value
rb_io_each_char(mrb_value io)
{
    mrb_value ch;

    //RETURN_ENUMERATOR(io, 0, 0);

    while (!mrb_nil_p(ch = mrb_io_getc(mrb, io))) {
      unsigned char c;
      int n;
      mrb_value str;

      c= mrb_fixnum_number(ch);
      n = mbclen(c);
      str = rb_tainted_str_new((const char *)&c, 1);

      while (--n > 0) {
          if (mrb_nil_p(ch = mrb_io_getc(mrb, io))) {
            rb_yield(str);
            return io;
          }
          c = mrb_fixnum_number(ch);
          rb_str_cat(mrb, str, (const char *)&c, 1);
      }
      rb_yield(str);
    }
    return io;
}

/*
 *  call-seq:
 *     ios.lines(sep=$/)     => anEnumerator
 *     ios.lines(limit)      => anEnumerator
 *     ios.lines(sep, limit) => anEnumerator
 *
 *  Returns an enumerator that gives each line in <em>ios</em>.
 *  The stream must be opened for reading or an <code>IOError</code>
 *  will be raised.
 *
 *     f = File.new("testfile")
 *     f.lines.to_a  #=> ["foo\n", "bar\n"]
 *     f.rewind
 *     f.lines.sort  #=> ["bar\n", "foo\n"]
 */

static mrb_value
rb_io_lines(int argc, mrb_value *argv, mrb_value io)
{
    return rb_enumeratorize(io, ID2SYM(rb_intern("each_line")), argc, argv);
}

/*
 *  call-seq:
 *     ios.bytes   => anEnumerator
 *
 *  Returns an enumerator that gives each byte (0..255) in <em>ios</em>.
 *  The stream must be opened for reading or an <code>IOError</code>
 *  will be raised.
 *
 *     f = File.new("testfile")
 *     f.bytes.to_a  #=> [104, 101, 108, 108, 111]
 *     f.rewind
 *     f.bytes.sort  #=> [101, 104, 108, 108, 111]
 */

static mrb_value
rb_io_bytes(mrb_value io)
{
    return rb_enumeratorize(io, ID2SYM(rb_intern("each_byte")), 0, 0);
}
#endif

/* 15.2.20.5.8  */
/*
 *  call-seq:
 *     ios.getc   => fixnum or nil
 *
 *  Gets the next 8-bit byte (0..255) from <em>ios</em>. Returns
 *  <code>nil</code> if called at end of file.
 *
 *     f = File.new("testfile")
 *     f.getc   #=> 84
 *     f.getc   #=> 104
 */

mrb_value
mrb_io_getc(mrb_state *mrb, mrb_value io)
{
  rb_io_t *fptr;
  FILE *f;
  int c;
  char buf[2];

  GetOpenFile(mrb, io, fptr);
  rb_io_check_readable(mrb, fptr);
  f = fptr->f;

retry:
  READ_CHECK(mrb, f);
  clearerr(f);
  TRAP_BEG(mrb);
  c = getc(f);
  TRAP_END(mrb);

  if (c == EOF) {
    if (ferror(f)) {
      clearerr(f);
      //if (!rb_io_wait_readable(mrb, fileno(f)))
      rb_sys_fail(fptr->path);
      goto retry;
    }
    return mrb_nil_value();
  }
  //return mrb_fixnum_value(c & 0xff);
  sprintf(buf, "%c", c & 0xff);
  return mrb_str_new_cstr(mrb, buf);
}

int
rb_getc(mrb_state *mrb, FILE *f)
{
    int c;

    clearerr(f);
    TRAP_BEG(mrb);
    c = getc(f);
    TRAP_END(mrb);

    return c;
}

/* 15.2.20.5.15 */
/*
 *  call-seq:
 *     ios.readchar   => fixnum
 *
 *  Reads a character as with <code>IO#getc</code>, but raises an
 *  <code>EOFError</code> on end of file.
 */

static mrb_value
mrb_io_readchar(mrb_state *mrb, mrb_value io)
{
    mrb_value c = mrb_io_getc(mrb, io);

    if (mrb_nil_p(c)) {
      rb_eof_error(mrb);
    }
    return c;
}

/*
 *  call-seq:
 *     ios.ungetc(integer)   => nil
 *
 *  Pushes back one character (passed as a parameter) onto <em>ios</em>,
 *  such that a subsequent buffered read will return it. Only one character
 *  may be pushed back before a subsequent read operation (that is,
 *  you will be able to read only the last of several characters that have been pushed
 *  back). Has no effect with unbuffered reads (such as <code>IO#sysread</code>).
 *
 *     f = File.new("testfile")   #=> #<File:testfile>
 *     c = f.getc                 #=> 84
 *     f.ungetc(c)                #=> nil
 *     f.getc                     #=> 84
 */

mrb_value
rb_io_ungetc(mrb_state *mrb, mrb_value io, mrb_value c)
{
    rb_io_t *fptr;
    int cc = mrb_fixnum_number(c);

    GetOpenFile(mrb, io, fptr);
    if (!(fptr->mode & FMODE_RBUF))
      mrb_raise(mrb, mrb->eIOError_class, "unread stream");
    rb_io_check_readable(mrb, fptr);

    if (ungetc(cc, fptr->f) == EOF && cc != EOF) {
      mrb_raise(mrb, mrb->eIOError_class, "ungetc failed");
    }
    return mrb_nil_value();
}

#if 0
/*
 *  call-seq:
 *     ios.isatty   => true or false
 *     ios.tty?     => true or false
 *
 *  Returns <code>true</code> if <em>ios</em> is associated with a
 *  terminal device (tty), <code>false</code> otherwise.
 *
 *     File.new("testfile").isatty   #=> false
 *     File.new("/dev/tty").isatty   #=> true
 */

static mrb_value
rb_io_isatty(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;

    GetOpenFile(mrb, io, fptr);
    if (isatty(fileno(fptr->f)) == 0)
      return mrb_false_value();
    return mrb_true_value();
}
#endif

void
rb_syswait(rb_pid_t pid)
{
  /* dummy */
}

static void
fptr_finalize(mrb_state *mrb, rb_io_t *fptr, int noraise)
{
    int n1 = 0, n2 = 0, f1, f2 = -1;

    errno = 0;
    if (fptr->f2) {
      f2 = fileno(fptr->f2);
      while (n2 = 0, fflush(fptr->f2) < 0) {
          n2 = errno;
          //if (!rb_io_wait_writable(mrb, f2)) {
            break;
          //}
          if (!fptr->f2) break;
      }
      if (fclose(fptr->f2) < 0 && n2 == 0) {
          n2 = errno;
      }
      fptr->f2 = 0;
    }
    if (fptr->f) {
      f1 = fileno(fptr->f);
      if ((f2 == -1) && (fptr->mode & FMODE_WBUF)) {
          while (n1 = 0, fflush(fptr->f) < 0) {
            n1 = errno;
            if (1/*!rb_io_wait_writable(mrb, f1)*/) break;
            if (!fptr->f) break;
          }
      }
      if (fclose(fptr->f) < 0 && n1 == 0) {
          n1 = errno;
      }
      fptr->f = 0;
      if (n1 == EBADF && f1 == f2) {
          n1 = 0;
      }
    }
    if (!noraise && (n1 || n2)) {
      errno = (n1 ? n1 : n2);
      rb_sys_fail(fptr->path);
    }
}

static void
rb_io_fptr_cleanup(mrb_state *mrb, rb_io_t *fptr, int noraise)
{
    if (fptr->finalize) {
      (*fptr->finalize)(fptr, noraise);
    }
    else {
      fptr_finalize(mrb, fptr, noraise);
    }
}

int
rb_io_fptr_finalize(mrb_state *mrb, rb_io_t *fptr)
{
  if (!fptr) return 0;
  if (fptr->path) {
    free(fptr->path);
  }
  if ((fptr->f && fileno(fptr->f) > 2) || fptr->f2) {
    rb_io_fptr_cleanup(mrb, fptr, TRUE);
  }
  xfree(fptr);
  return 1;
}

mrb_value
rb_io_close(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;
    int fd, fd2;

    fptr = RFILE(io)->fptr;
    if (!fptr) return mrb_nil_value();
    if (fptr->f2) {
      fd2 = fileno(fptr->f2);
    }
    else {
      if (!fptr->f) return mrb_nil_value();
      fd2 = -1;
    }

    fd = fileno(fptr->f);
    rb_io_fptr_cleanup(mrb, fptr, FALSE);

    if (fptr->pid) {
      rb_syswait(fptr->pid);
      fptr->pid = 0;
    }

    return mrb_nil_value();
}

/* 15.2.20.5.1  */
/*
 *  call-seq:
 *     ios.close   => nil
 *
 *  Closes <em>ios</em> and flushes any pending writes to the operating
 *  system. The stream is unavailable for any further data operations;
 *  an <code>IOError</code> is raised if such an attempt is made. I/O
 *  streams are automatically closed when they are claimed by the
 *  garbage collector.
 *
 *  If <em>ios</em> is opened by <code>IO.popen</code>,
 *  <code>close</code> sets <code>$?</code>.
 */

mrb_value
mrb_io_close_m(mrb_state *mrb, mrb_value io)
{
  //if (rb_safe_level() >= 4 && !OBJ_TAINTED(io)) {
  //  mrb_raise(mrb, mrb->eSecurityError, "Insecure: can't close");
  //}
  rb_io_check_closed(mrb, RFILE(io)->fptr);
  rb_io_close(mrb, io);
  return mrb_nil_value();
}

static mrb_value
io_call_close(mrb_state *mrb, mrb_value io)
{
  //return rb_funcall(io, rb_intern("close"), 0, 0);
  return mrb_funcall(mrb, io, "close", 0, 0);
}

static mrb_value
io_close(mrb_state *mrb, mrb_value io)
{
  //return rb_rescue(mrb, io_call_close, io, 0, 0);
  return rb_rescue(mrb, io_call_close, &io, 0, (mrb_value*)0);
}

/* 15.2.20.5.2  */
/*
 *  call-seq:
 *     ios.closed?    => true or false
 *
 *  Returns <code>true</code> if <em>ios</em> is completely closed (for
 *  duplex streams, both reader and writer), <code>false</code>
 *  otherwise.
 *
 *     f = File.new("testfile")
 *     f.close         #=> nil
 *     f.closed?       #=> true
 *     f = IO.popen("/bin/sh","r+")
 *     f.close_write   #=> nil
 *     f.closed?       #=> false
 *     f.close_read    #=> nil
 *     f.closed?       #=> true
 */

static mrb_value
mrb_io_closed(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;

    fptr = RFILE(io)->fptr;
    mrb_io_check_initialized(mrb, fptr);
    return (fptr->f || fptr->f2)? mrb_false_value() : mrb_true_value();
}

/*
 *  call-seq:
 *     ios.binmode    => ios
 *
 *  Puts <em>ios</em> into binary mode. This is useful only in
 *  MS-DOS/Windows environments. Once a stream is in binary mode, it
 *  cannot be reset to nonbinary mode.
 */

mrb_value
rb_io_binmode(mrb_state *mrb, mrb_value io)
{
    rb_io_t *fptr;

    GetOpenFile(mrb, io, fptr);
#if (defined(O_BINARY) && O_BINARY) || (defined(_IOBIN) && _IOBIN)
#if (defined(_IOBIN) && _IOBIN)      /* __human68k__ */
    if (fptr->f)
      fmode(fptr->f, _IOBIN);
    if (fptr->f2)
      fmode(fptr->f2, _IOBIN);
#else
    if (fptr->f && setmode(fileno(fptr->f), O_BINARY) == -1)
      rb_sys_fail(fptr->path);
    if (fptr->f2 && setmode(fileno(fptr->f2), O_BINARY) == -1)
      rb_sys_fail(fptr->path);
#endif

    fptr->mode |= FMODE_BINMODE;
#endif
    return io;
}

const char*
rb_io_flags_mode(mrb_state *mrb, int flags)
{
#ifdef O_BINARY
# define MODE_BINMODE(a,b) ((flags & FMODE_BINMODE) ? (b) : (a))
#else
# define MODE_BINMODE(a,b) (a)
#endif
    if (flags & FMODE_APPEND) {
      if ((flags & FMODE_READWRITE) == FMODE_READWRITE) {
          return MODE_BINMODE("a+", "ab+");
      }
      return MODE_BINMODE("a", "ab");
    }
    switch (flags & FMODE_READWRITE) {
      case FMODE_READABLE:
      return MODE_BINMODE("r", "rb");
      case FMODE_WRITABLE:
      return MODE_BINMODE("w", "wb");
      case FMODE_READWRITE:
      if (flags & FMODE_CREATE) {
          return MODE_BINMODE("w+", "wb+");
      }
      return MODE_BINMODE("r+", "rb+");
    }
    mrb_raise(mrb, mrb->eArgumentError_class, "illegal access modenum %o", flags);
    return NULL;            /* not reached */
}

int
rb_io_mode_flags(mrb_state *mrb, const char *mode)
{
    int flags = 0;
    const char *m = mode;

    switch (*m++) {
      case 'r':
        flags |= FMODE_READABLE;
        break;
      case 'w':
        flags |= FMODE_WRITABLE | FMODE_CREATE;
        break;
      case 'a':
        flags |= FMODE_WRITABLE | FMODE_APPEND | FMODE_CREATE;
        break;
      default:
error:
        mrb_raise(mrb, mrb->eArgumentError_class, "illegal access mode %s", mode);
    }

    while (*m) {
        switch (*m++) {
          case 'b':
            flags |= FMODE_BINMODE;
            break;
          case '+':
            flags |= FMODE_READWRITE;
            break;
          case ':':
            /* forward compatibility */
            rb_warning("encoding options not supported in 1.8: %s", mode);
            goto end;
          default:
            goto error;
        }
    }

 end:
    return flags;
}

int
rb_io_modenum_flags(int mode)
{
    int flags = 0;

    switch (mode & (O_RDONLY|O_WRONLY|O_RDWR)) {
      case O_RDONLY:
      flags = FMODE_READABLE;
      break;
      case O_WRONLY:
      flags = FMODE_WRITABLE;
      break;
      case O_RDWR:
      flags = FMODE_READWRITE;
      break;
    }

    if (mode & O_APPEND) {
      flags |= FMODE_APPEND;
    }
    if (mode & O_CREAT) {
      flags |= FMODE_CREATE;
    }
#ifdef O_BINARY
    if (mode & O_BINARY) {
      flags |= FMODE_BINMODE;
    }
#endif

    return flags;
}

static int
rb_io_mode_modenum(mrb_state *mrb, const char *mode)
{
    int flags = 0;
    const char *m = mode;

    switch (*m++) {
      case 'r':
      flags |= O_RDONLY;
      break;
      case 'w':
      flags |= O_WRONLY | O_CREAT | O_TRUNC;
      break;
      case 'a':
      flags |= O_WRONLY | O_CREAT | O_APPEND;
      break;
      default:
      error:
      mrb_raise(mrb, mrb->eArgumentError_class, "illegal access mode %s", mode);
    }

    while (*m) {
        switch (*m++) {
        case 'b':
#ifdef O_BINARY
            flags |= O_BINARY;
#endif
            break;
        case '+':
            flags = (flags & ~O_ACCMODE) | O_RDWR;
            break;
        case ':':
            /* forward compatibility */
            rb_warning("encoding options not supported in 1.8: %s", mode);
            goto end;
        default:
            goto error;
        }
    }

 end:
    return flags;
}

#define MODENUM_MAX 4

static const char*
rb_io_modenum_mode(mrb_state *mrb, int flags)
{
#ifdef O_BINARY
# define MODE_BINARY(a,b) ((flags & O_BINARY) ? (b) : (a))
#else
# define MODE_BINARY(a,b) (a)
#endif
    int accmode = flags & O_ACCMODE;
    if (flags & O_APPEND) {
      if (accmode == O_WRONLY) {
            return MODE_BINARY("a", "ab");
        }
      if (accmode == O_RDWR) {
          return MODE_BINARY("a+", "ab+");
      }
    }
    switch (accmode) {
      case O_RDONLY:
      return MODE_BINARY("r", "rb");
      case O_WRONLY:
      return MODE_BINARY("w", "wb");
      case O_RDWR:
      return MODE_BINARY("r+", "rb+");
    }
    mrb_raise(mrb, mrb->eArgumentError_class, "illegal access modenum %o", flags);
    return NULL;            /* not reached */
}

static int
rb_sysopen(char *fname, int flags, unsigned int mode)
{
    int fd;

#ifdef _WIN32
    errno = EINVAL;
#endif
    fd = open(fname, flags, mode);
    if (fd < 0) {
      if (errno == EMFILE || errno == ENFILE) {
          //rb_gc();
#ifdef _WIN32
          errno = EINVAL;
#endif
          fd = open(fname, flags, mode);
      }
      if (fd < 0) {
          rb_sys_fail(fname);
      }
    }
    return fd;
}

FILE *
rb_fopen(const char *fname, const char *mode)
{
    FILE *file;

    file = fopen(fname, mode);
    if (!file) {
      if (errno == EMFILE || errno == ENFILE) {
          //rb_gc();
          file = fopen(fname, mode);
      }
      if (!file) {
          rb_sys_fail(fname);
      }
    }
#ifdef USE_SETVBUF
    if (setvbuf(file, NULL, _IOFBF, 0) != 0)
      rb_warn("setvbuf() can't be honoured for %s", fname);
#endif
#ifdef __human68k__
    fmode(file, _IOTEXT);
#endif
    return file;
}

FILE *
rb_fdopen(int fd, const char *mode)
{
    FILE *file;

#if defined(sun)
    errno = 0;
#endif
    file = fdopen(fd, mode);
    if (!file) {
#if defined(sun)
      if (errno == 0 || errno == EMFILE || errno == ENFILE) {
#else
      if (errno == EMFILE || errno == ENFILE) {
#endif
          //rb_gc();
#if defined(sun)
          errno = 0;
#endif
          file = fdopen(fd, mode);
      }
      if (!file) {
#ifdef _WIN32
          if (errno == 0) errno = EINVAL;
#endif
#if defined(sun)
          if (errno == 0) errno = EMFILE;
#endif
          rb_sys_fail(0);
      }
    }

#ifdef USE_SETVBUF
    if (setvbuf(file, NULL, _IOFBF, 0) != 0)
      rb_warn("setvbuf() can't be honoured (fd=%d)", fd);
#endif
    return file;
}

static mrb_value
rb_file_open_internal(mrb_state *mrb, mrb_value io, const char *fname, const char *mode)
{
    rb_io_t *fptr;

    MakeOpenFile(mrb, io, fptr);

    fptr->mode = rb_io_mode_flags(mrb, mode);
    fptr->path = strdup(fname);
    fptr->f = rb_fopen(fptr->path, rb_io_flags_mode(mrb, fptr->mode));

    return io;
}

mrb_value
rb_file_open(mrb_state *mrb, const char *fname, const char *mode)
{
    return rb_file_open_internal(mrb, io_alloc(mrb, mrb->file_class), fname, mode);
}

static mrb_value
rb_file_sysopen_internal(mrb_state *mrb, mrb_value io, char *fname, int flags, int mode)
{
    rb_io_t *fptr;
    int fd;
    const char *m;

    MakeOpenFile(mrb, io, fptr);

    fptr->path = strdup(fname);
    m = rb_io_modenum_mode(mrb, flags);
    fptr->mode = rb_io_modenum_flags(flags);
    fd = rb_sysopen(fptr->path, flags, mode);
    fptr->f = rb_fdopen(fd, m);

    return io;
}

mrb_value
rb_file_sysopen(mrb_state *mrb, const char *fname, int flags, int mode)
{
    return rb_file_sysopen_internal(mrb, io_alloc(mrb, mrb->file_class), (char*)fname, flags, mode);
}

#if defined (_WIN32) || defined(DJGPP) || defined(__CYGWIN__) || defined(__human68k__) || defined(__VMS)
static struct pipe_list {
    rb_io_t *fptr;
    struct pipe_list *next;
} *pipe_list;

static void
pipe_add_fptr(rb_io_t *fptr)
{
    struct pipe_list *list;

    list = ALLOC(struct pipe_list);
    list->fptr = fptr;
    list->next = pipe_list;
    pipe_list = list;
}

static void
pipe_del_fptr(rb_io_t *fptr)
{
    struct pipe_list *list = pipe_list;
    struct pipe_list *tmp;

    if (list->fptr == fptr) {
      pipe_list = list->next;
      free(list);
      return;
    }

    while (list->next) {
      if (list->next->fptr == fptr) {
          tmp = list->next;
          list->next = list->next->next;
          free(tmp);
          return;
      }
      list = list->next;
    }
}

static void
pipe_atexit (mrb_state *mrb)
{
    struct pipe_list *list = pipe_list;
    struct pipe_list *tmp;

    while (list) {
      tmp = list->next;
      rb_io_fptr_finalize(mrb, list->fptr);
      list = tmp;
    }
}

static void
pipe_finalize(mrb_state *mrb, rb_io_t *fptr, int noraise)
{
#if !defined (__CYGWIN__) && !defined(_WIN32)
    extern mrb_value rb_last_status;
    int status;
    if (fptr->f) {
      status = pclose(fptr->f);
    }
    if (fptr->f2) {
      status = pclose(fptr->f2);
    }
    fptr->f = fptr->f2 = 0;
#if defined DJGPP
    status <<= 8;
#endif
    rb_last_status = mrb_fixnum_value(status);
#else
    fptr_finalize(mrb, fptr, noraise);
#endif
    pipe_del_fptr(fptr);
}
#endif

void
rb_io_synchronized(rb_io_t *fptr)
{
    fptr->mode |= FMODE_SYNC;
}

void
rb_io_unbuffered(rb_io_t *fptr)
{
    rb_io_synchronized(fptr);
}

#if 0
static mrb_value
pipe_open(mrb_state *mrb, mrb_value pstr, const char *pname, const char *mode)
{
    int modef = rb_io_mode_flags(mode);
    rb_io_t *fptr;
#if defined(DJGPP) || defined(__human68k__) || defined(__VMS)
    FILE *f;
#else
    int pid;
#ifdef _WIN32
    FILE *fpr, *fpw;
#else
    int pr[2], pw[2];
#endif
#endif
    volatile int doexec;

    //if (!pname) pname = StringValueCStr(pstr);
    if (!pname) pname = rb_string_value_cstr(mrb, &pstr);
    doexec = (strcmp("-", pname) != 0);

#if defined(DJGPP) || defined(__human68k__) || defined(__VMS) || defined(_WIN32)
    if (!doexec) {
      mrb_raise(mrb, mrb->eNotImpError_class,
             "fork() function is unimplemented on this machine");
    }
#endif

#if defined(DJGPP) || defined(__human68k__) || defined(__VMS)
    f = popen(pname, mode);

    if (!f) rb_sys_fail(pname);
    else {
      mrb_value port = io_alloc(mrb, mrb->io_class);

      MakeOpenFile(mrb, port, fptr);
      fptr->finalize = pipe_finalize;
      fptr->mode = modef;

      pipe_add_fptr(fptr);
      if (modef & FMODE_READABLE) fptr->f  = f;
      if (modef & FMODE_WRITABLE) {
          if (fptr->f) fptr->f2 = f;
          else fptr->f = f;
          rb_io_synchronized(fptr);
      }
      return (mrb_value)port;
    }
#else
#ifdef _WIN32
retry:
    pid = pipe_exec(pname, rb_io_mode_modenum(mrb, mode), &fpr, &fpw);
    if (pid == -1) {            /* exec failed */
      if (errno == EAGAIN) {
          rb_thread_sleep(1);
          goto retry;
      }
      rb_sys_fail(pname);
    }
    else {
        mrb_value port = io_alloc(mrb, mrb->io_class);

      MakeOpenFile(mrb, port, fptr);
      fptr->mode = modef;
      fptr->mode |= FMODE_SYNC;
      fptr->pid = pid;

      if (modef & FMODE_READABLE) {
          fptr->f = fpr;
      }
      if (modef & FMODE_WRITABLE) {
          if (fptr->f) fptr->f2 = fpw;
          else fptr->f = fpw;
      }
      fptr->finalize = pipe_finalize;
      pipe_add_fptr(fptr);
      return (mrb_value)port;
    }
#else
    if (((modef & FMODE_READABLE) && pipe(pr) == -1) ||
      ((modef & FMODE_WRITABLE) && pipe(pw) == -1))
      rb_sys_fail(pname);

    if (!doexec) {
      fflush(stdin);            /* is it really needed? */
      fflush(stdout);
      fflush(stderr);
    }

  retry:
    //rb_thread_stop_timer();
    switch ((pid = fork())) {
      case 0:                  /* child */
      if (modef & FMODE_READABLE) {
          close(pr[0]);
          if (pr[1] != 1) {
            dup2(pr[1], 1);
            close(pr[1]);
          }
      }
      if (modef & FMODE_WRITABLE) {
          close(pw[1]);
          if (pw[0] != 0) {
            dup2(pw[0], 0);
            close(pw[0]);
          }
      }

      if (doexec) {
          int fd;

          for (fd = 3; fd < NOFILE; fd++)
            close(fd);
          rb_proc_exec(pname);
          fprintf(stderr, "%s:%d: command not found: %s\n",
                ruby_sourcefile, ruby_sourceline, pname);
          _exit(127);
      }
      rb_thread_start_timer();
      rb_io_synchronized(RFILE(orig_stdout)->fptr);
      rb_io_synchronized(RFILE(orig_stderr)->fptr);
      return mrb_nil_value();

      case -1:                  /* fork failed */
      rb_thread_start_timer();
      if (errno == EAGAIN) {
          rb_thread_sleep(1);
          goto retry;
      }
      else {
          int e = errno;
          if ((modef & FMODE_READABLE)) {
            close(pr[0]);
            close(pr[1]);
          }
          if ((modef & FMODE_WRITABLE)) {
            close(pw[0]);
            close(pw[1]);
          }
          errno = e;
          rb_sys_fail(pname);
      }
      break;

      default:                  /* parent */
      rb_thread_start_timer();
      if (pid < 0) rb_sys_fail(pname);
      else {
          mrb_value port = io_alloc(mrb, mrb->io_class);

          MakeOpenFile(mrb, port, fptr);
          fptr->mode = modef;
          fptr->mode |= FMODE_SYNC;
          fptr->pid = pid;

          if (modef & FMODE_READABLE) {
            close(pr[1]);
            fptr->f  = rb_fdopen(pr[0], "r");
          }
          if (modef & FMODE_WRITABLE) {
            FILE *f = rb_fdopen(pw[1], "w");

            close(pw[0]);
            if (fptr->f) fptr->f2 = f;
            else fptr->f = f;
          }
#if defined (__CYGWIN__)
          fptr->finalize = pipe_finalize;
          pipe_add_fptr(fptr);
#endif
          return port;
      }
    }
#endif
#endif
}
#endif

#if 0
/*
 *  call-seq:
 *     IO.popen(cmd_string, mode="r" )               => io
 *     IO.popen(cmd_string, mode="r" ) {|io| block } => obj
 *
 *  Runs the specified command string as a subprocess; the subprocess's
 *  standard input and output will be connected to the returned
 *  <code>IO</code> object. If <i>cmd_string</i> starts with a
 *  ``<code>-</code>'', then a new instance of Ruby is started as the
 *  subprocess. The default mode for the new file object is ``r'', but
 *  <i>mode</i> may be set to any of the modes listed in the description
 *  for class IO.
 *
 *  If a block is given, Ruby will run the command as a child connected
 *  to Ruby with a pipe. Ruby's end of the pipe will be passed as a
 *  parameter to the block.
 *  At the end of block, Ruby close the pipe and sets <code>$?</code>.
 *  In this case <code>IO::popen</code> returns
 *  the value of the block.
 *
 *  If a block is given with a <i>cmd_string</i> of ``<code>-</code>'',
 *  the block will be run in two separate processes: once in the parent,
 *  and once in a child. The parent process will be passed the pipe
 *  object as a parameter to the block, the child version of the block
 *  will be passed <code>nil</code>, and the child's standard in and
 *  standard out will be connected to the parent through the pipe. Not
 *  available on all platforms.
 *
 *     f = IO.popen("uname")
 *     p f.readlines
 *     puts "Parent is #{Process.pid}"
 *     IO.popen ("date") { |f| puts f.gets }
 *     IO.popen("-") {|f| $stderr.puts "#{Process.pid} is here, f is #{f}"}
 *     p $?
 *
 *  <em>produces:</em>
 *
 *     ["Linux\n"]
 *     Parent is 26166
 *     Wed Apr  9 08:53:52 CDT 2003
 *     26169 is here, f is
 *     26166 is here, f is #<IO:0x401b3d44>
 *     #<Process::Status: pid=26166,exited(0)>
 */

static mrb_value
rb_io_s_popen(int argc, mrb_value *argv, mrb_value klass)
{
    const char *mode;
    mrb_value pname, pmode, port;

    if (rb_scan_args(argc, argv, "11", &pname, &pmode) == 1) {
      mode = "r";
    }
    else if (FIXNUM_P(pmode)) {
      mode = rb_io_modenum_mode(mrb, mrb_fixnum_number(pmode));
    }
    else {
      //mode = rb_io_flags_mode(rb_io_mode_flags(StringValueCStr(pmode)));
      mode = rb_io_flags_mode(rb_io_mode_flags(rb_string_value_cstr(mrb, &pmode)));
    }
    //SafeStringValue(pname);
    mrb_string_value(mrb, &pname);
    port = pipe_open(pname, 0, mode);
    if (mrb_nil_p(port)) {
      /* child */
      if (rb_block_given_p()) {
          rb_yield(mrb_nil_value());
          fflush(stdout);
          fflush(stderr);
          _exit(0);
      }
      return mrb_nil_value();
    }
    RBASIC(port)->klass = klass;
    if (rb_block_given_p()) {
      return rb_ensure(rb_yield, port, io_close, port);
    }
    return port;
}
#endif

mrb_value
rb_open_file(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
    mrb_value fname, vmode, perm;
    const char *path, *mode;
    int flags;
    unsigned int fmode;

    //rb_scan_args(argc, argv, "12", &fname, &vmode, &perm);
    fname = argv[0];
    vmode = argv[1];
    perm  = argv[2];
    //SafeStringValue(fname);
    mrb_string_value(mrb, &fname);

    //path = StringValueCStr(fname);
    path = rb_string_value_cstr(mrb, &fname);
    if (FIXNUM_P(vmode) || !mrb_nil_p(perm)) {
      if (FIXNUM_P(vmode)) {
          flags = mrb_fixnum_number(vmode);
      }
      else {
          //SafeStringValue(vmode);
          mrb_string_value(mrb, &vmode);
          flags = rb_io_mode_modenum(mrb, RSTRING(vmode)->buf);
      }
      fmode = mrb_nil_p(perm) ? 0666 :  mrb_fixnum_number(perm);

      rb_file_sysopen_internal(mrb, io, (char*)path, flags, fmode);
    }
    else {
      //mode = mrb_nil_p(vmode) ? "r" : StringValueCStr(vmode);
      mode = mrb_nil_p(vmode) ? "r" : rb_string_value_cstr(mrb, &vmode);
      rb_file_open_internal(mrb, io, path, mode);
    }
    return io;
}

/* 15.2.20.4.1  */
/*
 *  call-seq:
 *     IO.open(fd, mode_string="r" )               => io
 *     IO.open(fd, mode_string="r" ) {|io| block } => obj
 *
 *  With no associated block, <code>open</code> is a synonym for
 *  <code>IO::new</code>. If the optional code block is given, it will
 *  be passed <i>io</i> as an argument, and the IO object will
 *  automatically be closed when the block terminates. In this instance,
 *  <code>IO::open</code> returns the value of the block.
 *
 */

static mrb_value
mrb_io_s_open(mrb_state *mrb, /*int argc, mrb_value *argv, */mrb_value klass)
{
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);
#if 0 /* original */
  mrb_value io = rb_class_new_instance(mrb, argc, argv, mrb_class_get(mrb, klass));
#else
  //SCLASS mrb_value io = mrb_obj_value(mrb_obj_alloc(mrb, mrb_class_get(mrb, klass)->tt, mrb_class_get(mrb, klass)));
  mrb_value io = mrb_obj_value(mrb_obj_alloc(mrb, MRB_TT_FILE, mrb_class_ptr(klass)));
  rb_obj_call_init(mrb, io, argc, argv);
#endif

  if (rb_block_given_p()) {
    //return rb_ensure(rb_yield, io, io_close, io);
    return mrb_fixnum_value(rb_ensure(int_rb_yield, &io, io_close, io));
  }

  return io;
}

#if 0
/*
 *  call-seq:
 *     IO.sysopen(path, [mode, [perm]])  => fixnum
 *
 *  Opens the given path, returning the underlying file descriptor as a
 *  <code>Fixnum</code>.
 *
 *     IO.sysopen("testfile")   #=> 3
 *
 */

static mrb_value
rb_io_s_sysopen(int argc, mrb_value *argv)
{
    mrb_value fname, vmode, perm;
    int flags, fd;
    unsigned int fmode;
    char *path;

    rb_scan_args(argc, argv, "12", &fname, &vmode, &perm);
    //SafeStringValue(fname);
    mrb_string_value(mrb, &fname);

    if (mrb_nil_p(vmode)) flags = O_RDONLY;
    else if (FIXNUM_P(vmode)) flags = mrb_fixnum_number(vmode);
    else {
      //SafeStringValue(vmode);
      mrb_string_value(mrb, &vmode);
      flags = rb_io_mode_modenum(mrb, RSTRING(vmode)->buf);
    }
    if (mrb_nil_p(perm)) fmode = 0666;
    else                 fmode = mrb_fixnum_number(perm);

    //path = ALLOCA_N(char, strlen(RSTRING(fname)->buf)+1);
    path = mrb_malloc(mrb, strlen(RSTRING(fname)->buf)+1);
    strcpy(path, RSTRING(fname)->buf);
    fd = rb_sysopen(path, flags, fmode);
    return mrb_fixnum_value(fd);
}

/*
 *  call-seq:
 *     open(path [, mode [, perm]] )                => io or nil
 *     open(path [, mode [, perm]] ) {|io| block }  => obj
 *
 *  Creates an <code>IO</code> object connected to the given stream,
 *  file, or subprocess.
 *
 *  If <i>path</i> does not start with a pipe character
 *  (``<code>|</code>''), treat it as the name of a file to open using
 *  the specified mode (defaulting to ``<code>r</code>''). (See the table
 *  of valid modes on page 331.) If a file is being created, its initial
 *  permissions may be set using the integer third parameter.
 *
 *  If a block is specified, it will be invoked with the
 *  <code>File</code> object as a parameter, and the file will be
 *  automatically closed when the block terminates. The call
 *  returns the value of the block.
 *
 *  If <i>path</i> starts with a pipe character, a subprocess is
 *  created, connected to the caller by a pair of pipes. The returned
 *  <code>IO</code> object may be used to write to the standard input
 *  and read from the standard output of this subprocess. If the command
 *  following the ``<code>|</code>'' is a single minus sign, Ruby forks,
 *  and this subprocess is connected to the parent. In the subprocess,
 *  the <code>open</code> call returns <code>nil</code>. If the command
 *  is not ``<code>-</code>'', the subprocess runs the command. If a
 *  block is associated with an <code>open("|-")</code> call, that block
 *  will be run twice---once in the parent and once in the child. The
 *  block parameter will be an <code>IO</code> object in the parent and
 *  <code>nil</code> in the child. The parent's <code>IO</code> object
 *  will be connected to the child's <code>$stdin</code> and
 *  <code>$stdout</code>. The subprocess will be terminated at the end
 *  of the block.
 *
 *     open("testfile") do |f|
 *       print f.gets
 *     end
 *
 *  <em>produces:</em>
 *
 *     This is line one
 *
 *  Open a subprocess and read its output:
 *
 *     cmd = open("|date")
 *     print cmd.gets
 *     cmd.close
 *
 *  <em>produces:</em>
 *
 *     Wed Apr  9 08:56:31 CDT 2003
 *
 *  Open a subprocess running the same Ruby program:
 *
 *     f = open("|-", "w+")
 *     if f == nil
 *       puts "in Child"
 *       exit
 *     else
 *       puts "Got: #{f.gets}"
 *     end
 *
 *  <em>produces:</em>
 *
 *     Got: in Child
 *
 *  Open a subprocess using a block to receive the I/O object:
 *
 *     open("|-") do |f|
 *       if f == nil
 *         puts "in Child"
 *       else
 *         puts "Got: #{f.gets}"
 *       end
 *     end
 *
 *  <em>produces:</em>
 *
 *     Got: in Child
 */

static mrb_value
rb_f_open(mrb_state *mrb, int argc, mrb_value *argv)
{
    if (argc >= 1) {
      char *str = StringValuePtr(argv[0]);

      if (str[0] == '|') {
          mrb_value tmp = mrb_str_new(mrb, str+1, RSTRING(argv[0])->len-1);
          OBJ_INFECT(tmp, argv[0]);
          argv[0] = tmp;
          return rb_io_s_popen(argc, argv, rb_cIO);
      }
    }
    return rb_io_s_open(argc, argv, mrb->file_class);
}
#endif

#if 0
static mrb_value
rb_io_open(mrb_state *mrb, char *fname, char *mode)
{
    if (fname[0] == '|') {
      return pipe_open(0, fname+1, mode);
    }
    else {
      return rb_file_open(mrb, fname, mode);
    }
}
#endif

mrb_value
rb_io_get_io(mrb_state *mrb, mrb_value io)
{
    return mrb_convert_type(mrb, io, MRB_TT_FILE, "IO", "to_io");
}

#if 0
static mrb_value
rb_io_check_io(mrb_value io)
{
    return rb_check_convert_type(io, MRB_TT_FILE, "IO", "to_io");
}

static const char*
rb_io_mode_string(rb_io_t *fptr)
{
    switch (fptr->mode & FMODE_READWRITE) {
      case FMODE_READABLE:
      default:
      return "r";
      case FMODE_WRITABLE:
      return "w";
      case FMODE_READWRITE:
      return "r+";
    }
}
#endif


/* 15.2.20.5.10 */
/* :nodoc: */
static mrb_value
mrb_io_init_copy(mrb_state *mrb, mrb_value dest)
{
  rb_io_t *fptr, *orig;
  int fd;
  const char *mode;
  mrb_value io;

  mrb_get_args(mrb, "o", &io);

  io = rb_io_get_io(mrb, io);
  if (mrb_obj_equal(mrb, dest, io)) return dest;
  GetOpenFile(mrb, io, orig);
  MakeOpenFile(mrb, dest, fptr);

  if (orig->f2) {
    io_fflush(mrb, orig->f2, orig);
    fseeko(orig->f, 0L, SEEK_CUR);
  }
  else if (orig->mode & FMODE_WRITABLE) {
    io_fflush(mrb, orig->f, orig);
  }
  else {
    fseeko(orig->f, 0L, SEEK_CUR);
  }

  /* copy rb_io_t structure */
  fptr->mode = orig->mode;
  fptr->pid = orig->pid;
  fptr->lineno = orig->lineno;
  if (orig->path) fptr->path = strdup(orig->path);
  fptr->finalize = orig->finalize;

  switch (fptr->mode & FMODE_READWRITE) {
    case FMODE_READABLE:
    default:
    mode = "r"; break;
    case FMODE_WRITABLE:
    mode = "w"; break;
    case FMODE_READWRITE:
    if (orig->f2) mode = "r";
    else          mode = "r+";
    break;
  }
  fd = ruby_dup(fileno(orig->f));
  fptr->f = rb_fdopen(fd, mode);
  fseeko(fptr->f, ftello(orig->f), SEEK_SET);
  if (orig->f2) {
    if (fileno(orig->f) != fileno(orig->f2)) {
        fd = ruby_dup(fileno(orig->f2));
    }
    fptr->f2 = rb_fdopen(fd, "w");
    fseeko(fptr->f2, ftello(orig->f2), SEEK_SET);
  }
  if (fptr->mode & FMODE_BINMODE) {
    rb_io_binmode(mrb, dest);
  }

  return dest;
}

#if 0
/*
 *  call-seq:
 *     ios.printf(format_string [, obj, ...] )   => nil
 *
 *  Formats and writes to <em>ios</em>, converting parameters under
 *  control of the format string. See <code>Kernel#sprintf</code>
 *  for details.
 */

mrb_value
rb_io_printf(mrb_state *mrb, int argc, mrb_value argv[], mrb_value out)
{
    rb_io_write(mrb, out, rb_f_sprintf(argc, argv));
    return mrb_nil_value();
}

/*
 *  call-seq:
 *     printf(io, string [, obj ... ] )    => nil
 *     printf(string [, obj ... ] )        => nil
 *
 *  Equivalent to:
 *     io.write(sprintf(string, obj, ...)
 *  or
 *     $stdout.write(sprintf(string, obj, ...)
 */

static mrb_value
rb_f_printf(mrb_state *mrb, int argc, mrb_value *argv)
{
    mrb_value out;

    if (argc == 0) return mrb_nil_value();
    if (mrb_type(argv[0]) == T_STRING) {
      out = rb_stdout;
    }
    else {
      out = argv[0];
      argv++;
      argc--;
    }
    rb_io_write(mrb, out, rb_f_sprintf(argc, argv));

    return mrb_nil_value();
}
#endif
/* 15.2.20.5.11 */
/*
 *  call-seq:
 *     ios.print()             => nil
 *     ios.print(obj, ...)     => nil
 *
 *  Writes the given object(s) to <em>ios</em>. The stream must be
 *  opened for writing. If the output record separator (<code>$\\</code>)
 *  is not <code>nil</code>, it will be appended to the output. If no
 *  arguments are given, prints <code>$_</code>. Objects that aren't
 *  strings will be converted by calling their <code>to_s</code> method.
 *  With no argument, prints the contents of the variable <code>$_</code>.
 *  Returns <code>nil</code>.
 *
 *     $stdout.print("This is ", 100, " percent.\n")
 *
 *  <em>produces:</em>
 *
 *     This is 100 percent.
 */

mrb_value
mrb_io_print(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value out)
{
  int i;
  mrb_value line;
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);

  /* if no argument given, print `$_' */
  if (argc == 0) {
    argc = 1;
    line = rb_lastline_get(mrb);
    argv[0] = line;
  }
  for (i=0; i<argc; i++) {
    if (mrb_nil_p(argv[i])) {
        rb_io_write(mrb, out, mrb_str_new2(mrb, "nil"));
    }
    else {
        rb_io_write(mrb, out, argv[i]);
    }
  }

  return mrb_nil_value();
}

#if 0
/*
 *  call-seq:
 *     print(obj, ...)    => nil
 *
 *  Prints each object in turn to <code>$stdout</code>. If the output
 *  field separator (<code>$,</code>) is not +nil+, its
 *  contents will appear between each field. If the output record
 *  separator (<code>$\\</code>) is not +nil+, it will be
 *  appended to the output. If no arguments are given, prints
 *  <code>$_</code>. Objects that aren't strings will be converted by
 *  calling their <code>to_s</code> method.
 *
 *     print "cat", [1,2,3], 99, "\n"
 *     $, = ", "
 *     $\ = "\n"
 *     print "cat", [1,2,3], 99
 *
 *  <em>produces:</em>
 *
 *     cat12399
 *     cat, 1, 2, 3, 99
 */

static mrb_value
rb_f_print(int argc, mrb_value *argv)
{
    rb_io_print(argc, argv, rb_stdout);
    return mrb_nil_value();
}
#endif

/* 15.2.20.5.12 */
/*
 *  call-seq:
 *     ios.putc(obj)    => obj
 *
 *  If <i>obj</i> is <code>Numeric</code>, write the character whose
 *  code is <i>obj</i>, otherwise write the first character of the
 *  string representation of  <i>obj</i> to <em>ios</em>.
 *
 *     $stdout.putc "A"
 *     $stdout.putc 65
 *
 *  <em>produces:</em>
 *
 *     AA
 */

static mrb_value
mrb_io_putc(mrb_state *mrb, mrb_value io)
{
  char c;
  mrb_value ch;

  mrb_get_args(mrb, "o", &ch);
  c = NUM2CHR(ch);

  rb_io_write(mrb, io, mrb_str_new(mrb, &c, 1));
  return ch;
}

#if 0
/*
 *  call-seq:
 *     putc(int)   => int
 *
 *  Equivalent to:
 *
 *    $stdout.putc(int)
 */

static mrb_value
rb_f_putc(mrb_value recv, mrb_value ch)
{
    return rb_io_putc(rb_stdout, ch);
}
#endif

static mrb_value
io_puts_ary(mrb_state *mrb, mrb_value ary, mrb_value out, int recur)
{
    mrb_value tmp;
    long i;

    for (i=0; i<RARRAY(ary)->len; i++) {
      tmp = RARRAY(ary)->buf[i];
      if (recur /*rb_inspecting_p(tmp)*/) {
          tmp = mrb_str_new2(mrb, "[...]");
          recur = 0;
      }
      io_puts(mrb, 1, &tmp, out);
    }
    return mrb_nil_value();
}

/* 15.2.20.5.13 */
/*
 *  call-seq:
 *     ios.puts(obj, ...)    => nil
 *
 *  Writes the given objects to <em>ios</em> as with
 *  <code>IO#print</code>. Writes a record separator (typically a
 *  newline) after any that do not already end with a newline sequence.
 *  If called with an array argument, writes each element on a new line.
 *  If called without arguments, outputs a single record separator.
 *
 *     $stdout.puts("this", "is", "a", "test")
 *
 *  <em>produces:</em>
 *
 *     this
 *     is
 *     a
 *     test
 */

mrb_value
io_puts(mrb_state *mrb, int argc, mrb_value *argv, mrb_value out)
{
  int i;
  mrb_value line;

  /* if no argument given, print newline. */
  if (argc == 0) {
    rb_io_write(mrb, out, rb_default_rs);
    return mrb_nil_value();
  }
  for (i=0; i<argc; i++) {
    if (mrb_nil_p(argv[i])) {
        line = mrb_str_new2(mrb, "nil");
    }
    else {
      line = mrb_check_array_type(mrb, argv[i]);
        if (!mrb_nil_p(line)) {
          //1_8_7 rb_protect_inspect(io_puts_ary, line, out);
          rb_exec_recursive(mrb, io_puts_ary, line, &out);
          continue;
        }
        line = mrb_obj_as_string(mrb, argv[i]);
    }
    rb_io_write(mrb, out, line);
    if (RSTRING(line)->len == 0 ||
	RSTRING(line)->buf[RSTRING(line)->len-1] != '\n') {
      rb_io_write(mrb, out, rb_default_rs);
    }
  }

  return mrb_nil_value();
}

mrb_value
mrb_io_puts(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value out)
{
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);

  return io_puts(mrb, argc, argv, out);
}

#if 0
/*
 *  call-seq:
 *     puts(obj, ...)    => nil
 *
 *  Equivalent to
 *
 *      $stdout.puts(obj, ...)
 */

static mrb_value
rb_f_puts(mrb_state *mrb, int argc, mrb_value *argv)
{
    io_puts(mrb, argc, argv, rb_stdout);
    return mrb_nil_value();
}
#endif

void
rb_p(mrb_state *mrb, mrb_value obj)                  /* for debug print within C code */
{
    rb_io_write(mrb, rb_stdout, mrb_obj_as_string(mrb, mrb_inspect(mrb, obj)));
    rb_io_write(mrb, rb_stdout, rb_default_rs);
}

#if 0
/*
 *  call-seq:
 *     p(obj, ...)    => nil
 *
 *  For each object, directly writes
 *  _obj_.+inspect+ followed by the current output
 *  record separator to the program's standard output.
 *
 *     S = Struct.new(:name, :state)
 *     s = S['dave', 'TX']
 *     p s
 *
 *  <em>produces:</em>
 *
 *     #<S name="dave", state="TX">
 */

static mrb_value
rb_f_p(int argc, mrb_value *argv)
{
    int i;

    for (i=0; i<argc; i++) {
      rb_p(argv[i]);
    }
    if (mrb_type(rb_stdout) == MRB_TT_FILE) {
      mrb_io_flush(mrb, rb_stdout);
    }
    return mrb_nil_value();
}
#endif

#if 0
/*
 *  call-seq:
 *     obj.display(port=$>)    => nil
 *
 *  Prints <i>obj</i> on the given port (default <code>$></code>).
 *  Equivalent to:
 *
 *     def display(port=$>)
 *       port.write self
 *     end
 *
 *  For example:
 *
 *     1.display
 *     "cat".display
 *     [ 4, 5, 6 ].display
 *     puts
 *
 *  <em>produces:</em>
 *
 *     1cat456
 */

static mrb_value
rb_obj_display(mrb_state *mrb, int argc, mrb_value *argv, mrb_value self)
{
    mrb_value out;

    if (rb_scan_args(argc, argv, "01", &out) == 0) {
      out = rb_stdout;
    }

    rb_io_write(mrb, out, self);

    return mrb_nil_value();
}
#endif

void
rb_write_error2(mrb_state *mrb, const char *mesg, long len)
{
    rb_io_write(mrb, rb_stderr, mrb_str_new(mrb, mesg, len));
}

void
rb_write_error(mrb_state *mrb, const char *mesg)
{
    rb_write_error2(mrb, mesg, strlen(mesg));
}

static void
must_respond_to(mrb_state *mrb, mrb_sym mid, mrb_value val, mrb_sym id)
{
    if (!mrb_respond_to(mrb, val, mid)) {
      mrb_raise(mrb, mrb->eTypeError_class, "%s must have %s method, %s given",
             mrb_sym2name(mrb, id), mrb_sym2name(mrb, mid),//rb_id2name(id), rb_id2name(mid),
             mrb_obj_classname(mrb, val));
    }
}

static void
stdout_setter(mrb_state *mrb, mrb_value val, mrb_sym id, mrb_value *variable)
{
    must_respond_to(mrb, mrb_intern(mrb, "write"), val, id);
    *variable = val;
}

#if 0
static void
defout_setter(mrb_state *mrb, mrb_value val, mrb_sym id, mrb_value *variable)
{
    stdout_setter(mrb, val, id, variable);
    rb_warn("$defout is obsolete; use $stdout instead");
}

static void
deferr_setter(mrb_state *mrb, mrb_value val, mrb_sym id, mrb_value *variable)
{
    stdout_setter(mrb, val, id, variable);
    rb_warn("$deferr is obsolete; use $stderr instead");
}
#endif

static mrb_value
prep_stdio(mrb_state *mrb, FILE *f, int mode, struct RClass *klass, const char *path)
{
    rb_io_t *fp;
    mrb_value io = io_alloc(mrb, klass);

    MakeOpenFile(mrb, io, fp);
#ifdef __CYGWIN__
    if (!isatty(fileno(f))) {
      mode |= O_BINARY;
      setmode(fileno(f), O_BINARY);
    }
#endif
    fp->f = f;
    fp->mode = mode;

    return io;
}

#if 0
static void
prep_path(mrb_state *mrb, mrb_value io, char *path)
{
    rb_io_t *fptr;

    GetOpenFile(mrb, io, fptr);
    if (fptr->path) rb_bug("illegal prep_path() call");
    fptr->path = strdup(path);
}
#endif

/* 15.2.20.5.21 (x)*/
/*
 *  call-seq:
 *     IO.new(fd, mode)   => io
 *
 *  Returns a new <code>IO</code> object (a stream) for the given
 *  integer file descriptor and mode string. See also
 *  <code>IO#fileno</code> and <code>IO::for_fd</code>.
 *
 *     a = IO.new(2,"w")      # '2' is standard error
 *     $stderr.puts "Hello"
 *     a.puts "World"
 *
 *  <em>produces:</em>
 *
 *     Hello
 *     World
 */

mrb_value
rb_io_initialize(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
    mrb_value fnum, mode;
    rb_io_t *fp;
    int fd, flags;

    //rb_secure(4);
    //rb_scan_args(argc, argv, "11", &fnum, &mode);
    fnum = argv[0];
    mode = argv[1];
    fd = mrb_fixnum_number(fnum);
    if (argc == 2) {
      if (FIXNUM_P(mode)) {
          flags = mrb_fixnum_number(mode);
      }
      else {
          //SafeStringValue(mode);
          mrb_string_value(mrb, &mode);
          //flags = rb_io_mode_modenum(mrb, StringValueCStr(mode));
          flags = rb_io_mode_modenum(mrb, rb_string_value_cstr(mrb, &mode));
      }
    }
    else {
#if defined(HAVE_FCNTL) && defined(F_GETFL)
      flags = fcntl(fd, F_GETFL);
      if (flags == -1) rb_sys_fail(0);
#else
      flags = O_RDONLY;
#endif
    }
    MakeOpenFile(mrb, io, fp);
    fp->mode = rb_io_modenum_flags(flags);
    fp->f = rb_fdopen(fd, rb_io_modenum_mode(mrb, flags));

    return io;
}

mrb_value
mrb_io_initialize(mrb_state *mrb, mrb_value io)
{
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);
  return rb_io_initialize(mrb, argc, argv, io);
}

#if 0
/*
 *  call-seq:
 *     File.new(filename, mode="r")            => file
 *     File.new(filename [, mode [, perm]])    => file
 *

 *  Opens the file named by _filename_ according to
 *  _mode_ (default is ``r'') and returns a new
 *  <code>File</code> object. See the description of class +IO+ for
 *  a description of _mode_. The file mode may optionally be
 *  specified as a +Fixnum+ by _or_-ing together the
 *  flags (O_RDONLY etc, again described under +IO+). Optional
 *  permission bits may be given in _perm_. These mode and permission
 *  bits are platform dependent; on Unix systems, see
 *  <code>open(2)</code> for details.
 *
 *     f = File.new("testfile", "r")
 *     f = File.new("newfile",  "w+")
 *     f = File.new("newfile", File::CREAT|File::TRUNC|File::RDWR, 0644)
 */

static mrb_value
rb_file_initialize(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
    if (RFILE(io)->fptr) {
      mrb_raise(mrb, mrb->eRuntimeError_class, "reinitializing File");
    }
    if (0 < argc && argc < 3) {
      mrb_value fd = rb_check_convert_type(argv[0], T_FIXNUM, "Fixnum", "to_int");

      if (!mrb_nil_p(fd)) {
          argv[0] = fd;
          return rb_io_initialize(mrb, argc, argv, io);
      }
    }
    rb_open_file(argc, argv, io);

    return io;
}

/*
 *  call-seq:
 *     IO.new(fd, mode_string)   => io
 *
 *  Returns a new <code>IO</code> object (a stream) for the given
 *  integer file descriptor and mode string. See also
 *  <code>IO#fileno</code> and <code>IO::for_fd</code>.
 *
 *     a = IO.new(2,"w")      # '2' is standard error
 *     $stderr.puts "Hello"
 *     a.puts "World"
 *
 *  <em>produces:</em>
 *
 *     Hello
 *     World
 */

static mrb_value
rb_io_s_new(int argc, mrb_value *argv, klass)
    int argc;
    mrb_value *argv;
    mrb_value klass;
{
    if (rb_block_given_p()) {
      const char *cname = rb_class2name(klass);

      rb_warn("%s::new() does not take block; use %s::open() instead",
            cname, cname);
    }
    return rb_class_new_instance(argc, argv, klass);
}

/*
 *  call-seq:
 *     IO.for_fd(fd, mode)    => io
 *
 *  Synonym for <code>IO::new</code>.
 *
 */

static mrb_value
rb_io_s_for_fd(mrb_state *mrb, int argc, mrb_value *argv, mrb_value klass)
{
    mrb_value io = rb_obj_alloc(klass);
    rb_io_initialize(mrb, argc, argv, io);
    return io;
}
#endif

#if 0
static int binmode = 0;

static mrb_value
argf_forward(mrb_state *mrb, int argc, mrb_value *argv)
{
    //return rb_funcall3(current_file, ruby_frame->last_func, argc, argv);
    return mrb_funcall(mrb, current_file, ruby_frame->last_func, argc, argv);
}
#endif

#define ARGF_FORWARD(argc, argv) do {\
  if (mrb_type(current_file) != MRB_TT_FILE)\
     return argf_forward(argc, argv);\
} while (0)
#define NEXT_ARGF_FORWARD(argc, argv) do {\
     if (!next_argv(mrb)) return mrb_nil_value();\
     ARGF_FORWARD(argc, argv);\
} while (0)

#if 0
static void
argf_close(mrb_state *mrb, mrb_value file)
{
    //rb_funcall3(file, rb_intern("close"), 0, 0);
    mrb_funcall(mrb, file, "close", 0, 0);
}
#endif

#if 0
static int
next_argv(mrb_state *mrb)
{
    extern mrb_value rb_argv;
    char *fn;
    rb_io_t *fptr;
    int stdout_binmode = 0;

    if (mrb_type(rb_stdout) == MRB_TT_FILE) {
        GetOpenFile(mrb, rb_stdout, fptr);
        if (fptr->mode & FMODE_BINMODE)
            stdout_binmode = 1;
    }

    if (init_p == 0) {
      if (RARRAY(rb_argv)->len > 0) {
          next_p = 1;
      }
      else {
          next_p = -1;
      }
      init_p = 1;
      gets_lineno = 0;
    }

    if (next_p == 1) {
      next_p = 0;
      retry:
      if (RARRAY(rb_argv)->len > 0) {
          filename = mrb_ary_shift(mrb, rb_argv);
          //fn = StringValueCStr(filename);
          fn = rb_string_value_cstr(mrb, &filename);
          if (strlen(fn) == 1 && fn[0] == '-') {
            current_file = rb_stdin;
            if (ruby_inplace_mode) {
                rb_warn("Can't do inplace edit for stdio; skipping");
                goto retry;
            }
          }
          else {
            FILE *fr = rb_fopen(fn, "r");

            if (ruby_inplace_mode) {
                struct stat st;
#ifndef NO_SAFE_RENAME
                struct stat st2;
#endif
                mrb_value str;
                FILE *fw;

                if ((mrb_type(rb_stdout) == MRB_TT_FILE) && (!mrb_obj_equal(mrb, rb_stdout, orig_stdout))) {
                  rb_io_close(mrb, rb_stdout);
                }
                fstat(fileno(fr), &st);
                if (*ruby_inplace_mode) {
                  str = mrb_str_new2(mrb, fn);
#ifdef NO_LONG_FNAME
                        ruby_add_suffix(str, ruby_inplace_mode);
#else
                  rb_str_cat2(mrb, str, ruby_inplace_mode);
#endif
#ifdef NO_SAFE_RENAME
                  (void)fclose(fr);
                  (void)unlink(RSTRING(str)->buf);
                  (void)rename(fn, RSTRING(str)->buf);
                  fr = rb_fopen(RSTRING(str)->buf, "r");
#else
                  if (rename(fn, RSTRING(str)->buf) < 0) {
                      rb_warn("Can't rename %s to %s: %s, skipping file",
                            fn, RSTRING(str)->buf, strerror(errno));
                      fclose(fr);
                      goto retry;
                  }
#endif
                }
                else {
#ifdef NO_SAFE_RENAME
                  rb_fatal("Can't do inplace edit without backup");
#else
                  if (unlink(fn) < 0) {
                      rb_warn("Can't remove %s: %s, skipping file",
                            fn, strerror(errno));
                      fclose(fr);
                      goto retry;
                  }
#endif
                }
                fw = rb_fopen(fn, "w");
#ifndef NO_SAFE_RENAME
                fstat(fileno(fw), &st2);
#ifdef HAVE_FCHMOD
                fchmod(fileno(fw), st.st_mode);
#else
                chmod(fn, st.st_mode);
#endif
                if (st.st_uid!=st2.st_uid || st.st_gid!=st2.st_gid) {
                  fchown(fileno(fw), st.st_uid, st.st_gid);
                }
#endif
                rb_stdout = prep_stdio(mrb, fw, FMODE_WRITABLE, mrb->file_class, "<STDOUT>");
                prep_path(mrb, rb_stdout, fn);
                if (stdout_binmode) rb_io_binmode(mrb, rb_stdout);
            }
            current_file = prep_stdio(mrb, fr, FMODE_READABLE, mrb->file_class, "<STDIN>"); /* "<STDIN>"? */
            prep_path(mrb, current_file, fn);
          }
          if (binmode) rb_io_binmode(mrb, current_file);
      }
      else {
          next_p = 1;
          return FALSE;
      }
    }
    else if (next_p == -1) {
      current_file = rb_stdin;
      filename = mrb_str_new2(mrb, "-");
      if (ruby_inplace_mode) {
          rb_warn("Can't do inplace edit for stdio");
          rb_stdout = orig_stdout;
      }
    }
    return TRUE;
}
#endif

#if 0
static mrb_value
argf_getline(mrb_state *mrb, int argc, mrb_value *argv)
{
    mrb_value line;

  retry:
    if (!next_argv(mrb)) return mrb_nil_value();
    if (argc == 0) {
      line = rb_io_gets(mrb, current_file);
    }
    else {
      mrb_value rs;

      if (argc == 0) {
          rs = rb_rs;
      }
      else {
          rb_scan_args(argc, argv, "1", &rs);
          //if (!mrb_nil_p(rs)) StringValue(rs);
          if (!mrb_nil_p(rs)) mrb_string_value(mrb, &rs);
      }
      line = rb_io_getline(mrb, rs, current_file);
    }
    if (mrb_nil_p(line) && next_p != -1) {
      argf_close(mrb, current_file);
      next_p = 1;
      goto retry;
    }
    if (!mrb_nil_p(line)) {
      gets_lineno++;
      lineno = mrb_fixnum_value(gets_lineno);
    }
    return line;
}
#endif

/*
 *  call-seq:
 *     gets(separator=$/)    => string or nil
 *
 *  Returns (and assigns to <code>$_</code>) the next line from the list
 *  of files in +ARGV+ (or <code>$*</code>), or from standard
 *  input if no files are present on the command line. Returns
 *  +nil+ at end of file. The optional argument specifies the
 *  record separator. The separator is included with the contents of
 *  each record. A separator of +nil+ reads the entire
 *  contents, and a zero-length separator reads the input one paragraph
 *  at a time, where paragraphs are divided by two consecutive newlines.
 *  If multiple filenames are present in +ARGV+,
 *  +gets(nil)+ will read the contents one file at a time.
 *
 *     ARGV << "testfile"
 *     print while gets
 *
 *  <em>produces:</em>
 *
 *     This is line one
 *     This is line two
 *     This is line three
 *     And so on...
 *
 *  The style of programming using <code>$_</code> as an implicit
 *  parameter is gradually losing favor in the Ruby community.
 */
#if 0
static mrb_value
rb_f_gets(mrb_state *mrb, int argc, mrb_value *argv)
{
    mrb_value line;

    if (!next_argv(mrb)) return mrb_nil_value();
    if (mrb_type(current_file) != MRB_TT_FILE) {
      //line = rb_funcall3(current_file, rb_intern("gets"), argc, argv);
      line = mrb_funcall(mrb, current_file, "gets", argc, argv);
    }
    else {
      line = argf_getline(mrb, argc, argv);
    }
    rb_lastline_set(line);
    return line;
}
#endif

#if 0
mrb_value
rb_gets(mrb_state *mrb)
{
    mrb_value line;

  retry:
    if (!next_argv(mrb)) return mrb_nil_value();
    line = rb_io_gets(mrb, current_file);
    if (mrb_nil_p(line) && next_p != -1) {
      argf_close(mrb, current_file);
      next_p = 1;
      goto retry;
    }
    rb_lastline_set(line);
    if (!mrb_nil_p(line)) {
      gets_lineno++;
      lineno = mrb_fixnum_value(gets_lineno);
    }

    return line;
}
#endif

#if 0
/*
 *  call-seq:
 *     readline(separator=$/)   => string
 *
 *  Equivalent to <code>Kernel::gets</code>, except
 *  +readline+ raises +EOFError+ at end of file.
 */

static mrb_value
rb_f_readline(mrb_state *mrb, int argc, mrb_value *argv)
{
    mrb_value line;

    if (!next_argv(mrb)) rb_eof_error(mrb);
    ARGF_FORWARD(argc, argv);
    line = rb_f_gets(mrb, argc, argv);
    if (mrb_nil_p(line)) {
      rb_eof_error(mrb);
    }

    return line;
}

/*
 * obsolete
 */

static mrb_value
rb_f_getc()
{
    rb_warn("getc is obsolete; use STDIN.getc instead");
    if (mrb_type(rb_stdin) != MRB_TT_FILE) {
      //return rb_funcall3(rb_stdin, rb_intern("getc"), 0, 0);
      return mrb_funcall(mrb, rb_stdin, "getc", 0, 0);
    }
    return mrb_io_getc(mrb, rb_stdin);
}

/*
 *  call-seq:
 *     readlines(separator=$/)    => array
 *
 *  Returns an array containing the lines returned by calling
 *  <code>Kernel.gets(<i>separator</i>)</code> until the end of file.
 */

static mrb_value
rb_f_readlines(int argc, mrb_value *argv)
    int argc;
    mrb_value *argv;
{
    mrb_value line, ary;

    NEXT_ARGF_FORWARD(argc, argv);
    ary = mrb_ary_new(mrb);
    while (!mrb_nil_p(line = argf_getline(mrb, argc, argv))) {
      mrb_ary_push(mrb, ary, line);
    }

    return ary;
}

/*
 *  call-seq:
 *     `cmd`    => string
 *
 *  Returns the standard output of running _cmd_ in a subshell.
 *  The built-in syntax <code>%x{...}</code> uses
 *  this method. Sets <code>$?</code> to the process status.
 *
 *     `date`                   #=> "Wed Apr  9 08:56:30 CDT 2003\n"
 *     `ls testdir`.split[1]    #=> "main.rb"
 *     `echo oops && exit 99`   #=> "oops\n"
 *     $?.exitstatus            #=> 99
 */

static mrb_value
rb_f_backquote(mrb_state *mrb, mrb_value obj, mrb_value str)
{
    volatile mrb_value port;
    mrb_value result;
    rb_io_t *fptr;

    //SafeStringValue(str);
    mrb_string_value(mrb, &str);
    port = pipe_open(str, 0, "r");
    if (mrb_nil_p(port)) return mrb_str_new(mrb, 0,0);

    GetOpenFile(port, fptr);
    result = read_all(fptr, remain_size(fptr), mrb_nil_value());
    rb_io_close(mrb, port);

    return result;
}

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

/*
 *  call-seq:
 *     IO.select(read_array
 *               [, write_array
 *               [, error_array
 *               [, timeout]]] ) =>  array  or  nil
 *
 *  See <code>Kernel#select</code>.
 */

static mrb_value
rb_f_select(mrb_state *mrb, int argc, mrb_value *argv, mrb_value obj)
{
    mrb_value read, write, except, timeout, res, list;
    fd_set rset, wset, eset, pset;
    fd_set *rp, *wp, *ep;
    struct timeval *tp, timerec;
    rb_io_t *fptr;
    long i;
    int max = 0, n;
    int interrupt_flag = 0;
    int pending = 0;

    rb_scan_args(argc, argv, "13", &read, &write, &except, &timeout);
    if (mrb_nil_p(timeout)) {
      tp = 0;
    }
    else {
      timerec = rb_time_interval(timeout);
      tp = &timerec;
    }

    FD_ZERO(&pset);
    if (!mrb_nil_p(read)) {
      Check_Type(read, MRB_TT_ARRAY);
      rp = &rset;
      FD_ZERO(rp);
      for (i=0; i<RARRAY(read)->len; i++) {
          GetOpenFile(mrb, rb_io_get_io(mrb, RARRAY(read)->ptr[i]), fptr);
          FD_SET(fileno(fptr->f), rp);
          if (READ_DATA_PENDING(fptr->f)) { /* check for buffered data */
            pending++;
            FD_SET(fileno(fptr->f), &pset);
          }
          if (max < fileno(fptr->f)) max = fileno(fptr->f);
      }
      if (pending) {            /* no blocking if there's buffered data */
          timerec.tv_sec = timerec.tv_usec = 0;
          tp = &timerec;
      }
    }
    else
      rp = 0;

    if (!mrb_nil_p(write)) {
      Check_Type(write, MRB_TT_ARRAY);
      wp = &wset;
      FD_ZERO(wp);
      for (i=0; i<RARRAY(write)->len; i++) {
          GetOpenFile(mrb, rb_io_get_io(mrb, RARRAY(write)->ptr[i]), fptr);
          FD_SET(fileno(fptr->f), wp);
          if (max < fileno(fptr->f)) max = fileno(fptr->f);
          if (fptr->f2) {
            FD_SET(fileno(fptr->f2), wp);
            if (max < fileno(fptr->f2)) max = fileno(fptr->f2);
          }
      }
    }
    else
      wp = 0;

    if (!mrb_nil_p(except)) {
      Check_Type(except, MRB_TT_ARRAY);
      ep = &eset;
      FD_ZERO(ep);
      for (i=0; i<RARRAY(except)->len; i++) {
          GetOpenFile(mrb, rb_io_get_io(mrb, RARRAY(except)->ptr[i]), fptr);
          FD_SET(fileno(fptr->f), ep);
          if (max < fileno(fptr->f)) max = fileno(fptr->f);
          if (fptr->f2) {
            FD_SET(fileno(fptr->f2), ep);
            if (max < fileno(fptr->f2)) max = fileno(fptr->f2);
          }
      }
    }
    else {
      ep = 0;
    }

    max++;

    n = rb_thread_select(max, rp, wp, ep, tp);
    if (n < 0) {
      rb_sys_fail(0);
    }
    if (!pending && n == 0) return mrb_nil_value(); /* returns nil on timeout */

    res = mrb_ary_new_capa(mrb, 3);//rb_ary_new2(3);
    mrb_ary_push(mrb, res, rp?mrb_ary_new(mrb):mrb_ary_new_capa(mrb, 0));//:rb_ary_new2(0));
    mrb_ary_push(mrb, res, wp?mrb_ary_new(mrb):mrb_ary_new_capa(mrb, 0));//:rb_ary_new2(0));
    mrb_ary_push(mrb, res, ep?mrb_ary_new(mrb):mrb_ary_new_capa(mrb, 0));//:rb_ary_new2(0));

    if (interrupt_flag == 0) {
      if (rp) {
          list = RARRAY(res)->ptr[0];
          for (i=0; i< RARRAY(read)->len; i++) {
            GetOpenFile(mrb, rb_io_get_io(mrb, RARRAY(read)->ptr[i]), fptr);
            if (FD_ISSET(fileno(fptr->f), rp)
                || FD_ISSET(fileno(fptr->f), &pset)) {
                mrb_ary_push(mrb, list, rb_ary_entry(read, i));
            }
          }
      }

      if (wp) {
          list = RARRAY(res)->ptr[1];
          for (i=0; i< RARRAY(write)->len; i++) {
            GetOpenFile(mrb, rb_io_get_io(mrb, RARRAY(write)->ptr[i]), fptr);
            if (FD_ISSET(fileno(fptr->f), wp)) {
                mrb_ary_push(mrb, list, rb_ary_entry(write, i));
            }
            else if (fptr->f2 && FD_ISSET(fileno(fptr->f2), wp)) {
                mrb_ary_push(mrb, list, rb_ary_entry(write, i));
            }
          }
      }

      if (ep) {
          list = RARRAY(res)->ptr[2];
          for (i=0; i< RARRAY(except)->len; i++) {
            GetOpenFile(mrb, rb_io_get_io(mrb, RARRAY(except)->ptr[i]), fptr);
            if (FD_ISSET(fileno(fptr->f), ep)) {
                mrb_ary_push(mrb, list, rb_ary_entry(except, i));
            }
            else if (fptr->f2 && FD_ISSET(fileno(fptr->f2), ep)) {
                mrb_ary_push(mrb, list, rb_ary_entry(except, i));
            }
          }
      }
    }

    return res;                  /* returns an empty array on interrupt */
}

#if !defined(MSDOS) && !defined(__human68k__)
static int
io_cntl(fd, cmd, narg, io_p)
    int fd, cmd, io_p;
    long narg;
{
    int retval;

#ifdef HAVE_FCNTL
    TRAP_BEG(mrb);
# if defined(__CYGWIN__)
    retval = io_p?ioctl(fd, cmd, (void*)narg):fcntl(fd, cmd, narg);
# else
    retval = io_p?ioctl(fd, cmd, narg):fcntl(fd, cmd, narg);
# endif
    TRAP_END(mrb);
#else
    if (!io_p) {
      rb_notimplement();
    }
    TRAP_BEG(mrb);
    retval = ioctl(fd, cmd, narg);
    TRAP_END(mrb);
#endif
    return retval;
}
#endif

static mrb_value
rb_io_ctl(mrb_state *mrb, mrb_value io, mrb_value req, mrb_value arg, int io_p)
{
#if !defined(MSDOS) && !defined(__human68k__)
    int cmd = mrb_fixnum_number(req);
    rb_io_t *fptr;
    long len = 0;
    long narg = 0;
    int retval;

    rb_secure(2);

    if (mrb_nil_p(arg) || arg == mrb_false_value()) {
      narg = 0;
    }
    else if (FIXNUM_P(arg)) {
      narg = mrb_fixnum_number(arg);
    }
    else if (arg == mrb_true_value()) {
      narg = 1;
    }
    else {
      mrb_value tmp = rb_check_string_type(arg);

      if (mrb_nil_p(tmp)) {
          narg = mrb_fixnum_number(arg);
      }
      else {
          arg = tmp;
#ifdef IOCPARM_MASK
#ifndef IOCPARM_LEN
#define IOCPARM_LEN(x)  (((x) >> 16) & IOCPARM_MASK)
#endif
#endif
#ifdef IOCPARM_LEN
          len = IOCPARM_LEN(cmd);      /* on BSDish systems we're safe */
#else
          len = 256;            /* otherwise guess at what's safe */
#endif
          rb_str_modify(arg);

          if (len <= RSTRING(arg)->len) {
            len = RSTRING(arg)->len;
          }
          if (RSTRING(arg)->len < len) {
            mrb_str_resize(mrb, arg, len+1);
          }
          RSTRING(arg)->buf[len] = 17;      /* a little sanity check here */
          narg = (long)RSTRING(arg)->buf;
      }
    }
    GetOpenFile(mrb, io, fptr);
    retval = io_cntl(fileno(fptr->f), cmd, narg, io_p);
    if (retval < 0) rb_sys_fail(fptr->path);
    if (mrb_type(arg) == T_STRING && RSTRING(arg)->buf[len] != 17) {
      mrb_raise(mrb, mrb->eArgumentError_class, "return value overflowed string");
    }

    if (fptr->f2 && fileno(fptr->f) != fileno(fptr->f2)) {
      /* call on f2 too; ignore result */
      io_cntl(fileno(fptr->f2), cmd, narg, io_p);
    }

    if (!io_p && cmd == F_SETFL) {
      if (narg & O_NONBLOCK) {
        fptr->mode |= FMODE_WSPLIT_INITIALIZED;
        fptr->mode &= ~FMODE_WSPLIT;
      }
      else {
        fptr->mode &= ~(FMODE_WSPLIT_INITIALIZED|FMODE_WSPLIT);
      }
    }

    return mrb_fixnum_value(retval);
#else
    rb_notimplement();
    return mrb_nil_value();            /* not reached */
#endif
}

/*
 *  call-seq:
 *     ios.ioctl(integer_cmd, arg)    => integer
 *
 *  Provides a mechanism for issuing low-level commands to control or
 *  query I/O devices. Arguments and results are platform dependent. If
 *  <i>arg</i> is a number, its value is passed directly. If it is a
 *  string, it is interpreted as a binary sequence of bytes. On Unix
 *  platforms, see <code>ioctl(2)</code> for details. Not implemented on
 *  all platforms.
 */

static mrb_value
rb_io_ioctl(int argc, mrb_value *argv, mrb_value io)
{
    mrb_value req, arg;

    rb_scan_args(argc, argv, "11", &req, &arg);
    return rb_io_ctl(io, req, arg, 1);
}

/*
 *  call-seq:
 *     ios.fcntl(integer_cmd, arg)    => integer
 *
 *  Provides a mechanism for issuing low-level commands to control or
 *  query file-oriented I/O streams. Arguments and results are platform
 *  dependent. If <i>arg</i> is a number, its value is passed
 *  directly. If it is a string, it is interpreted as a binary sequence
 *  of bytes (<code>Array#pack</code> might be a useful way to build this
 *  string). On Unix platforms, see <code>fcntl(2)</code> for details.
 *  Not implemented on all platforms.
 */

static mrb_value
rb_io_fcntl(int argc, mrb_value *argv, mrb_value io)
{
#ifdef HAVE_FCNTL
    mrb_value req, arg;

    rb_scan_args(argc, argv, "11", &req, &arg);
    return rb_io_ctl(io, req, arg, 0);
#else
    rb_notimplement();
    return mrb_nil_value();            /* not reached */
#endif
}

/*
 *  call-seq:
 *     syscall(fixnum [, args...])   => integer
 *
 *  Calls the operating system function identified by _fixnum_,
 *  passing in the arguments, which must be either +String+
 *  objects, or +Integer+ objects that ultimately fit within
 *  a native +long+. Up to nine parameters may be passed (14
 *  on the Atari-ST). The function identified by _fixnum_ is system
 *  dependent. On some Unix systems, the numbers may be obtained from a
 *  header file called <code>syscall.h</code>.
 *
 *     syscall 4, 1, "hello\n", 6   # '4' is write(2) on our box
 *
 *  <em>produces:</em>
 *
 *     hello
 */

static mrb_value
rb_f_syscall(int argc, mrb_value *argv)
    int argc;
    mrb_value *argv;
{
#if defined(HAVE_SYSCALL) && !defined(__CHECKER__)
#ifdef atarist
    unsigned long arg[14]; /* yes, we really need that many ! */
#else
    unsigned long arg[8];
#endif
    int retval = -1;
    int i = 1;
    int items = argc - 1;

    /* This probably won't work on machines where sizeof(long) != sizeof(int)
     * or where sizeof(long) != sizeof(char*).  But such machines will
     * not likely have syscall implemented either, so who cares?
     */

    rb_secure(2);
    if (argc == 0)
      mrb_raise(mrb, mrb->eArgumentError_class, "too few arguments for syscall");
    if (argc > sizeof(arg) / sizeof(arg[0]))
      mrb_raise(mrb, mrb->eArgumentError_class, "too many arguments for syscall");
    arg[0] = mrb_fixnum_number(argv[0]); argv++;
    while (items--) {
      mrb_value v = rb_check_string_type(*argv);

      if (!mrb_nil_p(v)) {
          //StringValue(v);
          mrb_string_value(mrb, &v);
          rb_str_modify(v);
          //arg[i] = (unsigned long)StringValueCStr(v);
          arg[i] = (unsigned long)rb_string_value_cstr(mrb, &v);
      }
      else {
          arg[i] = (unsigned long)mrb_fixnum_number(*argv);
      }
      argv++;
      i++;
    }
    TRAP_BEG(mrb);
    switch (argc) {
      case 1:
      retval = syscall(arg[0]);
      break;
      case 2:
      retval = syscall(arg[0],arg[1]);
      break;
      case 3:
      retval = syscall(arg[0],arg[1],arg[2]);
      break;
      case 4:
      retval = syscall(arg[0],arg[1],arg[2],arg[3]);
      break;
      case 5:
      retval = syscall(arg[0],arg[1],arg[2],arg[3],arg[4]);
      break;
      case 6:
      retval = syscall(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5]);
      break;
      case 7:
      retval = syscall(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],arg[6]);
      break;
      case 8:
      retval = syscall(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],arg[6],
        arg[7]);
      break;
#ifdef atarist
      case 9:
      retval = syscall(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],arg[6],
        arg[7], arg[8]);
      break;
      case 10:
      retval = syscall(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],arg[6],
        arg[7], arg[8], arg[9]);
      break;
      case 11:
      retval = syscall(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],arg[6],
        arg[7], arg[8], arg[9], arg[10]);
      break;
      case 12:
      retval = syscall(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],arg[6],
        arg[7], arg[8], arg[9], arg[10], arg[11]);
      break;
      case 13:
      retval = syscall(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],arg[6],
        arg[7], arg[8], arg[9], arg[10], arg[11], arg[12]);
      break;
      case 14:
      retval = syscall(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],arg[6],
        arg[7], arg[8], arg[9], arg[10], arg[11], arg[12], arg[13]);
      break;
#endif /* atarist */
    }
    TRAP_END(mrb);
    if (retval < 0) rb_sys_fail(0);
    return mrb_fixnum_value(retval);
#else
    rb_notimplement();
    return mrb_nil_value();            /* not reached */
#endif
}
#endif //#if 0

#if 0
static mrb_value io_new_instance (mrb_value);
static mrb_value
io_new_instance(args)
    mrb_value args;
{
    return rb_class_new_instance(2, (mrb_value*)args+1, *(mrb_value*)args);
}

/*
 *  call-seq:
 *     IO.pipe -> array
 *
 *  Creates a pair of pipe endpoints (connected to each other) and
 *  returns them as a two-element array of <code>IO</code> objects:
 *  <code>[</code> <i>read_file</i>, <i>write_file</i> <code>]</code>. Not
 *  available on all platforms.
 *
 *  In the example below, the two processes close the ends of the pipe
 *  that they are not using. This is not just a cosmetic nicety. The
 *  read end of a pipe will not generate an end of file condition if
 *  there are any writers with the pipe still open. In the case of the
 *  parent process, the <code>rd.read</code> will never return if it
 *  does not first issue a <code>wr.close</code>.
 *
 *     rd, wr = IO.pipe
 *
 *     if fork
 *       wr.close
 *       puts "Parent got: <#{rd.read}>"
 *       rd.close
 *       Process.wait
 *     else
 *       rd.close
 *       puts "Sending message to parent"
 *       wr.write "Hi Dad"
 *       wr.close
 *     end
 *
 *  <em>produces:</em>
 *
 *     Sending message to parent
 *     Parent got: <Hi Dad>
 */

static mrb_value
rb_io_s_pipe(mrb_state *mrb, mrb_value klass)
{
#ifndef __human68k__
    int pipes[2], state;
    mrb_value r, w, args[3];

#ifdef _WIN32
    if (_pipe(pipes, 1024, O_BINARY) == -1)
#else
    if (pipe(pipes) == -1)
#endif
      rb_sys_fail(0);

    args[0] = klass;
    args[1] = mrb_fixnum_value(pipes[0]);
    args[2] = mrb_fixnum_value(O_RDONLY);
    r = rb_protect(io_new_instance, (mrb_value)args, &state);
    if (state) {
      close(pipes[0]);
      close(pipes[1]);
      rb_jump_tag(state);
    }
    args[1] = mrb_fixnum_value(pipes[1]);
    args[2] = mrb_fixnum_value(O_WRONLY);
    w = rb_protect(io_new_instance, (mrb_value)args, &state);
    if (state) {
      close(pipes[1]);
      if (!mrb_nil_p(r)) rb_io_close(mrb, r);
      rb_jump_tag(state);
    }
    rb_io_synchronized(RFILE(w)->fptr);

    return rb_assoc_new(r, w);
#else
    rb_notimplement();
    return mrb_nil_value();            /* not reached */
#endif
}
#endif

struct foreach_arg {
    int argc;
    mrb_value sep;
    mrb_value io;
};

#if 0
static mrb_value
io_s_foreach(mrb_state *mrb, struct foreach_arg *arg)
{
    mrb_value str;

    while (!mrb_nil_p(str = rb_io_getline(mrb, arg->sep, arg->io))) {
      rb_yield(str);
    }
    return mrb_nil_value();
}

#if 0
/*
 *  call-seq:
 *     IO.foreach(name, sep_string=$/) {|line| block }   => nil
 *
 *  Executes the block for every line in the named I/O port, where lines
 *  are separated by <em>sep_string</em>.
 *
 *     IO.foreach("testfile") {|x| print "GOT ", x }
 *
 *  <em>produces:</em>
 *
 *     GOT This is line one
 *     GOT This is line two
 *     GOT This is line three
 *     GOT And so on...
 */

static mrb_value
rb_io_s_foreach(int argc, mrb_value *argv, self)
    int argc;
    mrb_value *argv;
    mrb_value self;
{
    mrb_value fname;
    struct foreach_arg arg;

    rb_scan_args(argc, argv, "11", &fname, &arg.sep);
    //RETURN_ENUMERATOR(self, argc, argv);
    //SafeStringValue(fname);
    mrb_string_value(mrb, &fname);

    if (argc == 1) {
      arg.sep = mrb_nil_value();
    }
    else if (!mrb_nil_p(arg.sep)) {
      //StringValue(arg.sep);
      mrb_string_value(mrb, &arg.sep);
    }
    //arg.io = rb_io_open(StringValueCStr(fname), "r");
    arg.io = rb_io_open(rb_string_value_cstr(mrb, &fname), "r");
    if (mrb_nil_p(arg.io)) return mrb_nil_value();

    return rb_ensure(io_s_foreach, (mrb_value)&arg, rb_io_close, arg.io);
}

static mrb_value
io_s_readlines(arg)
    struct foreach_arg *arg;
{
    return rb_io_readlines(arg->argc, &arg->sep, arg->io);
}

/*
 *  call-seq:
 *     IO.readlines(name, sep_string=$/)   => array
 *
 *  Reads the entire file specified by <i>name</i> as individual
 *  lines, and returns those lines in an array. Lines are separated by
 *  <i>sep_string</i>.
 *
 *     a = IO.readlines("testfile")
 *     a[0]   #=> "This is line one\n"
 *
 */

static mrb_value
rb_io_s_readlines(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
    mrb_value fname;
    struct foreach_arg arg;

    rb_scan_args(argc, argv, "11", &fname, &arg.sep);
    //SafeStringValue(fname);
    mrb_string_value(mrb, &fname);

    arg.argc = argc - 1;
    //arg.io = rb_io_open(StringValueCStr(fname), "r");
    arg.io = rb_io_open(rb_string_value_cstr(mrb, &fname), "r");
    if (mrb_nil_p(arg.io)) return mrb_nil_value();
    return rb_ensure(io_s_readlines, (mrb_value)&arg, rb_io_close, arg.io);
}
#endif

static mrb_value
io_s_read(mrb_state *mrb, struct foreach_arg *arg)
{
    return io_read(mrb, arg->argc, &arg->sep, arg->io);
}
#endif

struct seek_arg {
    mrb_value io;
    mrb_value offset;
    int mode;
};

#if 0
static mrb_value
seek_before_read(struct seek_arg *arg)
{
    return rb_io_seek(arg->io, arg->offset, arg->mode);
}
#endif

#if 0
/*
 *  call-seq:
 *     IO.read(name, [length [, offset]] )   => string
 *
 *  Opens the file, optionally seeks to the given offset, then returns
 *  <i>length</i> bytes (defaulting to the rest of the file).
 *  <code>read</code> ensures the file is closed before returning.
 *
 *     IO.read("testfile")           #=> "This is line one\nThis is line two\nThis is line three\nAnd so on...\n"
 *     IO.read("testfile", 20)       #=> "This is line one\nThi"
 *     IO.read("testfile", 20, 10)   #=> "ne one\nThis is line "
 */

static mrb_value
rb_io_s_read(mrb_state *mrb, int argc, mrb_value *argv, mrb_value io)
{
    mrb_value fname, offset;
    struct foreach_arg arg;

    //rb_scan_args(argc, argv, "12", &fname, &arg.sep, &offset);
    mrb_get_args(mrb, "*", &length, &str);
    if (argc < 1) {
      length = mrb_nil_value();
      str = mrb_nil_value();
    }
    else {
      if (argc < 2) {
        length = argv[0];
        str = mrb_nil_value();
      }
      else {
        length = argv[0];
        str = argv[1];
      }
    }
    //SafeStringValue(fname);
    mrb_string_value(mrb, &fname);

    arg.argc = argc ? 1 : 0;
    //arg.io = rb_io_open(StringValueCStr(fname), "r");
    arg.io = rb_io_open(rb_string_value_cstr(mrb, &fname), "r");
    if (mrb_nil_p(arg.io)) return mrb_nil_value();
    if (!mrb_nil_p(offset)) {
      struct seek_arg sarg;
      int state = 0;
      sarg.io = arg.io;
      sarg.offset = offset;
      sarg.mode = SEEK_SET;
      rb_protect((mrb_value (*)(mrb_value))seek_before_read, (mrb_value)&sarg, &state);
      if (state) {
          rb_io_close(mrb, arg.io);
          rb_jump_tag(state);
      }
    }
    return rb_ensure(io_s_read, (mrb_value)&arg, rb_io_close, arg.io);
}

static mrb_value
argf_tell()
{
    if (!next_argv(mrb)) {
      mrb_raise(mrb, mrb->eArgumentError_class, "no stream to tell");
    }
    ARGF_FORWARD(0, 0);
    return rb_io_tell(current_file);
}

static mrb_value
argf_seek_m(int argc, mrb_value *argv, self)
    int argc;
    mrb_value *argv;
    mrb_value self;
{
    if (!next_argv(mrb)) {
      mrb_raise(mrb, mrb->eArgumentError_class, "no stream to seek");
    }
    ARGF_FORWARD(argc, argv);
    return rb_io_seek_m(argc, argv, current_file);
}

static mrb_value
argf_set_pos(self, offset)
     mrb_value self, offset;
{
    if (!next_argv(mrb)) {
      mrb_raise(mrb, mrb->eArgumentError_class, "no stream to set position");
    }
    ARGF_FORWARD(1, &offset);
    return rb_io_set_pos(current_file, offset);
}

static mrb_value
argf_rewind()
{
    if (!next_argv(mrb)) {
      mrb_raise(mrb, mrb->eArgumentError_class, "no stream to rewind");
    }
    ARGF_FORWARD(0, 0);
    return rb_io_rewind(current_file);
}

static mrb_value
argf_fileno()
{
    if (!next_argv(mrb)) {
      mrb_raise(mrb, mrb->eArgumentError_class, "no stream");
    }
    ARGF_FORWARD(0, 0);
    return rb_io_fileno(current_file);
}

static mrb_value
argf_to_io()
{
    next_argv(mrb);
    ARGF_FORWARD(0, 0);
    return current_file;
}

static mrb_value
argf_eof()
{
    if (current_file) {
      if (init_p == 0) return mrb_true_value();
      next_argv(mrb);
      ARGF_FORWARD(0, 0);
      if (rb_io_eof(current_file)) {
          return mrb_true_value();
      }
    }
    return mrb_false_value();
}

static mrb_value
argf_read(int argc, mrb_value *argv)
{
    mrb_value tmp, str, length;
    long len = 0;

    rb_scan_args(argc, argv, "02", &length, &str);
    if (!mrb_nil_p(length)) {
      len = mrb_fixnum_number(argv[0]);
    }
    if (!mrb_nil_p(str)) {
      //StringValue(str);
      mrb_string_value(mrb, &str);
      mrb_str_resize(mrb, str, 0);
      argv[1] = mrb_nil_value();
    }

  retry:
    if (!next_argv(mrb)) {
      return str;
    }
    if (mrb_type(current_file) != MRB_TT_FILE) {
      tmp = argf_forward(argc, argv);
    }
    else {
      tmp = io_read(mrb, argc, argv, current_file);
    }
    if (mrb_nil_p(str)) str = tmp;
    else if (!mrb_nil_p(tmp)) rb_str_append(str, tmp);
    if (mrb_nil_p(tmp) || mrb_nil_p(length)) {
      if (next_p != -1) {
          argf_close(mrb, current_file);
          next_p = 1;
          goto retry;
      }
    }
    else if (argc >= 1) {
      if (RSTRING(str)->len < len) {
          len -= RSTRING(str)->len;
          argv[0] = mrb_fixnum_value(len);
          goto retry;
      }
    }
    return str;
}

static mrb_value
argf_getc()
{
    mrb_value byte;

  retry:
    if (!next_argv(mrb)) return mrb_nil_value();
    if (mrb_type(current_file) != MRB_TT_FILE) {
      //byte = rb_funcall3(current_file, rb_intern("getc"), 0, 0);
      byte = mrb_funcall(mrb, current_file, "getc", 0, 0);
    }
    else {
      byte = mrb_io_getc(mrb, current_file);
    }
    if (mrb_nil_p(byte) && next_p != -1) {
      argf_close(mrb, current_file);
      next_p = 1;
      goto retry;
    }

    return byte;
}

static mrb_value
argf_readchar(mrb_state *mrb)
{
    mrb_value c;

    NEXT_ARGF_FORWARD(0, 0);
    c = argf_getc();
    if (mrb_nil_p(c)) {
      rb_eof_error(mrb);
    }
    return c;
}

static mrb_value
argf_each_line(int argc, mrb_value *argv, mrb_value argf)
{
    mrb_value str;

    //RETURN_ENUMERATOR(argf, argc, argv);
    if (!next_argv(mrb)) return mrb_nil_value();
    if (mrb_type(current_file) != MRB_TT_FILE) {
      for (;;) {
          if (!next_argv(mrb)) return argf;
          rb_iterate(rb_each, current_file, rb_yield, 0);
          next_p = 1;
      }
    }
    while (!mrb_nil_p(str = argf_getline(mrb, argc, argv))) {
      rb_yield(str);
    }
    return argf;
}

static mrb_value
argf_each_byte(mrb_value argf)
{
    mrb_value byte;

    //RETURN_ENUMERATOR(argf, 0, 0);
    while (!mrb_nil_p(byte = argf_getc())) {
      rb_yield(byte);
    }
    return argf;
}

static mrb_value
argf_each_char(mrb_value argf)
{
    mrb_value ch;

    //RETURN_ENUMERATOR(argf, 0, 0);

    while (!mrb_nil_p(ch = argf_getc())) {
      unsigned char c;
      int n;
      mrb_value str, file;

first_char:
      c = mrb_fixnum_number(ch);
      n = mbclen(c);
      str = rb_tainted_str_new((const char *)&c, 1);
      file = current_file;

      while (--n > 0) {
          if (mrb_nil_p(ch = argf_getc())) {
            rb_yield(str);
            return argf;
          }
          if (current_file != file) {
            rb_yield(str);
            goto first_char;
          }
          c = mrb_fixnum_number(ch);
          rb_str_cat(mrb, str, (const char *)&c, 1);
      }
      rb_yield(str);
    }
    return argf;
}

static mrb_value
argf_filename(mrb_state *mrb)
{
    next_argv(mrb);
    return filename;
}

static mrb_value
argf_file(mrb_state *mrb)
{
    next_argv(mrb);
    return current_file;
}

static mrb_value
argf_binmode(mrb_state *mrb)
{
    binmode = 1;
    next_argv(mrb);
    ARGF_FORWARD(0, 0);
    rb_io_binmode(mrb, current_file);
    return argf;
}

static mrb_value
argf_skip()
{
    if (init_p && next_p == 0) {
      argf_close(mrb, current_file);
      next_p = 1;
    }
    return argf;
}

#if 0
static mrb_value
argf_close_m(mrb_state *mrb)
{
    next_argv(mrb);
    argf_close(mrb, current_file);
    if (next_p != -1) {
      next_p = 1;
    }
    gets_lineno = 0;
    return argf;
}

static mrb_value
argf_closed(mrb_state *mrb)
{
    next_argv(mrb);
    ARGF_FORWARD(0, 0);
    return mrb_io_closed(mrb, current_file);
}

static mrb_value
argf_to_s()
{
    return mrb_str_new2(mrb, "ARGF");
}
#endif

static mrb_value
opt_i_get()
{
    if (!ruby_inplace_mode) return mrb_nil_value();
    return mrb_str_new2(mrb, ruby_inplace_mode);
}

static void
opt_i_set(mrb_value val)
{
    if (!RTEST(val)) {
      if (ruby_inplace_mode) free(ruby_inplace_mode);
      ruby_inplace_mode = 0;
      return;
    }
    //StringValue(val);
    mrb_string_value(mrb, &val);
    if (ruby_inplace_mode) free(ruby_inplace_mode);
    ruby_inplace_mode = 0;
    //ruby_inplace_mode = strdup(StringValueCStr(val));
    ruby_inplace_mode = strdup(rb_string_value_cstr(mrb, &val));
}
#endif

int
rb_ensure(int (*b_proc)(ANYARGS), void * data1, mrb_value (*e_proc)(ANYARGS), mrb_value data2)
{
#if 0 /* THREAD */
  int state;
  int result = (int)NULL;

  PUSH_TAG();
  if ((state = EXEC_TAG()) == 0) {
    result = (*b_proc) (data1);
  }
  POP_TAG();
  /* TODO: fix me */
  /* retval = prot_tag ? prot_tag->retval : mrb_nil_value(); */     /* save retval */
  (*e_proc) (data2);
  if (state)
    JUMP_TAG(state);
  return result;
#else
  int result;// = (int)NULL;

  result = (*b_proc) (data1);
  (*e_proc) (data2);
  return result;
#endif
}

/*
 *  Class <code>IO</code> is the basis for all input and output in Ruby.
 *  An I/O stream may be <em>duplexed</em> (that is, bidirectional), and
 *  so may use more than one native operating system stream.
 *
 *  Many of the examples in this section use class <code>File</code>,
 *  the only standard subclass of <code>IO</code>. The two classes are
 *  closely associated.
 *
 *  As used in this section, <em>portname</em> may take any of the
 *  following forms.
 *
 *  * A plain string represents a filename suitable for the underlying
 *    operating system.
 *
 *  * A string starting with ``<code>|</code>'' indicates a subprocess.
 *    The remainder of the string following the ``<code>|</code>'' is
 *    invoked as a process with appropriate input/output channels
 *    connected to it.
 *
 *  * A string equal to ``<code>|-</code>'' will create another Ruby
 *    instance as a subprocess.
 *
 *  Ruby will convert pathnames between different operating system
 *  conventions if possible. For instance, on a Windows system the
 *  filename ``<code>/gumby/ruby/test.rb</code>'' will be opened as
 *  ``<code>\gumby\ruby\test.rb</code>''. When specifying a
 *  Windows-style filename in a Ruby string, remember to escape the
 *  backslashes:
 *
 *     "c:\\gumby\\ruby\\test.rb"
 *
 *  Our examples here will use the Unix-style forward slashes;
 *  <code>File::SEPARATOR</code> can be used to get the
 *  platform-specific separator character.
 *
 *  I/O ports may be opened in any one of several different modes, which
 *  are shown in this section as <em>mode</em>. The mode may
 *  either be a Fixnum or a String. If numeric, it should be
 *  one of the operating system specific constants (O_RDONLY,
 *  O_WRONLY, O_RDWR, O_APPEND and so on). See man open(2) for
 *  more information.
 *
 *  If the mode is given as a String, it must be one of the
 *  values listed in the following table.
 *
 *    Mode |  Meaning
 *    -----+--------------------------------------------------------
 *    "r"  |  Read-only, starts at beginning of file  (default mode).
 *    -----+--------------------------------------------------------
 *    "r+" |  Read-write, starts at beginning of file.
 *    -----+--------------------------------------------------------
 *    "w"  |  Write-only, truncates existing file
 *         |  to zero length or creates a new file for writing.
 *    -----+--------------------------------------------------------
 *    "w+" |  Read-write, truncates existing file to zero length
 *         |  or creates a new file for reading and writing.
 *    -----+--------------------------------------------------------
 *    "a"  |  Write-only, starts at end of file if file exists,
 *         |  otherwise creates a new file for writing.
 *    -----+--------------------------------------------------------
 *    "a+" |  Read-write, starts at end of file if file exists,
 *         |  otherwise creates a new file for reading and
 *         |  writing.
 *    -----+--------------------------------------------------------
 *     "b" |  (DOS/Windows only) Binary file mode (may appear with
 *         |  any of the key letters listed above).
 *
 *
 *  The global constant ARGF (also accessible as $<) provides an
 *  IO-like stream which allows access to all files mentioned on the
 *  command line (or STDIN if no files are mentioned). ARGF provides
 *  the methods <code>#path</code> and <code>#filename</code> to access
 *  the name of the file currently being read.
 */

void
mrb_init_io(mrb_state *mrb)
{
#ifdef __CYGWIN__
#include <sys/cygwin.h>
    static struct __cygwin_perfile pf[] =
    {
      {"", O_RDONLY | O_BINARY},
      {"", O_WRONLY | O_BINARY},
      {"", O_RDWR | O_BINARY},
      {"", O_APPEND | O_BINARY},
      {NULL, 0}
    };
    cygwin_internal(CW_PERFILE, pf);
#endif

  struct RClass *c;
  c = mrb->io_class = mrb_define_class(mrb, "IO", mrb->object_class);
  mrb_define_class_method(mrb, c, "open",            mrb_io_s_open,     ARGS_ANY());    /* 15.2.20.4.1  */
  mrb_include_module(mrb, c, mrb->enumerable_module);
  mrb_define_method(mrb, c,       "close",           mrb_io_close_m,    ARGS_NONE());   /* 15.2.20.5.1  */
  mrb_define_method(mrb, c,       "closed?",         mrb_io_closed,     ARGS_NONE());   /* 15.2.20.5.2  */
  mrb_define_method(mrb, c,       "each",            mrb_io_each_line,  ARGS_ANY());    /* 15.2.20.5.3  */
  mrb_define_method(mrb, c,       "each_byte",       mrb_io_each_byte,  ARGS_NONE());   /* 15.2.20.5.4  */
  mrb_define_method(mrb, c,       "each_line",       mrb_io_each_line,  ARGS_ANY());    /* 15.2.20.5.5  */
  mrb_define_method(mrb, c,       "eof?",            mrb_io_eof,        ARGS_NONE());   /* 15.2.20.5.6  */
  mrb_define_method(mrb, c,       "flush",           mrb_io_flush,      ARGS_NONE());   /* 15.2.20.5.7  */
  mrb_define_method(mrb, c,       "getc",            mrb_io_getc,       ARGS_NONE());   /* 15.2.20.5.8  */
  mrb_define_method(mrb, c,       "gets",            mrb_io_gets_m,     ARGS_ANY());    /* 15.2.20.5.9  */
  mrb_define_method(mrb, c,       "initialize_copy", mrb_io_init_copy,  ARGS_REQ(1));   /* 15.2.20.5.10 */
  mrb_define_method(mrb, c,       "print",           mrb_io_print,      ARGS_ANY());    /* 15.2.20.5.11 */
  mrb_define_method(mrb, c,       "putc",            mrb_io_putc,       ARGS_REQ(1));   /* 15.2.20.5.12 */
  mrb_define_method(mrb, c,       "puts",            mrb_io_puts,       ARGS_ANY());    /* 15.2.20.5.13 */
  mrb_define_method(mrb, c,       "read",            mrb_io_read,       ARGS_ANY());    /* 15.2.20.5.14 */
  mrb_define_method(mrb, c,       "readchar",        mrb_io_readchar,   ARGS_NONE());   /* 15.2.20.5.15 */
  mrb_define_method(mrb, c,       "readline",        mrb_io_readline,   ARGS_ANY());    /* 15.2.20.5.16 */
  mrb_define_method(mrb, c,       "readlines",       mrb_io_readlines,  ARGS_ANY());    /* 15.2.20.5.17 */
  mrb_define_method(mrb, c,       "sync",            mrb_io_sync,       ARGS_NONE());   /* 15.2.20.5.18 */
  mrb_define_method(mrb, c,       "sync=",           mrb_io_set_sync,   ARGS_REQ(1));   /* 15.2.20.5.19 */
  mrb_define_method(mrb, c,       "write",           mrb_io_write_m,    ARGS_REQ(1));   /* 15.2.20.5.20 */

  mrb_define_method(mrb, c,       "initialize",      mrb_io_initialize, ARGS_ANY());    /* 15.2.20.5.21 (x)*/
  mrb_define_method(mrb, c,       "to_io",           mrb_io_to_io,      ARGS_NONE());   /* 15.2.20.5.22 (x)*/
  mrb_define_method(mrb, c,       "inspect",         mrb_io_inspect,    ARGS_NONE());   /* 15.2.20.5.23 (x)*/

  mrb_define_variable(mrb, "$stdin", &rb_stdin);
  rb_stdin  = prep_stdio(mrb, stdin, FMODE_READABLE, c, "<STDIN>");
  rb_define_hooked_variable(mrb, "$stdout", &rb_stdout, 0, stdout_setter);
  rb_stdout = prep_stdio(mrb, stdout, FMODE_WRITABLE, c, "<STDOUT>");
  rb_define_hooked_variable(mrb, "$stderr", &rb_stderr, 0, stdout_setter);
  rb_stderr = prep_stdio(mrb, stderr, FMODE_WRITABLE|FMODE_SYNC, c, "<STDERR>");
  rb_define_hooked_variable(mrb, "$>",  &rb_stdout,    0, stdout_setter);

  rb_define_hooked_variable(mrb, "$/",  &rb_rs,        0, rb_str_setter);
  rb_define_hooked_variable(mrb, "$-0", &rb_rs,        0, rb_str_setter);

  rb_define_hooked_variable(mrb, "$.",  &lineno,       0, lineno_setter);
  rb_define_virtual_variable(mrb, "$_", rb_lastline_get, rb_lastline_set);

  //argf = argf_alloc(mrb, mrb_nil_value());
  lineno = mrb_fixnum_value(0);
#if 0
  rb_define_global_function("syscall", rb_f_syscall, -1);

  rb_define_global_function("open", rb_f_open, -1);
  rb_define_global_function("printf", rb_f_printf, -1);
  rb_define_global_function("print", rb_f_print, -1);
  rb_define_global_function("putc", rb_f_putc, 1);
  rb_define_global_function("puts", rb_f_puts, -1);
  rb_define_global_function("gets", rb_f_gets, -1);
  rb_define_global_function("readline", rb_f_readline, -1);
  rb_define_global_function("getc", rb_f_getc, 0);
  rb_define_global_function("select", rb_f_select, -1);

  rb_define_global_function("readlines", rb_f_readlines, -1);

  rb_define_global_function("`", rb_f_backquote, 1);

  rb_define_global_function("p", rb_f_p, -1);
  rb_define_method(rb_mKernel, "display", rb_obj_display, -1);

  rb_cIO = rb_define_class("IO", rb_cObject);
  rb_include_module(rb_cIO, rb_mEnumerable);

  rb_define_alloc_func(rb_cIO, io_alloc);
  rb_define_singleton_method(rb_cIO, "new", rb_io_s_new, -1);
  rb_define_singleton_method(rb_cIO, "open",  rb_io_s_open, -1);
  rb_define_singleton_method(rb_cIO, "sysopen",  rb_io_s_sysopen, -1);
  rb_define_singleton_method(rb_cIO, "for_fd", rb_io_s_for_fd, -1);
  rb_define_singleton_method(rb_cIO, "popen", rb_io_s_popen, -1);
  rb_define_singleton_method(rb_cIO, "foreach", rb_io_s_foreach, -1);
  rb_define_singleton_method(rb_cIO, "readlines", rb_io_s_readlines, -1);
  rb_define_singleton_method(rb_cIO, "read", rb_io_s_read, -1);
  rb_define_singleton_method(rb_cIO, "select", rb_f_select, -1);
  rb_define_singleton_method(rb_cIO, "pipe", rb_io_s_pipe, 0);

  rb_define_method(rb_cIO, "initialize", rb_io_initialize, -1);

  rb_define_hooked_variable("$.", &lineno, 0, lineno_setter);
  rb_define_virtual_variable("$_", rb_lastline_get, rb_lastline_set);

  rb_define_method(rb_cIO, "initialize_copy", rb_io_init_copy, 1);
  rb_define_method(rb_cIO, "reopen", rb_io_reopen, -1);

  rb_define_method(rb_cIO, "print", rb_io_print, -1);
  rb_define_method(rb_cIO, "putc", rb_io_putc, 1);
  rb_define_method(rb_cIO, "puts", rb_io_puts, -1);
  rb_define_method(rb_cIO, "printf", rb_io_printf, -1);

  rb_define_method(rb_cIO, "each",  rb_io_each_line, -1);
  rb_define_method(rb_cIO, "each_line",  rb_io_each_line, -1);
  rb_define_method(rb_cIO, "each_byte",  rb_io_each_byte, 0);
  rb_define_method(rb_cIO, "each_char",  rb_io_each_char, 0);
  rb_define_method(rb_cIO, "lines",  rb_io_lines, -1);
  rb_define_method(rb_cIO, "bytes",  rb_io_bytes, 0);
  rb_define_method(rb_cIO, "chars",  rb_io_each_char, 0);

  rb_define_method(rb_cIO, "syswrite", rb_io_syswrite, 1);
  rb_define_method(rb_cIO, "sysread",  rb_io_sysread, -1);

  rb_define_method(rb_cIO, "fileno", rb_io_fileno, 0);
  rb_define_alias(rb_cIO, "to_i", "fileno");
  rb_define_method(rb_cIO, "to_io", rb_io_to_io, 0);

  rb_define_method(rb_cIO, "fsync",   rb_io_fsync, 0);
  rb_define_method(rb_cIO, "sync",   rb_io_sync, 0);
  rb_define_method(rb_cIO, "sync=",  rb_io_set_sync, 1);

  rb_define_method(rb_cIO, "lineno",   rb_io_lineno, 0);
  rb_define_method(rb_cIO, "lineno=",  rb_io_set_lineno, 1);

  rb_define_method(rb_cIO, "readlines",  rb_io_readlines, -1);

  rb_define_method(rb_cIO, "read_nonblock",  io_read_nonblock, -1);
  rb_define_method(rb_cIO, "write_nonblock", rb_io_write_nonblock, 1);
  rb_define_method(rb_cIO, "readpartial",  io_readpartial, -1);
  rb_define_method(rb_cIO, "read",  io_read, -1);
  rb_define_method(rb_cIO, "write", io_write, 1);
  rb_define_method(rb_cIO, "gets",  rb_io_gets_m, -1);
  rb_define_method(rb_cIO, "readline",  rb_io_readline, -1);
  rb_define_method(rb_cIO, "getc",  rb_io_getc, 0);
  rb_define_method(rb_cIO, "getbyte",  rb_io_getc, 0);
  rb_define_method(rb_cIO, "readchar",  rb_io_readchar, 0);
  rb_define_method(rb_cIO, "readbyte",  rb_io_readchar, 0);
  rb_define_method(rb_cIO, "ungetc",rb_io_ungetc, 1);
  rb_define_method(rb_cIO, "<<",    rb_io_addstr, 1);
  rb_define_method(rb_cIO, "flush", rb_io_flush, 0);
  rb_define_method(rb_cIO, "tell", rb_io_tell, 0);
  rb_define_method(rb_cIO, "seek", rb_io_seek_m, -1);
  rb_define_const(rb_cIO, "SEEK_SET", mrb_fixnum_value(SEEK_SET));
  rb_define_const(rb_cIO, "SEEK_CUR", mrb_fixnum_value(SEEK_CUR));
  rb_define_const(rb_cIO, "SEEK_END", mrb_fixnum_value(SEEK_END));
  rb_define_method(rb_cIO, "rewind", rb_io_rewind, 0);
  rb_define_method(rb_cIO, "pos", rb_io_tell, 0);
  rb_define_method(rb_cIO, "pos=", rb_io_set_pos, 1);
  rb_define_method(rb_cIO, "eof", rb_io_eof, 0);
  rb_define_method(rb_cIO, "eof?", rb_io_eof, 0);

  rb_define_method(rb_cIO, "close", rb_io_close_m, 0);
  rb_define_method(rb_cIO, "closed?", rb_io_closed, 0);
  rb_define_method(rb_cIO, "close_read", rb_io_close_read, 0);
  rb_define_method(rb_cIO, "close_write", rb_io_close_write, 0);

  rb_define_method(rb_cIO, "isatty", rb_io_isatty, 0);
  rb_define_method(rb_cIO, "tty?", rb_io_isatty, 0);
  rb_define_method(rb_cIO, "binmode",  rb_io_binmode, 0);
  rb_define_method(rb_cIO, "sysseek", rb_io_sysseek, -1);

  rb_define_method(rb_cIO, "ioctl", rb_io_ioctl, -1);
  rb_define_method(rb_cIO, "fcntl", rb_io_fcntl, -1);
  rb_define_method(rb_cIO, "pid", rb_io_pid, 0);
  rb_define_method(rb_cIO, "inspect",  rb_io_inspect, 0);

  rb_define_variable("$stdin", &rb_stdin);
  rb_stdin = prep_stdio(mrb, stdin, FMODE_READABLE, mrb->io_class);
  rb_define_hooked_variable("$stdout", &rb_stdout, 0, stdout_setter);
  rb_stdout = prep_stdio(mrb, stdout, FMODE_WRITABLE, mrb->io_class);
  rb_define_hooked_variable("$stderr", &rb_stderr, 0, stdout_setter);
  rb_stderr = prep_stdio(mrb, stderr, FMODE_WRITABLE, mrb->io_class);
  rb_define_hooked_variable("$>", &rb_stdout, 0, stdout_setter);
  orig_stdout = rb_stdout;
  rb_deferr = orig_stderr = rb_stderr;

  /* variables to be removed in 1.8.1 */
  rb_define_hooked_variable("$defout", &rb_stdout, 0, defout_setter);
  rb_define_hooked_variable("$deferr", &rb_stderr, 0, deferr_setter);

  /* constants to hold original stdin/stdout/stderr */
  rb_define_global_const("STDIN", rb_stdin);
  rb_define_global_const("STDOUT", rb_stdout);
  rb_define_global_const("STDERR", rb_stderr);

  rb_define_readonly_variable("$<", &argf);
  argf = rb_obj_alloc(rb_cObject);
  rb_extend_object(argf, rb_mEnumerable);
  rb_define_global_const("ARGF", argf);

  rb_define_singleton_method(argf, "to_s", argf_to_s, 0);

  rb_define_singleton_method(argf, "fileno", argf_fileno, 0);
  rb_define_singleton_method(argf, "to_i", argf_fileno, 0);
  rb_define_singleton_method(argf, "to_io", argf_to_io, 0);
  rb_define_singleton_method(argf, "each",  argf_each_line, -1);
  rb_define_singleton_method(argf, "each_line",  argf_each_line, -1);
  rb_define_singleton_method(argf, "each_byte",  argf_each_byte, 0);
  rb_define_singleton_method(argf, "each_char",  argf_each_char, 0);
  rb_define_singleton_method(argf, "lines",  argf_each_line, -1);
  rb_define_singleton_method(argf, "bytes",  argf_each_byte, 0);
  rb_define_singleton_method(argf, "chars",  argf_each_char, 0);

  rb_define_singleton_method(argf, "read",  argf_read, -1);
  rb_define_singleton_method(argf, "readlines", rb_f_readlines, -1);
  rb_define_singleton_method(argf, "to_a", rb_f_readlines, -1);
  rb_define_singleton_method(argf, "gets", rb_f_gets, -1);
  rb_define_singleton_method(argf, "readline", rb_f_readline, -1);
  rb_define_singleton_method(argf, "getc", argf_getc, 0);
  rb_define_singleton_method(argf, "getbyte", argf_getc, 0);
  rb_define_singleton_method(argf, "readchar", argf_readchar, 0);
  rb_define_singleton_method(argf, "readbyte", argf_readchar, 0);
  rb_define_singleton_method(argf, "tell", argf_tell, 0);
  rb_define_singleton_method(argf, "seek", argf_seek_m, -1);
  rb_define_singleton_method(argf, "rewind", argf_rewind, 0);
  rb_define_singleton_method(argf, "pos", argf_tell, 0);
  rb_define_singleton_method(argf, "pos=", argf_set_pos, 1);
  rb_define_singleton_method(argf, "eof", argf_eof, 0);
  rb_define_singleton_method(argf, "eof?", argf_eof, 0);
  rb_define_singleton_method(argf, "binmode", argf_binmode, 0);

  rb_define_singleton_method(argf, "filename", argf_filename, 0);
  rb_define_singleton_method(argf, "path", argf_filename, 0);
  rb_define_singleton_method(argf, "file", argf_file, 0);
  rb_define_singleton_method(argf, "skip", argf_skip, 0);
  rb_define_singleton_method(argf, "close", argf_close_m, 0);
  rb_define_singleton_method(argf, "closed?", argf_closed, 0);

  rb_define_singleton_method(argf, "lineno",   argf_lineno, 0);
  rb_define_singleton_method(argf, "lineno=",  argf_set_lineno, 1);

  rb_global_variable(&current_file);
  rb_define_readonly_variable("$FILENAME", &filename);
  filename = mrb_str_new2(mrb, "-");

  rb_define_virtual_variable("$-i", opt_i_get, opt_i_set);

#if defined (_WIN32) || defined(DJGPP) || defined(__CYGWIN__) || defined(__human68k__)
  atexit(pipe_atexit);
#endif

  Init_File();

  rb_define_method(mrb->file_class, "initialize",  rb_file_initialize, -1);

  rb_file_const("RDONLY", mrb_fixnum_value(O_RDONLY));
  rb_file_const("WRONLY", mrb_fixnum_value(O_WRONLY));
  rb_file_const("RDWR", mrb_fixnum_value(O_RDWR));
  rb_file_const("APPEND", mrb_fixnum_value(O_APPEND));
  rb_file_const("CREAT", mrb_fixnum_value(O_CREAT));
  rb_file_const("EXCL", mrb_fixnum_value(O_EXCL));
#if defined(O_NDELAY) || defined(O_NONBLOCK)
#   ifdef O_NONBLOCK
    rb_file_const("NONBLOCK", mrb_fixnum_value(O_NONBLOCK));
#   else
    rb_file_const("NONBLOCK", mrb_fixnum_value(O_NDELAY));
#   endif
#endif
    rb_file_const("TRUNC", mrb_fixnum_value(O_TRUNC));
#ifdef O_NOCTTY
    rb_file_const("NOCTTY", mrb_fixnum_value(O_NOCTTY));
#endif
#ifdef O_BINARY
    rb_file_const("BINARY", mrb_fixnum_value(O_BINARY));
#endif
#ifdef O_SYNC
    rb_file_const("SYNC", mrb_fixnum_value(O_SYNC));
#endif
#endif //#if 0
}

/* for IO */

#if defined(NFDBITS) && defined(HAVE_RB_FD_INIT)

/*
 * several Unix platforms support file descriptors bigger than FD_SETSIZE
 * in select(2) system call.
 *
 * - Linux 2.2.12 (?)
 * - NetBSD 1.2 (src/sys/kern/sys_generic.c:1.25)
 *   select(2) documents how to allocate fd_set dynamically.
 *   http://netbsd.gw.com/cgi-bin/man-cgi?select++NetBSD-4.0
 * - FreeBSD 2.2 (src/sys/kern/sys_generic.c:1.19)
 * - OpenBSD 2.0 (src/sys/kern/sys_generic.c:1.4)
 *   select(2) documents how to allocate fd_set dynamically.
 *   http://www.openbsd.org/cgi-bin/man.cgi?query=select&manpath=OpenBSD+4.4
 * - HP-UX documents how to allocate fd_set dynamically.
 *   http://docs.hp.com/en/B2355-60105/select.2.html
 * - Solaris 8 has select_large_fdset
 *
 * When fd_set is not big enough to hold big file descriptors,
 * it should be allocated dynamically.
 * Note that this assumes fd_set is structured as bitmap.
 *
 * rb_fd_init allocates the memory.
 * rb_fd_term free the memory.
 * rb_fd_set may re-allocates bitmap.
 *
 * So rb_fd_set doesn't reject file descriptors bigger than FD_SETSIZE.
 */

void
rb_fd_zero(rb_fdset_t *fds)
{
  if (fds->fdset) {
      memset(fds->fdset, 0, sizeof(fd_mask)*howmany(fds->maxfd, NFDBITS));
      FD_ZERO(fds->fdset);
  }
}

void
rb_fd_init(rb_fdset_t *fds)
{
    fds->maxfd = 0;
    fds->fdset = malloc(sizeof(fd_set));
    FD_ZERO(fds->fdset);
}

void
rb_fd_term(rb_fdset_t *fds)
{
    if (fds->fdset) xfree(fds->fdset);
    fds->maxfd = 0;
    fds->fdset = 0;
}

static void
rb_fd_resize(int n, rb_fdset_t *fds)
{
    size_t m = howmany(n + 1, NFDBITS) * sizeof(fd_mask);
    size_t o = howmany(fds->maxfd, NFDBITS) * sizeof(fd_mask);

    if (m < sizeof(fd_set)) m = sizeof(fd_set);
    if (o < sizeof(fd_set)) o = sizeof(fd_set);

    if (m > o) {
    fds->fdset = xrealloc(fds->fdset, m);
    memset((char *)fds->fdset + o, 0, m - o);
    }
    if (n >= fds->maxfd) fds->maxfd = n + 1;
}

void
rb_fd_set(int n, rb_fdset_t *fds)
{
    rb_fd_resize(n, fds);
    FD_SET(n, fds->fdset);
}

void
rb_fd_clr(int n, rb_fdset_t *fds)
{
    if (n >= fds->maxfd) return;
    FD_CLR(n, fds->fdset);
}

int
rb_fd_isset(int n, const rb_fdset_t *fds)
{
    if (n >= fds->maxfd) return 0;
    return FD_ISSET(n, (const rb_fdset_t *)fds->fdset) != 0; /* "!= 0" avoids FreeBSD PR 91421 */
}

void
rb_fd_copy(rb_fdset_t *dst, const fd_set *src, int max)
{
    size_t size = howmany(max, NFDBITS) * sizeof(fd_mask);

    if (size < sizeof(fd_set)) size = sizeof(fd_set);
    dst->maxfd = max;
    dst->fdset = xrealloc(dst->fdset, size);
    memcpy(dst->fdset, src, size);
}
#endif

#endif //#ifndef INCLUDE_REGEXP
