/**********************************************************************

  time.c -

  $Author: yugui $
  created at: Tue Dec 28 14:31:59 JST 1993

  Copyright (C) 1993-2007 Yukihiro Matsumoto

**********************************************************************/

#include "mruby.h"
#include <sys/types.h>
#include <sys/time.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <string.h>
#include "encoding.h"
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <float.h>
#include <math.h>

#include "timev.h"
#include "st.h"
#include "numeric.h"
#include "error.h"
#include "transcode_data.h"
#include "mstring.h"
#include "variable.h"
#include "class.h"
#include "array.h"
#include "mio.h"
#include "hash.h"

#define rb_long2int(n) ((int)(n))

#ifndef FALSE
#define FALSE   0
#endif

#ifndef TRUE
#define TRUE    1
#endif

static mrb_sym id_divmod, id_mul, id_submicro, id_nano_num, id_nano_den, id_offset;
static mrb_sym id_eq, id_ne, id_quo, id_div, id_cmp, id_lshift;

#define NDIV(x,y) (-(-((x)+1)/(y))-1)
#define NMOD(x,y) ((y)-(-((x)+1)%(y))-1)
#define DIV(n,d) ((n)<0 ? NDIV((n),(d)) : (n)/(d))
#define MOD(n,d) ((n)<0 ? NMOD((n),(d)) : (n)%(d))

#if 0 /* 2011_0518 */
#define TIMET2NUM(n) mrb_fixnum_value(n)
#define NUM2TIMET(n) mrb_fixnum_number(n)
#else
#define TIMET2NUM(n) ((long long)n)
#define NUM2TIMET(n) mrb_fixnum_number(n)
#endif

#ifndef HAVE_UNISTD_H
struct timespec {
  time_t  tv_sec;   /* Seconds */
  long    tv_nsec;  /* Nanoseconds */
};
#endif

#if 0 /* 2011_0518 */
static int
eq(mrb_state *mrb, mrb_value x, mrb_value y)
{
    if (FIXNUM_P(x) && FIXNUM_P(y)) {
        return mrb_obj_equal(mrb, x, y);
    }
    return mrb_test(mrb_funcall(mrb, x, "eq", 1, y));
}
#else
static int
eq(mrb_state *mrb, long x, long y)
{
    return (x == y);
}
#endif

#if 0 /* 2011_0518 */
static int
cmp(mrb_state *mrb, mrb_value x, mrb_value y)
{
  if (FIXNUM_P(x) && FIXNUM_P(y)) {
    if ((long)mrb_fixnum_number(x) < (long)mrb_fixnum_number(y))
        return -1;
    if ((long)mrb_fixnum_number(x) > (long)mrb_fixnum_number(y))
        return 1;
    return 0;
  }
  return rb_cmpint(mrb, mrb_funcall(mrb, x, "<=>", 1, y), x, y);
}
#endif

#if 0
#define ne(mrb,x,y) (!eq(mrb, (x),(y)))
#define lt(mrb,x,y) (cmp((mrb),(x),(y)) < 0)
#define gt(mrb,x,y) (cmp((mrb),(x),(y)) > 0)
#define le(mrb,x,y) (cmp((mrb),(x),(y)) <= 0)
#define ge(mrb,x,y) (cmp((mrb),(x),(y)) >= 0)
#else
#define ne(mrb,x,y) ((x) !=  (y))
#define lt(mrb,x,y) ((x) <   (y))
#define gt(mrb,x,y) ((x) >   (y))
#define le(mrb,x,y) ((x) <=  (y))
#define ge(mrb,x,y) ((x) >=  (y))
#endif

#if 0 /* 2011_0518 */
static mrb_value
add(mrb_state *mrb, mrb_value x, mrb_value y)
{
    if (FIXNUM_P(x) && FIXNUM_P(y)) {
        long l = mrb_fixnum_number(x) + mrb_fixnum_number(y);
        if (FIXABLE(l)) return mrb_fixnum_value(l);
        return mrb_fixnum_value(l);
    }
    return mrb_funcall(mrb, x, "+", 1, y);
}

static mrb_value
sub(mrb_state *mrb, mrb_value x, mrb_value y)
{
    if (FIXNUM_P(x) && FIXNUM_P(y)) {
        long l = mrb_fixnum_number(x) - mrb_fixnum_number(y);
        if (FIXABLE(l)) return mrb_fixnum_value(l);
        return mrb_fixnum_value(l);
    }
    return mrb_funcall(mrb, x, "-", 1, y);
}
#else
static long
add(mrb_state *mrb, long x, long y)
{
  return x + y;
}
static long
sub(mrb_state *mrb, long x, long y)
{
  return x - y;
}
#endif

#if 0
#if !(HAVE_LONG_LONG && SIZEOF_LONG * 2 <= SIZEOF_LONG_LONG)
static int
long_mul(long x, long y, long *z)
{
    unsigned long a, b, c;
    int s;
    if (x == 0 || y == 0) {
        *z = 0;
        return 1;
    }
    if (x < 0) {
        s = -1;
        a = (unsigned long)-x;
    }
    else {
        s = 1;
        a = (unsigned long)x;
    }
    if (y < 0) {
        s = -s;
        b = (unsigned long)-y;
    }
    else {
        b = (unsigned long)y;
    }
    if (a <= ULONG_MAX / b) {
        c = a * b;
        if (s < 0) {
            if (c <= (unsigned long)LONG_MAX + 1) {
                *z = -(long)c;
                return 1;
            }
        }
        else {
            if (c <= (unsigned long)LONG_MAX) {
                *z = (long)c;
                return 1;
            }
        }
    }
    return 0;
}
#endif
#endif

#if 0 /* 2011_0518 */
static mrb_value
mul(mrb_state *mrb, mrb_value x, mrb_value y)
{
    if (FIXNUM_P(x) && FIXNUM_P(y)) {
#if HAVE_LONG_LONG && SIZEOF_LONG * 2 <= SIZEOF_LONG_LONG
        LONG_LONG ll = (LONG_LONG)mrb_fixnum_number(x) * mrb_fixnum_number(y);
        if (FIXABLE(ll))
            return mrb_fixnum_value(ll);
        return LL2NUM(ll);
#else
        long z;
        if (long_mul(mrb_fixnum_number(x), mrb_fixnum_number(y), &z))
            return mrb_fixnum_value(z);
#endif
    }
    return mrb_funcall(mrb, x, "*", 1, y);
}
#else
static long long
mul(mrb_state *mrb, long long x, long long y)
{
  return x * y;
}
#endif

#if 0 /* 2011_0518 */
#define div(mrb,x,y) (mrb_funcall(mrb, (x), "div", 1, (y)))

static mrb_value
mod(mrb_state *mrb, mrb_value x, mrb_value y)
{
    switch (mrb_type(x)) {
      default: return mrb_funcall(mrb, x, "%", 1, y);
    }
}
#else
#define div(mrb,x,y) ((x) / (y))
static long
mod(mrb_state *mrb, long x, long y)
{
  return x % y;
}
#endif

#if 0 /* 2011_0518 */
#define neg(mrb,x) (sub(mrb, mrb_fixnum_value(0), (x)))
#define lshift(mrb,x,y) (mrb_funcall(mrb, (x), id_lshift, 1, (y)))
#else
#define neg(mrb,x) (sub(mrb, (0), (x)))
#define lshift(mrb,x,y) ((x) << (y))
#endif

#if 0 /* 2011_0518 */
static mrb_value
quo(mrb_state *mrb, mrb_value x, mrb_value y)
{
    mrb_value ret;
    if (FIXNUM_P(x) && FIXNUM_P(y)) {
        long a, b, c;
        a = mrb_fixnum_number(x);
        b = mrb_fixnum_number(y);
        if (b == 0) rb_num_zerodiv(mrb);
        c = a / b;
        if (c * b == a) {
            return mrb_fixnum_value(c);
        }
    }
    ret = mrb_funcall(mrb, x, "quo", 1, y);
#if 0
    if (mrb_type(ret) == T_RATIONAL &&
        RRATIONAL(ret)->den == mrb_fixnum_value(1)) {
        ret = RRATIONAL(ret)->num;
    }
#endif
    return ret;
}
#else
static mrb_float
quo(mrb_state *mrb, long long x, long long y)
{
  return (mrb_float)x / y;
}
#endif

#if 0 /* 2011_0518 */
#define mulquo(mrb,x,y,z) ((mrb_fixnum_number(y) == mrb_fixnum_number(z)) ? (x) : quo(mrb, mul(mrb,(x),(y)),(z)))
#else
#define mulquo(mrb,x,y,z) (((y) == (z)) ? (x) : quo(mrb, mul(mrb,(x),(y)),(z)))
#endif

#if 0 /* 2011_0518 */
static void
divmodv(mrb_state *mrb, mrb_value n, mrb_value d, mrb_value *q, mrb_value *r)
{
    mrb_value tmp, ary;
    tmp = mrb_funcall(mrb, n, "divmod", 1, d);
    ary = mrb_check_array_type(mrb, tmp);
    if (mrb_nil_p(ary)) {
        mrb_raise(mrb, mrb->eTypeError_class, "unexpected divmod result: into %s",
                 mrb_obj_classname(mrb, tmp));
    }
    *q = mrb_ary_ref(mrb, ary, 0);
    *r = mrb_ary_ref(mrb, ary, 1);
}
#else
static void
divmodv(mrb_state *mrb, long n, long d, long *q, long *r)
{
  *q = n / d;
  *r = n - (*q * d);
  if ((*r < 0 && d > 0) || (*r > 0 && d < 0)) {
    *r += d;
    *q -= 1;
  }
}

#endif

#if SIZEOF_LONG == 8
# define INT64toNUM(x) mrb_fixnum_value(x)
# define UINT64toNUM(x) ULONG2NUM(x)
#elif defined(HAVE_LONG_LONG) && SIZEOF_LONG_LONG == 8
# define INT64toNUM(x) LL2NUM(x)
# define UINT64toNUM(x) ULL2NUM(x)
#endif

#if defined(HAVE_UINT64_T) && SIZEOF_LONG*2 <= SIZEOF_UINT64_T
    typedef uint64_t uwideint_t;
    typedef int64_t wideint_t;
    typedef uint64_t WIDEVALUE;
    typedef int64_t SIGNED_WIDEVALUE;
#   define WIDEVALUE_IS_WIDER 1
#   define UWIDEINT_MAX UINT64_MAX
#   define WIDEINT_MAX INT64_MAX
#   define WIDEINT_MIN INT64_MIN
#   define FIXWINT_P(tv) ((tv) & 1)
#   define FIXWVtoINT64(tv) RSHIFT((SIGNED_WIDEVALUE)(tv), 1)
#   define INT64toFIXWV(wi) ((WIDEVALUE)((SIGNED_WIDEVALUE)(wi) << 1 | FIXNUM_FLAG))
#   define FIXWV_MAX (((int64_t)1 << 62) - 1)
#   define FIXWV_MIN (-((int64_t)1 << 62))
#   define FIXWVABLE(wi) (POSFIXWVABLE(wi) && NEGFIXWVABLE(wi))
#   define WINT2FIXWV(i) WIDEVAL_WRAP(INT64toFIXWV(i))
#   define FIXWV2WINT(w) FIXWVtoINT64(WIDEVAL_GET(w))
#else
    typedef unsigned long long uwideint_t;
    typedef long long wideint_t;
    typedef long long WIDEVALUE;
    typedef long long SIGNED_WIDEVALUE;
#   define WIDEVALUE_IS_WIDER 0
#   define UWIDEINT_MAX ULONG_MAX
#   define WIDEINT_MAX LONG_MAX
#   define WIDEINT_MIN LONG_MIN
#   define FIXWINT_P(v) FIXNUM_P(v)
#   define FIXWV_MAX FIXNUM_MAX
#   define FIXWV_MIN FIXNUM_MIN
#   define FIXWVABLE(i) FIXABLE(i)
  #if 0 /* 2011_0518 */
#   define WINT2FIXWV(i) WIDEVAL_WRAP(mrb_fixnum_value(i))
#   define FIXWV2WINT(w) mrb_fixnum_number(WIDEVAL_GET(w))
  #else
    #define WINT2FIXWV(i) WIDEVAL_WRAP(i)
    #define FIXWV2WINT(w) WIDEVAL_GET(w)
  #endif
#endif

#define POSFIXWVABLE(wi) ((wi) < FIXWV_MAX+1)
#define NEGFIXWVABLE(wi) ((wi) >= FIXWV_MIN)
#define FIXWV_P(w) FIXWINT_P(WIDEVAL_GET(w))

/* #define STRUCT_WIDEVAL */
#ifdef STRUCT_WIDEVAL
    /* for type checking */
    typedef struct {
        WIDEVALUE value;
    } wideval_t;
    static inline wideval_t WIDEVAL_WRAP(WIDEVALUE v) { wideval_t w = { v }; return w; }
#   define WIDEVAL_GET(w) ((w).value)
#else
    typedef WIDEVALUE wideval_t;
//#   define WIDEVAL_WRAP(v) (v)
  #if 0 /* 2011_0518 */
    #define WIDEVAL_WRAP(v) mrb_fixnum_number(v)
  #else
    #define WIDEVAL_WRAP(v) (v)
  #endif
//#   define WIDEVAL_GET(w) (w)
  #if 0 /* 2011_0518 */
    #define WIDEVAL_GET(w) mrb_fixnum_value(w)
  #else
    #define WIDEVAL_GET(w) (w)
  #endif
#endif

#if WIDEVALUE_IS_WIDER
    static inline wideval_t
    wint2wv(wideint_t wi)
    {
        if (FIXWVABLE(wi))
            return WINT2FIXWV(wi);
        else
            return WIDEVAL_WRAP(INT64toNUM(wi));
    }
#   define WINT2WV(wi) wint2wv(wi)
#else
//#   define WINT2WV(wi) WIDEVAL_WRAP(mrb_fixnum_value(wi))
    #define WINT2WV(wi) (wi)
#endif

#if 0 /* 2011_0518 */
static inline mrb_value
w2v(wideval_t w)
{
#if WIDEVALUE_IS_WIDER
    if (FIXWV_P(w))
        return INT64toNUM(FIXWV2WINT(w));
    return (mrb_value)WIDEVAL_GET(w);
#else
    return WIDEVAL_GET(w);
#endif
}
#else
#define w2v(w) (w)
#endif

#if WIDEVALUE_IS_WIDER
static int
bdigit_find_maxbit(BDIGIT d)
{
    int res = 0;
    if (d & ~(BDIGIT)0xffff) {
        d >>= 16;
        res += 16;
    }
    if (d & ~(BDIGIT)0xff) {
        d >>= 8;
        res += 8;
    }
    if (d & ~(BDIGIT)0xf) {
        d >>= 4;
        res += 4;
    }
    if (d & ~(BDIGIT)0x3) {
        d >>= 2;
        res += 2;
    }
    if (d & ~(BDIGIT)0x1) {
        d >>= 1;
        res += 1;
    }
    return res;
}

static mrb_value
rb_big_abs_find_maxbit(mrb_value big)
{
    BDIGIT *ds = RBIGNUM_DIGITS(big);
    BDIGIT d;
    long len = RBIGNUM_LEN(big);
    mrb_value res;
    while (0 < len && ds[len-1] == 0)
        len--;
    if (len == 0)
        return mrb_nil_value();
    res = mul(mrb, mrb_fixnum_value(len-1), mrb_fixnum_value(SIZEOF_BDIGITS * CHAR_BIT));
    d = ds[len-1];
    res = add(mrb, res, mrb_fixnum_value(bdigit_find_maxbit(d)));
    return res;
}

static mrb_value
rb_big_abs_find_minbit(mrb_value big)
{
    BDIGIT *ds = RBIGNUM_DIGITS(big);
    BDIGIT d;
    long len = RBIGNUM_LEN(big);
    long i;
    mrb_value res;
    for (i = 0; i < len; i++)
        if (ds[i])
            break;
    if (i == len)
        return mrb_nil_value();
    res = mul(mrb, mrb_fixnum_value(i), mrb_fixnum_value(SIZEOF_BDIGITS * CHAR_BIT));
    d = ds[i];
    res = add(mrb, res, mrb_fixnum_value(bdigit_find_maxbit(d & (~d-1))));
    return res;
}

static wideval_t
v2w_bignum(mrb_value v)
{
    long len = RBIGNUM_LEN(v);
    BDIGIT *ds;
    wideval_t w;
    mrb_value maxbit;
    ds = RBIGNUM_DIGITS(v);
    w = WIDEVAL_WRAP(v);
    maxbit = rb_big_abs_find_maxbit(v);
    if (mrb_nil_p(maxbit))
        return WINT2FIXWV(0);
    if (lt(mrb, maxbit, mrb_fixnum_value(sizeof(wideint_t) * CHAR_BIT - 2)) ||
        (eq(maxbit, mrb_fixnum_value(sizeof(wideint_t) * CHAR_BIT - 2)) &&
         RBIGNUM_NEGATIVE_P(v) &&
         eq(rb_big_abs_find_minbit(v), mrb_fixnum_value(sizeof(wideint_t) * CHAR_BIT - 2)))) {
        wideint_t i;
        i = 0;
        while (len)
            i = (i << sizeof(BDIGIT)*CHAR_BIT) | ds[--len];
        if (RBIGNUM_NEGATIVE_P(v)) {
            i = -i;
        }
        w = WINT2FIXWV(i);
    }
    return w;
}
#endif

#if 0 /* 2011_0518 */
static inline wideval_t
v2w(mrb_value v)
{
#if WIDEVALUE_IS_WIDER
    if (FIXNUM_P(v)) {
        return WIDEVAL_WRAP((WIDEVALUE)(SIGNED_WIDEVALUE)(long)v);
    }
#endif
  //return WIDEVAL_WRAP(v);
  if (mrb_nil_p(v)) {
    //mrb_raise(mrb, mrb->eTypeError_class, "can't convert %s into an exact number",
    //                            "nil" );
    return 0; /* not reach */
  }
  if (FIXNUM_P(v)) {
    return mrb_fixnum_number(v);
  }
  else {
    return mrb_float_number(v);
  }
}
#else
  #define v2w(v) v
#endif

static int
weq(mrb_state *mrb, wideval_t wx, wideval_t wy)
{
#if WIDEVALUE_IS_WIDER
    if (FIXWV_P(wx) && FIXWV_P(wy)) {
        return WIDEVAL_GET(wx) == WIDEVAL_GET(wy);
    }
    return RTEST(mrb_funcall(mrb, w2v(wx), id_eq, 1, w2v(wy)));
#else
    return eq(mrb, WIDEVAL_GET(wx), WIDEVAL_GET(wy));
#endif
}

static int
wcmp(mrb_state *mrb, wideval_t wx, wideval_t wy)
{
#if 0 /* 2011_0518 */
    mrb_value x, y;
#if WIDEVALUE_IS_WIDER
    if (FIXWV_P(wx) && FIXWV_P(wy)) {
        wideint_t a, b;
        a = FIXWV2WINT(wx);
        b = FIXWV2WINT(wy);
        if (a < b)
            return -1;
        if (a > b)
            return 1;
        return 0;
    }
#endif
    x = w2v(wx);
    y = w2v(wy);
    return rb_cmpint(mrb, mrb_funcall(mrb, x, "<=>", 1, y), x, y);
#else
  if (wx < wy) return -1;
  if (wx > wy) return 1;
  return 0;
#endif
}

#define wne(mrb,x,y) (!weq((mrb),(x),(y)))
#define wlt(mrb,x,y) (wcmp((mrb),(x),(y)) < 0)
#define wgt(mrb,x,y) (wcmp((mrb),(x),(y)) > 0)
#define wle(mrb,x,y) (wcmp((mrb),(x),(y)) <= 0)
#define wge(mrb,x,y) (wcmp((mrb),(x),(y)) >= 0)

static wideval_t
wadd(mrb_state *mrb, wideval_t wx, wideval_t wy)
{
#if 0 /* 2011_0518 */
    mrb_value x;
#if WIDEVALUE_IS_WIDER
    if (FIXWV_P(wx) && FIXWV_P(wy)) {
        wideint_t r = FIXWV2WINT(wx) + FIXWV2WINT(wy);
        return WINT2WV(r);
    }
    else
#endif
    x = w2v(wx);
    return v2w(mrb_funcall(mrb, x, "+", 1, w2v(wy)));
#else
    return wx+wy;
#endif
}

static wideval_t
wsub(mrb_state *mrb, wideval_t wx, wideval_t wy)
{
#if 0 /* 2011_0518 */
    mrb_value x;
#if WIDEVALUE_IS_WIDER
    if (FIXWV_P(wx) && FIXWV_P(wy)) {
        wideint_t r = FIXWV2WINT(wx) - FIXWV2WINT(wy);
        return WINT2WV(r);
    }
    else
#endif
    x = w2v(wx);
    return v2w(mrb_funcall(mrb, x, "-", 1, w2v(wy)));
#else
    return wx-wy;
#endif
}

#if 0
static int
wi_mul(wideint_t x, wideint_t y, wideint_t *z)
{
    uwideint_t a, b, c;
    int s;
    if (x == 0 || y == 0) {
        *z = 0;
        return 1;
    }
    if (x < 0) {
        s = -1;
        a = (uwideint_t)-x;
    }
    else {
        s = 1;
        a = (uwideint_t)x;
    }
    if (y < 0) {
        s = -s;
        b = (uwideint_t)-y;
    }
    else {
        b = (uwideint_t)y;
    }
    if (a <= UWIDEINT_MAX / b) {
        c = a * b;
        if (s < 0) {
            if (c <= (uwideint_t)WIDEINT_MAX + 1) {
                *z = -(wideint_t)c;
                return 1;
            }
        }
        else {
            if (c <= (uwideint_t)WIDEINT_MAX) {
                *z = (wideint_t)c;
                return 1;
            }
        }
    }
    return 0;
}
#endif

static wideval_t
wmul(mrb_state *mrb, wideval_t wx, wideval_t wy)
{
#if 0 /* 2011_0518 */
    mrb_value x, z;
#if WIDEVALUE_IS_WIDER
    if (FIXWV_P(wx) && FIXWV_P(wy)) {
        wideint_t z;
        if (wi_mul(FIXWV2WINT(wx), FIXWV2WINT(wy), &z))
            return WINT2WV(z);
    }
#endif
    x = w2v(wx);
    z = mrb_funcall(mrb, x, "*", 1, w2v(wy));
#if 0
    if (mrb_type(z) == T_RATIONAL && RRATIONAL(z)->den == mrb_fixnum_value(1)) {
        z = RRATIONAL(z)->num;
    }
#endif
    return v2w(z);
#else
    return wx*wy;
#endif
}

static mrb_float
wmulFloat(mrb_state *mrb, mrb_float wx, mrb_float wy)
{
  return wx*wy;
}

static wideval_t
wquo(mrb_state *mrb, wideval_t wx, wideval_t wy)
{
#if 0 /* 2011_0518 */
    mrb_value x, y, ret;
#if WIDEVALUE_IS_WIDER
    if (FIXWV_P(wx) && FIXWV_P(wy)) {
        wideint_t a, b, c;
        a = FIXWV2WINT(wx);
        b = FIXWV2WINT(wy);
        if (b == 0) rb_num_zerodiv(mrb);
        c = a / b;
        if (c * b == a) {
            return WINT2WV(c);
        }
    }
#endif
    x = w2v(wx);
    y = w2v(wy);
    ret = mrb_funcall(mrb, x, "quo", 1, y);
#if 0
    if (mrb_type(ret) == T_RATIONAL &&
        RRATIONAL(ret)->den == mrb_fixnum_value(1)) {
        ret = RRATIONAL(ret)->num;
    }
#endif
    return v2w(ret);
#else
    return wx/wy;
#endif
}

#define wmulquo(mrb,x,y,z) ((WIDEVAL_GET(y) == WIDEVAL_GET(z)) ? (x) : wquo(mrb, wmul(mrb, (x),(y)),(z)))
#define wmulquoll(mrb,x,y,z) (((y) == (z)) ? (x) : wquo(mrb, wmul(mrb, (x),WINT2WV(y)),WINT2WV(z)))

static void
wdivmod(mrb_state *mrb, wideval_t wn, wideval_t wd, wideval_t *wq, wideval_t *wr)
{
#if 0 /* 2011_0518 */
    mrb_value tmp, ary;
#if WIDEVALUE_IS_WIDER
    if (FIXWV_P(wn) && FIXWV_P(wd)) {
        wideint_t n, d, q, r;
        d = FIXWV2WINT(wd);
        if (d == 0) rb_num_zerodiv(mrb);
        if (d == 1) {
            *wq = wn;
            *wr = WINT2FIXWV(0);
            return;
        }
        if (d == -1) {
            wideint_t xneg = -FIXWV2WINT(wn);
            *wq = WINT2WV(xneg);
            *wr = WINT2FIXWV(0);
            return;
        }
        n = FIXWV2WINT(wn);
        if (n == 0) {
            *wq = WINT2FIXWV(0);
            *wr = WINT2FIXWV(0);
            return;
        }
        if (d < 0) {
            if (n < 0) {
                q = ((-n) / (-d));
                r = ((-n) % (-d));
                if (r != 0) {
                    q -= 1;
                    r += d;
                }
            }
            else { /* 0 < n */
                q = -(n / (-d));
                r = -(n % (-d));
            }
        }
        else { /* 0 < d */
            if (n < 0) {
                q = -((-n) / d);
                r = -((-n) % d);
                if (r != 0) {
                    q -= 1;
                    r += d;
                }
            }
            else { /* 0 < n */
                q = n / d;
                r = n % d;
            }
        }
        *wq = WINT2FIXWV(q);
        *wr = WINT2FIXWV(r);
        return;
    }
#endif
    tmp = mrb_funcall(mrb, w2v(wn), "divmod", 1, w2v(wd));
    ary = mrb_check_array_type(mrb, tmp);
    if (mrb_nil_p(ary)) {
        mrb_raise(mrb, mrb->eTypeError_class, "unexpected divmod result: into %s",
                 mrb_obj_classname(mrb, tmp));
    }
    *wq = v2w(mrb_ary_ref(mrb, ary, 0));
    *wr = v2w(mrb_ary_ref(mrb, ary, 1));
#else
    *wq = wn / wd;
    *wr = wn - (*wq * wd);
    if ((*wr < 0 && wd > 0) || (*wr > 0 && wd < 0)) {
      *wr += wd;
      *wq -= 1;
    }
#endif
}

static void
wmuldivmod(mrb_state *mrb,
           wideval_t wx, wideval_t wy, wideval_t wz, wideval_t *wq, wideval_t *wr)
{
#if 0 /* 2011_0518 */
    if (mrb_obj_equal(mrb, WIDEVAL_GET(wy), WIDEVAL_GET(wz))) {
        *wq = wx;
        *wr = WINT2FIXWV(0);
        return;
    }
    wdivmod(mrb, wmul(mrb, wx,wy), wz, wq, wr);
#else
    if (wy == wz) {
        *wq = wx;
        *wr = 0;
        return;
    }
    *wq = (wx * wy) / wz;
    *wr = (wx * wy) % wz;
#endif
}

static wideval_t
wdiv(mrb_state *mrb, wideval_t wx, wideval_t wy)
{
    wideval_t q, r;
    wdivmod(mrb, wx, wy, &q, &r);
    return q;
}

static wideval_t
wmod(mrb_state *mrb, wideval_t wx, wideval_t wy)
{
    wideval_t q, r;
    wdivmod(mrb, wx, wy, &q, &r);
    return r;
}

static mrb_value
num_exact(mrb_state *mrb, mrb_value v)
{
    mrb_value tmp;
    int t;

    if (mrb_nil_p(v)) goto typeerror;
    t = mrb_type(v);
    switch (t) {
      case MRB_TT_FIXNUM:
      case MRB_TT_FLOAT:/* kusuda */
        return (v);
#if 0
      case MRB_TT_RATIONAL:
        break;
#endif
      case MRB_TT_STRING:
      //case MRB_TT_NIL:
        goto typeerror;

      default:
        //if ((tmp = rb_check_funcall(mrb, v, mrb_intern(mrb, "to_r"), 0, NULL)) != Qundef) {
        //    if (rb_respond_to(v, mrb_intern(mrb, "to_s"))) goto typeerror;
        //    v = tmp;
        //    break;
        //}
        if (mrb_respond_to(mrb, v, mrb_intern(mrb, "to_r"))) {
          v = mrb_funcall(mrb, v, "to_r", 0);
        }
        if (!mrb_nil_p(tmp = rb_check_to_integer(mrb, v, "to_i"))) {
            v = tmp;
            break;
        }
        goto typeerror;
    }

    t = mrb_type(v);
    switch (t) {
      case MRB_TT_FIXNUM:
        return (v);
#if 0
      case MRB_TT_RATIONAL:
        if (RRATIONAL(v)->den == mrb_fixnum_value(1))
            v = RRATIONAL(v)->num;
        break;
#endif
      default:
typeerror:
        mrb_raise(mrb, mrb->eTypeError_class, "can't convert %s into an exact number",
                                mrb_nil_p(v) ? "nil" : mrb_obj_classname(mrb, v));
    }
    return (v);
}

static long long
long_num_exact(mrb_state *mrb, mrb_value v)
{
  mrb_value tmp;

  tmp = num_exact(mrb, v);
  if (mrb_nil_p(tmp)) {
    mrb_raise(mrb, mrb->eTypeError_class, "can't convert %s into an exact number",
                                "nil" );
  }
  if (FIXNUM_P(tmp)) {
    return mrb_fixnum_number(tmp);
  }
  else {
    return (long long)(mrb_float_number(tmp));
  }
}
/* time_t */

#ifndef TYPEOF_TIMEVAL_TV_SEC
# define TYPEOF_TIMEVAL_TV_SEC time_t
#endif
#ifndef TYPEOF_TIMEVAL_TV_USEC
# if INT_MAX >= 1000000
# define TYPEOF_TIMEVAL_TV_USEC int
# else
# define TYPEOF_TIMEVAL_TV_USEC long
# endif
#endif

#if SIZEOF_TIME_T == SIZEOF_LONG
typedef unsigned long unsigned_time_t;
#elif SIZEOF_TIME_T == SIZEOF_INT
typedef unsigned int unsigned_time_t;
#elif SIZEOF_TIME_T == SIZEOF_LONG_LONG
typedef unsigned LONG_LONG unsigned_time_t;
#else
# error cannot find integer type which size is same as time_t.
#endif

#define TIMET_MAX (~(time_t)0 <= 0 ? (time_t)((~(unsigned_time_t)0U) >> 1) : (unsigned_time_t)(~(unsigned_time_t)0U))
#define TIMET_MIN (~(time_t)0 <= 0 ? (time_t)(((unsigned_time_t)1U) << (sizeof(unsigned_time_t) * CHAR_BIT - 1)) : (unsigned_time_t)0U)

static wideval_t
rb_time_magnify(mrb_state *mrb, wideval_t w)
{
#if 0 /* 2011_0518 */
    if (FIXWV_P(w)) {
        wideint_t z;
        if (wi_mul(FIXWV2WINT(w), TIME_SCALE, &z))
            return WINT2WV(z);
    }
    return wmul(mrb, w, WINT2FIXWV(TIME_SCALE));
#else
    return wmul(mrb, w, TIME_SCALE);
#endif
}

static wideval_t
rb_time_magnifyFromMrbValue(mrb_state *mrb, mrb_value vw)
{
#if 0 /* 2011_0518 */
    if (FIXWV_P(w)) {
        wideint_t z;
        if (wi_mul(FIXWV2WINT(w), TIME_SCALE, &z))
            return WINT2WV(z);
    }
    return wmul(mrb, w, WINT2FIXWV(TIME_SCALE));
#else
  mrb_float w;
  if (FIXNUM_P(vw)) {
    w = mrb_fixnum_number(vw);
  }
  else {
    w = mrb_float_number(vw);
  }
  return wmulFloat(mrb, w, TIME_SCALE);
#endif
}
static wideval_t
rb_time_unmagnify(mrb_state *mrb, wideval_t w)
{
#if 0 /* 2011_0518 */
#if WIDEVALUE_IS_WIDER
    if (FIXWV_P(w)) {
        wideint_t a, b, c;
        a = FIXWV2WINT(w);
        b = TIME_SCALE;
        c = a / b;
        if (c * b == a) {
            return WINT2FIXWV(c);
        }
    }
#endif
    return wquo(mrb, w, WINT2FIXWV(TIME_SCALE));
#else
    return wquo(mrb, w, TIME_SCALE);
#endif
}

static mrb_value
rb_time_unmagnify_to_float(mrb_state *mrb, wideval_t w)
{
#if 0 /* 2011_0518 */
    mrb_value v;
#if WIDEVALUE_IS_WIDER
    if (FIXWV_P(w)) {
        wideint_t a, b, c;
        a = FIXWV2WINT(w);
        b = TIME_SCALE;
        c = a / b;
        if (c * b == a) {
            return DBL2NUM((double)c);
        }
        v = mrb_float_value((double)FIXWV2WINT(w));
        return quo(mrb, v, mrb_float_value(TIME_SCALE));
    }
#endif
    v = w2v(w);
    return quo(mrb, v, mrb_float_value(TIME_SCALE));
#else
    return mrb_float_value(quo(mrb, w, TIME_SCALE));
#endif
}

static void
split_second(mrb_state *mrb, wideval_t timew, wideval_t *timew_p, int *subsecx_p)
{
    wideval_t q, r;
    wdivmod(mrb, timew, WINT2FIXWV(TIME_SCALE), &q, &r);
    *timew_p = q;
    *subsecx_p = w2v(r);
}

static wideval_t
timet2wv(time_t t)
{
#if WIDEVALUE_IS_WIDER
    if (TIMET_MIN == 0) {
        uwideint_t wi = (uwideint_t)t;
        if (wi <= FIXWV_MAX) {
            return WINT2FIXWV(wi);
        }
    }
    else {
        wideint_t wi = (wideint_t)t;
        if (FIXWV_MIN <= wi && wi <= FIXWV_MAX) {
            return WINT2FIXWV(wi);
        }
    }
#endif
    //return v2w(TIMET2NUM(t));
    return t;
}
#define TIMET2WV(t) timet2wv(t)

static time_t
wv2timet(wideval_t w)
{
#if WIDEVALUE_IS_WIDER
    if (FIXWV_P(w)) {
        wideint_t wi = FIXWV2WINT(w);
        if (TIMET_MIN == 0) {
            if (wi < 0)
                mrb_raise(mrb, mrb->eRangeError_class, "negative value to convert into `time_t'");
            if (TIMET_MAX < (uwideint_t)wi)
                mrb_raise(mrb, mrb->eRangeError_class, "too big to convert into `time_t'");
        }
        else {
            if (wi < TIMET_MIN || TIMET_MAX < wi)
                mrb_raise(mrb, mrb->eRangeError_class, "too big to convert into `time_t'");
        }
        return (time_t)wi;
    }
#endif
    //return NUM2TIMET(w2v(w));
    return (w2v(w));
}
#define WV2TIMET(t) wv2timet(t)

//mrb_value rb_cTime;
static mrb_value time_utc_offset _((mrb_state *mrb, mrb_value));

static int obj2int(mrb_state *mrb, mrb_value obj);
static mrb_value obj2vint(mrb_state *mrb, mrb_value obj);
static int month_arg(mrb_state *mrb, mrb_value arg);
static void validate_utc_offset(mrb_state *mrb, mrb_value utc_offset);
static void validate_vtm(mrb_state *mrb, struct vtm *vtm);

static mrb_value time_gmtime(mrb_state *mrb, mrb_value);
static mrb_value time_localtime(mrb_state *mrb, mrb_value);
static mrb_value time_fixoff(mrb_state *mrb, mrb_value);

static time_t timegm_noleapsecond(struct tm *tm);
static int tmcmp(struct tm *a, struct tm *b);
static int vtmcmp(mrb_state *mrb, struct vtm *a, struct vtm *b);
static const char *find_time_t(struct tm *tptr, int utc_p, time_t *tp);

static struct vtm *localtimew(mrb_state *mrb, wideval_t timew, struct vtm *result);

static int leap_year_p(long y);
#define leap_year_v_p(mrb,y) leap_year_p((mod((mrb), mrb_fixnum_number(y), (400))))

#ifdef HAVE_GMTIME_R
#define rb_gmtime_r(t, tm) gmtime_r((t), (tm))
#define rb_localtime_r(t, tm) localtime_r((t), (tm))
#else
static inline struct tm *
rb_gmtime_r(const time_t *tp, struct tm *result)
{
    struct tm *t = gmtime(tp);
    if (t) *result = *t;
    return t;
}

static inline struct tm *
rb_localtime_r(const time_t *tp, struct tm *result)
{
    struct tm *t = localtime(tp); /* *tp=0 means 1970-01-01 00:00:00(GMT) ...if *tp < 0, localtime() can't conver. */
    if (t) *result = *t;
    return t;
}
#endif

static struct tm *
rb_localtime_r2(const time_t *t, struct tm *result)
{
#if defined __APPLE__ && defined __LP64__
    if (*t != (time_t)(int)*t) return NULL;
#endif
    result = rb_localtime_r(t, result);
#if defined(HAVE_MKTIME) && defined(LOCALTIME_OVERFLOW_PROBLEM)
    if (result) {
        int gmtoff1 = 0;
        int gmtoff2 = 0;
        struct tm tmp = *result;
        time_t t2;
#  if defined(HAVE_STRUCT_TM_TM_GMTOFF)
        gmtoff1 = result->tm_gmtoff;
#  endif
        t2 = mktime(&tmp);
#  if defined(HAVE_STRUCT_TM_TM_GMTOFF)
        gmtoff2 = tmp.tm_gmtoff;
#  endif
        if (*t + gmtoff1 != t2 + gmtoff2)
            result = NULL;
    }
#endif
    return result;
}
#define LOCALTIME(tm, result) (tzset(),rb_localtime_r2((tm), &(result)))

#if !defined(HAVE_STRUCT_TM_TM_GMTOFF)
    static struct tm *
    rb_gmtime_r2(const time_t *t, struct tm *result)
    {
        result = rb_gmtime_r(t, result);
#if defined(HAVE_TIMEGM) && defined(LOCALTIME_OVERFLOW_PROBLEM)
        if (result) {
            struct tm tmp = *result;
            time_t t2 = timegm(&tmp);
            if (*t != t2)
                result = NULL;
        }
#endif
        return result;
    }
#   define GMTIME(tm, result) rb_gmtime_r2((tm), &(result))
#endif

static const int common_year_yday_offset[] = {
    -1,
    -1 + 31,
    -1 + 31 + 28,
    -1 + 31 + 28 + 31,
    -1 + 31 + 28 + 31 + 30,
    -1 + 31 + 28 + 31 + 30 + 31,
    -1 + 31 + 28 + 31 + 30 + 31 + 30,
    -1 + 31 + 28 + 31 + 30 + 31 + 30 + 31,
    -1 + 31 + 28 + 31 + 30 + 31 + 30 + 31 + 31,
    -1 + 31 + 28 + 31 + 30 + 31 + 30 + 31 + 31 + 30,
    -1 + 31 + 28 + 31 + 30 + 31 + 30 + 31 + 31 + 30 + 31,
    -1 + 31 + 28 + 31 + 30 + 31 + 30 + 31 + 31 + 30 + 31 + 30
      /* 1    2    3    4    5    6    7    8    9    10   11 */
};
static const int leap_year_yday_offset[] = {
    -1,
    -1 + 31,
    -1 + 31 + 29,
    -1 + 31 + 29 + 31,
    -1 + 31 + 29 + 31 + 30,
    -1 + 31 + 29 + 31 + 30 + 31,
    -1 + 31 + 29 + 31 + 30 + 31 + 30,
    -1 + 31 + 29 + 31 + 30 + 31 + 30 + 31,
    -1 + 31 + 29 + 31 + 30 + 31 + 30 + 31 + 31,
    -1 + 31 + 29 + 31 + 30 + 31 + 30 + 31 + 31 + 30,
    -1 + 31 + 29 + 31 + 30 + 31 + 30 + 31 + 31 + 30 + 31,
    -1 + 31 + 29 + 31 + 30 + 31 + 30 + 31 + 31 + 30 + 31 + 30
      /* 1    2    3    4    5    6    7    8    9    10   11 */
};

static const int common_year_days_in_month[] = {
    31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};
static const int leap_year_days_in_month[] = {
    31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

static int
calc_tm_yday(long tm_year, int tm_mon, int tm_mday)
{
    int tm_year_mod400;
    int tm_yday = tm_mday;

    tm_year_mod400 = MOD(tm_year, 400);

    if (leap_year_p(tm_year_mod400 + 1900))
        tm_yday += leap_year_yday_offset[tm_mon];
    else
        tm_yday += common_year_yday_offset[tm_mon];

    return tm_yday;
}

static wideval_t
timegmw_noleapsecond(mrb_state *mrb, struct vtm *vtm)
{
    int year1900;
    long q400, r400;
    int year_mod400;
    int yday;
    long days_in400;
    long vdays, ret;
    wideval_t wret;

    year1900 = sub(mrb, mrb_fixnum_number(vtm->year), (1900));

    divmodv(mrb, year1900, (400), &q400, &r400);
    year_mod400 = (r400);

    yday = calc_tm_yday(year_mod400, vtm->mon-1, vtm->mday);

    /*
     *  `Seconds Since the Epoch' in SUSv3:
     *  tm_sec + tm_min*60 + tm_hour*3600 + tm_yday*86400 +
     *  (tm_year-70)*31536000 + ((tm_year-69)/4)*86400 -
     *  ((tm_year-1)/100)*86400 + ((tm_year+299)/400)*86400
     */
    ret = (vtm->sec
                 + vtm->min*60
                 + vtm->hour*3600);
    days_in400 = yday
               - 70*365
               + DIV(year_mod400 - 69, 4)
               - DIV(year_mod400 - 1, 100)
               + (year_mod400 + 299) / 400;
    vdays = (days_in400);
    vdays = add(mrb, vdays, mul(mrb, q400, (97)));
    vdays = add(mrb, vdays, mul(mrb, year1900, (365)));
    wret = wadd(mrb, rb_time_magnify(mrb, v2w(ret)), wmul(mrb, rb_time_magnify(mrb, v2w(vdays)), WINT2FIXWV(86400)));
    wret = wadd(mrb, wret, v2w(vtm->subsecx));

    return wret;
}

static st_table *zone_table;

static const char *
zone_str(const char *s)
{
    st_data_t k, v;

    if (!zone_table)
        zone_table = st_init_strtable();

    k = (st_data_t)s;
    if (st_lookup(zone_table, k, &v)) {
        return (const char *)v;
    }
    s = ruby_strdup(s);
    k = (st_data_t)s;
    st_add_direct(zone_table, k, k);

    return s;
}

static void
gmtimew_noleapsecond(mrb_state *mrb, wideval_t timew, struct vtm *vtm)
{
    long v;
    int i, n, x, y;
    const int *yday_offset;
    int wday;
    long timev;
    wideval_t timew2, w, w2;

    vtm->isdst = 0;

    split_second(mrb, timew, &timew2, &vtm->subsecx);

    wdivmod(mrb, timew2, WINT2FIXWV(86400), &w2, &w);
    timev = w2v(w2);
    v = w2v(w);

    wday = (mod(mrb, timev, (7)));
    vtm->wday = (wday + 4) % 7;

    n = (v);
    vtm->sec = n % 60; n = n / 60;
    vtm->min = n % 60; n = n / 60;
    vtm->hour = n;

    /* 97 leap days in the 400 year cycle */
    divmodv(mrb, timev, (400*365 + 97), &timev, &v);
    vtm->year = mrb_fixnum_value(mul(mrb, timev, (400)));

    /* n is the days in the 400 year cycle.
     * the start of the cycle is 1970-01-01. */

    n = (v);
    y = 1970;

    /* 30 years including 7 leap days (1972, 1976, ... 1996),
     * 31 days in January 2000 and
     * 29 days in February 2000
     * from 1970-01-01 to 2000-02-29 */
    if (30*365+7+31+29-1 <= n) {
        /* 2000-02-29 or after */
        if (n < 31*365+8) {
            /* 2000-02-29 to 2000-12-31 */
            y += 30;
            n -= 30*365+7;
            goto found;
        }
        else {
            /* 2001-01-01 or after */
            n -= 1;
        }
    }

    x = n / (365*100 + 24);
    n = n % (365*100 + 24);
    y += x * 100;
    if (30*365+7+31+29-1 <= n) {
        if (n < 31*365+7) {
            y += 30;
            n -= 30*365+7;
            goto found;
        }
        else
            n += 1;
    }

    x = n / (365*4 + 1);
    n = n % (365*4 + 1);
    y += x * 4;
    if (365*2+31+29-1 <= n) {
        if (n < 365*2+366) {
            y += 2;
            n -= 365*2;
            goto found;
        }
        else
            n -= 1;
    }

    x = n / 365;
    n = n % 365;
    y += x;

  found:
    vtm->yday = n+1;
    vtm->year = mrb_fixnum_value(add(mrb, mrb_fixnum_number(vtm->year), (y)));

    if (leap_year_p(y))
        yday_offset = leap_year_yday_offset;
    else
        yday_offset = common_year_yday_offset;

    for (i = 0; i < 12; i++) {
        if (yday_offset[i] < n) {
            vtm->mon = i+1;
            vtm->mday = n - yday_offset[i];
        }
        else
            break;
    }

    vtm->utc_offset = mrb_fixnum_value(0);
    vtm->zone = "UTC";
}

static struct tm *
gmtime_with_leapsecond(const time_t *timep, struct tm *result)
{
#if defined(HAVE_STRUCT_TM_TM_GMTOFF)
    /* 4.4BSD counts leap seconds only with localtime, not with gmtime. */
    struct tm *t;
    int sign;
    int gmtoff_sec, gmtoff_min, gmtoff_hour, gmtoff_day;
    long gmtoff;
    t = LOCALTIME(timep, *result);
    if (t == NULL)
        return NULL;

    /* subtract gmtoff */
    if (t->tm_gmtoff < 0) {
        sign = 1;
        gmtoff = -t->tm_gmtoff;
    }
    else {
        sign = -1;
        gmtoff = t->tm_gmtoff;
    }
    gmtoff_sec = (int)(gmtoff % 60);
    gmtoff = gmtoff / 60;
    gmtoff_min = (int)(gmtoff % 60);
    gmtoff = gmtoff / 60;
    gmtoff_hour = (int)gmtoff;        /* <= 12 */

    gmtoff_sec *= sign;
    gmtoff_min *= sign;
    gmtoff_hour *= sign;

    gmtoff_day = 0;

    if (gmtoff_sec) {
        /* If gmtoff_sec == 0, don't change result->tm_sec.
         * It may be 60 which is a leap second. */
        result->tm_sec += gmtoff_sec;
        if (result->tm_sec < 0) {
            result->tm_sec += 60;
            gmtoff_min -= 1;
        }
        if (60 <= result->tm_sec) {
            result->tm_sec -= 60;
            gmtoff_min += 1;
        }
    }
    if (gmtoff_min) {
        result->tm_min += gmtoff_min;
        if (result->tm_min < 0) {
            result->tm_min += 60;
            gmtoff_hour -= 1;
        }
        if (60 <= result->tm_min) {
            result->tm_min -= 60;
            gmtoff_hour += 1;
        }
    }
    if (gmtoff_hour) {
        result->tm_hour += gmtoff_hour;
        if (result->tm_hour < 0) {
            result->tm_hour += 24;
            gmtoff_day = -1;
        }
        if (24 <= result->tm_hour) {
            result->tm_hour -= 24;
            gmtoff_day = 1;
        }
    }

    if (gmtoff_day) {
        if (gmtoff_day < 0) {
            if (result->tm_yday == 0) {
                result->tm_mday = 31;
                result->tm_mon = 11; /* December */
                result->tm_year--;
                result->tm_yday = leap_year_p(result->tm_year + 1900) ? 365 : 364;
            }
            else if (result->tm_mday == 1) {
                const int *days_in_month = leap_year_p(result->tm_year + 1900) ?
                                           leap_year_days_in_month :
                                           common_year_days_in_month;
                result->tm_mon--;
                result->tm_mday = days_in_month[result->tm_mon];
                result->tm_yday--;
            }
            else {
                result->tm_mday--;
                result->tm_yday--;
            }
            result->tm_wday = (result->tm_wday + 6) % 7;
        }
        else {
            int leap = leap_year_p(result->tm_year + 1900);
            if (result->tm_yday == (leap ? 365 : 364)) {
                result->tm_year++;
                result->tm_mon = 0; /* January */
                result->tm_mday = 1;
                result->tm_yday = 0;
            }
            else if (result->tm_mday == (leap ? leap_year_days_in_month :
                                                common_year_days_in_month)[result->tm_mon]) {
                result->tm_mon++;
                result->tm_mday = 1;
                result->tm_yday++;
            }
            else {
                result->tm_mday++;
                result->tm_yday++;
            }
            result->tm_wday = (result->tm_wday + 1) % 7;
        }
    }
    result->tm_isdst = 0;
    result->tm_gmtoff = 0;
#if defined(HAVE_TM_ZONE)
    result->tm_zone = (char *)"UTC";
#endif
    return result;
#else
    return GMTIME(timep, *result);
#endif
}

static long this_year = 0;
static time_t known_leap_seconds_limit;
static int number_of_leap_seconds_known;

static void
init_leap_second_info(mrb_state *mrb)
{
    /*
     * leap seconds are determined by IERS.
     * It is announced 6 months before the leap second.
     * So no one knows leap seconds in the future after the next year.
     */
    if (this_year == 0) {
        time_t now;
        struct tm *tm, result;
        struct vtm vtm;
        wideval_t timew;
        now = time(NULL);
        gmtime(&now);
        tm = gmtime_with_leapsecond(&now, &result);
        if (!tm) return;
        this_year = tm->tm_year;

        if (TIMET_MAX - now < (time_t)(366*86400))
            known_leap_seconds_limit = TIMET_MAX;
        else
            known_leap_seconds_limit = now + (time_t)(366*86400);

        if (!gmtime_with_leapsecond(&known_leap_seconds_limit, &result))
            return;

        vtm.year = mrb_fixnum_value(result.tm_year + 1900);
        vtm.mon = result.tm_mon + 1;
        vtm.mday = result.tm_mday;
        vtm.hour = result.tm_hour;
        vtm.min = result.tm_min;
        vtm.sec = result.tm_sec;
        vtm.subsecx = (0);
        vtm.utc_offset = mrb_fixnum_value(0);

        timew = timegmw_noleapsecond(mrb, &vtm);

        number_of_leap_seconds_known = (w2v(wsub(mrb, TIMET2WV(known_leap_seconds_limit), rb_time_unmagnify(mrb, timew))));
    }
}

static wideval_t
timegmw(mrb_state *mrb, struct vtm *vtm)
{
    wideval_t timew;
    struct tm tm;
    time_t t;
    const char *errmsg;

    /* The first leap second is 1972-06-30 23:59:60 UTC.
     * No leap seconds before. */
    if (gt(mrb, (1972), mrb_fixnum_number(vtm->year)))
        return timegmw_noleapsecond(mrb, vtm);

    init_leap_second_info(mrb);

    timew = timegmw_noleapsecond(mrb, vtm);

    if (wlt(mrb, rb_time_magnify(mrb, TIMET2WV(known_leap_seconds_limit)), timew)) {
        return wadd(mrb, timew, rb_time_magnify(mrb, WINT2WV(number_of_leap_seconds_known)));
    }

    tm.tm_year = rb_long2int(mrb_fixnum_number(vtm->year) - 1900);
    tm.tm_mon = vtm->mon - 1;
    tm.tm_mday = vtm->mday;
    tm.tm_hour = vtm->hour;
    tm.tm_min = vtm->min;
    tm.tm_sec = vtm->sec;
    tm.tm_isdst = 0;

    errmsg = find_time_t(&tm, 1, &t);
    if (errmsg)
        mrb_raise(mrb, mrb->eArgumentError_class, "%s", errmsg);
    return wadd(mrb, rb_time_magnify(mrb, TIMET2WV(t)), v2w(vtm->subsecx));
}

static struct vtm *
gmtimew(mrb_state *mrb, wideval_t timew, struct vtm *result)
{
    time_t t;
    struct tm tm;
    int subsecx;
    wideval_t timew2;

    if (wlt(mrb, timew, WINT2FIXWV(0))) {
        gmtimew_noleapsecond(mrb, timew, result);
        return result;
    }

    init_leap_second_info(mrb);

    if (wlt(mrb, rb_time_magnify(mrb, TIMET2WV(known_leap_seconds_limit)), timew)) {
        timew = wsub(mrb, timew, rb_time_magnify(mrb, WINT2WV(number_of_leap_seconds_known)));
        gmtimew_noleapsecond(mrb, timew, result);
        return result;
    }

    split_second(mrb, timew, &timew2, &subsecx);

    t = WV2TIMET(timew2);
    if (!gmtime_with_leapsecond(&t, &tm))
        return NULL;

    result->year = mrb_fixnum_value((long)tm.tm_year + 1900);
    result->mon = tm.tm_mon + 1;
    result->mday = tm.tm_mday;
    result->hour = tm.tm_hour;
    result->min = tm.tm_min;
    result->sec = tm.tm_sec;
    result->subsecx = subsecx;
    result->utc_offset = mrb_fixnum_value(0);
    result->wday = tm.tm_wday;
    result->yday = tm.tm_yday+1;
    result->isdst = tm.tm_isdst;
    result->zone = "UTC";

    return result;
}

static struct tm *localtime_with_gmtoff(const time_t *t, struct tm *result, long *gmtoff);

/*
 * The idea is come from Perl:
 * http://use.perl.org/articles/08/02/07/197204.shtml
 *
 * compat_common_month_table is generated by following program.
 * This table finds the last month which start the same day of a week.
 * The year 2037 is not used because
 * http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=522949
 *
 *  #!/usr/bin/ruby
 *
 *  require 'date'
 *
 *  h = {}
 *  2036.downto(2010) {|y|
 *    1.upto(12) {|m|
 *      next if m == 2 && y % 4 == 0
 *      d = Date.new(y,m,1)
 *      h[m] ||= {}
 *      h[m][d.wday] ||= y
 *    }
 *  }
 *
 *  1.upto(12) {|m|
 *    print "{"
 *    0.upto(6) {|w|
 *      y = h[m][w]
 *      print " #{y},"
 *    }
 *    puts "},"
 *  }
 *
 */
static int compat_common_month_table[12][7] = {
  /* Sun   Mon   Tue   Wed   Thu   Fri   Sat */
  { 2034, 2035, 2036, 2031, 2032, 2027, 2033 }, /* January */
  { 2026, 2027, 2033, 2034, 2035, 2030, 2031 }, /* February */
  { 2026, 2032, 2033, 2034, 2035, 2030, 2036 }, /* March */
  { 2035, 2030, 2036, 2026, 2032, 2033, 2034 }, /* April */
  { 2033, 2034, 2035, 2030, 2036, 2026, 2032 }, /* May */
  { 2036, 2026, 2032, 2033, 2034, 2035, 2030 }, /* June */
  { 2035, 2030, 2036, 2026, 2032, 2033, 2034 }, /* July */
  { 2032, 2033, 2034, 2035, 2030, 2036, 2026 }, /* August */
  { 2030, 2036, 2026, 2032, 2033, 2034, 2035 }, /* September */
  { 2034, 2035, 2030, 2036, 2026, 2032, 2033 }, /* October */
  { 2026, 2032, 2033, 2034, 2035, 2030, 2036 }, /* November */
  { 2030, 2036, 2026, 2032, 2033, 2034, 2035 }, /* December */
};

/*
 * compat_leap_month_table is generated by following program.
 *
 *  #!/usr/bin/ruby
 *
 *  require 'date'
 *
 *  h = {}
 *  2037.downto(2010) {|y|
 *    1.upto(12) {|m|
 *      next unless m == 2 && y % 4 == 0
 *      d = Date.new(y,m,1)
 *      h[m] ||= {}
 *      h[m][d.wday] ||= y
 *    }
 *  }
 *
 *  2.upto(2) {|m|
 *    0.upto(6) {|w|
 *      y = h[m][w]
 *      print " #{y},"
 *    }
 *    puts
 *  }
 */
static int compat_leap_month_table[7] = {
/* Sun   Mon   Tue   Wed   Thu   Fri   Sat */
  2032, 2016, 2028, 2012, 2024, 2036, 2020, /* February */
};

static int
calc_wday(int year, int month, int day)
{
    int a, y, m;
    int wday;

    a = (14 - month) / 12;
    y = year + 4800 - a;
    m = month + 12 * a - 3;
    wday = day + (153*m+2)/5 + 365*y + y/4 - y/100 + y/400 + 2;
    wday = wday % 7;
    return wday;
}

static wideval_t
guess_local_offset(mrb_state *mrb, struct vtm *vtm_utc)
{
    wideval_t off = (0);
    struct tm tm;
    long gmtoff;
    time_t t;
    struct vtm vtm2;
    //mrb_value timev;
    int y, wday;

# if defined(NEGATIVE_TIME_T)
    /* 1901-12-13 20:45:52 UTC : The oldest time in 32-bit signed time_t. */
    if (localtime_with_gmtoff((t = (time_t)0x80000000, &t), &tm, &gmtoff))
        off = (gmtoff);
    else
# endif
    /* 1970-01-01 00:00:00 UTC : The Unix epoch - the oldest time in portable time_t. */
    if (localtime_with_gmtoff((t = 0, &t), &tm, &gmtoff))
        off = (gmtoff);

    /* The first DST is at 1916 in German.
     * So we don't need to care DST before that. */
    if (lt(mrb, mrb_fixnum_number(vtm_utc->year), (1916)))
        return off;

    /* It is difficult to guess future. */

    vtm2 = *vtm_utc;

    /* guess using a year before 2038. */
    y = (mod(mrb, mrb_fixnum_number(vtm_utc->year), (400)));
    wday = calc_wday(y, vtm_utc->mon, 1);
    if (vtm_utc->mon == 2 && leap_year_p(y))
        vtm2.year = mrb_fixnum_value(compat_leap_month_table[wday]);
    else
        vtm2.year = mrb_fixnum_value(compat_common_month_table[vtm_utc->mon-1][wday]);

    //timev = w2v(rb_time_unmagnify(mrb, timegmw(mrb, &vtm2)));
    //t = NUM2TIMET(timev);
    t = w2v(rb_time_unmagnify(mrb, timegmw(mrb, &vtm2)));

    if (localtime_with_gmtoff(&t, &tm, &gmtoff))
        return (gmtoff);

    {
        /* Use the current time offset as a last resort. */
        static time_t now = 0;
        static long now_gmtoff = 0;
        if (now == 0) {
            now = time(NULL);
            localtime_with_gmtoff(&now, &tm, &now_gmtoff);
        }
        return (now_gmtoff);
    }
}

static wideval_t
small_vtm_sub(mrb_state *mrb, struct vtm *vtm1, struct vtm *vtm2)
{
    int off;
    int year1 = mrb_fixnum_number(vtm1->year);
    int year2 = mrb_fixnum_number(vtm2->year);

    off = vtm1->sec - vtm2->sec;
    off += (vtm1->min - vtm2->min) * 60;
    off += (vtm1->hour - vtm2->hour) * 3600;
    if (ne(mrb, year1, year2))
        off += lt(mrb, year1, year2) ? -24*3600 : 24*3600;
    else if (vtm1->mon != vtm2->mon)
        off += vtm1->mon < vtm2->mon ? -24*3600 : 24*3600;
    else if (vtm1->mday != vtm2->mday)
        off += vtm1->mday < vtm2->mday ? -24*3600 : 24*3600;

    return (off);
}

static wideval_t
timelocalw(mrb_state *mrb, struct vtm *vtm)
{
    time_t t;
    struct tm tm;
    long v;
    wideval_t timew1, timew2;
    struct vtm vtm1, vtm2;
    int n;

    if (FIXNUM_P(vtm->year)) {
        long l = mrb_fixnum_number(vtm->year) - 1900;
        if (l < INT_MIN || INT_MAX < l)
            goto no_localtime;
        tm.tm_year = (int)l;
    }
    else {
        v = sub(mrb, mrb_fixnum_number(vtm->year), (1900));
        if (lt(mrb, v, (INT_MIN)) || lt(mrb, (INT_MAX), v))
            goto no_localtime;
        tm.tm_year = (v);
    }

    tm.tm_mon = vtm->mon-1;
    tm.tm_mday = vtm->mday;
    tm.tm_hour = vtm->hour;
    tm.tm_min = vtm->min;
    tm.tm_sec = vtm->sec;
    tm.tm_isdst = vtm->isdst;

    if (find_time_t(&tm, 0, &t))
        goto no_localtime;
    return wadd(mrb, rb_time_magnify(mrb, TIMET2WV(t)), v2w(vtm->subsecx));

  no_localtime:
    timew1 = timegmw(mrb, vtm);

    if (!localtimew(mrb, timew1, &vtm1))
        mrb_raise(mrb, mrb->eArgumentError_class, "localtimew error");

    n = vtmcmp(mrb, vtm, &vtm1);
    if (n == 0) {
        timew1 = wsub(mrb, timew1, rb_time_magnify(mrb, WINT2FIXWV(12*3600)));
        if (!localtimew(mrb, timew1, &vtm1))
            mrb_raise(mrb, mrb->eArgumentError_class, "localtimew error");
        n = 1;
    }

    if (n < 0) {
        timew2 = timew1;
        vtm2 = vtm1;
        timew1 = wsub(mrb, timew1, rb_time_magnify(mrb, WINT2FIXWV(24*3600)));
        if (!localtimew(mrb, timew1, &vtm1))
            mrb_raise(mrb, mrb->eArgumentError_class, "localtimew error");
    }
    else {
        timew2 = wadd(mrb, timew1, rb_time_magnify(mrb, WINT2FIXWV(24*3600)));
        if (!localtimew(mrb, timew2, &vtm2))
            mrb_raise(mrb, mrb->eArgumentError_class, "localtimew error");
    }
    timew1 = wadd(mrb, timew1, rb_time_magnify(mrb, v2w(small_vtm_sub(mrb, vtm, &vtm1))));
    timew2 = wadd(mrb, timew2, rb_time_magnify(mrb, v2w(small_vtm_sub(mrb, vtm, &vtm2))));

    if (weq(mrb, timew1, timew2))
        return timew1;

    if (!localtimew(mrb, timew1, &vtm1))
        mrb_raise(mrb, mrb->eArgumentError_class, "localtimew error");
    if (vtm->hour != vtm1.hour || vtm->min != vtm1.min || vtm->sec != vtm1.sec)
        return timew2;

    if (!localtimew(mrb, timew2, &vtm2))
        mrb_raise(mrb, mrb->eArgumentError_class, "localtimew error");
    if (vtm->hour != vtm2.hour || vtm->min != vtm2.min || vtm->sec != vtm2.sec)
        return timew1;

    if (vtm->isdst)
        return lt(mrb, mrb_fixnum_number(vtm1.utc_offset), mrb_fixnum_number(vtm2.utc_offset)) ? timew2 : timew1;
    else
        return lt(mrb, mrb_fixnum_number(vtm1.utc_offset), mrb_fixnum_number(vtm2.utc_offset)) ? timew1 : timew2;
}

static struct tm *
localtime_with_gmtoff(const time_t *t, struct tm *result, long *gmtoff)
{
    struct tm tm;

    if (LOCALTIME(t, tm)) {
#if defined(HAVE_STRUCT_TM_TM_GMTOFF)
        *gmtoff = tm.tm_gmtoff;
#else
        struct tm *u, *l;
        long off;
        struct tm tmbuf;
        l = &tm;
        u = GMTIME(t, tmbuf);
        if (!u)
            return NULL;
        if (l->tm_year != u->tm_year)
            off = l->tm_year < u->tm_year ? -1 : 1;
        else if (l->tm_mon != u->tm_mon)
            off = l->tm_mon < u->tm_mon ? -1 : 1;
        else if (l->tm_mday != u->tm_mday)
            off = l->tm_mday < u->tm_mday ? -1 : 1;
        else
            off = 0;
        off = off * 24 + l->tm_hour - u->tm_hour;
        off = off * 60 + l->tm_min - u->tm_min;
        off = off * 60 + l->tm_sec - u->tm_sec;
        *gmtoff = off;
#endif
        *result = tm;
        return result;
    }
    return NULL;
}

static int
timew_out_of_timet_range(mrb_state *mrb, wideval_t timew)
{
    long long timexv;
    long long time_minimum;
    long long time_maximum;
#if WIDEVALUE_IS_WIDER && SIZEOF_TIME_T < SIZEOF_INT64_T
    if (FIXWV_P(timew)) {
        wideint_t t = FIXWV2WINT(timew);
        if (t < TIME_SCALE * (wideint_t)TIMET_MIN ||
            TIME_SCALE * (1 + (wideint_t)TIMET_MAX) <= t)
            return 1;
        return 0;
    }
#endif
    timexv = w2v(timew);
    time_minimum = TIME_SCALE*TIMET_MIN;
    time_maximum = TIME_SCALE*(TIMET_MAX+1);
    if (timexv < time_minimum)
        return 1; /* NG */
    if (time_maximum <= timexv)
        return 1;
    return 0; /* OK */
}

static struct vtm *
localtimew(mrb_state *mrb, wideval_t timew, struct vtm *result)
{
    int subsecx;
    wideval_t offset;

    if (!timew_out_of_timet_range(mrb, timew)) {
        time_t t;
        struct tm tm;
        long gmtoff;
        wideval_t timew2;

        split_second(mrb, timew, &timew2, &subsecx);

        t = WV2TIMET(timew2);

        if (localtime_with_gmtoff(&t, &tm, &gmtoff)) {
            result->year = mrb_fixnum_value((long)tm.tm_year + 1900);
            result->mon = tm.tm_mon + 1;
            result->mday = tm.tm_mday;
            result->hour = tm.tm_hour;
            result->min = tm.tm_min;
            result->sec = tm.tm_sec;
            result->subsecx = subsecx;
            result->wday = tm.tm_wday;
            result->yday = tm.tm_yday+1;
            result->isdst = tm.tm_isdst;
            result->utc_offset = mrb_fixnum_value(gmtoff);
#if defined(HAVE_TM_ZONE)
            result->zone = zone_str(tm.tm_zone);
#elif defined(HAVE_TZNAME) && defined(HAVE_DAYLIGHT)
            /* this needs tzset or localtime, instead of localtime_r */
            result->zone = zone_str(tzname[daylight && tm.tm_isdst]);
#else
            {
                char buf[64];
                strftime(buf, sizeof(buf), "%Z", &tm);
                result->zone = zone_str(buf);
            }
#endif

            return result;
        }
    }

    if (!gmtimew(mrb, timew, result))
        return NULL;

    offset = guess_local_offset(mrb, result);

    if (!gmtimew(mrb, wadd(mrb, timew, rb_time_magnify(mrb, v2w(offset))), result))
        return NULL;

    result->utc_offset = mrb_fixnum_value(offset);

    return result;
}

#define GetTimeval(mrb, obj, tobj) \
    TypedData_Get_Struct((mrb), (obj), struct time_object, &time_data_type, (tobj))

#define IsTimeval(obj) rb_typeddata_is_kind_of((obj), &time_data_type)

#define TIME_UTC_P(tobj) ((tobj)->gmt == 1)
#define TIME_SET_UTC(tobj) ((tobj)->gmt = 1)

#define TIME_LOCALTIME_P(tobj) ((tobj)->gmt == 0)
#define TIME_SET_LOCALTIME(tobj) ((tobj)->gmt = 0)

#define TIME_FIXOFF_P(tobj) ((tobj)->gmt == 2)
#define TIME_SET_FIXOFF(tobj, off) \
    ((tobj)->gmt = 2, \
     (tobj)->vtm.utc_offset = (off), \
     (tobj)->vtm.zone = NULL)

#define TIME_COPY_GMT(tobj1, tobj2) ((tobj1)->gmt = (tobj2)->gmt)

static mrb_value time_get_tm(mrb_state *mrb, mrb_value, struct time_object *);
#define MAKE_TM(mrb, time, tobj) \
  do { \
    if ((tobj)->tm_got == 0) { \
        time_get_tm((mrb), (time), (tobj)); \
    } \
  } while (0)

static void
time_mark(mrb_state *mrb, void *ptr)
{
    struct time_object *tobj = ptr;
    if (!tobj) return;
    //if (!FIXWV_P(tobj->timew))
    //  mrb_gc_mark(mrb, mrb_object(w2v(tobj->timew)));
    mrb_gc_mark(mrb, mrb_object(tobj->vtm.year));
    //mrb_gc_mark(mrb, mrb_object(tobj->vtm.subsecx));
    mrb_gc_mark(mrb, mrb_object(tobj->vtm.utc_offset));
}

static void
time_free(void *tobj)
{
    if (tobj) xfree(tobj);
}

static size_t
time_memsize(const void *tobj)
{
    return tobj ? sizeof(struct time_object) : 0;
}

static const rb_data_type_t time_data_type = {
    "time",
    time_mark, time_free, time_memsize,
};

struct RTypedData *
time_s_alloc(mrb_state *mrb, struct RClass *klass)
{
    struct RTypedData * obj;
    struct time_object *tobj;

    obj = TypedData_Make_Struct(mrb, klass, struct time_object, &time_data_type, tobj);
    tobj->tm_got=0;
    tobj->timew = WINT2FIXWV(0);

    return obj;
}

static void
time_modify(mrb_value time)
{
    //rb_check_frozen(time);
    //if (!OBJ_UNTRUSTED(time) && rb_safe_level() >= 4)
    //    mrb_raise(mrb, mrb->eSecurityError_class, "Insecure: can't modify Time");
}

static wideval_t
timespec2timew(mrb_state *mrb, struct timespec *ts)
{
    wideval_t timew;

    timew = rb_time_magnify(mrb, TIMET2WV(ts->tv_sec));
    if (ts->tv_nsec)
        timew = wadd(mrb, timew, wmulquoll(mrb, WINT2WV(ts->tv_nsec), TIME_SCALE, 1000000000));
    return timew;
}

static struct timespec
timew2timespec(mrb_state *mrb, wideval_t timew)
{
    int subsecx;
    struct timespec ts;
    wideval_t timew2;

    if (timew_out_of_timet_range(mrb, timew))
        mrb_raise(mrb, mrb->eArgumentError_class, "time out of system range");
    split_second(mrb, timew, &timew2, &subsecx);
    ts.tv_sec = WV2TIMET(timew2);
    ts.tv_nsec = (mulquo(mrb, subsecx, (1000000000), (TIME_SCALE)));
    return ts;
}

static struct timespec *
timew2timespec_exact(mrb_state *mrb, wideval_t timew, struct timespec *ts)
{
    int subsecx;
    wideval_t timew2;
    long nsecv;

    if (timew_out_of_timet_range(mrb, timew))
        return NULL;
    split_second(mrb, timew, &timew2, &subsecx);
    ts->tv_sec = WV2TIMET(timew2);
    nsecv = mulquo(mrb, subsecx, (1000000000), (TIME_SCALE));
    //if (!FIXNUM_P(nsecv))
    //    return NULL;
    ts->tv_nsec = (nsecv);
    return ts;
}

#ifdef INCLUDE_REGEXP
/*
 *  Document-method: now
 *
 *  Synonym for <code>Time.new</code>. Returns a +Time+ object
 *  initialized to the current system time.
 */

static mrb_value
time_init_0(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;
    struct timespec ts;

    time_modify(time);

    struct RTypedData *ttime = time_s_alloc(mrb, mrb->time_class); /* kusuda */
    time.value.p = ttime; /* kusuda */

    GetTimeval(mrb, time, tobj);
    tobj->tm_got=0;
    tobj->timew = WINT2FIXWV(0);
#ifdef HAVE_CLOCK_GETTIME
    if (clock_gettime(CLOCK_REALTIME, &ts) == -1) {
        rb_sys_fail("clock_gettime");
    }
#else
    {
        struct timeval tv;
        int ans;
        ans = gettimeofday(&tv, 0);
        if (ans < 0) {
            rb_sys_fail("gettimeofday");
        }
#if 0 /* 1:Test Timing */
        tv.tv_sec =978253200; /* 2000/12/31 18:00*/
        tv.tv_usec=100;
#endif
        ts.tv_sec = tv.tv_sec;
        ts.tv_nsec = tv.tv_usec * 1000;
    }
#endif
    tobj->timew = timespec2timew(mrb, &ts);

    return time;
}
#endif

static mrb_value
time_set_utc_offset(mrb_state *mrb, mrb_value time, mrb_value voff)
{
    struct time_object *tobj;
    voff = (num_exact(mrb, voff));

    time_modify(time);
    GetTimeval(mrb, time, tobj);

    tobj->tm_got = 0;
    TIME_SET_FIXOFF(tobj, voff);

    return time;
}

static void
vtm_add_offset(mrb_state *mrb, struct vtm *vtm, mrb_value voff)
{
    int sign;
    long subsec, v;
    int sec, min, hour;
    int day;
    long off = mrb_fixnum_number(voff);

    vtm->utc_offset = mrb_fixnum_value(sub(mrb, mrb_fixnum_number(vtm->utc_offset), off));

    if (lt(mrb, off, (0))) {
        sign = -1;
        off = neg(mrb, off);
    }
    else {
        sign = 1;
    }
    divmodv(mrb, off, (1), &off, &subsec);
    divmodv(mrb, off, (60), &off, &v);
    sec = (v);
    divmodv(mrb, off, (60), &off, &v);
    min = (v);
    divmodv(mrb, off, (24), &off, &v);
    hour = (v);

    if (sign < 0) {
        subsec = neg(mrb, subsec);
        sec = -sec;
        min = -min;
        hour = -hour;
    }

    day = 0;

    if (subsec != (0)) {
        vtm->subsecx = add(mrb, vtm->subsecx, w2v(rb_time_magnify(mrb, v2w(subsec))));
        if (lt(mrb, vtm->subsecx, (0))) {
            vtm->subsecx = add(mrb, vtm->subsecx, (TIME_SCALE));
            sec -= 1;
        }
        if (le(mrb, (TIME_SCALE), vtm->subsecx)) {
            vtm->subsecx = sub(mrb, vtm->subsecx, (TIME_SCALE));
            sec += 1;
        }
        goto not_zero_sec;
    }
    if (sec) {
      not_zero_sec:
        /* If sec + subsec == 0, don't change vtm->sec.
         * It may be 60 which is a leap second. */
        vtm->sec += sec;
        if (vtm->sec < 0) {
            vtm->sec += 60;
            min -= 1;
        }
        if (60 <= vtm->sec) {
            vtm->sec -= 60;
            min += 1;
        }
    }
    if (min) {
        vtm->min += min;
        if (vtm->min < 0) {
            vtm->min += 60;
            hour -= 1;
        }
        if (60 <= vtm->min) {
            vtm->min -= 60;
            hour += 1;
        }
    }
    if (hour) {
        vtm->hour += hour;
        if (vtm->hour < 0) {
            vtm->hour += 24;
            day = -1;
        }
        if (24 <= vtm->hour) {
            vtm->hour -= 24;
            day = 1;
        }
    }

    if (day) {
        if (day < 0) {
            if (vtm->mon == 1 && vtm->mday == 1) {
                vtm->mday = 31;
                vtm->mon = 12; /* December */
                vtm->year = mrb_fixnum_value(sub(mrb, mrb_fixnum_number(vtm->year), (1)));
                vtm->yday = leap_year_v_p(mrb, vtm->year) ? 365 : 364;
            }
            else if (vtm->mday == 1) {
                const int *days_in_month = leap_year_v_p(mrb, vtm->year) ?
                                           leap_year_days_in_month :
                                           common_year_days_in_month;
                vtm->mon--;
                vtm->mday = days_in_month[vtm->mon-1];
                vtm->yday--;
            }
            else {
                vtm->mday--;
                vtm->yday--;
            }
            vtm->wday = (vtm->wday + 6) % 7;
        }
        else {
            int leap = leap_year_v_p(mrb, vtm->year);
            if (vtm->mon == 12 && vtm->mday == 31) {
                vtm->year = mrb_fixnum_value(add(mrb, mrb_fixnum_number(vtm->year), (1)));
                vtm->mon = 1; /* January */
                vtm->mday = 1;
                vtm->yday = 1;
            }
            else if (vtm->mday == (leap ? leap_year_days_in_month :
                                          common_year_days_in_month)[vtm->mon-1]) {
                vtm->mon++;
                vtm->mday = 1;
                vtm->yday++;
            }
            else {
                vtm->mday++;
                vtm->yday++;
            }
            vtm->wday = (vtm->wday + 1) % 7;
        }
    }
}

static mrb_value
utc_offset_arg(mrb_state *mrb, mrb_value arg)
{
    mrb_value tmp;
    if (!mrb_nil_p(tmp = mrb_check_string_type(mrb, arg))) {
        int n;
        char *s = RSTRING_PTR(tmp);
        if (//!rb_enc_str_asciicompat_p(mrb, tmp) ||
            RSTRING_LEN(tmp) != 6 ||
            (s[0] != '+' && s[0] != '-') ||
            !ISDIGIT(s[1]) ||
            !ISDIGIT(s[2]) ||
            s[3] != ':' ||
            !ISDIGIT(s[4]) ||
            !ISDIGIT(s[5]))
          mrb_raise(mrb, mrb->eArgumentError_class, "\"+HH:MM\" or \"-HH:MM\" expected for utc_offset");
        n = (s[1] * 10 + s[2] - '0' * 11) * 3600;
        n += (s[4] * 10 + s[5] - '0' * 11) * 60;
        if (s[0] == '-')
          n = -n;
        return mrb_fixnum_value(n);
    }
    else {
        return (num_exact(mrb, arg));
    }
}

#ifdef INCLUDE_REGEXP
static mrb_value
time_init_1(mrb_state *mrb, int argc, mrb_value *argv, mrb_value time)
{
    struct vtm vtm;
    //mrb_value v[7];
    struct time_object *tobj;

    vtm.wday = -1;
    vtm.yday = 0;
    vtm.zone = "";

      /*                             year     mon      mday     hour     min      sec      off */
    //rb_scan_args(argc, argv, "16", &argv[0],&argv[1],&argv[2],&argv[3],&argv[4],&argv[5],&argv[6]);

    vtm.year = obj2vint(mrb, argv[0]);

    vtm.mon = mrb_nil_p(argv[1]) ? 1 : month_arg(mrb, argv[1]);

    vtm.mday = mrb_nil_p(argv[2]) ? 1 : obj2int(mrb, argv[2]);

    vtm.hour = mrb_nil_p(argv[3]) ? 0 : obj2int(mrb, argv[3]);

    vtm.min  = mrb_nil_p(argv[4]) ? 0 : obj2int(mrb, argv[4]);

    vtm.sec = 0;
    vtm.subsecx = (0);
    if (!mrb_nil_p(argv[5])) {
        long sec = long_num_exact(mrb, argv[5]);
        long subsec;
        divmodv(mrb, sec, (1), &sec, &subsec);
        vtm.sec = (sec);
        vtm.subsecx = w2v(rb_time_magnify(mrb, v2w(subsec)));
    }

    vtm.isdst = -1;
    vtm.utc_offset = mrb_nil_value();
    if (!mrb_nil_p(argv[6])) {
        mrb_value arg = argv[6];
        if (mrb_obj_equal(mrb, arg, mrb_symbol_value(mrb_intern(mrb, "dst"))))
            vtm.isdst = 1;
        else if (mrb_obj_equal(mrb, arg, mrb_symbol_value(mrb_intern(mrb, "std"))))
            vtm.isdst = 0;
        else
            vtm.utc_offset = utc_offset_arg(mrb, arg);
    }

    validate_vtm(mrb, &vtm);

    time_modify(time);

    struct RTypedData *ttime = time_s_alloc(mrb, mrb->time_class); /* kusuda */
    time.value.p = ttime; /* kusuda */

    GetTimeval(mrb, time, tobj);
    tobj->tm_got=0;
    tobj->timew = WINT2FIXWV(0);

    if (!mrb_nil_p(vtm.utc_offset)) {
        long off = mrb_fixnum_number(vtm.utc_offset);
        vtm_add_offset(mrb, &vtm, mrb_fixnum_value(neg(mrb, off)));
        vtm.utc_offset = mrb_nil_value();
        tobj->timew = timegmw(mrb, &vtm);
        return time_set_utc_offset(mrb, time, mrb_fixnum_value(off));
    }
    else {
        tobj->timew = timelocalw(mrb, &vtm);
        return time_localtime(mrb, time);
    }
}
#endif

/* 15.2.19.7.16 */
/*
 *  call-seq:
 *     Time.new -> time
 *     Time.new(year, month=nil, day=nil, hour=nil, min=nil, sec=nil, utc_offset=nil) -> time
 *
 *  Returns a <code>Time</code> object.
 *
 *  It is initialized to the current system time if no argument.
 *  <b>Note:</b> The object created will be created using the

 *  resolution available on your system clock, and so may include
 *  fractional seconds.
 *
 *  If one or more arguments specified, the time is initialized
 *  to the specified time.
 *  _sec_ may have fraction if it is a rational.
 *
 *  _utc_offset_ is the offset from UTC.
 *  It is a string such as "+09:00" or a number of seconds such as 32400.
 *
 *     a = Time.new      #=> 2007-11-19 07:50:02 -0600
 *     b = Time.new      #=> 2007-11-19 07:50:02 -0600
 *     a == b            #=> false
 *     "%.6f" % a.to_f   #=> "1195480202.282373"
 *     "%.6f" % b.to_f   #=> "1195480202.283415"
 *
 *     Time.new(2008,6,21, 13,30,0, "+09:00") #=> 2008-06-21 13:30:00 +0900
 *
 *     # A trip for RubyConf 2007
 *     t1 = Time.new(2007,11,1,15,25,0, "+09:00") # JST (Narita)
 *     t2 = Time.new(2007,11,1,12, 5,0, "-05:00") # CDT (Minneapolis)
 *     t3 = Time.new(2007,11,1,13,25,0, "-05:00") # CDT (Minneapolis)
 *     t4 = Time.new(2007,11,1,16,53,0, "-04:00") # EDT (Charlotte)
 *     t5 = Time.new(2007,11,5, 9,24,0, "-05:00") # EST (Charlotte)
 *     t6 = Time.new(2007,11,5,11,21,0, "-05:00") # EST (Detroit)
 *     t7 = Time.new(2007,11,5,13,45,0, "-05:00") # EST (Detroit)
 *     t8 = Time.new(2007,11,6,17,10,0, "+09:00") # JST (Narita)
 *     p((t2-t1)/3600.0)                          #=> 10.666666666666666
 *     p((t4-t3)/3600.0)                          #=> 2.466666666666667
 *     p((t6-t5)/3600.0)                          #=> 1.95
 *     p((t8-t7)/3600.0)                          #=> 13.416666666666666
 *
 */

static mrb_value
time_init(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value time)
{
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);

#ifdef INCLUDE_REGEXP
  if (argc == 0)
    return time_init_0(mrb, time);
  else
    return time_init_1(mrb, argc, argv, time);
#else
  struct time_object *tobj;

  time_modify(time);

  struct RTypedData *ttime = time_s_alloc(mrb, mrb->time_class); /* kusuda */
  time.value.p = ttime; /* kusuda */

  GetTimeval(mrb, time, tobj);
  struct timeval tv;
  struct timespec ts;
  tobj->tm_got=0;
  tv.tv_sec = 0;
  tv.tv_usec = 0;
  if (gettimeofday(&tv, 0) < 0) {
    rb_sys_fail("gettimeofday");
  }
  ts.tv_sec = tv.tv_sec;
  ts.tv_nsec = tv.tv_usec * 1000;
  tobj->timew = timespec2timew(mrb, &ts);
  return time;
#endif
}

static void
time_overflow_p(mrb_state *mrb, time_t *secp, long *nsecp)
{
    time_t tmp, sec = *secp;
    long nsec = *nsecp;

    if (nsec >= 1000000000) {        /* nsec positive overflow */
        tmp = sec + nsec / 1000000000;
        nsec %= 1000000000;
        if (sec > 0 && tmp < 0) {
            mrb_raise(mrb, mrb->eRangeError_class, "out of Time range");
        }
        sec = tmp;
    }
    if (nsec < 0) {                /* nsec negative overflow */
        tmp = sec + NDIV(nsec,1000000000); /* negative div */
        nsec = NMOD(nsec,1000000000);      /* negative mod */
        if (sec < 0 && tmp > 0) {
            mrb_raise(mrb, mrb->eRangeError_class, "out of Time range");
        }
        sec = tmp;
    }
#ifndef NEGATIVE_TIME_T
    if (sec < 0)
        mrb_raise(mrb, mrb->eArgumentError_class, "time must be positive");
#endif
    *secp = sec;
    *nsecp = nsec;
}

static wideval_t
nsec2timew(mrb_state *mrb, time_t sec, long nsec)
{
    struct timespec ts;
    time_overflow_p(mrb, &sec, &nsec);
    ts.tv_sec = sec;
    ts.tv_nsec = nsec;
    return timespec2timew(mrb, &ts);
}

static mrb_value
time_new_timew(mrb_state *mrb, struct RClass* klass, wideval_t timew)
{
    mrb_value time = mrb_obj_value(time_s_alloc(mrb, klass));
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    tobj->timew = timew;

    return time;
}

mrb_value
rb_time_new(mrb_state *mrb, time_t sec, long usec)
{
    return time_new_timew(mrb, mrb->time_class, nsec2timew(mrb, sec, usec * 1000));
}

mrb_value
rb_time_nano_new(mrb_state *mrb, time_t sec, long nsec)
{
    return time_new_timew(mrb, mrb->time_class, nsec2timew(mrb, sec, nsec));
}

mrb_value
rb_time_num_new(mrb_state *mrb, long timev, mrb_value off)
{
    mrb_value time = time_new_timew(mrb, mrb->time_class, rb_time_magnify(mrb, v2w(timev)));

    if (!mrb_nil_p(off)) {
        off = utc_offset_arg(mrb, off);
        validate_utc_offset(mrb, off);
        time_set_utc_offset(mrb, time, off);
        return time;
    }

    return time;
}

static struct timespec
time_timespec(mrb_state *mrb, mrb_value num, int interval)
{
    struct timespec t;
    const char *tstr = interval ? "time interval" : "time";
    long i, f;
    //mrb_value ary;

#ifndef NEGATIVE_TIME_T
    interval = 1;
#endif

    switch (mrb_type(num)) {
      case MRB_TT_FIXNUM:
        t.tv_sec = NUM2TIMET(num);
        if (interval && t.tv_sec < 0)
            mrb_raise(mrb, mrb->eArgumentError_class, "%s must be positive", tstr);
        t.tv_nsec = 0;
        break;

      case MRB_TT_FLOAT:
        if (interval && mrb_float_number(num) < 0.0)
            mrb_raise(mrb, mrb->eArgumentError_class, "%s must be positive", tstr);
        else {
            double f, d;

            d = modf(mrb_float_number(num), &f);
            if (d >= 0) {
                t.tv_nsec = (int)(d*1e9+0.5);
            }
            else if ((t.tv_nsec = (int)(-d*1e9+0.5)) > 0) {
                t.tv_nsec = 1000000000 - t.tv_nsec;
                f -= 1;
            }
            t.tv_sec = (time_t)f;
            if (f != t.tv_sec) {
                mrb_raise(mrb, mrb->eRangeError_class, "%f out of Time range", mrb_float_number(num));
            }
        }
        break;

      default:
        i = (1);
        //ary = rb_check_funcall(mrb, num, mrb_intern(mrb, "divmod"), 1, &i);
        if (mrb_respond_to(mrb, num, mrb_intern(mrb, "divmod"))) {
          //ary = mrb_funcall(mrb, num, "divmod", 1, &i);
          //if ((ary.tt > 0) && !mrb_nil_p(ary = mrb_check_array_type(mrb, ary))) {
          //  i = mrb_ary_ref(mrb, ary, 0);
          //  f = mrb_ary_ref(mrb, ary, 1);
          //  t.tv_sec = NUM2TIMET(i);
          t.tv_sec = mrb_fixnum_number(num) / i;
          f = mrb_fixnum_number(num) % i;
          if (interval && t.tv_sec < 0)
              mrb_raise(mrb, mrb->eArgumentError_class, "%s must be positive", tstr);
          //f = mrb_funcall(mrb, f, "*", 1, mrb_fixnum_value(1000000000));
          //t.tv_nsec = mrb_fixnum_number(f);
          t.tv_nsec = f * 1000000000;
        }
        else {
            mrb_raise(mrb, mrb->eTypeError_class, "can't convert %s into %s",
                     mrb_obj_classname(mrb, num), tstr);
        }
        break;
    }
    return t;
}

static struct timeval
time_timeval(mrb_state *mrb, mrb_value num, int interval)
{
    struct timespec ts;
    struct timeval tv;

    ts = time_timespec(mrb, num, interval);
    tv.tv_sec = (TYPEOF_TIMEVAL_TV_SEC)ts.tv_sec;
    tv.tv_usec = (TYPEOF_TIMEVAL_TV_USEC)(ts.tv_nsec / 1000);

    return tv;
}

struct timeval
rb_time_interval(mrb_state *mrb, mrb_value num)
{
    return time_timeval(mrb, num, TRUE);
}

struct timeval
rb_time_timeval(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;
    struct timeval t;
    struct timespec ts;

    if (IsTimeval(time)) {
        GetTimeval(mrb, time, tobj);
        ts = timew2timespec(mrb, tobj->timew);
        t.tv_sec = (TYPEOF_TIMEVAL_TV_SEC)ts.tv_sec;
        t.tv_usec = (TYPEOF_TIMEVAL_TV_USEC)(ts.tv_nsec / 1000);
        return t;
    }
    return time_timeval(mrb, time, FALSE);
}

struct timespec
rb_time_timespec(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;
    struct timespec t;

    if (IsTimeval(time)) {
        GetTimeval(mrb, time, tobj);
        t = timew2timespec(mrb, tobj->timew);
        return t;
    }
    return time_timespec(mrb, time, FALSE);
}

/* 15.2.19.6.5  */
/*
 *  call-seq:
 *     Time.now -> time
 *
 *  Creates a new time object for the current time.
 *
 *     Time.now            #=> 2009-06-24 12:39:54 +0900
 */

static mrb_value
time_s_now(mrb_state *mrb, mrb_value klass)
{
#if 0 /* original */
  return rb_class_new_instance(mrb, 0, NULL, mrb_class_get(mrb, klass));
#else
  //SCLASS mrb_value io = mrb_obj_value(mrb_obj_alloc(mrb, mrb_class_get(mrb, klass)->tt, mrb_class_get(mrb, klass)));
  //mrb_value time = mrb_obj_value(mrb_obj_alloc(mrb, MRB_TT_TIME, mrb_class_get(mrb, klass)));
  //mrb_value time = mrb_obj_value(mrb_obj_alloc(mrb, MRB_TT_DATA, mrb_class_get(mrb, klass)));
  mrb_value time = mrb_obj_value(mrb_obj_alloc(mrb, MRB_TT_DATA, mrb_class_ptr(klass)));
  time = rb_obj_call_init(mrb, time, 0, NULL);
  return time;
#endif
}

/* 15.2.19.6.7(x) */
static mrb_value
time_s_new(mrb_state *mrb, mrb_value klass)
{
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);
  //mrb_value time = mrb_obj_value(mrb_obj_alloc(mrb, MRB_TT_DATA, mrb_class_get(mrb, klass)));
  mrb_value time = mrb_obj_value(mrb_obj_alloc(mrb, MRB_TT_DATA, mrb_class_ptr(klass)));
  time = rb_obj_call_init(mrb, time, argc, argv);
  return time;
}

/* 15.2.19.6.1  */
/*
 *  call-seq:
 *     Time.at(time) -> time
 *     Time.at(seconds_with_frac) -> time
 *     Time.at(seconds, microseconds_with_frac) -> time
 *
 *  Creates a new time object with the value given by <i>time</i>,
 *  the given number of <i>seconds_with_frac</i>, or
 *  <i>seconds</i> and <i>microseconds_with_frac</i> from the Epoch.
 *  <i>seconds_with_frac</i> and <i>microseconds_with_frac</i>
 *  can be Integer, Float, Rational, or other Numeric.
 *  non-portable feature allows the offset to be negative on some systems.
 *
 *     Time.at(0)            #=> 1969-12-31 18:00:00 -0600
 *     Time.at(Time.at(0))   #=> 1969-12-31 18:00:00 -0600
 *     Time.at(946702800)    #=> 1999-12-31 23:00:00 -0600
 *     Time.at(-284061600)   #=> 1960-12-31 00:00:00 -0600
 *     Time.at(946684800.2).usec #=> 200000
 *     Time.at(946684800, 123456.789).nsec #=> 123456789
 */

static mrb_value
time_s_at(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value klass)
{
    mrb_value time, t;
    wideval_t timew;
    wideval_t wktime;
    wideval_t wkt;

  mrb_value argv[16];
  int argc;

  time = mrb_nil_value();
  t    = mrb_nil_value();
  mrb_get_args(mrb, "*", &argv, &argc);
  if (argc > 0) time = argv[0];
  if (argc > 1) t    = argv[1];
  //if (rb_scan_args(argc, argv, "11", &time, &t) == 2) {
  if (argc == 2) {
    /* ex. Time.at(946684800, 123456.789) */
    wktime = long_num_exact(mrb, time);
    wkt = long_num_exact(mrb, t);
    timew = wadd(mrb, rb_time_magnify(mrb, v2w(wktime)), wmulquoll(mrb, v2w(wkt), TIME_SCALE, 1000000));
    t = time_new_timew(mrb, mrb_class_get(mrb, klass), timew);
  }
  else if (IsTimeval(time)) {
    /* Time.at(Time.at(0)) */
    struct time_object *tobj, *tobj2;
    GetTimeval(mrb, time, tobj);
    t = time_new_timew(mrb, mrb_class_get(mrb, klass), tobj->timew);
    GetTimeval(mrb, t, tobj2);
    TIME_COPY_GMT(tobj2, tobj);
  }
  else {
    /* Time.at(946684800.2) */
    timew = rb_time_magnifyFromMrbValue(mrb, v2w(num_exact(mrb, time)));
    t = time_new_timew(mrb, mrb_class_get(mrb, klass), timew);
  }

  return t;
}

static const char months[][4] = {
    "jan", "feb", "mar", "apr", "may", "jun",
    "jul", "aug", "sep", "oct", "nov", "dec",
};

static int
obj2int(mrb_state *mrb, mrb_value obj)
{
    if (mrb_type(obj) == MRB_TT_STRING) {
        obj = rb_str_to_inum(mrb, obj, 10, FALSE);
    }

    return mrb_fixnum_number(obj);
}

static mrb_value
obj2vint(mrb_state *mrb, mrb_value obj)
{
    if (mrb_type(obj) == MRB_TT_STRING) {
        obj = rb_str_to_inum(mrb, obj, 10, FALSE);
    }
    else {
        obj = rb_to_int(mrb, obj);
    }

    return obj;
}

static void
obj2subsecx(mrb_state *mrb, mrb_value obj, int *sec, int *subsecx)
{
    wideval_t secw;

    if (mrb_type(obj) == MRB_TT_STRING) {
        obj = rb_str_to_inum(mrb, obj, 10, FALSE);
        *subsecx = (0);
        *sec = mrb_fixnum_number(obj);
    }
    else {
#if 0
    long q;
    long subsec;
    divmodv(mrb, long_num_exact(mrb, obj), (1), &q, &subsec);
    *subsecx = mrb_fixnum_value(w2v(rb_time_magnify(mrb, v2w(subsec))));
#else
    secw = rb_time_magnifyFromMrbValue(mrb, v2w(num_exact(mrb, obj)));
    wideval_t q, r;
    wdivmod(mrb, secw, WINT2FIXWV(TIME_SCALE), &q, &r);
    *sec = q;
    *subsecx = w2v(r);
#endif
    }
}

static long
usec2subsecx(mrb_state *mrb, mrb_value obj)
{
    if (mrb_type(obj) == MRB_TT_STRING) {
        obj = rb_str_to_inum(mrb, obj, 10, FALSE);
    }

    return mulquo(mrb, long_num_exact(mrb, obj), (TIME_SCALE), (1000000));
}

static int
month_arg(mrb_state *mrb, mrb_value arg)
{
    int i, mon;

    mrb_value s = mrb_check_string_type(mrb, arg);
    if (!mrb_nil_p(s)) {
        mon = 0;
        for (i=0; i<12; i++) {
            if (RSTRING_LEN(s) == 3 &&
                STRCASECMP(months[i], RSTRING_PTR(s)) == 0) {
                mon = i+1;
                break;
            }
        }
        if (mon == 0) {
            char c = RSTRING_PTR(s)[0];

            if ('0' <= c && c <= '9') {
                mon = obj2int(mrb, s);
            }
        }
    }
    else {
        mon = obj2int(mrb, arg);
    }
    return mon;
}

static void
validate_utc_offset(mrb_state *mrb, mrb_value utc_offset)
{
    if (le(mrb, mrb_fixnum_number(utc_offset), (-86400)) || ge(mrb, mrb_fixnum_number(utc_offset), (86400)))
        mrb_raise(mrb, mrb->eArgumentError_class, "utc_offset out of range");
}

static void
validate_vtm(mrb_state *mrb, struct vtm *vtm)
{
    if (   vtm->mon  < 1 || vtm->mon  > 12
        || vtm->mday < 1 || vtm->mday > 31
        || vtm->hour < 0 || vtm->hour > 24
        || (vtm->hour == 24 && (vtm->min > 0 || vtm->sec > 0))
        || vtm->min  < 0 || vtm->min  > 59
        || vtm->sec  < 0 || vtm->sec  > 60
        || lt(mrb, vtm->subsecx, (0)) || ge(mrb, vtm->subsecx, (TIME_SCALE))
        || (!mrb_nil_p(vtm->utc_offset) && (validate_utc_offset(mrb, vtm->utc_offset), 0)))
        mrb_raise(mrb, mrb->eArgumentError_class, "argument out of range");
}

static void
time_arg(mrb_state *mrb, int argc, mrb_value *argv, struct vtm *vtm)
{
    mrb_value v[8];

    int i;
    if (argc < 9) { /* use max 8elements */
        for (i=argc; i <9; i++) {
            argv[i] = mrb_nil_value();
        }
    }
    vtm->year = mrb_fixnum_value(0);
    vtm->mon = 0;
    vtm->mday = 0;
    vtm->hour = 0;
    vtm->min = 0;
    vtm->sec = 0;
    vtm->subsecx = (0);
    vtm->utc_offset = mrb_nil_value();
    vtm->wday = 0;
    vtm->yday = 0;
    vtm->isdst = 0;
    vtm->zone = "";

    if (argc == 10) {
        v[0] = argv[5];
        v[1] = argv[4];
        v[2] = argv[3];
        v[3] = argv[2];
        v[4] = argv[1];
        v[5] = argv[0];
        v[6] = mrb_nil_value();
        vtm->isdst = RTEST(argv[8]) ? 1 : 0;
    }
    else {
        //rb_scan_args(argc, argv, "17", &v[0],&v[1],&v[2],&v[3],&v[4],&v[5],&v[6],&v[7]);
        /* v[6] may be usec or zone (parsedate) */
        /* v[7] is wday (parsedate; ignored) */
        v[0] = argv[0];
        v[1] = argv[1];
        v[2] = argv[2];
        v[3] = argv[3];
        v[4] = argv[4];
        v[5] = argv[5];
        v[6] = argv[6];
        v[7] = argv[7];
        vtm->wday = -1;
        vtm->isdst = -1;
    }

    vtm->year = obj2vint(mrb, v[0]);

    if (mrb_nil_p(v[1])) {
        vtm->mon = 1;
    }
    else {
        vtm->mon = month_arg(mrb, v[1]);
    }

    if (mrb_nil_p(v[2])) {
        vtm->mday = 1;
    }
    else {
        vtm->mday = obj2int(mrb, v[2]);
    }

    vtm->hour = mrb_nil_p(v[3])?0:obj2int(mrb, v[3]);

    vtm->min  = mrb_nil_p(v[4])?0:obj2int(mrb, v[4]);

    if (!mrb_nil_p(v[6]) && argc == 7) {
        vtm->sec  = mrb_nil_p(v[5])?0:obj2int(mrb, v[5]);
        vtm->subsecx  = usec2subsecx(mrb, v[6]);
    }
    else {
        /* when argc == 8, v[6] is timezone, but ignored */
        if (mrb_nil_p(v[5])) {
          vtm->sec  = 0;
        }
        else {
#if 0
          mrb_value subsecx;
          obj2subsecx(mrb, v[5], &vtm->sec);
          vtm->subsecx = mrb_fixnum_number(subsecx);
#else
          obj2subsecx(mrb, v[5], &vtm->sec, &vtm->subsecx);
#endif
        }
    }

    validate_vtm(mrb, vtm);
}

static int
leap_year_p(long y)
{
    return ((y % 4 == 0) && (y % 100 != 0)) || (y % 400 == 0);
}

static time_t
timegm_noleapsecond(struct tm *tm)
{
    long tm_year = tm->tm_year;
    int tm_yday = tm->tm_mday;
    if (leap_year_p(tm_year + 1900))
        tm_yday += leap_year_yday_offset[tm->tm_mon];
    else
        tm_yday += common_year_yday_offset[tm->tm_mon];

    /*
     *  `Seconds Since the Epoch' in SUSv3:
     *  tm_sec + tm_min*60 + tm_hour*3600 + tm_yday*86400 +
     *  (tm_year-70)*31536000 + ((tm_year-69)/4)*86400 -
     *  ((tm_year-1)/100)*86400 + ((tm_year+299)/400)*86400
     */
    return tm->tm_sec + tm->tm_min*60 + tm->tm_hour*3600 +
           (time_t)(tm_yday +
                    (tm_year-70)*365 +
                    DIV(tm_year-69,4) -
                    DIV(tm_year-1,100) +
                    DIV(tm_year+299,400))*86400;
}

#if 0
#define DEBUG_FIND_TIME_NUMGUESS
#define DEBUG_GUESSRANGE
#endif

#ifdef DEBUG_GUESSRANGE
#define DEBUG_REPORT_GUESSRANGE fprintf(stderr, "find time guess range: %ld - %ld : %lu\n", guess_lo, guess_hi, (unsigned_time_t)(guess_hi-guess_lo))
#else
#define DEBUG_REPORT_GUESSRANGE
#endif

#ifdef DEBUG_FIND_TIME_NUMGUESS
#define DEBUG_FIND_TIME_NUMGUESS_INC find_time_numguess++,
static unsigned long long find_time_numguess;

static mrb_value find_time_numguess_getter(void)
{
    return ULL2NUM(find_time_numguess);
}
#else
#define DEBUG_FIND_TIME_NUMGUESS_INC
#endif

static const char *
find_time_t(struct tm *tptr, int utc_p, time_t *tp)
{
    time_t guess, guess0, guess_lo, guess_hi;
    struct tm *tm, tm0, tm_lo, tm_hi;
    int d;
    int find_dst;
    struct tm result;
    int status;
    int tptr_tm_yday;

#define GUESS(p) (DEBUG_FIND_TIME_NUMGUESS_INC (utc_p ? gmtime_with_leapsecond((p), &result) : LOCALTIME((p), result)))

    guess_lo = TIMET_MIN;
    guess_hi = TIMET_MAX;

    //#define TIMET_MIN (~(time_t)0 <= 0 ? (time_t)(((unsigned_time_t)1) << (sizeof(time_t) * CHAR_BIT - 1)) : (time_t)0)
    guess_lo = (time_t)(((unsigned_time_t)1) << (sizeof(time_t) * CHAR_BIT - 1));
    guess_lo = (time_t)(((unsigned_time_t)1) << 1);
    guess_lo =  (time_t)0;


    find_dst = 0 < tptr->tm_isdst;

#if defined(HAVE_MKTIME)
    tm0 = *tptr;
    if (!utc_p && (guess = mktime(&tm0)) != -1) {
        tm = GUESS(&guess);
        if (tm && tmcmp(tptr, tm) == 0) {
            goto found;
        }
    }
#endif

    tm0 = *tptr;
    if (tm0.tm_mon < 0) {
      tm0.tm_mon = 0;
      tm0.tm_mday = 1;
      tm0.tm_hour = 0;
      tm0.tm_min = 0;
      tm0.tm_sec = 0;
    }
    else if (11 < tm0.tm_mon) {
      tm0.tm_mon = 11;
      tm0.tm_mday = 31;
      tm0.tm_hour = 23;
      tm0.tm_min = 59;
      tm0.tm_sec = 60;
    }
    else if (tm0.tm_mday < 1) {
      tm0.tm_mday = 1;
      tm0.tm_hour = 0;
      tm0.tm_min = 0;
      tm0.tm_sec = 0;
    }
    else if ((d = (leap_year_p(1900 + tm0.tm_year) ?
                   leap_year_days_in_month :
                   common_year_days_in_month)[tm0.tm_mon]) < tm0.tm_mday) {
      tm0.tm_mday = d;
      tm0.tm_hour = 23;
      tm0.tm_min = 59;
      tm0.tm_sec = 60;
    }
    else if (tm0.tm_hour < 0) {
      tm0.tm_hour = 0;
      tm0.tm_min = 0;
      tm0.tm_sec = 0;
    }
    else if (23 < tm0.tm_hour) {
      tm0.tm_hour = 23;
      tm0.tm_min = 59;
      tm0.tm_sec = 60;
    }
    else if (tm0.tm_min < 0) {
      tm0.tm_min = 0;
      tm0.tm_sec = 0;
    }
    else if (59 < tm0.tm_min) {
      tm0.tm_min = 59;
      tm0.tm_sec = 60;
    }
    else if (tm0.tm_sec < 0) {
      tm0.tm_sec = 0;
    }
    else if (60 < tm0.tm_sec) {
      tm0.tm_sec = 60;
    }

    DEBUG_REPORT_GUESSRANGE;
    guess0 = guess = timegm_noleapsecond(&tm0);
    tm = GUESS(&guess);
    if (tm) {
        d = tmcmp(tptr, tm);
        if (d == 0) { goto found; }
        if (d < 0) {
            guess_hi = guess;
            guess -= 24 * 60 * 60;
        }
        else {
            guess_lo = guess;
            guess += 24 * 60 * 60;
        }
        DEBUG_REPORT_GUESSRANGE;
        if (guess_lo < guess && guess < guess_hi && (tm = GUESS(&guess)) != NULL) {
            d = tmcmp(tptr, tm);
            if (d == 0) { goto found; }
            if (d < 0)
                guess_hi = guess;
            else
                guess_lo = guess;
            DEBUG_REPORT_GUESSRANGE;
        }
    }

    tm = GUESS(&guess_lo);
    if (!tm) goto error;
    d = tmcmp(tptr, tm);
    if (d < 0) goto out_of_range;
    if (d == 0) { guess = guess_lo; goto found; }
    tm_lo = *tm;

    tm = GUESS(&guess_hi);
    if (!tm) goto error;
    d = tmcmp(tptr, tm);
    if (d > 0) goto out_of_range;
    if (d == 0) { guess = guess_hi; goto found; }
    tm_hi = *tm;

    DEBUG_REPORT_GUESSRANGE;

    status = 1;

    while (guess_lo + 1 < guess_hi) {
        if (status == 0) {
          binsearch:
            guess = guess_lo / 2 + guess_hi / 2;
            if (guess <= guess_lo)
                guess = guess_lo + 1;
            else if (guess >= guess_hi)
                guess = guess_hi - 1;
            status = 1;
        }
        else {
            if (status == 1) {
                time_t guess0_hi = timegm_noleapsecond(&tm_hi);
                guess = guess_hi - (guess0_hi - guess0);
                if (guess == guess_hi) /* hh:mm:60 tends to cause this condition. */
                    guess--;
                status = 2;
            }
            else if (status == 2) {
                time_t guess0_lo = timegm_noleapsecond(&tm_lo);
                guess = guess_lo + (guess0 - guess0_lo);
                if (guess == guess_lo)
                    guess++;
                status = 0;
            }
            if (guess <= guess_lo || guess_hi <= guess) {
                /* Precious guess is invalid. try binary search. */
#ifdef DEBUG_GUESSRANGE
                if (guess <= guess_lo) fprintf(stderr, "too small guess: %ld <= %ld\n", guess, guess_lo);
                if (guess_hi <= guess) fprintf(stderr, "too big guess: %ld <= %ld\n", guess_hi, guess);
#endif
                goto binsearch;
            }
        }

        tm = GUESS(&guess);
        if (!tm) goto error;

        d = tmcmp(tptr, tm);

        if (d < 0) {
            guess_hi = guess;
            tm_hi = *tm;
            DEBUG_REPORT_GUESSRANGE;
        }
        else if (d > 0) {
            guess_lo = guess;
            tm_lo = *tm;
            DEBUG_REPORT_GUESSRANGE;
        }
        else {
          found:
            if (!utc_p) {
                /* If localtime is nonmonotonic, another result may exist. */
                time_t guess2;
                if (find_dst) {
                    guess2 = guess - 2 * 60 * 60;
                    tm = LOCALTIME(&guess2, result);
                    if (tm) {
                        if (tptr->tm_hour != (tm->tm_hour + 2) % 24 ||
                            tptr->tm_min != tm->tm_min ||
                            tptr->tm_sec != tm->tm_sec) {
                            guess2 -= (tm->tm_hour - tptr->tm_hour) * 60 * 60 +
                                      (tm->tm_min - tptr->tm_min) * 60 +
                                      (tm->tm_sec - tptr->tm_sec);
                            if (tptr->tm_mday != tm->tm_mday)
                                guess2 += 24 * 60 * 60;
                            if (guess != guess2) {
                                tm = LOCALTIME(&guess2, result);
                                if (tm && tmcmp(tptr, tm) == 0) {
                                    if (guess < guess2)
                                        *tp = guess;
                                    else
                                        *tp = guess2;
                                    return NULL;
                                }
                            }
                        }
                    }
                }
                else {
                    guess2 = guess + 2 * 60 * 60;
                    tm = LOCALTIME(&guess2, result);
                    if (tm) {
                        if ((tptr->tm_hour + 2) % 24 != tm->tm_hour ||
                            tptr->tm_min != tm->tm_min ||
                            tptr->tm_sec != tm->tm_sec) {
                            guess2 -= (tm->tm_hour - tptr->tm_hour) * 60 * 60 +
                                      (tm->tm_min - tptr->tm_min) * 60 +
                                      (tm->tm_sec - tptr->tm_sec);
                            if (tptr->tm_mday != tm->tm_mday)
                                guess2 -= 24 * 60 * 60;
                            if (guess != guess2) {
                                tm = LOCALTIME(&guess2, result);
                                if (tm && tmcmp(tptr, tm) == 0) {
                                    if (guess < guess2)
                                        *tp = guess2;
                                    else
                                        *tp = guess;
                                    return NULL;
                                }
                            }
                        }
                    }
                }
            }
            *tp = guess;
            return NULL;
        }
    }
    /* Given argument has no corresponding time_t. Let's outerpolation. */
    /*
     *  `Seconds Since the Epoch' in SUSv3:
     *  tm_sec + tm_min*60 + tm_hour*3600 + tm_yday*86400 +
     *  (tm_year-70)*31536000 + ((tm_year-69)/4)*86400 -
     *  ((tm_year-1)/100)*86400 + ((tm_year+299)/400)*86400
     */

    tptr_tm_yday = calc_tm_yday(tptr->tm_year, tptr->tm_mon, tptr->tm_mday);

    *tp = guess_lo +
          ((tptr->tm_year - tm_lo.tm_year) * 365 +
           ((tptr->tm_year-69)/4) -
           ((tptr->tm_year-1)/100) +
           ((tptr->tm_year+299)/400) -
           ((tm_lo.tm_year-69)/4) +
           ((tm_lo.tm_year-1)/100) -
           ((tm_lo.tm_year+299)/400) +
           tptr_tm_yday -
           tm_lo.tm_yday) * 86400 +
          (tptr->tm_hour - tm_lo.tm_hour) * 3600 +
          (tptr->tm_min - tm_lo.tm_min) * 60 +
          (tptr->tm_sec - tm_lo.tm_sec);

    return NULL;

  out_of_range:
    return "time out of range";

  error:
    return "gmtime/localtime error";
}

static int
vtmcmp(mrb_state *mrb, struct vtm *a, struct vtm *b)
{
    if (ne(mrb, mrb_fixnum_number(a->year), mrb_fixnum_number(b->year)))
        return lt(mrb, mrb_fixnum_number(a->year), mrb_fixnum_number(b->year)) ? -1 : 1;
    else if (a->mon != b->mon)
        return a->mon < b->mon ? -1 : 1;
    else if (a->mday != b->mday)
        return a->mday < b->mday ? -1 : 1;
    else if (a->hour != b->hour)
        return a->hour < b->hour ? -1 : 1;
    else if (a->min != b->min)
        return a->min < b->min ? -1 : 1;
    else if (a->sec != b->sec)
        return a->sec < b->sec ? -1 : 1;
    else if (ne(mrb, a->subsecx, b->subsecx))
        return lt(mrb, a->subsecx, b->subsecx) ? -1 : 1;
    else
        return 0;
}

static int
tmcmp(struct tm *a, struct tm *b)
{
    if (a->tm_year != b->tm_year)
        return a->tm_year < b->tm_year ? -1 : 1;
    else if (a->tm_mon != b->tm_mon)
        return a->tm_mon < b->tm_mon ? -1 : 1;
    else if (a->tm_mday != b->tm_mday)
        return a->tm_mday < b->tm_mday ? -1 : 1;
    else if (a->tm_hour != b->tm_hour)
        return a->tm_hour < b->tm_hour ? -1 : 1;
    else if (a->tm_min != b->tm_min)
        return a->tm_min < b->tm_min ? -1 : 1;
    else if (a->tm_sec != b->tm_sec)
        return a->tm_sec < b->tm_sec ? -1 : 1;
    else
        return 0;
}

static mrb_value
time_utc_or_local(mrb_state *mrb, int argc, mrb_value *argv, int utc_p, mrb_value klass)
{
    struct vtm vtm;
    mrb_value time;

    time_arg(mrb, argc, argv, &vtm);
    if (utc_p) {
        time = time_new_timew(mrb, mrb_class_get(mrb, klass), timegmw(mrb, &vtm));
        return time_gmtime(mrb, time);
    }
    else {
        time = time_new_timew(mrb, mrb_class_get(mrb, klass), timelocalw(mrb, &vtm));
        return time_localtime(mrb, time);
    }
}

/* 15.2.19.6.2  */
/* 15.2.19.6.6  */
/*
 *  call-seq:
 *    Time.utc(year) -> time
 *    Time.utc(year, month) -> time
 *    Time.utc(year, month, day) -> time
 *    Time.utc(year, month, day, hour) -> time
 *    Time.utc(year, month, day, hour, min) -> time
 *    Time.utc(year, month, day, hour, min, sec_with_frac) -> time
 *    Time.utc(year, month, day, hour, min, sec, usec_with_frac) -> time
 *    Time.utc(sec, min, hour, day, month, year, wday, yday, isdst, tz) -> time
 *    Time.gm(year) -> time
 *    Time.gm(year, month) -> time
 *    Time.gm(year, month, day) -> time
 *    Time.gm(year, month, day, hour) -> time
 *    Time.gm(year, month, day, hour, min) -> time
 *    Time.gm(year, month, day, hour, min, sec_with_frac) -> time
 *    Time.gm(year, month, day, hour, min, sec, usec_with_frac) -> time
 *    Time.gm(sec, min, hour, day, month, year, wday, yday, isdst, tz) -> time
 *
 *  Creates a time based on given values, interpreted as UTC (GMT). The
 *  year must be specified. Other values default to the minimum value
 *  for that field (and may be <code>nil</code> or omitted). Months may
 *  be specified by numbers from 1 to 12, or by the three-letter English
 *  month names. Hours are specified on a 24-hour clock (0..23). Raises
 *  an <code>ArgumentError</code> if any values are out of range. Will
 *  also accept ten arguments in the order output by
 *  <code>Time#to_a</code>.
 *  <i>sec_with_frac</i> and <i>usec_with_frac</i> can have a fractional part.
 *
 *     Time.utc(2000,"jan",1,20,15,1)  #=> 2000-01-01 20:15:01 UTC
 *     Time.gm(2000,"jan",1,20,15,1)   #=> 2000-01-01 20:15:01 UTC
 */
static mrb_value
time_s_mkutc(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value klass)
{
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);
  return time_utc_or_local(mrb, argc, argv, TRUE, klass);
}

/* 15.2.19.6.3  */
/* 15.2.19.6.4  */
/*
 *  call-seq:
 *   Time.local(year) -> time
 *   Time.local(year, month) -> time
 *   Time.local(year, month, day) -> time
 *   Time.local(year, month, day, hour) -> time
 *   Time.local(year, month, day, hour, min) -> time
 *   Time.local(year, month, day, hour, min, sec_with_frac) -> time
 *   Time.local(year, month, day, hour, min, sec, usec_with_frac) -> time
 *   Time.local(sec, min, hour, day, month, year, wday, yday, isdst, tz) -> time
 *   Time.mktime(year) -> time
 *   Time.mktime(year, month) -> time
 *   Time.mktime(year, month, day) -> time
 *   Time.mktime(year, month, day, hour) -> time
 *   Time.mktime(year, month, day, hour, min) -> time
 *   Time.mktime(year, month, day, hour, min, sec_with_frac) -> time
 *   Time.mktime(year, month, day, hour, min, sec, usec_with_frac) -> time
 *   Time.mktime(sec, min, hour, day, month, year, wday, yday, isdst, tz) -> time
 *
 *  Same as <code>Time::gm</code>, but interprets the values in the
 *  local time zone.
 *
 *     Time.local(2000,"jan",1,20,15,1)   #=> 2000-01-01 20:15:01 -0600
 */

static mrb_value
time_s_mktime(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value klass)
{
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);
  return time_utc_or_local(mrb, argc, argv, FALSE, klass);
}

/* 15.2.19.7.25 */
/*
 *  call-seq:
 *     time.to_i   -> int
 *     time.tv_sec -> int
 *
 *  Returns the value of <i>time</i> as an integer number of seconds
 *  since the Epoch.
 *
 *     t = Time.now
 *     "%10.5f" % t.to_f   #=> "1270968656.89607"
 *     t.to_i              #=> 1270968656
 */

static mrb_value
time_to_i(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    return mrb_fixnum_value(w2v(wdiv(mrb, tobj->timew, WINT2FIXWV(TIME_SCALE))));
}

/* 15.2.19.7.24 */
/*
 *  call-seq:
 *     time.to_f -> float
 *
 *  Returns the value of <i>time</i> as a floating point number of
 *  seconds since the Epoch.
 *
 *     t = Time.now
 *     "%10.5f" % t.to_f   #=> "1270968744.77658"
 *     t.to_i              #=> 1270968744
 *
 *  Note that IEEE 754 double is not accurate enough to represent
 *  number of nanoseconds from the Epoch.
 */

static mrb_value
time_to_f(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    return rb_Float(mrb, rb_time_unmagnify_to_float(mrb, tobj->timew));
}

/*
 *  call-seq:
 *     time.to_r -> a_rational
 *
 *  Returns the value of <i>time</i> as a rational number of seconds
 *  since the Epoch.
 *
 *     t = Time.now
 *     p t.to_r            #=> (1270968792716287611/1000000000)
 *
 *  This methods is intended to be used to get an accurate value
 *  representing nanoseconds from the Epoch.  You can use this
 *  to convert time to another Epoch.
 */
#if 0
static mrb_value
time_to_r(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;
    mrb_value v;

    GetTimeval(mrb, time, tobj);
    v = w2v(rb_time_unmagnify(mrb, tobj->timew));
    //if (mrb_type(v) != T_RATIONAL) {
        v = rb_Rational1(v);
    //}
    return v;
}
#endif

/* 15.2.19.7.26 */
/*
 *  call-seq:
 *     time.usec    -> int
 *     time.tv_usec -> int
 *
 *  Returns just the number of microseconds for <i>time</i>.
 *
 *     t = Time.now        #=> 2007-11-19 08:03:26 -0600
 *     "%10.6f" % t.to_f   #=> "1195481006.775195"
 *     t.usec              #=> 775195
 */

static mrb_value
time_usec(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;
    wideval_t w, q, r;

    GetTimeval(mrb, time, tobj);

    w = wmod(mrb, tobj->timew, WINT2WV(TIME_SCALE));
    wmuldivmod(mrb, w, WINT2FIXWV(1000000), WINT2FIXWV(TIME_SCALE), &q, &r);
    //return rb_to_int(mrb, w2v(q));
    return mrb_fixnum_value(q);
}

/*
 *  call-seq:
 *     time.nsec    -> int
 *     time.tv_nsec -> int
 *
 *  Returns just the number of nanoseconds for <i>time</i>.
 *
 *     t = Time.now        #=> 2007-11-17 15:18:03 +0900
 *     "%10.9f" % t.to_f   #=> "1195280283.536151409"
 *     t.nsec              #=> 536151406
 *
 *  The lowest digit of to_f and nsec is different because
 *  IEEE 754 double is not accurate enough to represent
 *  nanoseconds from the Epoch.
 *  The accurate value is returned by nsec.
 */

static mrb_value
time_nsec(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    //return rb_to_int(mrb, w2v(wmulquoll(mrb, wmod(mrb, tobj->timew, WINT2WV(TIME_SCALE)), 1000000000, TIME_SCALE)));
    return mrb_fixnum_value(wmulquoll(mrb, wmod(mrb, tobj->timew, WINT2WV(TIME_SCALE)), 1000000000, TIME_SCALE));
}

/*
 *  call-seq:
 *     time.subsec    -> number
 *
 *  Returns just the fraction for <i>time</i>.
 *
 *  The result is possibly rational.
 *
 *     t = Time.now        #=> 2009-03-26 22:33:12 +0900
 *     "%10.9f" % t.to_f   #=> "1238074392.940563917"
 *     t.subsec            #=> (94056401/100000000)
 *
 *  The lowest digit of to_f and subsec is different because
 *  IEEE 754 double is not accurate enough to represent
 *  the rational.
 *  The accurate value is returned by subsec.
 */

static mrb_value
time_subsec(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    return mrb_fixnum_value(quo(mrb, w2v(wmod(mrb, tobj->timew, WINT2FIXWV(TIME_SCALE))), (TIME_SCALE)));
}

/* 15.2.19.7.3  */
/*
 *  call-seq:
 *     time <=> other_time -> -1, 0, +1 or nil
 *
 *  Comparison---Compares <i>time</i> with <i>other_time</i>.
 *
 *     t = Time.now       #=> 2007-11-19 08:12:12 -0600
 *     t2 = t + 2592000   #=> 2007-12-19 08:12:12 -0600
 *     t <=> t2           #=> -1
 *     t2 <=> t           #=> 1
 *
 *     t = Time.now       #=> 2007-11-19 08:13:38 -0600
 *     t2 = t + 0.1       #=> 2007-11-19 08:13:38 -0600
 *     t.nsec             #=> 98222999
 *     t2.nsec            #=> 198222999
 *     t <=> t2           #=> -1
 *     t2 <=> t           #=> 1
 *     t <=> t            #=> 0
 */

static mrb_value
time_cmp(mrb_state *mrb, mrb_value time1)
{
  struct time_object *tobj1, *tobj2;
  int n;
  mrb_value time2;

  mrb_get_args(mrb, "o", &time2);

  GetTimeval(mrb, time1, tobj1);
  if (IsTimeval(time2)) {
      GetTimeval(mrb, time2, tobj2);
      n = wcmp(mrb, tobj1->timew, tobj2->timew);
  }
  else {
      mrb_value tmp;

      tmp = mrb_funcall(mrb, time2, "<=>", 1, time1);
      if (mrb_nil_p(tmp)) return mrb_nil_value();

      n = -rb_cmpint(mrb, tmp, time1, time2);
  }
  if (n == 0) return mrb_fixnum_value(0);
  if (n > 0) return mrb_fixnum_value(1);
  return mrb_fixnum_value(-1);
}

/*
 * call-seq:
 *  time.eql?(other_time)
 *
 * Return <code>true</code> if <i>time</i> and <i>other_time</i> are
 * both <code>Time</code> objects with the same seconds and fractional
 * seconds.
 */

static mrb_value
time_eql(mrb_state *mrb, mrb_value time1)
{
  struct time_object *tobj1, *tobj2;
  mrb_value time2;

  mrb_get_args(mrb, "o", &time2);

  GetTimeval(mrb, time1, tobj1);
  if (IsTimeval(time2)) {
      GetTimeval(mrb, time2, tobj2);
      if (w2v(tobj1->timew) == w2v(tobj2->timew))
        return mrb_true_value();
  }
  return mrb_false_value();
}

/* 15.2.19.7.11 */
/* 15.2.19.7.28 */
/*
 *  call-seq:
 *     time.utc? -> true or false
 *     time.gmt? -> true or false
 *
 *  Returns <code>true</code> if <i>time</i> represents a time in UTC
 *  (GMT).
 *
 *     t = Time.now                        #=> 2007-11-19 08:15:23 -0600
 *     t.utc?                              #=> false
 *     t = Time.gm(2000,"jan",1,20,15,1)   #=> 2000-01-01 20:15:01 UTC
 *     t.utc?                              #=> true
 *
 *     t = Time.now                        #=> 2007-11-19 08:16:03 -0600
 *     t.gmt?                              #=> false
 *     t = Time.gm(2000,1,1,20,15,1)       #=> 2000-01-01 20:15:01 UTC
 *     t.gmt?                              #=> true
 */

static mrb_value
time_utc_p(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    if (TIME_UTC_P(tobj)) return mrb_true_value();
    return mrb_false_value();
}

#if 0
/*
 * call-seq:
 *   time.hash   -> fixnum
 *
 * Return a hash code for this time object.
 */

static mrb_value
time_hash(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    return rb_hash(mrb, w2v(tobj->timew));
}
#endif

static mrb_value
time_init_copytime(mrb_state *mrb, mrb_value copy, mrb_value time)
{
  struct time_object *tobj, *tcopy;
  if (mrb_obj_equal(mrb, copy, time)) return copy;

  time_modify(copy);
  //ttime = time_s_alloc(mrb, mrb->time_class); /* kusuda */
  tcopy = malloc(sizeof(struct time_object));
  memset(tcopy, 0, sizeof(struct time_object));
  RTYPEDDATA(copy)->data = tcopy;
  RTYPEDDATA(copy)->typed_flag = 1;
  RTYPEDDATA(copy)->type = &time_data_type;

  tcopy->tm_got=0;
  tcopy->timew = WINT2FIXWV(0);

/*********/
  GetTimeval(mrb, time, tobj);
  GetTimeval(mrb, copy, tcopy);
  //MEMCPY(tcopy, tobj, struct time_object, 1);
  memcpy(tcopy, tobj, sizeof(struct time_object));

  return copy;
}
/* 15.2.19.7.17 */
/* :nodoc: */
static mrb_value
time_init_copy(mrb_state *mrb, mrb_value copy)
{
  mrb_value time;

  mrb_get_args(mrb, "o", &time);

  return time_init_copytime(mrb, copy, time);
}

static mrb_value
time_dup(mrb_state *mrb, mrb_value time)
{
    mrb_value dup = mrb_obj_value(time_s_alloc(mrb, mrb_class_get(mrb, time))); //CLASS_OF(time));
    time_init_copytime(mrb, dup, time);
    return dup;
}

static mrb_value
time_localtime(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;
    struct vtm vtm;

    GetTimeval(mrb, time, tobj);
    if (TIME_LOCALTIME_P(tobj)) {
        if (tobj->tm_got)
            return time;
    }
    else {
        time_modify(time);
    }

    if (!localtimew(mrb, tobj->timew, &vtm))
        mrb_raise(mrb, mrb->eArgumentError_class, "localtime error");
    tobj->vtm = vtm;

    tobj->tm_got = 1;
    TIME_SET_LOCALTIME(tobj);
    return time;
}

mrb_value glb_time_localtime(mrb_state *mrb, mrb_value time)
{
  return time_localtime(mrb, time);
}

/* 15.2.19.7.18 */
/*
 *  call-seq:
 *     time.localtime -> time
 *     time.localtime(utc_offset) -> time
 *
 *  Converts <i>time</i> to local time (using the local time zone in
 *  effect for this process) modifying the receiver.
 *
 *  If _utc_offset_ is given, it is used instead of the local time.
 *
 *     t = Time.utc(2000, "jan", 1, 20, 15, 1) #=> 2000-01-01 20:15:01 UTC
 *     t.utc?                                  #=> true
 *
 *     t.localtime                             #=> 2000-01-01 14:15:01 -0600
 *     t.utc?                                  #=> false
 *
 *     t.localtime("+09:00")                   #=> 2000-01-02 05:15:01 +0900
 *     t.utc?                                  #=> false
 */

static mrb_value
time_localtime_m(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value time)
{
  mrb_value off;
  mrb_value argv[16];
  int argc;

  //rb_scan_args(argc, argv, "01", &off);
  off = mrb_nil_value();
  mrb_get_args(mrb, "*", &argv, &argc);
  if (argc > 0) off = argv[0];

  if (!mrb_nil_p(off)) {
    off = utc_offset_arg(mrb, off);
    validate_utc_offset(mrb, off);

    time_set_utc_offset(mrb, time, off);
    return time_fixoff(mrb, time);
  }

  return time_localtime(mrb, time);
}

/* 15.2.19.7.13 */
/* 15.2.19.7.27 */
/*
 *  call-seq:
 *     time.gmtime    -> time
 *     time.utc       -> time
 *
 *  Converts <i>time</i> to UTC (GMT), modifying the receiver.
 *
 *     t = Time.now   #=> 2007-11-19 08:18:31 -0600
 *     t.gmt?         #=> false
 *     t.gmtime       #=> 2007-11-19 14:18:31 UTC
 *     t.gmt?         #=> true
 *
 *     t = Time.now   #=> 2007-11-19 08:18:51 -0600
 *     t.utc?         #=> false
 *     t.utc          #=> 2007-11-19 14:18:51 UTC
 *     t.utc?         #=> true
 */

static mrb_value
time_gmtime(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;
    struct vtm vtm;

    GetTimeval(mrb, time, tobj);
    if (TIME_UTC_P(tobj)) {
        if (tobj->tm_got)
            return time;
    }
    else {
        time_modify(time);
    }

    if (!gmtimew(mrb, tobj->timew, &vtm))
        mrb_raise(mrb, mrb->eArgumentError_class, "gmtime error");
    tobj->vtm = vtm;

    tobj->tm_got = 1;
    TIME_SET_UTC(tobj);
    return time;
}

static mrb_value
time_fixoff(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;
    struct vtm vtm;
    mrb_value off;

    GetTimeval(mrb, time, tobj);
    if (TIME_FIXOFF_P(tobj)) {
       if (tobj->tm_got)
           return time;
    }
    else {
       time_modify(time);
    }

    if (TIME_FIXOFF_P(tobj))
        off = tobj->vtm.utc_offset;
    else
        off = mrb_fixnum_value(0);

    if (!gmtimew(mrb, tobj->timew, &vtm))
       mrb_raise(mrb, mrb->eArgumentError_class, "gmtime error");

    tobj->vtm = vtm;
    vtm_add_offset(mrb, &tobj->vtm, off);

    tobj->tm_got = 1;
    TIME_SET_FIXOFF(tobj, off);
    return time;
}

/* 15.2.19.7.9  */
/*
 *  call-seq:
 *     time.getlocal -> new_time
 *     time.getlocal(utc_offset) -> new_time
 *
 *  Returns a new <code>new_time</code> object representing <i>time</i> in
 *  local time (using the local time zone in effect for this process).
 *
 *  If _utc_offset_ is given, it is used instead of the local time.
 *
 *     t = Time.utc(2000,1,1,20,15,1)  #=> 2000-01-01 20:15:01 UTC
 *     t.utc?                          #=> true
 *
 *     l = t.getlocal                  #=> 2000-01-01 14:15:01 -0600
 *     l.utc?                          #=> false
 *     t == l                          #=> true
 *
 *     j = t.getlocal("+09:00")        #=> 2000-01-02 05:15:01 +0900
 *     j.utc?                          #=> false
 *     t == j                          #=> true
 */

static mrb_value
time_getlocaltime(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value time)
{
  mrb_value off;
  mrb_value argv[16];
  int argc;

  //rb_scan_args(argc, argv, "01", &off);
  off = mrb_nil_value();
  mrb_get_args(mrb, "*", &argv, &argc);
  if (argc > 0) off = argv[0];

    if (!mrb_nil_p(off)) {
        off = utc_offset_arg(mrb, off);
        validate_utc_offset(mrb, off);

        time = time_dup(mrb, time);
        time_set_utc_offset(mrb, time, off);
        return time_fixoff(mrb, time);
    }

    return time_localtime(mrb, time_dup(mrb, time));
}

/* 15.2.19.7.8  */
/* 15.2.19.7.10 */
/*
 *  call-seq:
 *     time.getgm  -> new_time
 *     time.getutc -> new_time
 *
 *  Returns a new <code>new_time</code> object representing <i>time</i> in
 *  UTC.
 *
 *     t = Time.local(2000,1,1,20,15,1)   #=> 2000-01-01 20:15:01 -0600
 *     t.gmt?                             #=> false
 *     y = t.getgm                        #=> 2000-01-02 02:15:01 UTC
 *     y.gmt?                             #=> true
 *     t == y                             #=> true
 */

static mrb_value
time_getgmtime(mrb_state *mrb, mrb_value time)
{
    return time_gmtime(mrb, time_dup(mrb, time));
}

static mrb_value
time_get_tm(mrb_state *mrb, mrb_value time, struct time_object *tobj)
{
    if (TIME_UTC_P(tobj)) return time_gmtime(mrb, time);
    if (TIME_FIXOFF_P(tobj)) return time_fixoff(mrb, time);
    return time_localtime(mrb, time);
}

static mrb_value strftimev(mrb_state *mrb, const char *fmt, mrb_value time);

/* 15.2.19.7.4  */
/* 15.2.19.7.5  */
/*
 *  call-seq:
 *     time.asctime -> string
 *     time.ctime   -> string
 *
 *  Returns a canonical string representation of <i>time</i>.
 *
 *     Time.now.asctime   #=> "Wed Apr  9 08:56:03 2003"
 */

static mrb_value
time_asctime(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    return strftimev(mrb, "%a %b %e %T %Y", time);
}

/*
 *  call-seq:
 *     time.inspect -> string
 *     time.to_s    -> string
 *
 *  Returns a string representing <i>time</i>. Equivalent to calling
 *  <code>Time#strftime</code> with a format string of
 *  ``<code>%Y-%m-%d</code> <code>%H:%M:%S</code> <code>%z</code>''
 *  for a local time and
 *  ``<code>%Y-%m-%d</code> <code>%H:%M:%S</code> <code>UTC</code>''
 *  for a UTC time.
 *
 *     Time.now.to_s       #=> "2007-10-05 16:09:51 +0900"
 *     Time.now.utc.to_s   #=> "2007-10-05 07:09:51 UTC"
 */

static mrb_value
time_to_s(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    if (TIME_UTC_P(tobj))
        return strftimev(mrb, "%Y-%m-%d %H:%M:%S UTC", time);
    else
        return strftimev(mrb, "%Y-%m-%d %H:%M:%S %z", time);
}

static mrb_value
time_add(mrb_state *mrb, struct time_object *tobj, mrb_value voffset, int sign)
{
  mrb_value result;

  voffset = num_exact(mrb, (voffset));
  if (sign < 0)
    result = time_new_timew(mrb, mrb->time_class, wsub(mrb, tobj->timew, rb_time_magnifyFromMrbValue(mrb, v2w(voffset))));
  else
    result = time_new_timew(mrb, mrb->time_class, wadd(mrb, tobj->timew, rb_time_magnifyFromMrbValue(mrb, v2w(voffset))));

  if (TIME_UTC_P(tobj)) {
    GetTimeval(mrb, result, tobj);
    TIME_SET_UTC(tobj);
  }
  else if (TIME_FIXOFF_P(tobj)) {
    mrb_value off = tobj->vtm.utc_offset;
    GetTimeval(mrb, result, tobj);
    TIME_SET_FIXOFF(tobj, off);
  }
  return result;
}

/* 15.2.19.7.1  */
/*
 *  call-seq:
 *     time + numeric -> time
 *
 *  Addition---Adds some number of seconds (possibly fractional) to
 *  <i>time</i> and returns that value as a new time.
 *
 *     t = Time.now         #=> 2007-11-19 08:22:21 -0600
 *     t + (60 * 60 * 24)   #=> 2007-11-20 08:22:21 -0600
 */

static mrb_value
time_plus(mrb_state *mrb, mrb_value time1)
{
  struct time_object *tobj;
  mrb_value time2;

  mrb_get_args(mrb, "o", &time2);
  GetTimeval(mrb, time1, tobj);
  if (IsTimeval(time2)) {
    mrb_raise(mrb, mrb->eTypeError_class, "time + time?");
  }
  return time_add(mrb, tobj, (time2), 1);
}

/* 15.2.19.7.2  */
/*
 *  call-seq:
 *     time - other_time -> float
 *     time - numeric    -> time
 *
 *  Difference---Returns a new time that represents the difference
 *  between two times, or subtracts the given number of seconds in
 *  <i>numeric</i> from <i>time</i>.
 *
 *     t = Time.now       #=> 2007-11-19 08:23:10 -0600
 *     t2 = t + 2592000   #=> 2007-12-19 08:23:10 -0600
 *     t2 - t             #=> 2592000.0
 *     t2 - 2592000       #=> 2007-11-19 08:23:10 -0600
 */

static mrb_value
time_minus(mrb_state *mrb, mrb_value time1)
{
  struct time_object *tobj;
  mrb_value time2;

  mrb_get_args(mrb, "o", &time2);

  GetTimeval(mrb, time1, tobj);
  if (IsTimeval(time2)) {
      struct time_object *tobj2;

      GetTimeval(mrb, time2, tobj2);
      return rb_Float(mrb, rb_time_unmagnify_to_float(mrb, wsub(mrb, tobj->timew, tobj2->timew)));
  }
  return time_add(mrb, tobj, (time2), -1);
}

/*
 * call-seq:
 *   time.succ   -> new_time
 *
 * Return a new time object, one second later than <code>time</code>.
 * Time#succ is obsolete since 1.9.2 for time is not a discrete value.
 *
 *     t = Time.now       #=> 2007-11-19 08:23:57 -0600
 *     t.succ             #=> 2007-11-19 08:23:58 -0600
 */

mrb_value
rb_time_succ(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;
    struct time_object *tobj2;

    rb_warn("Time#succ is obsolete; use time + 1");
    GetTimeval(mrb, time, tobj);
    time = time_new_timew(mrb, mrb->time_class, wadd(mrb, tobj->timew, WINT2FIXWV(TIME_SCALE)));
    GetTimeval(mrb, time, tobj2);
    TIME_COPY_GMT(tobj2, tobj);
    return time;
}

#define time_succ rb_time_succ

/*
 * call-seq:
 *   time.round([ndigits])   -> new_time
 *
 * Rounds sub seconds to a given precision in decimal digits (0 digits by default).
 * It returns a new time object.
 * _ndigits_ should be zero or positive integer.
 *
 *     require 'time'
 *
 *     t = Time.utc(2010,3,30, 5,43,"25.123456789".to_r)
 *     p t.iso8601(10)           #=> "2010-03-30T05:43:25.1234567890Z"
 *     p t.round.iso8601(10)     #=> "2010-03-30T05:43:25.0000000000Z"
 *     p t.round(0).iso8601(10)  #=> "2010-03-30T05:43:25.0000000000Z"
 *     p t.round(1).iso8601(10)  #=> "2010-03-30T05:43:25.1000000000Z"
 *     p t.round(2).iso8601(10)  #=> "2010-03-30T05:43:25.1200000000Z"
 *     p t.round(3).iso8601(10)  #=> "2010-03-30T05:43:25.1230000000Z"
 *     p t.round(4).iso8601(10)  #=> "2010-03-30T05:43:25.1235000000Z"
 *     p t.round(5).iso8601(10)  #=> "2010-03-30T05:43:25.1234600000Z"
 *     p t.round(6).iso8601(10)  #=> "2010-03-30T05:43:25.1234570000Z"
 *     p t.round(7).iso8601(10)  #=> "2010-03-30T05:43:25.1234568000Z"
 *     p t.round(8).iso8601(10)  #=> "2010-03-30T05:43:25.1234567900Z"
 *     p t.round(9).iso8601(10)  #=> "2010-03-30T05:43:25.1234567890Z"
 *     p t.round(10).iso8601(10) #=> "2010-03-30T05:43:25.1234567890Z"
 *
 *     t = Time.utc(1999,12,31, 23,59,59)
 *     p((t + 0.4).round.iso8601(3))    #=> "1999-12-31T23:59:59.000Z"
 *     p((t + 0.49).round.iso8601(3))   #=> "1999-12-31T23:59:59.000Z"
 *     p((t + 0.5).round.iso8601(3))    #=> "2000-01-01T00:00:00.000Z"
 *     p((t + 1.4).round.iso8601(3))    #=> "2000-01-01T00:00:00.000Z"
 *     p((t + 1.49).round.iso8601(3))   #=> "2000-01-01T00:00:00.000Z"
 *     p((t + 1.5).round.iso8601(3))    #=> "2000-01-01T00:00:01.000Z"
 *
 *     t = Time.utc(1999,12,31, 23,59,59)
 *     p (t + 0.123456789).round(4).iso8601(6)  #=> "1999-12-31T23:59:59.123500Z"
 */

static mrb_value
time_round(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value time)
{
  int ndigits, a, b;
  wideval_t v, den, va, vb;
  long nd;
  struct time_object *tobj;
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);

  //rb_scan_args(argc, argv, "01", &ndigits);
  ndigits = 0;
  mrb_get_args(mrb, "*", &argv, &argc);
  if (argc > 0) ndigits = mrb_fixnum_number(argv[0]);


  nd = ndigits;
  if (nd < 0)
      mrb_raise(mrb, mrb->eArgumentError_class, "negative ndigits given");

  GetTimeval(mrb, time, tobj);
  v = w2v(rb_time_unmagnify(mrb, tobj->timew));

  a = 1;
  b = 10;
  while (0 < nd) {
      if (nd & 1)
          va = mul(mrb, (a), (b));
      vb = mul(mrb, (b), (b));
      nd = nd >> 1;
  }
  den = quo(mrb, (1), va);
  v = mod(mrb, v, den);
  if (lt(mrb, v, quo(mrb, den, (2))))
      return time_add(mrb, tobj, mrb_fixnum_value(v), -1);
  else
      return time_add(mrb, tobj, mrb_fixnum_value(sub(mrb, den, v)), 1);
}

/* 15.2.19.7.23 */
/*
 *  call-seq:
 *     time.sec -> fixnum
 *
 *  Returns the second of the minute (0..60)<em>[Yes, seconds really can
 *  range from zero to 60. This allows the system to inject leap seconds
 *  every now and then to correct for the fact that years are not really
 *  a convenient number of hours long.]</em> for <i>time</i>.
 *
 *     t = Time.now   #=> 2007-11-19 08:25:02 -0600
 *     t.sec          #=> 2
 */

static mrb_value
time_sec(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    MAKE_TM(mrb, time, tobj);
    return mrb_fixnum_value(tobj->vtm.sec);
}

/* 15.2.19.7.20 */
/*
 *  call-seq:
 *     time.min -> fixnum
 *
 *  Returns the minute of the hour (0..59) for <i>time</i>.
 *
 *     t = Time.now   #=> 2007-11-19 08:25:51 -0600
 *     t.min          #=> 25
 */

static mrb_value
time_min(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    MAKE_TM(mrb, time, tobj);
    return mrb_fixnum_value(tobj->vtm.min);
}

/* 15.2.19.7.15 */
/*
 *  call-seq:
 *     time.hour -> fixnum
 *
 *  Returns the hour of the day (0..23) for <i>time</i>.
 *
 *     t = Time.now   #=> 2007-11-19 08:26:20 -0600
 *     t.hour         #=> 8
 */

static mrb_value
time_hour(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    MAKE_TM(mrb, time, tobj);
    return mrb_fixnum_value(tobj->vtm.hour);
}

int int_time_hour(mrb_state *mrb, mrb_value time)
{
  return mrb_fixnum_number(time_hour(mrb, time));
}
int int_time_min(mrb_state *mrb, mrb_value time)
{
  return mrb_fixnum_number(time_min(mrb, time));
}
int int_time_sec(mrb_state *mrb, mrb_value time)
{
  return mrb_fixnum_number(time_sec(mrb, time));
}

/* 15.2.19.7.6  */
/* 15.2.19.7.19 */
/*
 *  call-seq:
 *     time.day  -> fixnum
 *     time.mday -> fixnum
 *
 *  Returns the day of the month (1..n) for <i>time</i>.
 *
 *     t = Time.now   #=> 2007-11-19 08:27:03 -0600
 *     t.day          #=> 19
 *     t.mday         #=> 19
 */

static mrb_value
time_mday(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    MAKE_TM(mrb, time, tobj);
    return mrb_fixnum_value(tobj->vtm.mday);
}

/* 15.2.19.7.21 */
/* 15.2.19.7.22 */
/*
 *  call-seq:
 *     time.mon   -> fixnum
 *     time.month -> fixnum
 *
 *  Returns the month of the year (1..12) for <i>time</i>.
 *
 *     t = Time.now   #=> 2007-11-19 08:27:30 -0600
 *     t.mon          #=> 11
 *     t.month        #=> 11
 */

static mrb_value
time_mon(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    MAKE_TM(mrb, time, tobj);
    return mrb_fixnum_value(tobj->vtm.mon);
}

/* 15.2.19.7.32 */
/*
 *  call-seq:
 *     time.year -> fixnum
 *
 *  Returns the year for <i>time</i> (including the century).
 *
 *     t = Time.now   #=> 2007-11-19 08:27:51 -0600
 *     t.year         #=> 2007
 */

static mrb_value
time_year(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    MAKE_TM(mrb, time, tobj);
    return tobj->vtm.year;
}

int int_time_year(mrb_state *mrb, mrb_value time)
{
  return mrb_fixnum_number(time_year(mrb, time));
}
int int_time_month(mrb_state *mrb, mrb_value time)
{
  return mrb_fixnum_number(time_mon(mrb, time));
}
int int_time_mday(mrb_state *mrb, mrb_value time)
{
  return mrb_fixnum_number(time_mday(mrb, time));
}
/* 15.2.19.7.30 */
/*
 *  call-seq:
 *     time.wday -> fixnum
 *
 *  Returns an integer representing the day of the week, 0..6, with
 *  Sunday == 0.
 *
 *     t = Time.now   #=> 2007-11-20 02:35:35 -0600
 *     t.wday         #=> 2
 *     t.sunday?      #=> false
 *     t.monday?      #=> false
 *     t.tuesday?     #=> true
 *     t.wednesday?   #=> false
 *     t.thursday?    #=> false
 *     t.friday?      #=> false
 *     t.saturday?    #=> false
 */

static mrb_value
time_wday(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    MAKE_TM(mrb, time, tobj);
    return mrb_fixnum_value(tobj->vtm.wday);
}

#define wday_p(mrb, n) {\
    struct time_object *tobj;\
    GetTimeval(mrb, time, tobj);\
    MAKE_TM(mrb, time, tobj);\
    return (tobj->vtm.wday == (n)) ? mrb_true_value() : mrb_false_value();\
}

/*
 *  call-seq:
 *     time.sunday? -> true or false
 *
 *  Returns <code>true</code> if <i>time</i> represents Sunday.
 *
 *     t = Time.local(1990, 4, 1)       #=> 1990-04-01 00:00:00 -0600
 *     t.sunday?                        #=> true
 */

static mrb_value
time_sunday(mrb_state *mrb, mrb_value time)
{
    wday_p(mrb, 0);
}

/*
 *  call-seq:
 *     time.monday? -> true or false
 *
 *  Returns <code>true</code> if <i>time</i> represents Monday.
 *
 *     t = Time.local(2003, 8, 4)       #=> 2003-08-04 00:00:00 -0500
 *     p t.monday?                      #=> true
 */

static mrb_value
time_monday(mrb_state *mrb, mrb_value time)
{
    wday_p(mrb, 1);
}

/*
 *  call-seq:
 *     time.tuesday? -> true or false
 *
 *  Returns <code>true</code> if <i>time</i> represents Tuesday.
 *
 *     t = Time.local(1991, 2, 19)      #=> 1991-02-19 00:00:00 -0600
 *     p t.tuesday?                     #=> true
 */

static mrb_value
time_tuesday(mrb_state *mrb, mrb_value time)
{
    wday_p(mrb, 2);
}

/*
 *  call-seq:
 *     time.wednesday? -> true or false
 *
 *  Returns <code>true</code> if <i>time</i> represents Wednesday.
 *
 *     t = Time.local(1993, 2, 24)      #=> 1993-02-24 00:00:00 -0600
 *     p t.wednesday?                   #=> true
 */

static mrb_value
time_wednesday(mrb_state *mrb, mrb_value time)
{
    wday_p(mrb, 3);
}

/*
 *  call-seq:
 *     time.thursday? -> true or false
 *
 *  Returns <code>true</code> if <i>time</i> represents Thursday.
 *
 *     t = Time.local(1995, 12, 21)     #=> 1995-12-21 00:00:00 -0600
 *     p t.thursday?                    #=> true
 */

static mrb_value
time_thursday(mrb_state *mrb, mrb_value time)
{
    wday_p(mrb, 4);
}

/*
 *  call-seq:
 *     time.friday? -> true or false
 *
 *  Returns <code>true</code> if <i>time</i> represents Friday.
 *
 *     t = Time.local(1987, 12, 18)     #=> 1987-12-18 00:00:00 -0600
 *     t.friday?                        #=> true
 */

static mrb_value
time_friday(mrb_state *mrb, mrb_value time)
{
    wday_p(mrb, 5);
}

/*
 *  call-seq:
 *     time.saturday? -> true or false
 *
 *  Returns <code>true</code> if <i>time</i> represents Saturday.
 *
 *     t = Time.local(2006, 6, 10)      #=> 2006-06-10 00:00:00 -0500
 *     t.saturday?                      #=> true
 */

static mrb_value
time_saturday(mrb_state *mrb, mrb_value time)
{
    wday_p(mrb, 6);
}

/* 15.2.19.7.31 */
/*
 *  call-seq:
 *     time.yday -> fixnum
 *
 *  Returns an integer representing the day of the year, 1..366.
 *
 *     t = Time.now   #=> 2007-11-19 08:32:31 -0600
 *     t.yday         #=> 323
 */

static mrb_value
time_yday(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    MAKE_TM(mrb, time, tobj);
    return mrb_fixnum_value(tobj->vtm.yday);
}

/* 15.2.19.7.7  */
/*
 *  call-seq:
 *     time.isdst -> true or false
 *     time.dst?  -> true or false
 *
 *  Returns <code>true</code> if <i>time</i> occurs during Daylight
 *  Saving Time in its time zone.
 *
 *   # CST6CDT:
 *     Time.local(2000, 1, 1).zone    #=> "CST"
 *     Time.local(2000, 1, 1).isdst   #=> false
 *     Time.local(2000, 1, 1).dst?    #=> false
 *     Time.local(2000, 7, 1).zone    #=> "CDT"
 *     Time.local(2000, 7, 1).isdst   #=> true
 *     Time.local(2000, 7, 1).dst?    #=> true
 *
 *   # Asia/Tokyo:
 *     Time.local(2000, 1, 1).zone    #=> "JST"
 *     Time.local(2000, 1, 1).isdst   #=> false
 *     Time.local(2000, 1, 1).dst?    #=> false
 *     Time.local(2000, 7, 1).zone    #=> "JST"
 *     Time.local(2000, 7, 1).isdst   #=> false
 *     Time.local(2000, 7, 1).dst?    #=> false
 */

static mrb_value
time_isdst(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    MAKE_TM(mrb, time, tobj);
    return tobj->vtm.isdst ? mrb_true_value() : mrb_false_value();
}

/* 15.2.19.7.33 */
/*
 *  call-seq:
 *     time.zone -> string
 *
 *  Returns the name of the time zone used for <i>time</i>. As of Ruby
 *  1.8, returns ``UTC'' rather than ``GMT'' for UTC times.
 *
 *     t = Time.gm(2000, "jan", 1, 20, 15, 1)
 *     t.zone   #=> "UTC"
 *     t = Time.local(2000, "jan", 1, 20, 15, 1)
 *     t.zone   #=> "CST"
 */

static mrb_value
time_zone(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    MAKE_TM(mrb, time, tobj);

    if (TIME_UTC_P(tobj)) {
        return mrb_str_new_cstr(mrb, "UTC");//rb_str_new2("UTC");
    }
    if (tobj->vtm.zone == NULL)
        return mrb_nil_value();
    return mrb_str_new_cstr(mrb, tobj->vtm.zone);//rb_str_new2(tobj->vtm.zone);
}

/* 15.2.19.7.12 */
/* 15.2.19.7.14 */
/* 15.2.19.7.29 */
/*
 *  call-seq:
 *     time.gmt_offset -> fixnum
 *     time.gmtoff     -> fixnum
 *     time.utc_offset -> fixnum
 *
 *  Returns the offset in seconds between the timezone of <i>time</i>
 *  and UTC.
 *
 *     t = Time.gm(2000,1,1,20,15,1)   #=> 2000-01-01 20:15:01 UTC
 *     t.gmt_offset                    #=> 0
 *     l = t.getlocal                  #=> 2000-01-01 14:15:01 -0600
 *     l.gmt_offset                    #=> -21600
 */

static mrb_value
time_utc_offset(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;

    GetTimeval(mrb, time, tobj);
    MAKE_TM(mrb, time, tobj);

    if (TIME_UTC_P(tobj)) {
        return mrb_fixnum_value(0);
    }
    else {
        return tobj->vtm.utc_offset;
    }
}

int int_time_utc_offset(mrb_state *mrb, mrb_value time)
{
  return mrb_fixnum_number(time_utc_offset(mrb, time));
}


/*
 *  call-seq:
 *     time.to_a -> array
 *
 *  Returns a ten-element <i>array</i> of values for <i>time</i>:
 *  {<code>[ sec, min, hour, day, month, year, wday, yday, isdst, zone
 *  ]</code>}. See the individual methods for an explanation of the
 *  valid ranges of each value. The ten elements can be passed directly
 *  to <code>Time::utc</code> or <code>Time::local</code> to create a
 *  new <code>Time</code>.
 *
 *     t = Time.now     #=> 2007-11-19 08:36:01 -0600
 *     now = t.to_a     #=> [1, 36, 8, 19, 11, 2007, 1, 323, false, "CST"]
 */

static mrb_value
time_to_a(mrb_state *mrb, mrb_value time)
{
  struct time_object *tobj;
  mrb_value arv[16];

  GetTimeval(mrb, time, tobj);
  MAKE_TM(mrb, time, tobj);
  //return rb_ary_new3(10,
  //                  mrb_fixnum_value(tobj->vtm.sec),
  //                  mrb_fixnum_value(tobj->vtm.min),
  //                  mrb_fixnum_value(tobj->vtm.hour),
  //                  mrb_fixnum_value(tobj->vtm.mday),
  //                  mrb_fixnum_value(tobj->vtm.mon),
  //                  tobj->vtm.year,
  //                  mrb_fixnum_value(tobj->vtm.wday),
  //                  mrb_fixnum_value(tobj->vtm.yday),
  //                  tobj->vtm.isdst?mrb_true_value():mrb_false_value(),
  //                  time_zone(mrb, time));
  arv[0] = mrb_fixnum_value(tobj->vtm.sec);
  arv[1] = mrb_fixnum_value(tobj->vtm.min);
  arv[2] = mrb_fixnum_value(tobj->vtm.hour);
  arv[3] = mrb_fixnum_value(tobj->vtm.mday);
  arv[4] = mrb_fixnum_value(tobj->vtm.mon);
  arv[5] = tobj->vtm.year;
  arv[6] = mrb_fixnum_value(tobj->vtm.wday);
  arv[7] = mrb_fixnum_value(tobj->vtm.yday);
  arv[8] = tobj->vtm.isdst?mrb_true_value():mrb_false_value();
  arv[9] = time_zone(mrb, time);
  return mrb_ary_new_from_values(mrb, arv, 10);
}

size_t
rb_strftime(char *s, size_t maxsize, const char *format,
            const struct vtm *vtm, wideval_t timev,
            int gmt);

size_t
rb_strftime_timespec(mrb_state *mrb, char *s, size_t maxsize, const char *format, const struct vtm *vtm, struct timespec *ts, int gmt);

#define SMALLBUF 100
static size_t
rb_strftime_alloc(mrb_state *mrb, char **buf, const char *format,
                  struct vtm *vtm, wideval_t timew, int gmt)
{
    size_t size, len, flen;
#if 0 /* 2011_0518 */
    mrb_value timev = mrb_nil_value();
#else
    #define Qnil 0xffff
    wideval_t timev = Qnil;
#endif
    struct timespec ts;

    if (!timew2timespec_exact(mrb, timew, &ts))
        timev = w2v(rb_time_unmagnify(mrb, timew));

    (*buf)[0] = '\0';
    flen = strlen(format);
    if (flen == 0) {
        return 0;
    }
    errno = 0;
#if 0 /* 2011_0518 */
    if (mrb_nil_p(timev))
#else
    if (timev == Qnil)
#endif
        len = rb_strftime_timespec(mrb, *buf, SMALLBUF, format, vtm, &ts, gmt);
    else
        len = rb_strftime(*buf, SMALLBUF, format, vtm, timev, gmt);
    if (len != 0 || (**buf == '\0' && errno != ERANGE)) return len;
    for (size=1024; ; size*=2) {
        *buf = xmalloc(size);
        (*buf)[0] = '\0';
#if 0 /* 2011_0518 */
        if (mrb_nil_p(timev))
#else
        if (timev == Qnil)
#endif
            len = rb_strftime_timespec(mrb, *buf, size, format, vtm, &ts, gmt);
        else
            len = rb_strftime(*buf, size, format, vtm, timev, gmt);
        /*
         * buflen can be zero EITHER because there's not enough
         * room in the string, or because the control command
         * goes to the empty string. Make a reasonable guess that
         * if the buffer is 1024 times bigger than the length of the
         * format string, it's not failing for lack of room.
         */
        if (len > 0 || size >= 1024 * flen) break;
        xfree(*buf);
    }
    return len;
}

static mrb_value
strftimev(mrb_state *mrb, const char *fmt, mrb_value time)
{
    struct time_object *tobj;
    char buffer[SMALLBUF], *buf = buffer;
    long len;
    mrb_value str;

    GetTimeval(mrb, time, tobj);
    MAKE_TM(mrb, time, tobj);
    len = rb_strftime_alloc(mrb, &buf, fmt, &tobj->vtm, tobj->timew, TIME_UTC_P(tobj));
    str = mrb_str_new(mrb, buf, len);
    if (buf != buffer) xfree(buf);
    return str;
}

/*
 *  call-seq:
 *     time.strftime( string ) -> string
 *
 *  Formats <i>time</i> according to the directives in the given format
 *  string. Any text not listed as a directive will be passed through
 *  to the output string.
 *
 *  Format meaning:
 *    %a - The abbreviated weekday name (``Sun'')
 *    %A - The  full  weekday  name (``Sunday'')
 *    %b - The abbreviated month name (``Jan'')
 *    %B - The  full  month  name (``January'')
 *    %c - The preferred local date and time representation
 *    %C - Century (20 in 2009)
 *    %d - Day of the month (01..31)
 *    %D - Date (%m/%d/%y)
 *    %e - Day of the month, blank-padded ( 1..31)
 *    %F - Equivalent to %Y-%m-%d (the ISO 8601 date format)
 *    %h - Equivalent to %b
 *    %H - Hour of the day, 24-hour clock (00..23)
 *    %I - Hour of the day, 12-hour clock (01..12)
 *    %j - Day of the year (001..366)
 *    %k - hour, 24-hour clock, blank-padded ( 0..23)
 *    %l - hour, 12-hour clock, blank-padded ( 0..12)
 *    %L - Millisecond of the second (000..999)
 *    %m - Month of the year (01..12)
 *    %M - Minute of the hour (00..59)
 *    %n - Newline (\n)
 *    %N - Fractional seconds digits, default is 9 digits (nanosecond)
 *            %3N  millisecond (3 digits)
 *            %6N  microsecond (6 digits)
 *            %9N  nanosecond (9 digits)
 *    %p - Meridian indicator (``AM''  or  ``PM'')
 *    %P - Meridian indicator (``am''  or  ``pm'')
 *    %r - time, 12-hour (same as %I:%M:%S %p)
 *    %R - time, 24-hour (%H:%M)
 *    %s - Number of seconds since 1970-01-01 00:00:00 UTC.
 *    %S - Second of the minute (00..60)
 *    %t - Tab character (\t)
 *    %T - time, 24-hour (%H:%M:%S)
 *    %u - Day of the week as a decimal, Monday being 1. (1..7)
 *    %U - Week  number  of the current year,
 *            starting with the first Sunday as the first
 *            day of the first week (00..53)
 *    %v - VMS date (%e-%b-%Y)
 *    %V - Week number of year according to ISO 8601 (01..53)
 *    %W - Week  number  of the current year,
 *            starting with the first Monday as the first
 *            day of the first week (00..53)
 *    %w - Day of the week (Sunday is 0, 0..6)
 *    %x - Preferred representation for the date alone, no time
 *    %X - Preferred representation for the time alone, no date
 *    %y - Year without a century (00..99)
 *    %Y - Year with century
 *    %z - Time zone as  hour offset from UTC (e.g. +0900)
 *    %Z - Time zone name
 *    %% - Literal ``%'' character
 *
 *     t = Time.now                        #=> 2007-11-19 08:37:48 -0600
 *     t.strftime("Printed on %m/%d/%Y")   #=> "Printed on 11/19/2007"
 *     t.strftime("at %I:%M%p")            #=> "at 08:37AM"
 */

static mrb_value
time_strftime(mrb_state *mrb, mrb_value time)
{
  void rb_enc_copy(mrb_state *mrb, mrb_value, mrb_value);
  struct time_object *tobj;
  char buffer[SMALLBUF], *buf = buffer;
  const char *fmt;
  long len;
  mrb_value str;
  mrb_value format;

  mrb_get_args(mrb, "o", &format);

    GetTimeval(mrb, time, tobj);
    MAKE_TM(mrb, time, tobj);
    //StringValue(format);
    mrb_string_value(mrb, &format);
#ifdef INCLUDE_REGEXP
    if (!rb_enc_str_asciicompat_p(mrb, format)) {
        mrb_raise(mrb, mrb->eArgumentError_class, "format should have ASCII compatible encoding");
    }
#endif
    format = rb_str_new4(mrb, format);
    fmt = RSTRING_PTR(format);
    len = RSTRING_LEN(format);
    if (len == 0) {
        rb_warning("strftime called with empty format string");
    }
    else if (memchr(fmt, '\0', len)) {
        /* Ruby string may contain \0's. */
        const char *p = fmt, *pe = fmt + len;

        str = mrb_str_new(mrb, 0, 0);
        while (p < pe) {
            len = rb_strftime_alloc(mrb, &buf, p, &tobj->vtm, tobj->timew, TIME_UTC_P(tobj));
            rb_str_cat(mrb, str, buf, len);
            p += strlen(p);
            if (buf != buffer) {
                xfree(buf);
                buf = buffer;
            }
            for (fmt = p; p < pe && !*p; ++p);
            if (p > fmt) rb_str_cat(mrb, str, fmt, p - fmt);
        }
        return str;
    }
    else {
        len = rb_strftime_alloc(mrb, &buf, RSTRING_PTR(format),
                                &tobj->vtm, tobj->timew, TIME_UTC_P(tobj));
    }
    str = mrb_str_new(mrb, buf, len);
    if (buf != buffer) xfree(buf);
#ifdef INCLUDE_REGEXP
    rb_enc_copy(mrb, str, format);
#endif
    return str;
}

#if 0
/*
 * undocumented
 */

static mrb_value
time_mdump(mrb_state *mrb, mrb_value time)
{
    struct time_object *tobj;
    unsigned long p, s;
    char buf[8];
    int i;
    mrb_value str;

    struct vtm vtm;
    long year;
    long usec, nsec;
    unsigned long subsecx, nano, subnano, v;

    GetTimeval(mrb, time, tobj);

    gmtimew(mrb, tobj->timew, &vtm);

    if (FIXNUM_P(vtm.year)) {
        year = mrb_fixnum_number(vtm.year);
        if (year < 1900 || 1900+0xffff < year)
            mrb_raise(mrb, mrb->eArgumentError_class, "year too big to marshal: %ld UTC", year);
    }
    else {
        mrb_raise(mrb, mrb->eArgumentError_class, "year too big to marshal");
    }

    subsecx = vtm.subsecx;

    nano = mulquo(mrb, subsecx, (1000000000), (TIME_SCALE));
    divmodv(mrb, nano, (1), &v, &subnano);
    nsec = (v);
    usec = nsec / 1000;
    nsec = nsec % 1000;

    nano = add(mrb, (nsec), subnano);

    p = 0x1UL            << 31 | /*  1 */
        TIME_UTC_P(tobj) << 30 | /*  1 */
        (year-1900)      << 14 | /* 16 */
        (vtm.mon-1)      << 10 | /*  4 */
        vtm.mday         <<  5 | /*  5 */
        vtm.hour;                /*  5 */
    s = vtm.min          << 26 | /*  6 */
        vtm.sec          << 20 | /*  6 */
        usec;    /* 20 */

    for (i=0; i<4; i++) {
        buf[i] = (unsigned char)p;
        p = RSHIFT(p, 8);
    }
    for (i=4; i<8; i++) {
        buf[i] = (unsigned char)s;
        s = RSHIFT(s, 8);
    }

    str = mrb_str_new(mrb, buf, 8);
    rb_copy_generic_ivar(str, time);
    if (nano != (0)) {
        /*if (mrb_type(nano) == T_RATIONAL) {
            mrb_iv_set(mrb, str, id_nano_num, RRATIONAL(nano)->num);
            mrb_iv_set(mrb, str, id_nano_den, RRATIONAL(nano)->den);
        }
        else {*/
            mrb_iv_set(mrb, str, id_nano_num, mrb_fixnum_value(nano));
            mrb_iv_set(mrb, str, id_nano_den, mrb_fixnum_value(1));
        /*}*/
    }
    if (nsec) { /* submicro is only for Ruby 1.9.1 compatibility */
        /*
         * submicro is formatted in fixed-point packed BCD (without sign).
         * It represent digits under microsecond.
         * For nanosecond resolution, 3 digits (2 bytes) are used.
         * However it can be longer.
         * Extra digits are ignored for loading.
         */
        char buf[2];
        int len = (int)sizeof(buf);
        buf[1] = (char)((nsec % 10) << 4);
        nsec /= 10;
        buf[0] = (char)(nsec % 10);
        nsec /= 10;
        buf[0] |= (char)((nsec % 10) << 4);
        if (buf[1] == 0)
            len = 1;
        mrb_iv_set(mrb, str, id_submicro, mrb_str_new(mrb, buf, len));
    }
    if (!TIME_UTC_P(tobj)) {
        mrb_value off = time_utc_offset(mrb, time);
        long div, mod;
        divmodv(mrb, mrb_fixnum_number(off), (1), &div, &mod);
        if (mod == (0))
            off = rb_Integer(mrb, mrb_fixnum_value(div));
        mrb_iv_set(mrb, str, id_offset, off);
    }
    return str;
}

/*
 * call-seq:
 *   time._dump   -> string
 *
 * Dump _time_ for marshaling.
 */

static mrb_value
time_dump(mrb_state *mrb, /*int argc, mrb_value *argv,*/ mrb_value time)
{
  mrb_value str;
  mrb_value argv[16];
  int argc;

  //rb_scan_args(argc, argv, "01", 0);
  mrb_get_args(mrb, "*", &argv, &argc);

  str = time_mdump(mrb, time);

  return str;
}

/*
 * undocumented
 */
static mrb_value
time_mload_str(mrb_state *mrb, mrb_value time, mrb_value str)
{
  struct time_object *tobj;
  unsigned long p, s;
  time_t sec;
  long usec;
  unsigned char *buf;
  struct vtm vtm;
  int i, gmt;
  long nsec;
  mrb_value submicro, nano_num, nano_den, offset;
  wideval_t timew;

  time_modify(time);

  nano_num = rb_attr_get(mrb, str, id_nano_num);
  if (!mrb_nil_p(nano_num)) {
      st_delete((st_table *)RSTRING(str)->iv, (st_data_t*)&id_nano_num, 0);
  }
  nano_den = rb_attr_get(mrb, str, id_nano_den);
  if (!mrb_nil_p(nano_den)) {
      st_delete((st_table *)RSTRING(str)->iv, (st_data_t*)&id_nano_den, 0);
  }
  submicro = rb_attr_get(mrb, str, id_submicro);
  if (!mrb_nil_p(submicro)) {
      st_delete((st_table *)RSTRING(str)->iv, (st_data_t*)&id_submicro, 0);
  }
  offset = rb_attr_get(mrb, str, id_offset);
  if (!mrb_nil_p(offset)) {
      validate_utc_offset(mrb, offset);
      st_delete((st_table *)RSTRING(str)->iv, (st_data_t*)&id_offset, 0);
  }
  rb_copy_generic_ivar(time, str);

  //StringValue(str);
  mrb_string_value(mrb, &str);
  buf = (unsigned char *)RSTRING_PTR(str);
  if (RSTRING_LEN(str) != 8) {
      mrb_raise(mrb, mrb->eTypeError_class, "marshaled time format differ");
  }

  p = s = 0;
  for (i=0; i<4; i++) {
      p |= buf[i]<<(8*i);
  }
  for (i=4; i<8; i++) {
      s |= buf[i]<<(8*(i-4));
  }

  if ((p & (1UL<<31)) == 0) {
      gmt = 0;
      offset = mrb_nil_value();
      sec = p;
      usec = s;
      nsec = usec * 1000;
      timew = wadd(mrb, rb_time_magnify(mrb, TIMET2WV(sec)), wmulquoll(mrb, WINT2FIXWV(usec), TIME_SCALE, 1000000));
  }
  else {
      p &= ~(1UL<<31);
      gmt        = (int)((p >> 30) & 0x1);

      vtm.year = mrb_fixnum_value(((int)(p >> 14) & 0xffff) + 1900);
      vtm.mon  = ((int)(p >> 10) & 0xf) + 1;
      vtm.mday = (int)(p >>  5) & 0x1f;
      vtm.hour = (int) p        & 0x1f;
      vtm.min  = (int)(s >> 26) & 0x3f;
      vtm.sec  = (int)(s >> 20) & 0x3f;
      vtm.utc_offset = mrb_fixnum_value(0);
      vtm.yday = vtm.wday = 0;
      vtm.isdst = 0;
      vtm.zone = "";

      usec = (long)(s & 0xfffff);
      nsec = usec * 1000;


      vtm.subsecx = mulquo(mrb, (nsec), (TIME_SCALE), (1000000000));
      if (!mrb_nil_p(nano_num)) {
          wideval_t nano = quo(mrb, (long_num_exact(mrb, nano_num)), (long_num_exact(mrb, nano_den)));
          vtm.subsecx = add(mrb, vtm.subsecx, mulquo(mrb, nano, (TIME_SCALE), (1000000000)));
      }
      else if (!mrb_nil_p(submicro)) { /* for Ruby 1.9.1 compatibility */
          unsigned char *ptr;
          long len;
          int digit;
          //ptr = (unsigned char*)StringValuePtr(submicro);
          ptr = (unsigned char*)mrb_string_value_ptr(mrb, submicro);
          len = RSTRING_LEN(submicro);
          nsec = 0;
          if (0 < len) {
              if (10 <= (digit = ptr[0] >> 4)) goto end_submicro;
              nsec += digit * 100;
              if (10 <= (digit = ptr[0] & 0xf)) goto end_submicro;
              nsec += digit * 10;
          }
          if (1 < len) {
              if (10 <= (digit = ptr[1] >> 4)) goto end_submicro;
              nsec += digit;
          }
          vtm.subsecx = add(mrb, vtm.subsecx, mulquo(mrb, (nsec), (TIME_SCALE), (1000000000)));
end_submicro: ;
      }
      timew = timegmw(mrb, &vtm);
  }

  GetTimeval(mrb, time, tobj);
  tobj->tm_got = 0;
  tobj->timew = timew;
  if (gmt) {
      TIME_SET_UTC(tobj);
  }
  else if (!mrb_nil_p(offset)) {
      time_set_utc_offset(mrb, time, offset);
      time_fixoff(mrb, time);
  }

  return time;
}

static mrb_value
time_mload(mrb_state *mrb, mrb_value time)
{
  mrb_value str;

  mrb_get_args(mrb, "o", &str);
  return time_mload_str(mrb, time, str);
}

/*
 * call-seq:
 *   Time._load(string)   -> time
 *
 * Unmarshal a dumped +Time+ object.
 */

static mrb_value
time_load(mrb_state *mrb, mrb_value klass)
{
  mrb_value time = mrb_obj_value(time_s_alloc(mrb, mrb_class_get(mrb, klass)));
  mrb_value str;

  mrb_get_args(mrb, "o", &str);

  time_mload_str(mrb, time, str);
  return time;
}
#endif

/*
 *  <code>Time</code> is an abstraction of dates and times. Time is
 *  stored internally as the number of seconds with fraction since
 *  the <em>Epoch</em>, January 1, 1970 00:00 UTC.
 *  Also see the library modules <code>Date</code>.
 *  The <code>Time</code> class treats GMT (Greenwich Mean Time) and
 *  UTC (Coordinated Universal Time)<em>[Yes, UTC really does stand for
 *  Coordinated Universal Time. There was a committee involved.]</em>
 *  as equivalent.  GMT is the older way of referring to these
 *  baseline times but persists in the names of calls on POSIX
 *  systems.
 *
 *  All times may have fraction. Be aware of
 *  this fact when comparing times with each other---times that are
 *  apparently equal when displayed may be different when compared.
 */

void
mrb_init_time(mrb_state *mrb)
{
  struct RClass *t;

  id_eq = mrb_intern(mrb, "==");
  id_ne = mrb_intern(mrb, "!=");
  id_quo = mrb_intern(mrb, "quo");
  id_div = mrb_intern(mrb, "div");
  id_cmp = mrb_intern(mrb, "<=>");
  id_lshift = mrb_intern(mrb, "<<");
  id_divmod = mrb_intern(mrb, "divmod");
  id_mul = mrb_intern(mrb, "*");
  id_submicro = mrb_intern(mrb, "submicro");
  id_nano_num = mrb_intern(mrb, "nano_num");
  id_nano_den = mrb_intern(mrb, "nano_den");
  id_offset = mrb_intern(mrb, "offset");

  t = mrb->time_class = mrb_define_class(mrb, "Time", mrb->object_class);

  //rb_define_alloc_func(mrb, t, time_s_alloc);
  mrb_define_class_method(mrb, t, "at",                time_s_at,              ARGS_ANY());  /* 15.2.19.6.1  */
  mrb_define_class_method(mrb, t, "gm",                time_s_mkutc,           ARGS_ANY());  /* 15.2.19.6.2  */
  mrb_define_class_method(mrb, t, "local",             time_s_mktime,          ARGS_ANY());  /* 15.2.19.6.3  */
  mrb_define_class_method(mrb, t, "mktime",            time_s_mktime,          ARGS_ANY());  /* 15.2.19.6.4  */
  mrb_define_class_method(mrb, t, "now",               time_s_now,             ARGS_NONE()); /* 15.2.19.6.5  */
  mrb_define_class_method(mrb, t, "utc",               time_s_mkutc,           ARGS_ANY());  /* 15.2.19.6.6  */
  mrb_define_class_method(mrb, t, "new",               time_s_new,             ARGS_ANY());  /* 15.2.19.6.7(x) */
  mrb_include_module(mrb, t, mrb->comparable_module);

  mrb_define_method(mrb, t,       "+",                 time_plus,              ARGS_REQ(1)); /* 15.2.19.7.1  */
  mrb_define_method(mrb, t,       "-",                 time_minus,             ARGS_REQ(1)); /* 15.2.19.7.2  */
  mrb_define_method(mrb, t,       "<=>",               time_cmp,               ARGS_REQ(1)); /* 15.2.19.7.3  */
  mrb_define_method(mrb, t,       "asctime",           time_asctime,           ARGS_NONE()); /* 15.2.19.7.4  */
  mrb_define_method(mrb, t,       "ctime",             time_asctime,           ARGS_NONE()); /* 15.2.19.7.5  */
  mrb_define_method(mrb, t,       "day",               time_mday,              ARGS_NONE()); /* 15.2.19.7.6  */
  mrb_define_method(mrb, t,       "dst?",              time_isdst,             ARGS_NONE()); /* 15.2.19.7.7  */
  mrb_define_method(mrb, t,       "getgm",             time_getgmtime,         ARGS_NONE()); /* 15.2.19.7.8  */
  mrb_define_method(mrb, t,       "getlocal",          time_getlocaltime,      ARGS_ANY());  /* 15.2.19.7.9  */
  mrb_define_method(mrb, t,       "getutc",            time_getgmtime,         ARGS_NONE()); /* 15.2.19.7.10 */
  mrb_define_method(mrb, t,       "gmt?",              time_utc_p,             ARGS_NONE()); /* 15.2.19.7.11 */
  mrb_define_method(mrb, t,       "gmt_offset",        time_utc_offset,        ARGS_NONE()); /* 15.2.19.7.12 */
  mrb_define_method(mrb, t,       "gmtime",            time_gmtime,            ARGS_NONE()); /* 15.2.19.7.13 */
  mrb_define_method(mrb, t,       "gmtoff",            time_utc_offset,        ARGS_NONE()); /* 15.2.19.7.14 */
  mrb_define_method(mrb, t,       "hour",              time_hour,              ARGS_NONE()); /* 15.2.19.7.15 */
  mrb_define_method(mrb, t,       "initialize",        time_init,              ARGS_ANY());  /* 15.2.19.7.16 */
  mrb_define_method(mrb, t,       "initialize_copy",   time_init_copy,         ARGS_REQ(1)); /* 15.2.19.7.17 */
  mrb_define_method(mrb, t,       "localtime",         time_localtime_m,       ARGS_ANY());  /* 15.2.19.7.18 */
  mrb_define_method(mrb, t,       "mday",              time_mday,              ARGS_NONE()); /* 15.2.19.7.19 */
  mrb_define_method(mrb, t,       "min",               time_min,               ARGS_NONE()); /* 15.2.19.7.20 */
  mrb_define_method(mrb, t,       "mon",               time_mon,               ARGS_NONE()); /* 15.2.19.7.21 */
  mrb_define_method(mrb, t,       "month",             time_mon,               ARGS_NONE()); /* 15.2.19.7.22 */
  mrb_define_method(mrb, t,       "sec",               time_sec,               ARGS_NONE()); /* 15.2.19.7.23 */
  mrb_define_method(mrb, t,       "to_f",              time_to_f,              ARGS_NONE()); /* 15.2.19.7.24 */
  mrb_define_method(mrb, t,       "to_i",              time_to_i,              ARGS_NONE()); /* 15.2.19.7.25 */
  mrb_define_method(mrb, t,       "usec",              time_usec,              ARGS_NONE()); /* 15.2.19.7.26 */
  mrb_define_method(mrb, t,       "utc",               time_gmtime,            ARGS_NONE()); /* 15.2.19.7.27 */
  mrb_define_method(mrb, t,       "utc?",              time_utc_p,             ARGS_NONE()); /* 15.2.19.7.28 */
  mrb_define_method(mrb, t,       "utc_offset",        time_utc_offset,        ARGS_NONE()); /* 15.2.19.7.29 */
  mrb_define_method(mrb, t,       "wday",              time_wday,              ARGS_NONE()); /* 15.2.19.7.30 */
  mrb_define_method(mrb, t,       "yday",              time_yday,              ARGS_NONE()); /* 15.2.19.7.31 */
  mrb_define_method(mrb, t,       "year",              time_year,              ARGS_NONE()); /* 15.2.19.7.32 */
  mrb_define_method(mrb, t,       "zone",              time_zone,              ARGS_NONE()); /* 15.2.19.7.33 */


#if 0
  mrb_define_method(mrb, t,       "to_r",              time_to_r,              ARGS_NONE()); /* 15.2.19.7.xx */
#endif
  mrb_define_method(mrb, t,       "eql?",              time_eql,               ARGS_REQ(1)); /* 15.2.19.7.xx */
//  mrb_define_method(mrb, t,       "hash",              time_hash,              ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "to_s",              time_to_s,              ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "inspect",           time_to_s,              ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "to_a",              time_to_a,              ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "succ",              time_succ,              ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "round",             time_round,             ARGS_ANY());  /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "isdst",             time_isdst,             ARGS_NONE()); /* 15.2.19.7.xx */


  mrb_define_method(mrb, t,       "sunday?",           time_sunday,            ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "monday?",           time_monday,            ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "tuesday?",          time_tuesday,           ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "wednesday?",        time_wednesday,         ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "thursday?",         time_thursday,          ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "friday?",           time_friday,            ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "saturday?",         time_saturday,          ARGS_NONE()); /* 15.2.19.7.xx */

  mrb_define_method(mrb, t,       "tv_sec",            time_to_i,              ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "tv_usec",           time_usec,              ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "tv_nsec",           time_nsec,              ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "nsec",              time_nsec,              ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "subsec",            time_subsec,            ARGS_NONE()); /* 15.2.19.7.xx */

  mrb_define_method(mrb, t,       "strftime",          time_strftime,          ARGS_REQ(1)); /* 15.2.19.7.xx */

#if 0
    /* methods for marshaling */
  mrb_define_method(mrb, t,       "_dump",             time_dump,              ARGS_ANY());  /* 15.2.19.7.xx */
  mrb_define_class_method(mrb, t, "_load",             time_load,              ARGS_REQ(1)); /* 15.2.19.7.xx */
    /* Time will support marshal_dump and marshal_load in the future (1.9 maybe) */
  mrb_define_method(mrb, t,       "marshal_dump",      time_mdump,             ARGS_NONE()); /* 15.2.19.7.xx */
  mrb_define_method(mrb, t,       "marshal_load",      time_mload,             ARGS_REQ(1)); /* 15.2.19.7.xx */
#endif

#ifdef DEBUG_FIND_TIME_NUMGUESS
  rb_define_virtual_variable(mrb, "$find_time_numguess",    find_time_numguess_getter, NULL);
#endif
}
